@@ -44,11 +44,17 @@
 #include "db_con.h"
 #include "db_res.h"
 #include "db_cap.h"
-#include "db_con.h"
 #include "db_row.h"
 #include "db_ps.h"
+#include "db_pool.h"
 #include "../globals.h"
 
+extern stat_var *sql_total_queries;
+extern stat_var *sql_slow_queries;
+
+/* to be called in the pre-fork phase */
+int init_db_support(void);
+
 /**
  * \brief Specify table name that will be used for subsequent operations.
  *
@@ -388,7 +394,7 @@ int db_bind_mod(const str* mod, db_func_t* dbf);
  * \return returns a pointer to the db_con_t representing the connection if it was
    successful, otherwise 0 is returned.
  */
-db_con_t* db_do_init(const str* url, void* (*new_connection)());
+db_con_t* db_do_init(const str* url, void* (*new_connection)(const struct db_id *));
 
 
 /**
@@ -399,7 +405,7 @@ db_con_t* db_do_init(const str* url, void* (*new_connection)());
  * \param _h database connection handle
  * \param (*free_connection) Pointer to the db specifc free_connection method
  */
-void db_do_close(db_con_t* _h, void (*free_connection)());
+void db_do_close(db_con_t* _h, void (*free_connection)(struct pool_con*));
 
 
 /**
@@ -467,7 +473,7 @@ int estimate_available_rows( int payload_size, int column_count);
 	do{\
 		if (_db_url.s==NULL) {\
 			if (db_default_url==NULL) { \
-				if (!_can_be_null) {\
+				if (!(_can_be_null)) {\
 					LM_ERR("DB URL is not defined!\n"); \
 					return -1; \
 				} \
