@@ -90,6 +90,7 @@
 #include <grp.h>
 #include <signal.h>
 #include <time.h>
+#include <locale.h>
 
 #include <sys/ioctl.h>
 #include <net/if.h>
@@ -99,6 +100,8 @@
 
 #include "help_msg.h"
 #include "config.h"
+#include "cfg_pp.h"
+#include "cfg_reload.h"
 #include "dprint.h"
 #include "daemonize.h"
 #include "route.h"
@@ -106,8 +109,10 @@
 #include "globals.h"
 #include "mem/mem.h"
 #include "mem/shm_mem.h"
+#include "mem/rpm_mem.h"
 #include "sr_module.h"
 #include "timer.h"
+#include "ipc.h"
 #include "parser/msg_parser.h"
 #include "ip_addr.h"
 #include "resolve.h"
@@ -119,13 +124,18 @@
 #include "script_cb.h"
 #include "dset.h"
 #include "blacklists.h"
+#include "xlog.h"
+#include "ipc.h"
 
 #include "pt.h"
 #include "ut.h"
 #include "serialize.h"
 #include "statistics.h"
+#include "status_report.h"
 #include "core_stats.h"
 #include "pvar.h"
+#include "signals.h"
+#include "shutdown.h"
 #include "poll_types.h"
 #include "net/net_tcp.h"
 #include "net/net_udp.h"
@@ -133,24 +143,44 @@
 #include "version.h"
 #include "mi/mi_core.h"
 #include "db/db_insertq.h"
+#include "cachedb/cachedb.h"
 #include "net/trans.h"
 
+#include "test/unit_tests.h"
+#include "lib/dbg/profiling.h"
+
+#include "ssl_tweaks.h"
+
+/* global vars */
+
+#ifdef VERSION_NODATE
+static char compiled[] =  "" ;
+#else
+#ifdef VERSION_DATE
+static const char compiled[] =  VERSION_DATE ;
+#else
+static const char compiled[] =  __TIME__ " " __DATE__ ;
+#endif
+#endif
+
+static int own_pgid = 0; /* whether or not we have our own pgid (and it's ok
+			    to use kill(0, sig) */
+
 static char* version=OPENSIPS_FULL_VERSION;
 static char* flags=OPENSIPS_COMPILE_FLAGS;
-#ifdef VERSION_NODATE
-char compiled[] =  "" ;
-#else
-#ifdef VERSION_DATE
-const char compiled[] =  VERSION_DATE ;
-#else
-const char compiled[] =  __TIME__ " " __DATE__ ;
-#endif
-#endif
+
+static int user_id = 0;
+static int group_id = 0;
+
+const struct scm_version core_scm_ver = {
+	.type = VERSIONTYPE,
+	.rev = THISREVISION
+};
 
 /**
  * Print compile-time constants
  */
-void print_ct_constants(void)
+static void print_ct_constants(void)
 {
 #ifdef ADAPTIVE_WAIT
 	printf("ADAPTIVE_WAIT_LOOPS=%d, ", ADAPTIVE_WAIT_LOOPS);
@@ -160,478 +190,7 @@ void print_ct_constants(void)
 		MAX_RECV_BUFFER_SIZE, MAX_LISTEN, MAX_URI_SIZE,
 		BUF_SIZE );
 	printf("poll method support: %s.\n", poll_support);
-#ifdef VERSIONTYPE
-	printf("%s revision: %s\n", VERSIONTYPE, THISREVISION);
-#endif
-}
-
-/* global vars */
-
-int own_pgid = 0; /* whether or not we have our own pgid (and it's ok
-					 to use kill(0, sig) */
-char* cfg_file = 0;
-unsigned int maxbuffer = MAX_RECV_BUFFER_SIZE; /* maximum buffer size we do
-												  not want to exceed during the
-												  auto-probing procedure; may
-												  be re-configured */
-int children_no = CHILD_NO;		/* number of children processing requests */
-enum poll_types io_poll_method=0; 	/*!< by default choose the best method */
-int sig_flag = 0;              /* last signal received */
-
-/* activate debug mode */
-int debug_mode = 0;
-/* do not become daemon, stay attached to the console */
-int no_daemon_mode = 0;
-/* assertion statements in script. disabled by default */
-int enable_asserts = 0;
-/* abort process on failed assertion. disabled by default */
-int abort_on_assert = 0;
-/* start by logging to stderr */
-int log_stderr = 1;
-/* log facility (see syslog(3)) */
-int log_facility = LOG_DAEMON;
-/* the id to be printed in syslog */
-char *log_name = 0;
-int config_check = 0;
-/* check if reply first via host==us */
-int check_via =  0;
-/* debugging level for memory stats */
-int memlog = L_DBG + 10;
-int memdump = L_DBG + 10;
-/* debugging in case msg processing takes. too long disabled by default */
-int execmsgthreshold = 0;
-/* debugging in case dns takes too long. disabled by default */
-int execdnsthreshold = 0;
-/* debugging in case tcp stuff take too long. disabled by default */
-int tcpthreshold = 0;
-/* should replies include extensive warnings? by default yes,
-   good for trouble-shooting
-*/
-int sip_warning = 0;
-/* should localy-generated messages include server's signature? */
-int server_signature=1;
-/* Server header to be used when proxy generates request as UAS.
-   Default is to use SERVER_HDR CRLF (assigned later).
-*/
-str server_header = {SERVER_HDR,sizeof(SERVER_HDR)-1};
-/* User-Agent header to be used when proxy generates request as UAC.
-   Default is to use USER_AGENT CRLF (assigned later).
-*/
-str user_agent_header = {USER_AGENT,sizeof(USER_AGENT)-1};
-/* should opensips try to locate outbound interface on multihomed
- * host? by default not -- too expensive
- */
-int mhomed=0;
-/* use dns and/or rdns or to see if we need to add
-   a ;received=x.x.x.x to via: */
-int received_dns = 0;
-char* working_dir = 0;
-char* chroot_dir = 0;
-char* user=0;
-char* group=0;
-int uid = 0;
-int gid = 0;
-
-/* more config stuff */
-int disable_core_dump=0; /* by default enabled */
-int open_files_limit=-1; /* don't touch it by default */
-
-#ifdef USE_MCAST
-int mcast_loopback = 0;
-int mcast_ttl = -1; /* if -1, don't touch it, use the default (usually 1) */
-#endif /* USE_MCAST */
-
-int tos = IPTOS_LOWDELAY;
-
-struct socket_info* bind_address=0; /* pointer to the crt. proc.
-									 listening address*/
-
-
-/* if aliases should be automatically discovered and added
- * during fixing listening sockets */
-int auto_aliases=1;
-
-/* if the stateless forwarding support in core should be
- * disabled or not */
-int sl_fwd_disabled=-1;
-
-/* process number - 0 is the main process */
-int process_no = 0;
-
-/* cfg parsing */
-int cfg_errors=0;
-
-/* start-up time */
-time_t startup_time = 0;
-
-/* shared memory (in MB) */
-unsigned long shm_mem_size=SHM_MEM_SIZE * 1024 * 1024;
-unsigned int shm_hash_split_percentage = DEFAULT_SHM_HASH_SPLIT_PERCENTAGE;
-unsigned int shm_secondary_hash_size = DEFAULT_SHM_SECONDARY_HASH_SIZE;
-
-/* packaged memory (in MB) */
-unsigned long pkg_mem_size=PKG_MEM_SIZE * 1024 * 1024;
-
-
-/* export command-line to anywhere else */
-int my_argc;
-char **my_argv;
-
-extern FILE* yyin;
-extern int yyparse();
-#ifdef DEBUG_PARSER
-extern int yydebug;
-#endif
-
-
-int is_main = 1; /* flag = is this the  "main" process? */
-
-char* pid_file = 0; /* filename as asked by user */
-char* pgid_file = 0;
-
-
-/**
- * Clean up on exit. This should be called before exiting.
- * \param show_status set to one to display the mem status
- */
-void cleanup(int show_status)
-{
-	LM_INFO("cleanup\n");
-	/*clean-up*/
-
-	/* hack: force-unlock the shared memory lock in case
-	   		 some process crashed and let it locked; this will
-	   		 allow an almost gracious shutdown */
-	if (mem_lock)
-#ifdef HP_MALLOC
-	{
-		int i;
-
-		for (i = 0; i < HP_HASH_SIZE; i++)
-			shm_unlock(i);
-	}
-#else
-		shm_unlock();
-#endif
-
-	handle_ql_shutdown();
-	destroy_modules();
-	udp_destroy();
-	tcp_destroy();
-	destroy_timer();
-	destroy_stats_collector();
-	destroy_script_cb();
-	pv_free_extra_list();
-	destroy_argv_list();
-	destroy_black_lists();
-#ifdef PKG_MALLOC
-	if (show_status){
-		LM_GEN1(memdump, "Memory status (pkg):\n");
-		pkg_status();
-	}
-#endif
-	cleanup_log_level();
-
-	if (pt) shm_free(pt);
-	pt=0;
-	if (show_status){
-			LM_GEN1(memdump, "Memory status (shm):\n");
-			shm_status();
-	}
-
-	/* zero all shmem alloc vars that we still use */
-	shm_mem_destroy();
-	if (pid_file) unlink(pid_file);
-	if (pgid_file) unlink(pgid_file);
-}
-
-
-/**
- * Tries to send a signal to all our processes
- * If daemonized  is ok to send the signal to all the process group,
- * however if not daemonized we might end up sending the signal also
- * to the shell which launched us => most signals will kill it if
- * it's not in interactive mode and we don't want this. The non-daemonized
- * case can occur when an error is encountered before daemonize is called
- * (e.g. when parsing the config file) or when opensips is started in
- * "dont-fork" mode.
- * \param signum signal for killing the children
- */
-static void kill_all_children(int signum)
-{
-	int r;
-	if (own_pgid) {
-		/* check that all processes initialized properly */
-		for (r=1; r<counted_processes; r++) {
-			while (pt[r].pid==0){
-				usleep(1000);
-			}
-		}
-		kill(0, signum);
-	} else if (pt)
-		for (r=1; r<counted_processes; r++) {
-			if (pt[r].pid==-1) continue;
-			/* as the PIDs are filled in by child processes, a 0 PID means
-			 * an un-initalized procees; killing an uninitialized proc is
-			 * very dangerous, so better wait for it to finish its init
-			 * sequance by checking when the pid is populated */
-			while (pt[r].pid==0) usleep(1000);
-			kill(pt[r].pid, signum);
-		}
-}
-
-
-
-/**
- * Timeout handler during wait for children exit.
- * If this handler is called, a critical timeout has occurred while
- * waiting for the children to finish => we should kill everything and exit
- * \param signo signal for killing the children
- */
-static void sig_alarm_kill(int signo)
-{
-	kill_all_children(SIGKILL); /* this will kill the whole group
-								  including "this" process;
-								  for debugging replace with SIGABRT
-								  (but warning: it might generate lots
-								   of cores) */
-}
-
-
-/**
- * Timeout handler during wait for children exit.
- * like sig_alarm_kill, but the timeout has occurred when cleaning up,
- * try to leave a core for future diagnostics
- * \param signo signal for killing the children
- * \see sig_alarm_kill
- */
-static void sig_alarm_abort(int signo)
-{
-	/* LOG is not signal safe, but who cares, we are abort-ing anyway :-) */
-	LM_CRIT("BUG - shutdown timeout triggered, dying...");
-	abort();
-}
-
-
-/**
- * Signal handler for the server.
- */
-void handle_sigs(void)
-{
-	pid_t  chld;
-	int    chld_status,overall_status=0;
-	int    i;
-	int    do_exit;
-	const unsigned int shutdown_time = 60; /* one minute close timeout */
-
-	switch(sig_flag){
-		case 0: break; /* do nothing*/
-		case SIGPIPE:
-				/* SIGPIPE might be rarely received on use of
-				   exec module; simply ignore it
-				 */
-				LM_WARN("SIGPIPE received and ignored\n");
-				break;
-		case SIGINT:
-		case SIGTERM:
-			/* we end the program in all these cases */
-			if (sig_flag==SIGINT)
-				LM_DBG("INT received, program terminates\n");
-			else
-				LM_DBG("SIGTERM received, program terminates\n");
-
-			/* first of all, kill the children also */
-			kill_all_children(SIGTERM);
-			if (signal(SIGALRM, sig_alarm_kill) == SIG_ERR ) {
-				LM_ERR("could not install SIGALARM handler\n");
-				/* continue, the process will die anyway if no
-				 * alarm is installed which is exactly what we want */
-			}
-			alarm(shutdown_time);
-			while(wait(0) > 0); /* Wait for all the children to terminate */
-			signal(SIGALRM, sig_alarm_abort);
-
-			cleanup(1); /* cleanup & show status*/
-			alarm(0);
-			signal(SIGALRM, SIG_IGN);
-			dprint("Thank you for flying " NAME "\n");
-			exit(0);
-			break;
-
-		case SIGUSR1:
-#ifdef PKG_MALLOC
-			LM_GEN1(memdump, "Memory status (pkg):\n");
-			pkg_status();
-#endif
-			LM_GEN1(memdump, "Memory status (shm):\n");
-			shm_status();
-			break;
-
-		case SIGUSR2:
-#ifdef PKG_MALLOC
-			set_pkg_stats( get_pkg_status_holder(process_no) );
-#endif
-			break;
-
-		case SIGCHLD:
-			do_exit = 0;
-			while ((chld=waitpid( -1, &chld_status, WNOHANG ))>0) {
-				/* is it a process we know about? */
-				for( i=0 ; i<counted_processes ; i++ )
-					if (pt[i].pid==chld) break;
-				if (i==counted_processes) {
-					LM_DBG("unknown child process %d ended. Ignoring\n",chld);
-					continue;
-				}
-				do_exit = 1;
-				/* process the signal */
-				overall_status |= chld_status;
-				LM_DBG("status = %d\n",overall_status);
-
-				if (WIFEXITED(chld_status))
-					LM_INFO("child process %d exited normally,"
-							" status=%d\n", chld,
-							WEXITSTATUS(chld_status));
-				else if (WIFSIGNALED(chld_status)) {
-					LM_INFO("child process %d exited by a signal"
-							" %d\n", chld, WTERMSIG(chld_status));
-#ifdef WCOREDUMP
-					LM_INFO("core was %sgenerated\n",
-							 WCOREDUMP(chld_status) ?  "" : "not " );
-#endif
-				}else if (WIFSTOPPED(chld_status))
-					LM_INFO("child process %d stopped by a"
-								" signal %d\n", chld,
-								 WSTOPSIG(chld_status));
-			}
-			if (!do_exit)
-				break;
-			LM_INFO("terminating due to SIGCHLD\n");
-			/* exit */
-			kill_all_children(SIGTERM);
-			if (signal(SIGALRM, sig_alarm_kill) == SIG_ERR ) {
-				LM_ERR("could not install SIGALARM handler\n");
-				/* continue, the process will die anyway if no
-				 * alarm is installed which is exactly what we want */
-			}
-			alarm(shutdown_time);
-			while(wait(0) > 0); /* wait for all the children to terminate*/
-			signal(SIGALRM, sig_alarm_abort);
-			cleanup(1); /* cleanup & show status*/
-			alarm(0);
-			signal(SIGALRM, SIG_IGN);
-			LM_DBG("terminating due to SIGCHLD\n");
-			exit(overall_status ? -1 : 0);
-			break;
-
-		case SIGHUP: /* ignoring it*/
-			LM_DBG("SIGHUP received, ignoring it\n");
-			break;
-		default:
-			LM_CRIT("unhandled signal %d\n", sig_flag);
-	}
-	sig_flag=0;
-}
-
-
-
-/**
- * Exit regulary on a specific signal.
- * This is good for profiling which only works if exited regularly
- * and not by default signal handlers
- * \param signo The signal that should be handled
- */
-static void sig_usr(int signo)
-{
-	int status;
-	pid_t pid;
-	UNUSED(pid);
-
-	if (is_main){
-		if (sig_flag==0) sig_flag=signo;
-		else /*  previous sig. not processed yet, ignoring? */
-			return; ;
-	}else{
-		/* process the important signals */
-		switch(signo){
-			case SIGPIPE:
-					LM_INFO("signal %d received\n", signo);
-				break;
-			case SIGINT:
-			case SIGTERM:
-					LM_INFO("signal %d received\n", signo);
-					/* print memory stats for non-main too */
-					#ifdef PKG_MALLOC
-					LM_GEN1(memdump, "Memory status (pkg):\n");
-					pkg_status();
-					#endif
-					exit(0);
-					break;
-			case SIGUSR1:
-					/* statistics -> show only pkg mem */
-					#ifdef PKG_MALLOC
-					LM_GEN1(memdump, "Memory status (pkg):\n");
-					pkg_status();
-					#endif
-					break;
-			case SIGUSR2:
-					#ifdef PKG_MALLOC
-					set_pkg_stats( get_pkg_status_holder(process_no) );
-					#endif
-					break;
-			case SIGHUP:
-					/* ignored*/
-					break;
-			case SIGCHLD:
-					pid = waitpid(-1, &status, WNOHANG);
-					LM_DBG("SIGCHLD received from %ld (status=%d), ignoring\n",
-						(long)pid,status);
-		}
-	}
-}
-
-
-
-/**
- * Install the signal handlers.
- * \return 0 on success, -1 on error
- */
-int install_sigs(void)
-{
-	/* added by jku: add exit handler */
-	if (signal(SIGINT, sig_usr) == SIG_ERR ) {
-		LM_ERR("no SIGINT signal handler can be installed\n");
-		goto error;
-	}
-	/* if we debug and write to a pipe, we want to exit nicely too */
-	if (signal(SIGPIPE, sig_usr) == SIG_ERR ) {
-		LM_ERR("no SIGINT signal handler can be installed\n");
-		goto error;
-	}
-
-	if (signal(SIGUSR1, sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGUSR1 signal handler can be installed\n");
-		goto error;
-	}
-	if (signal(SIGCHLD , sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGCHLD signal handler can be installed\n");
-		goto error;
-	}
-	if (signal(SIGTERM , sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGTERM signal handler can be installed\n");
-		goto error;
-	}
-	if (signal(SIGHUP , sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGHUP signal handler can be installed\n");
-		goto error;
-	}
-	if (signal(SIGUSR2 , sig_usr)  == SIG_ERR ) {
-		LM_ERR("no SIGUSR2 signal handler can be installed\n");
-		goto error;
-	}
-	return 0;
-error:
-	return -1;
+	printf("%s revision: %s\n", core_scm_ver.type, core_scm_ver.rev);
 }
 
 
@@ -644,15 +203,22 @@ static int main_loop(void)
 {
 	static int chd_rank;
 	int* startup_done = NULL;
+	utime_t last_check = 0;
+	int rc;
+	static struct internal_fork_handler profiling_handler = {
+		.desc = "_ProfilerStart_child()",
+		.on_child_init = _ProfilerStart_child,
+	};
 
 	chd_rank=0;
+	register_fork_handler(&profiling_handler);
 
 	if (start_module_procs()!=0) {
 		LM_ERR("failed to fork module processes\n");
 		goto error;
 	}
 
-	if(startup_rlist.a) {/* if a startup route was defined */
+	if(sroutes->startup.a) {/* if a startup route was defined */
 		startup_done = (int*)shm_malloc(sizeof(int));
 		if(startup_done == NULL) {
 			LM_ERR("No more shared memory\n");
@@ -669,7 +235,7 @@ static int main_loop(void)
 
 	/* fork all processes required by UDP network layer */
 	if (udp_start_processes( &chd_rank, startup_done)<0) {
-		LM_CRIT("cannot start TCP processes\n");
+		LM_CRIT("cannot start UDP processes\n");
 		goto error;
 	}
 
@@ -700,21 +266,63 @@ static int main_loop(void)
 		shm_free(startup_done);
 	}
 
+	sr_set_core_status( STATE_RUNNING, MI_SSTR("running"));
+	sr_add_core_report( MI_SSTR("initialization completed, ready now") );
+
 	/* main process left */
 	is_main=1;
 	set_proc_attrs("attendant");
+	pt[process_no].flags |= OSS_PROC_IS_RUNNING|OSS_PROC_NO_IPC|OSS_PROC_NO_LOAD;
 
-	if (init_child(PROC_MAIN) < 0) {
-		LM_ERR("error in init_child for PROC_MAIN\n");
-		report_failure_status();
-		goto error;
+	if (testing_framework) {
+		if (init_child(1) < 0) {
+			LM_ERR("error in init_child for First Worker\n");
+			report_failure_status();
+			goto error;
+		}
+
+		rc = run_unit_tests();
+		shutdown_opensips(rc);
 	}
 
 	report_conditional_status( (!no_daemon_mode), 0);
 
+	if (auto_scaling_enabled) {
+		/* re-create the status pipes to collect the status of the
+		 * dynamically forked processes */
+		if (create_status_pipe() < 0) {
+			LM_ERR("failed to create status pipe\n");
+			goto error;
+		}
+		/* keep both ends on the status pipe as we will keep forking 
+		 * processes, so we will need to pass write-end to the new children;
+		 * of course, we will need the read-end, here in the main proc */
+		last_check = get_uticks();
+	}
+
+	if (set_log_event_cons_cfg_state() < 0) {
+		LM_ERR("Failed to set the config state for event consumer\n");
+		goto error;
+	}
+
+	if (_ProfilerStart(0, "attendant") != 0) {
+		LM_ERR("failed to start profiling\n");
+		goto error;
+	}
+
 	for(;;){
 			handle_sigs();
-			pause();
+			if (auto_scaling_enabled) {
+				sleep( auto_scaling_cycle );
+				if ( (get_uticks()-last_check) >= (utime_t)auto_scaling_cycle*1000000) {
+					do_workers_auto_scaling();
+					last_check = get_uticks();
+				} else {
+					sleep_us( last_check + auto_scaling_cycle*1000000 -
+						get_uticks() );
+				}
+			} else
+				pause();
 	}
 
 	/*return 0; */
@@ -738,9 +346,6 @@ error:
  */
 int main(int argc, char** argv)
 {
-	/* configure by default logging to syslog */
-	int cfg_log_stderr = 1;
-	FILE* cfg_stream;
 	int c,r;
 	char *tmp;
 	int tmp_len;
@@ -751,6 +356,7 @@ int main(int argc, char** argv)
 	int ret;
 	unsigned int seed;
 	int rfd;
+	int procs_no;
 
 	/*init*/
 	ret=-1;
@@ -758,7 +364,12 @@ int main(int argc, char** argv)
 
 	/* process pkg mem size from command line */
 	opterr=0;
-	options="f:cCm:M:b:l:n:N:rRvdDFETSVhw:t:u:g:P:G:W:o:";
+
+	options="f:cCm:M:b:l:n:N:rRvdDFEVhw:t:u:g:p:P:G:W:o:a:k:s:"
+#ifdef UNIT_TESTS
+	"T:"
+#endif
+	;
 
 	while((c=getopt(argc,argv,options))!=-1){
 		switch(c){
@@ -767,16 +378,80 @@ int main(int argc, char** argv)
 					if (tmp &&(*tmp)){
 						LM_ERR("bad pkgmem size number: -m %s\n", optarg);
 						goto error00;
-					};
-
+					}
+					break;
+			case 'm':
+					shm_mem_size=strtol(optarg, &tmp, 10) * 1024 * 1024;
+					if (tmp &&(*tmp)){
+						LM_ERR("bad shmem size number: -m %s\n", optarg);
+						goto error00;
+					}
+					break;
+			case 'd':
+					*log_level = debug_mode ? L_DBG : (*log_level)+1;
+					break;
+			case 'u':
+					user=optarg;
+					break;
+			case 'g':
+					group=optarg;
+					break;
+			case 'a':
+					if (set_global_mm(optarg) < 0) {
+						LM_ERR("current build does not support "
+						       "this allocator (-a %s)\n", optarg);
+						goto error00;
+					}
+					break;
+			case 'k':
+					if (set_pkg_mm(optarg) < 0) {
+						LM_ERR("current build does not support "
+						       "this allocator (-k %s)\n", optarg);
+						goto error00;
+					}
+					break;
+			case 's':
+					if (set_shm_mm(optarg) < 0) {
+						LM_ERR("current build does not support "
+						       "this allocator (-s %s)\n", optarg);
+						goto error00;
+					}
+					break;
+			case 'e':
+					if (set_rpm_mm(optarg) < 0) {
+						LM_ERR("current build does not support "
+						       "this allocator (-e %s)\n", optarg);
+						goto error00;
+					}
 					break;
 		}
 	}
+
+	/* get uid/gid */
+	if (user){
+		if (user2uid(&user_id, &group_id, user)<0){
+			LM_ERR("bad user name/uid number: -u %s\n", user);
+			goto error00;
+		}
+	}
+	if (group){
+		if (group2gid(&group_id, group)<0){
+			LM_ERR("bad group name/gid number: -u %s\n", group);
+			goto error00;
+		}
+	}
+
 	/*init pkg mallocs (before parsing cfg but after parsing cmd line !)*/
 	if (init_pkg_mallocs()==-1)
 		goto error00;
 
-	init_route_lists();
+	if ( (sroutes=new_sroutes_holder(0))==NULL )
+		goto error00;
+
+	/* we want to be sure that from now on, all the floating numbers are 
+	 * using the dot as separator. This is a real issue when printing the
+	 * floats for SQL ops, where the dot must be used */
+	setlocale( LC_NUMERIC, "POSIX");
 
 	/* process command line (get port no, cfg. file path etc) */
 	/* first reset getopt */
@@ -792,17 +467,15 @@ int main(int argc, char** argv)
 					if (config_check==3)
 						break;
 					config_check |= 1;
-					cfg_log_stderr=1; /* force stderr logging */
 					break;
 			case 'm':
-					shm_mem_size=strtol(optarg, &tmp, 10) * 1024 * 1024;
-					if (tmp &&(*tmp)){
-						LM_ERR("bad shmem size number: -m %s\n", optarg);
-						goto error00;
-					};
-					break;
 			case 'M':
-					/* ignoring it, parsed previously */
+			case 'd':
+			case 'a':
+			case 'k':
+			case 's':
+			case 'e':
+					/* ignoring, parsed previously */
 					break;
 			case 'b':
 					maxbuffer=strtol(optarg, &tmp, 10);
@@ -819,15 +492,15 @@ int main(int argc, char** argv)
 					}
 					tmp[tmp_len]=0; /* null terminate the host */
 					/* add a new addr. to our address list */
-					if (add_cmd_listener(tmp, port, proto)!=0){
+					if (add_cmd_listening_socket(tmp, port, proto)!=0){
 						LM_ERR("failed to add new listen address\n");
 						goto error00;
 					}
 					break;
 			case 'n':
-					children_no=strtol(optarg, &tmp, 10);
+					udp_workers_no=strtol(optarg, &tmp, 10);
 					if ((tmp==0) ||(*tmp)){
-						LM_ERR("bad process number: -n %s\n", optarg);
+						LM_ERR("bad UDP workers number: -n %s\n", optarg);
 						goto error00;
 					}
 					break;
@@ -840,9 +513,6 @@ int main(int argc, char** argv)
 			case 'R':
 					received_dns|=DO_REV_DNS;
 				    break;
-			case 'd':
-					*log_level = debug_mode ? L_DBG : (*log_level)+1;
-					break;
 			case 'D':
 					debug_mode=1;
 					*log_level = L_DBG;
@@ -851,12 +521,13 @@ int main(int argc, char** argv)
 					no_daemon_mode=1;
 					break;
 			case 'E':
-					cfg_log_stderr=1;
-					break;
+					LM_ERR("-E option deprecated since 3.4, set \"stderror_enabled=yes\" "
+					    "at the script level instead\n");
+					goto error00;
 			case 'N':
-					tcp_children_no=strtol(optarg, &tmp, 10);
+					tcp_workers_no=strtol(optarg, &tmp, 10);
 					if ((tmp==0) ||(*tmp)){
-						LM_ERR("bad process number: -N %s\n", optarg);
+						LM_ERR("bad TCP workers number: -N %s\n", optarg);
 						goto error00;
 					}
 					break;
@@ -888,10 +559,13 @@ int main(int argc, char** argv)
 					chroot_dir=optarg;
 					break;
 			case 'u':
-					user=optarg;
+					/* ignoring it, parsed previously */
 					break;
 			case 'g':
-					group=optarg;
+					/* ignoring it, parsed previously */
+					break;
+			case 'p':
+					preproc=optarg;
 					break;
 			case 'P':
 					pid_file=optarg;
@@ -903,6 +577,18 @@ int main(int argc, char** argv)
 					if (add_arg_var(optarg) < 0)
 						LM_ERR("cannot add option %s\n", optarg);
 					break;
+#ifdef UNIT_TESTS
+			case 'T':
+					LM_INFO("running in testing framework mode, for '%s'\n", optarg);
+					testing_framework = 1;
+					testing_module = optarg;
+					if (strcmp(testing_module, "core")) {
+						cfg_file = malloc(100);
+						snprintf(cfg_file, 100, "modules/%s/test/opensips.cfg",
+						         testing_module);
+					}
+					break;
+#endif
 			case '?':
 					if (isprint(optopt))
 						LM_ERR("Unknown option `-%c`.\n", optopt);
@@ -917,19 +603,6 @@ int main(int argc, char** argv)
 		}
 	}
 
-	log_stderr = cfg_log_stderr;
-
-	/* fill missing arguments with the default values*/
-	if (cfg_file==0) cfg_file=CFG_FILE;
-
-	/* load config file or die */
-	cfg_stream=fopen (cfg_file, "r");
-	if (cfg_stream==0){
-		LM_ERR("loading config file(%s): %s\n", cfg_file,
-				strerror(errno));
-		goto error00;
-	}
-
 	/* seed the prng, try to use /dev/urandom if possible */
 	/* no debugging information is logged, because the standard
 	   log level prior the config file parsing is L_NOTICE */
@@ -962,26 +635,7 @@ try_again:
 		goto error;
 	}
 
-	/* used for parser debugging */
-#ifdef DEBUG_PARSER
-	yydebug = 1;
-#endif
-
-		/* get uid/gid */
-	if (user){
-		if (user2uid(&uid, &gid, user)<0){
-			LM_ERR("bad user name/uid number: -u %s\n", user);
-			goto error00;
-		}
-	}
-	if (group){
-		if (group2gid(&gid, group)<0){
-			LM_ERR("bad group name/gid number: -u %s\n", group);
-			goto error00;
-		}
-	}
-
-		/*init shm mallocs
+	/*  init shm mallocs
 	 *  this must be here
 	 *     -to allow setting shm mem size from the command line
 	 *     -it must be also before init_timer and init_tcp
@@ -996,14 +650,34 @@ try_again:
 		goto error;
 	}
 
-	/* parse the config file, prior to this only default values
-	   e.g. for debugging settings will be used */
-	yyin=cfg_stream;
-	if ((yyparse()!=0)||(cfg_errors)){
-		LM_ERR("bad config file (%d errors)\n", cfg_errors);
+	if (init_status_report() < 0) {
+		LM_ERR("failed to initialize status-report support\n");
+		goto error;
+	}
+
+	sr_set_core_status( STATE_INITIALIZING, MI_SSTR("initializing"));
+	sr_add_core_report( MI_SSTR("initializing") );
+
+	if ((!testing_framework || strcmp(testing_module, "core"))
+	        && parse_opensips_cfg(cfg_file, preproc, NULL) < 0) {
+		LM_ERR("failed to parse config file %s\n", cfg_file);
 		goto error00;
 	}
 
+	if (shm_memlog_size && init_dbg_shm_mallocs()==-1)
+		goto error;
+
+	/* shm statistics, module stat groups, memory warming */
+	if (init_shm_post_yyparse() != 0) {
+		LM_ERR("failed to init SHM memory\n");
+		return ret;
+	}
+
+	if (init_log_cons_shm_table() < 0) {
+		LM_ERR("Failed to initialize shm log consumers table\n");
+		goto error;
+	}
+
 	if (config_check>1 && check_rls()!=0) {
 		LM_ERR("bad function call in config file\n");
 		return ret;
@@ -1015,12 +689,15 @@ try_again:
 	}
 
 	/* init the resolver, before fixing the config */
-	resolv_init();
+	if (resolv_init() != 0) {
+		LM_ERR("failed to init DNS resolver\n");
+		return -1;
+	}
 
 	fix_poll_method( &io_poll_method );
 
-	/* fix temporary listeners added in the cmd line */
-	if (fix_cmd_listeners() < 0) {
+	/* fix temporary listening sockets added in the cmd line */
+	if (fix_cmd_listening_sockets() < 0) {
 		LM_ERR("cannot add temproray listeners\n");
 		return ret;
 	}
@@ -1030,7 +707,7 @@ try_again:
 	if (protos_no < 0) {
 		LM_ERR("cannot load transport protocols\n");
 		goto error;
-	} else if (protos_no == 0) {
+	} else if (protos_no == 0 && !testing_framework) {
 		LM_ERR("no transport protocol loaded\n");
 		goto error;
 	} else
@@ -1058,9 +735,6 @@ try_again:
 
 	time(&startup_time);
 
-	/* Init statistics */
-	init_shm_statistics();
-
 	/*init UDP networking layer*/
 	if (udp_init()<0){
 		LM_CRIT("could not initialize tcp\n");
@@ -1077,18 +751,27 @@ try_again:
 		goto error;
 	}
 
-	if (debug_mode) {
+	if (testing_framework)
+		debug_mode = 1;
 
+	if (debug_mode) {
 		LM_NOTICE("DEBUG MODE activated\n");
 		if (no_daemon_mode==0) {
 			LM_NOTICE("disabling daemon mode (found enabled)\n");
 			no_daemon_mode = 1;
 		}
-		if (log_stderr==0) {
+		if (stderr_enabled==0) {
 			LM_NOTICE("enabling logging to standard error (found disabled)\n");
-			log_stderr = 1;
+			stderr_enabled = 1;
+			set_log_consumer_mute_state(&str_init(STDERR_CONSUMER_NAME), 0);
 		}
-		if (*log_level<L_DBG) {
+		if (syslog_enabled) {
+			LM_NOTICE("disabling logging to syslog (found enabled)\n");
+			syslog_enabled = 0;
+			set_log_consumer_mute_state(&str_init(SYSLOG_CONSUMER_NAME), 1);
+		}
+		if (*log_level < L_DBG && (!testing_framework ||
+		                           !strcmp(testing_module, "core"))) {
 			LM_NOTICE("setting logging to debug level (found on %d)\n",
 				*log_level);
 			*log_level = L_DBG;
@@ -1097,27 +780,26 @@ try_again:
 			LM_NOTICE("enabling core dumping (found off)\n");
 			disable_core_dump = 0;
 		}
-		if (udp_count_processes()!=0) {
-			if (children_no!=2) {
+		if (udp_count_processes(NULL)!=0) {
+			if (udp_workers_no!=2) {
 				LM_NOTICE("setting UDP children to 2 (found %d)\n",
-					children_no);
-				children_no = 2;
+					udp_workers_no);
+				udp_workers_no = 2;
 			}
 		}
-		if (tcp_count_processes()!=0) {
-			if (tcp_children_no!=2) {
-				LM_NOTICE("setting TCP children to 2 (found %d)\n",
-					tcp_children_no);
-				tcp_children_no = 2;
+		if (tcp_count_processes(NULL)!=0) {
+			procs_no = (tcp_workers_max_no>=2) ? 2 : tcp_workers_max_no;
+			if (tcp_workers_no != procs_no) {
+				LM_NOTICE("setting TCP children to %d (found %d)\n",
+					procs_no, tcp_workers_no);
+				tcp_workers_no = procs_no;
 			}
 		}
 
 	} else { /* debug_mode */
-
 		/* init_daemon */
 		if ( daemonize((log_name==0)?argv[0]:log_name, &own_pgid) <0 )
 			goto error;
-
 	}
 
 	/* install signal handlers */
@@ -1139,9 +821,14 @@ try_again:
 	LM_NOTICE("version: %s\n", version);
 
 	/* print some data about the configuration */
-	LM_INFO("using %ld Mb shared memory\n", ((shm_mem_size/1024)/1024));
-	LM_INFO("using %ld Mb private memory per process\n",
-		((pkg_mem_size/1024)/1024));
+	LM_NOTICE("using %ld MB of shared memory, allocator: %s\n",
+	          shm_mem_size/1024/1024, mm_str(mem_allocator_shm));
+#if defined(PKG_MALLOC)
+	LM_NOTICE("using %ld MB of private process memory, allocator: %s\n",
+	          pkg_mem_size/1024/1024, mm_str(mem_allocator_pkg));
+#else
+	LM_NOTICE("using system memory for private process memory\n");
+#endif
 
 	/* init async reactor */
 	if (init_reactor_size()<0) {
@@ -1155,6 +842,15 @@ try_again:
 		goto error;
 	}
 
+	if (shm_memlog_size)
+		shm_mem_enable_dbg();
+
+	/* init IPC */
+	if (init_ipc()<0){
+		LM_CRIT("could not initialize IPC support, exiting...\n");
+		goto error;
+	}
+
 	/* init serial forking engine */
 	if (init_serialization()!=0) {
 		LM_ERR("failed to initialize serialization\n");
@@ -1188,49 +884,48 @@ try_again:
 		goto error;
 	}
 
+	/* init SQL DB support */
+	if (init_db_support() != 0) {
+		LM_ERR("failed to initialize SQL database support\n");
+		goto error;
+	}
+
+	/* init CacheDB support */
+	if (init_cdb_support() != 0) {
+		LM_ERR("failed to initialize CacheDB support\n");
+		goto error;
+	}
+
+	/* init pseudo-variable support */
+	if (init_pvar_support() != 0) {
+		LM_ERR("failed to init pvar support\n");
+		goto error;
+	}
+
 	/* init modules */
 	if (init_modules() != 0) {
 		LM_ERR("error while initializing modules\n");
 		goto error;
 	}
 
+	/* init xlog */
+	if (init_xlog() < 0) {
+		LM_ERR("error while initializing xlog!\n");
+		goto error;
+	}
+
 	/* register route timers */
 	if(register_route_timers() < 0) {
 		LM_ERR("Failed to register timer\n");
 		goto error;
 	}
 
-	/* check pv context list */
-	if(pv_contextlist_check() != 0) {
-		LM_ERR("used pv context that was not defined\n");
-		goto error;
-	}
-
-	/* init query list now in shm
-	 * so all processes that will be forked from now on
-	 * will have access to it
-	 *
-	 * if it fails, give it a try and carry on */
-	if (init_ql_support() != 0) {
-		LM_ERR("failed to initialise buffering query list\n");
-		query_buffer_size = 0;
-		*query_list = NULL;
-	}
-
 	/* init multi processes support */
 	if (init_multi_proc_support()!=0) {
 		LM_ERR("failed to init multi-proc support\n");
 		goto error;
 	}
 
-	#ifdef PKG_MALLOC
-	/* init stats support for pkg mem */
-	if (init_pkg_stats(counted_processes)!=0) {
-		LM_ERR("failed to init stats for pkg\n");
-		goto error;
-	}
-	#endif
-
 	/* init avps */
 	if (init_extra_avps() != 0) {
 		LM_ERR("error while initializing avps\n");
@@ -1248,14 +943,24 @@ try_again:
 		goto error;
 	}
 
+	if (init_log_event_cons() < 0) {
+		LM_ERR("Failed to initialize log event consumer\n");
+		goto error;
+	}
+
 	if (trans_init_all_listeners()<0) {
 		LM_ERR("failed to init all SIP listeners, aborting\n");
 		goto error;
 	}
 
+	if (init_script_reload()<0) {
+		LM_ERR("failed to init cfg reload ctx, aborting\n");
+		goto error;
+	}
+
 	/* all processes should have access to all the sockets (for sending)
 	 * so we open all first*/
-	if (do_suid(uid, gid)==-1)
+	if (do_suid(user_id, group_id)==-1)
 		goto error;
 
 	ret = main_loop();
@@ -1267,5 +972,6 @@ error:
 	cleanup(0);
 error00:
 	LM_NOTICE("Exiting....\n");
+	_ProfilerStop();
 	return ret;
 }
