@@ -60,9 +60,6 @@ static db_key_t  store_keys[6];
 static db_val_t  store_vals[6];
 static str      empty={"",0};
 
-extern unsigned buf_size;
-extern char *printbuf;
-
 void init_store_avps(str **db_columns)
 {
 	/* unique user id */
@@ -220,37 +217,6 @@ static int dbrow2avp(struct db_row *row, struct db_param *dbp, int attr,
 }
 
 
-inline static str* get_source_uri(struct sip_msg* msg,int source)
-{
-	/* which uri will be used? */
-	if (source&AVPOPS_USE_FROM)
-	{ /* from */
-		if (parse_from_header( msg )<0 )
-		{
-			LM_ERR("failed to parse from\n");
-			goto error;
-		}
-		return &(get_from(msg)->uri);
-	} else if (source&AVPOPS_USE_TO)
-	{  /* to */
-		if (parse_headers( msg, HDR_TO_F, 0)<0)
-		{
-			LM_ERR("failed to parse to\n");
-			goto error;
-		}
-		return &(get_to(msg)->uri);
-	} else if (source&AVPOPS_USE_RURI) {  /* RURI */
-		if(msg->new_uri.s!=NULL && msg->new_uri.len>0)
-			return &(msg->new_uri);
-		return &(msg->first_line.u.request.uri);
-	} else {
-		LM_ERR("unknown source <%d>\n", source);
-		goto error;
-	}
-error:
-	return 0;
-}
-
 static inline void int_str2db_val( int_str is_val, str *val, int is_s)
 {
 	if (is_s)
@@ -589,7 +555,6 @@ int ops_dbstore_avps (struct sip_msg* msg, struct fis_param *sp,
 	struct sip_uri   uri;
 	struct usr_avp   **avp_list;
 	struct usr_avp   *avp;
-	unsigned short   name_type;
 	int              avp_name;
 	int_str          i_s;
 	str              uuid;
@@ -600,7 +565,6 @@ int ops_dbstore_avps (struct sip_msg* msg, struct fis_param *sp,
 	str *sn;
 
 	s0 = s1 = s2 = NULL;
-	name_type = 0;
 	if (!((sp->opd&AVPOPS_VAL_PVAR)||(sp->opd&AVPOPS_VAL_STR))) {
 		LM_CRIT("invalid flag combination (%d/%d)\n", sp->opd, sp->ops);
 		goto error;
@@ -691,10 +655,8 @@ int ops_dbstore_avps (struct sip_msg* msg, struct fis_param *sp,
 			}
 			if(xvalue.flags&PV_TYPE_INT)
 			{
-				name_type = 0;
 				avp_name = xvalue.ri;
 			} else {
-				name_type = AVP_NAME_STR;
 				avp_name = -1;
 			}
 			if(xvalue.flags&PV_VAL_STR)
@@ -720,7 +682,6 @@ int ops_dbstore_avps (struct sip_msg* msg, struct fis_param *sp,
 				goto error;
 			}
 		} else {
-			name_type = dbp->a.u.sval.pvp.pvn.u.isname.type;
 			avp_name = dbp->a.u.sval.pvp.pvn.u.isname.name.n;
 		}
 	} else {
@@ -728,10 +689,6 @@ int ops_dbstore_avps (struct sip_msg* msg, struct fis_param *sp,
 		avp_name = -1;
 	}
 
-	/* set the script flags */
-	if(dbp->a.type==AVPOPS_VAL_PVAR)
-		name_type |= dbp->a.u.sval.pvp.pvn.u.isname.type&0xff00;
-
 	/* set uuid/(username and domain) fields */
 
 	n =0 ;
@@ -813,12 +770,10 @@ error:
 
 
 /* @return : non-zero */
-int ops_dbquery_avps(struct sip_msg* msg, pv_elem_t* query,
-									struct db_url *url, pvname_list_t* dest)
+int ops_dbquery_avps(struct sip_msg* msg, str* query,
+                     struct db_url *url, pvname_list_t* dest)
 {
-	int printbuf_len;
 	int ret;
-	str qstr;
 
 	if(msg==NULL || query==NULL)
 	{
@@ -826,18 +781,8 @@ int ops_dbquery_avps(struct sip_msg* msg, pv_elem_t* query,
 		return -1;
 	}
 
-	printbuf_len = buf_size-1;
-	if(pv_printf(msg, query, printbuf, &printbuf_len)<0 || printbuf_len<=0)
-	{
-		LM_ERR("cannot print the query\n");
-		return -1;
-	}
-
-	qstr.s = printbuf;
-	qstr.len = printbuf_len;
-
-	LM_DBG("query [%.*s]\n", printbuf_len, printbuf);
-	ret = db_query_avp(url, msg, &qstr, dest);
+	LM_DBG("query [%.*s]\n", query->len, query->s);
+	ret = db_query_avp(url, msg, query, dest);
 
 	//Empty return set
 	if(ret==1)
@@ -851,13 +796,11 @@ int ops_dbquery_avps(struct sip_msg* msg, pv_elem_t* query,
 	return 1;
 }
 
-int ops_async_dbquery(struct sip_msg* msg, async_resume_module **rfunc,
-		void **rparam,  pv_elem_t *query, struct db_url *url, pvname_list_t *dest)
+int ops_async_dbquery(struct sip_msg* msg, async_ctx *ctx,
+		str *query, struct db_url *url, pvname_list_t *dest)
 {
-	int printbuf_len;
 	int rc, read_fd;
 	query_async_param *param;
-	str qstr;
 
 	void *_priv;
 
@@ -867,37 +810,27 @@ int ops_async_dbquery(struct sip_msg* msg, async_resume_module **rfunc,
 		return -1;
 	}
 
-	printbuf_len = buf_size - 1;
-	if (pv_printf(msg, query, printbuf, &printbuf_len) < 0 || printbuf_len <= 0)
-	{
-		LM_ERR("cannot print the query\n");
-		return -1;
-	}
-
-	LM_DBG("query [%s]\n", printbuf);
-
-	qstr.s = printbuf;
-	qstr.len = printbuf_len;
+	LM_DBG("query [%.*s]\n", query->len, query->s);
 
 	/* No async capabilities - just run it in blocking mode */
 	if (!DB_CAPABILITY(url->dbf, DB_CAP_ASYNC_RAW_QUERY))
 	{
-		rc = db_query_avp(url, msg, &qstr, dest);
-		LM_DBG("sync query \"%.*s\" returned: %d\n", qstr.len, qstr.s, rc);
+		rc = db_query_avp(url, msg, query, dest);
+		LM_DBG("sync query \"%.*s\" returned: %d\n", query->len, query->s, rc);
 
-		*rparam = NULL;
-		*rfunc = NULL;
+		ctx->resume_param = NULL;
+		ctx->resume_f = NULL;
 		async_status = ASYNC_NO_IO;
 
 		/* Empty_set / Other_errors / Success */
 		return rc == 1 ? -2 : (rc != 0 ? -1 : 1);
 	}
 
-	read_fd = url->dbf.async_raw_query(url->hdl, &qstr, &_priv);
+	read_fd = url->dbf.async_raw_query(url->hdl, query, &_priv);
 	if (read_fd < 0)
 	{
-		*rparam = NULL;
-		*rfunc = NULL;
+		ctx->resume_param = NULL;
+		ctx->resume_f = NULL;
 		return -1;
 	}
 
@@ -909,8 +842,8 @@ int ops_async_dbquery(struct sip_msg* msg, async_resume_module **rfunc,
 	}
 	memset(param, '\0', sizeof *param);
 
-	*rparam = param;
-	*rfunc = resume_async_dbquery;
+	ctx->resume_param = param;
+	ctx->resume_f = resume_async_dbquery;
 
 	param->output_avps = dest;
 	param->hdl = url->hdl;
@@ -1094,6 +1027,49 @@ error:
 	return -1;
 }
 
+int ops_shuffle_avp( struct sip_msg* msg, struct fis_param* src)
+{
+	struct usr_avp *src_avp, *rnd_avp;
+	struct usr_avp *temp_avp = NULL;
+	int_str src_val, rnd_val;
+	unsigned short avp_type;
+	int avp_name;
+	int n, rnd_idx;
+
+	/* get avp src name */
+	if(avpops_get_aname(msg, src, &avp_name, &avp_type)!=0)
+	{
+		LM_ERR("failed to get src AVP name\n");
+		goto error;
+	}
+
+	/* count AVPs */
+	n = 0;
+	while ((temp_avp=search_first_avp(avp_type, avp_name, NULL, temp_avp)) != 0)
+		n++;
+
+	/* randomize AVPs */
+	for ( ; n>1; n-- ) {
+		rnd_idx = random() % n;
+		if (rnd_idx == (n-1))
+			continue;
+
+		LM_DBG("swapping [%d] <--> [%d]\n", (n-1), rnd_idx);
+
+		src_avp = search_index_avp(avp_type, avp_name, &src_val, (n-1));
+		rnd_avp = search_index_avp(avp_type, avp_name, &rnd_val, rnd_idx);
+
+		if ( replace_avp(avp_type|(rnd_avp->flags&AVP_VAL_STR), avp_name, rnd_val, (n-1))==-1 ||
+					replace_avp(avp_type|(src_avp->flags&AVP_VAL_STR), avp_name, src_val, rnd_idx)==-1 ) {
+			LM_ERR("failed to swap avp\n");
+			goto error;
+		}
+	}
+
+	return 1;
+error:
+	return -1;
+}
 
 #define STR_BUF_SIZE  1024
 static char str_buf[STR_BUF_SIZE];
@@ -1113,7 +1089,6 @@ inline static int append_0(str *in, str *out)
 int ops_pushto_avp (struct sip_msg* msg, struct fis_param* dst,
 													struct fis_param* src)
 {
-	struct action  act;
 	struct usr_avp *avp;
 	unsigned short name_type;
 	int_str        avp_val;
@@ -1165,16 +1140,16 @@ int ops_pushto_avp (struct sip_msg* msg, struct fis_param* dst,
 			val.s = int2str((unsigned long)avp_val.n, &val.len);
 		}
 
-		act_type = 0;
+		act_type = -1;
 		/* push the value into right position */
 		if (dst->opd&AVPOPS_USE_RURI)
 		{
 			if (dst->opd&AVPOPS_FLAG_USER0)
-				act_type = SET_USER_T;
+				act_type = RW_RURI_USER;
 			else if (dst->opd&AVPOPS_FLAG_DOMAIN0)
-				act_type = SET_HOST_T;
+				act_type = RW_RURI_HOST;
 			else
-				act_type = SET_URI_T;
+				act_type = 0; /* entire RURI */
 			if ( flags&AVP_VAL_STR && append_0( &val, &val)!=0 ) {
 				LM_ERR("failed to make 0 term.\n");
 				goto error;
@@ -1192,7 +1167,7 @@ int ops_pushto_avp (struct sip_msg* msg, struct fis_param* dst,
 			goto error;
 		}
 
-		if ( act_type )
+		if ( act_type != -1 )
 		{
 			/* rewrite part of ruri */
 			if (n)
@@ -1205,14 +1180,16 @@ int ops_pushto_avp (struct sip_msg* msg, struct fis_param* dst,
 					goto error;
 				}
 			}
-			memset(&act, 0, sizeof(act));
-			act.elem[0].type = STR_ST;
-			act.elem[0].u.s = val;
-			act.type = act_type;
-			if (do_action(&act, msg)<0)
-			{
-				LM_ERR("SET_XXXX_T action failed\n");
-				goto error;
+			if (act_type == 0) {
+				if (set_ruri(msg, &val) < 0) {
+					LM_ERR("Failed to set RURI\n");
+					goto error;
+				}
+			} else {
+				if (rewrite_ruri(msg, &val, 0, act_type) < 0) {
+					LM_ERR("Failed to set user or host\n");
+					goto error;
+				}
 			}
 		} else if (dst->opd&AVPOPS_USE_DURI) {
 			if(set_dst_uri(msg, &val)!=0)
@@ -1519,22 +1496,24 @@ int ops_print_avp(void)
 	avp_list = get_avp_list();
 	avp = *avp_list;
 
+	LM_INFO("----------- All AVPs in this context --------\n");
+	LM_INFO("  (SIP txn, script event, timer route, etc.)\n");
 	for ( ; avp ; avp=avp->next)
 	{
 		LM_INFO("p=%p, flags=0x%04X\n",avp, avp->flags);
 		name = get_avp_name(avp);
-		LM_INFO("\t\t\tname=<%.*s>\n",name->len,name->s);
-		LM_INFO("\t\t\tid=<%d>\n",avp->id);
+		LM_INFO("    name=<%.*s>\n",name->len,name->s);
+		LM_INFO("    id=<%d>\n",avp->id);
 		get_avp_val( avp, &val);
 		if (avp->flags&AVP_VAL_STR)
 		{
-			LM_INFO("\t\t\tval_str=<%.*s / %d>\n",val.s.len,val.s.s,
+			LM_INFO("    val_str=<%.*s / %d>\n",val.s.len,val.s.s,
 					val.s.len);
 		} else {
-			LM_INFO("\t\t\tval_int=<%d>\n",val.n);
+			LM_INFO("    val_int=<%d>\n",val.n);
 		}
 	}
-
+	LM_INFO("---------------- END ALL AVPs ---------------\n");
 
 	return 1;
 }
@@ -1899,101 +1878,3 @@ int ops_is_avp_set(struct sip_msg* msg, struct fis_param *ap)
 
 	return -1;
 }
-
-int w_insert_avp(struct sip_msg* msg, char* name, char* value,
-		char *index_char)
-{
-	int              index = *(int*)index_char;
-	int              avp_name;
-	struct usr_avp   *avp= NULL, *prev_avp= NULL;
-	struct usr_avp   *avp_new;
-	unsigned short   name_type;
-	pv_value_t       xvalue;
-	int              flags = 0;
-	int_str          avp_val;
-	pv_elem_t*       pv_dest = (pv_elem_t*)name;
-	pv_elem_t*       pv_src = (pv_elem_t*)value;
-
-	/* get avp name */
-	if(pv_get_avp_name(msg, &pv_dest->spec.pvp, &avp_name, &name_type)< 0)
-	{
-		LM_ERR("failed to get src AVP name\n");
-		return -1;
-	}
-
-	/* get value to be inserted */
-	avp = NULL;
-	flags = 0;
-	if(pv_src->spec.type == PVT_NONE)
-	{
-		avp_val.s = pv_src->text;
-		flags = AVP_VAL_STR;
-	}
-	else
-	{
-		if(pv_get_spec_value(msg, &(pv_src->spec), &xvalue)!=0)
-		{
-			LM_ERR("cannot get src value\n");
-			return -1;
-		}
-		if(xvalue.flags&PV_TYPE_INT)
-		{
-			avp_val.n = xvalue.ri;
-		} else {
-			flags = AVP_VAL_STR;
-			avp_val.s = xvalue.rs;
-		}
-	}
-	name_type |= flags;
-	/* insert it at the right place */
-	if(index == 0)
-	{
-		if(add_avp(name_type, avp_name, avp_val) < 0)
-		{
-			LM_ERR("Failed to add new avp\n");
-			return -1;
-		}
-		return 1;
-	}
-
-	/* search the previous avp */
-	index--;
-	avp = NULL;
-	while ( (avp=search_first_avp( name_type, avp_name, 0, avp))!=0 )
-	{
-		if( index == 0 )
-		{
-			break;
-		}
-		index--;
-		prev_avp = avp;
-	}
-
-	/* if the index is greater then the count */
-	if(avp == NULL)
-	{
-		if(prev_avp == NULL)
-		{
-			if(add_avp(name_type, avp_name, avp_val) < 0)
-			{
-				LM_ERR("Failed to add new avp\n");
-				return -1;
-			}
-			return 1;
-		}
-		avp = prev_avp;
-	}
-
-	/* if a previous record was found -> insert the new avp after it */
-	avp_new = new_avp(name_type, avp_name, avp_val);
-	if(avp_new == NULL)
-	{
-		LM_ERR("Failed to allocate new avp structure\n");
-		return -1;
-	}
-	LM_DBG("am alocat un avp nou\n");
-	avp_new->next = avp->next;
-	avp->next = avp_new;
-
-	return 1;
-}
