@@ -48,9 +48,10 @@ extern str app_state[];
 static int mod_init(void);
 static void mod_destroy(void);
 static int child_init(int rank);
-int sca_init_request(struct sip_msg* msg);
-int sca_bridge_request(struct sip_msg* msg, str* arg1, str* arg2);
-static struct mi_root* mi_sca_list(struct mi_root* cmd, void* param);
+int sca_init_request(struct sip_msg* msg, int *shared_entity);
+int sca_bridge_request(struct sip_msg* msg, str* arg1);
+mi_response_t *mi_sca_list(const mi_params_t *params,
+								struct mi_handler *async_hdl);
 
 /** Global variables */
 struct tm_binds tmb;
@@ -83,15 +84,19 @@ str presence_server = {NULL, 0};
 
 
 /** Exported functions */
-static cmd_export_t cmds[]=
+static const cmd_export_t cmds[]=
 {
-	{"sca_init_request"  ,(cmd_function)sca_init_request  ,1,fixup_pvar_pvar,0,REQUEST_ROUTE},
-	{"sca_bridge_request",(cmd_function)sca_bridge_request,1,fixup_pvar_pvar,0,REQUEST_ROUTE},
-	{ 0,                 0,                              0 , 0 , 0,  0}
+	{"sca_init_request"  ,(cmd_function)sca_init_request, {
+		{CMD_PARAM_INT,0,0}, {0,0,0}},
+		REQUEST_ROUTE},
+	{"sca_bridge_request",(cmd_function)sca_bridge_request, {
+		{CMD_PARAM_STR,0,0}, {0,0,0}},
+		REQUEST_ROUTE},
+	{0,0,{{0,0,0}},0}
 };
 
 /** Exported parameters */
-static param_export_t params[]=
+static const param_export_t params[]=
 {
 	{"hash_size",				INT_PARAM,&b2b_sca_hsize			},
 	{"presence_server",			STR_PARAM,&presence_server.s			},
@@ -157,12 +162,15 @@ static param_export_t params[]=
 };
 
 /** MI commands */
-static mi_export_t mi_cmds[] = {
-	{ "sca_list",   0, mi_sca_list, 0,  0,  0},
-	{ 0,            0,           0, 0,  0,  0}
+static const mi_export_t mi_cmds[] = {
+	{ "sca_list", 0,0,0,{
+		{mi_sca_list, {0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{EMPTY_MI_EXPORT}
 };
 
-static dep_export_t deps = {
+static const dep_export_t deps = {
 	{ /* OpenSIPS module dependencies */
 		{ MOD_TYPE_DEFAULT, "pua",       DEP_ABORT },
 		{ MOD_TYPE_DEFAULT, "b2b_logic", DEP_ABORT },
@@ -180,6 +188,7 @@ struct module_exports exports= {
         MOD_TYPE_DEFAULT,               /* class of this module */
         MODULE_VERSION,                 /* module version */
         DEFAULT_DLFLAGS,                /* dlopen flags */
+        0,				                /* load function */
         &deps,                          /* OpenSIPS module dependencies */
         cmds,                           /* exported functions */
         0,                              /* exported async functions */
@@ -187,11 +196,14 @@ struct module_exports exports= {
         0,                              /* exported statistics */
         mi_cmds,                        /* exported MI functions */
         0,                              /* exported pseudo-variables */
+		0,								/* exported transformations */
         0,                              /* extra processes */
+        0,                              /* module pre-initialization function */
         mod_init,                       /* module initialization function */
         (response_function) 0,          /* response handling function */
         (destroy_function) mod_destroy, /* destroy function */
-        child_init                      /* per-child init function */
+        child_init,                     /* per-child init function */
+		0                               /* reload confirm function */
 };
 
 /** Module init function */
@@ -343,7 +355,7 @@ static int mod_init(void)
 
 static int child_init(int rank)
 {
-	if (sca_db_mode==DB_MODE_REALTIME &&  (rank>PROC_MAIN || rank==PROC_MODULE)) {
+	if (sca_db_mode==DB_MODE_REALTIME &&  (rank>=1 || rank==PROC_MODULE)) {
 		if (connect_sca_db(&db_url)) {
 			LM_ERR("failed to connect to database (rank=%d)\n",rank);
 			return -1;
@@ -356,9 +368,6 @@ static int child_init(int rank)
 
 static void mod_destroy(void)
 {
-	//if (sca_db_mode != DB_MODE_NONE) {
-	//	sca_update_db();
-	//}
 	destroy_b2b_sca_handlers();
 	destroy_b2b_sca_htable();
 
@@ -479,55 +488,72 @@ struct to_body* get_appearance_name_addr(struct sip_msg* msg)
 	return msg->from->parsed;
 }
 
-
-static struct mi_root* mi_sca_list(struct mi_root* cmd, void* param)
+mi_response_t *mi_sca_list(const mi_params_t *params,
+								struct mi_handler *async_hdl)
 {
 	int i, index;
-	struct mi_root *rpl_tree;
-	struct mi_node *node=NULL, *node1=NULL, *rpl=NULL;
-	struct mi_attr* attr;
 	b2b_sca_record_t *rec;
 	b2b_sca_call_t *call;
 	str_lst_t *watcher;
+	mi_response_t *resp;
+	mi_item_t *resp_arr;
+	mi_item_t *resp_item, *watchers_arr, *apps_arr, *app_item;
 
-	rpl_tree = init_mi_tree( 200, MI_OK_S, MI_OK_LEN);
-	if (rpl_tree==NULL) return NULL;
-	rpl = &rpl_tree->node;
-	rpl->flags |= MI_IS_ARRAY;
+	resp = init_mi_result_array(&resp_arr);
+	if (!resp)
+		return 0;
 
 	for(index = 0; index<b2b_sca_hsize; index++) {
 		lock_get(&b2b_sca_htable[index].lock);
 		rec = b2b_sca_htable[index].first;
 		while(rec) {
-			node = add_mi_node_child(rpl, MI_IS_ARRAY|MI_DUP_VALUE,
-				"shared_line", 11,rec->shared_line.s, rec->shared_line.len);
-			if(node == NULL) goto error;
+			resp_item = add_mi_object(resp_arr, NULL, 0);
+			if (!resp_item)
+				goto error;
+
+			if (add_mi_string(resp_item, MI_SSTR("shared_line"),
+				rec->shared_line.s, rec->shared_line.len) < 0)
+				goto error;
+
+			watchers_arr = add_mi_array(resp_item, MI_SSTR("watchers"));
+			if (!watchers_arr)
+				goto error;
 			watcher = rec->watchers;
 			while (watcher) {
-				attr = add_mi_attr(node, MI_DUP_VALUE, "watcher", 7,
-					watcher->watcher.s, watcher->watcher.len);
-				if(attr == NULL) goto error;
+				if (add_mi_string(watchers_arr, 0, 0,
+					watcher->watcher.s, watcher->watcher.len) < 0)
+					goto error;
 				watcher = watcher->next;
 			}
+
+			apps_arr = add_mi_array(resp_item, MI_SSTR("appearances"));
+			if (!apps_arr)
+				goto error;
 			for (i=0; i<MAX_APPEARANCE_INDEX; i++) {
 				if (rec->call[i]) {
 					call = rec->call[i];
-					node1 = add_mi_node_child(node, MI_DUP_VALUE,
-							"appearance", 10,
-							call->appearance_index_str.s,
-							call->appearance_index_str.len);
-					if(node1 == NULL) goto error;
-					attr = add_mi_attr(node1, MI_DUP_VALUE, "state", 5,
-							app_state[call->call_state].s,
-							app_state[call->call_state].len);
-					if(attr == NULL) goto error;
-					attr = add_mi_attr(node1, MI_DUP_VALUE, "b2b_key", 7,
-							call->b2bl_key.s, call->b2bl_key.len);
-					if(attr == NULL) goto error;
-					attr = add_mi_attr(node1, MI_DUP_VALUE, "app_uri", 7,
-							call->call_info_apperance_uri.s,
-							call->call_info_apperance_uri.len);
-					if(attr == NULL) goto error;
+					app_item = add_mi_object(apps_arr, NULL, 0);
+					if (!app_item)
+						goto error;
+
+					if (add_mi_string(app_item, MI_SSTR("index"),
+						call->appearance_index_str.s,
+						call->appearance_index_str.len) < 0)
+						goto error;
+
+					if (add_mi_string(app_item, MI_SSTR("state"),
+						app_state[call->call_state].s,
+						app_state[call->call_state].len) < 0)
+						goto error;
+
+					if (add_mi_string(app_item, MI_SSTR("b2b_key"),
+						call->b2bl_key.s, call->b2bl_key.len) < 0)
+						goto error;
+
+					if (add_mi_string(app_item, MI_SSTR("app_uri"),
+						call->call_info_apperance_uri.s,
+						call->call_info_apperance_uri.len) < 0)
+						goto error;
 				}
 			}
 
@@ -535,11 +561,13 @@ static struct mi_root* mi_sca_list(struct mi_root* cmd, void* param)
 		}
 		lock_release(&b2b_sca_htable[index].lock);
 	}
-	return rpl_tree;
+
+	return resp;
+
 error:
 	lock_release(&b2b_sca_htable[index].lock);
 	LM_ERR("Unable to create reply\n");
-	free_mi_tree(rpl_tree);
+	free_mi_response(resp);
 	return NULL;
 }
 
