@@ -683,7 +683,6 @@ static inline char *run_time_switch( struct cpl_interpreter *intr )
 	int nr_attrs;
 	int i,j;
 	str user_tz = {0,0};
-	ac_tm_t att;
 	tmrec_p trt = 0;
 
 	LM_DBG("checking recv. time stamp <%d>\n",
@@ -735,8 +734,7 @@ static inline char *run_time_switch( struct cpl_interpreter *intr )
 					LM_ERR("no more pkg error\n");
 					goto script_error;
 				}
-				if(ac_tm_set_time( &att, intr->recv_time))
-					goto runtime_error;
+
 				/* let's see how many attributes we have */
 				nr_attrs = NR_OF_ATTR(kid);
 				/* get the attributes */
@@ -813,7 +811,7 @@ static inline char *run_time_switch( struct cpl_interpreter *intr )
 					goto script_error;
 				}
 				/* does the recv_time match the specified interval?  */
-				j = check_tmrec( trt, &att, 0);
+				j = _tmrec_check(trt, intr->recv_time);
 				/* restore the orig TZ */
 				if ( flags&(1<<7) )
 					set_TZ(cpl_env.orig_tz.s);
@@ -912,18 +910,19 @@ inline static int is_lang_tag_matching(str *range,str *cpl_tag,str *cpl_subtag)
 			/* do compare */
 			LM_DBG("testing range [%.*s]-[%.*s] against tag [%.*s]-[%.*s]\n",
 				tag.len,tag.s,subtag.len,subtag.s,
-				cpl_tag->len,cpl_tag->s,cpl_subtag->len,cpl_subtag->s);
+				cpl_tag->len,(cpl_tag->s?cpl_tag->s:""),
+				cpl_subtag->len,(cpl_subtag->s?cpl_subtag->s:""));
 			/* language range of "*" is ignored for the purpose of matching*/
 			if ( !(tag.len==1 && *tag.s=='*') ) {
 				/* does the language tag matches ? */
-				if (tag.len==cpl_tag->len && !strncasecmp(tag.s,cpl_tag->s,
-				tag.len)) {
+				if (cpl_tag->len && cpl_tag->s && tag.len==cpl_tag->len &&
+						!strncasecmp(tag.s,cpl_tag->s, tag.len)) {
 					/* if the subtag of the range is void -> matche */
 					if (subtag.len==0)
 						return 1;
 					/* the subtags equals -> matche */
-					if (subtag.len==cpl_subtag->len &&
-					!strncasecmp(subtag.s,cpl_subtag->s,subtag.len) )
+					if (cpl_subtag->len && cpl_subtag->s && subtag.len==cpl_subtag->len &&
+							!strncasecmp(subtag.s,cpl_subtag->s,subtag.len) )
 						return 1;
 				}
 			}
