@@ -32,10 +32,9 @@
 
 #include <stdio.h>
 #include <string.h>
-#include <mysql/mysql.h>
-#include <mysql/errmsg.h>
-#include <mysql/mysqld_error.h>
-#include <mysql/mysql_version.h>
+#include <mysql.h>
+#include <errmsg.h>
+#include <mysqld_error.h>
 #include "../../mem/mem.h"
 #include "../../dprint.h"
 #include "../../db/db_query.h"
@@ -128,6 +127,11 @@ static inline int wrapper_single_mysql_stmt_prepare(const db_con_t *conn,
 		case CR_SERVER_GONE_ERROR:
 		case CR_SERVER_LOST:
 		case CR_COMMANDS_OUT_OF_SYNC:
+		/* UGLY magic number here - unfortunately MySQL & MariaDB
+		 * are diverging. MySQL has this as ER_CLIENT_INTERACTION_TIMEOUT
+		 * while MariaDB has it as ER_REFERENCED_TRG_DOES_NOT_EXIST
+		 */
+		case 4031:
 			return -1; /* reconnection error -> <0 */
 		default:
 			LM_CRIT("driver error (%i): %s\n",
@@ -155,6 +159,11 @@ static inline int wrapper_single_mysql_stmt_execute(const db_con_t *conn,
 		case CR_SERVER_GONE_ERROR:
 		case CR_SERVER_LOST:
 		case CR_COMMANDS_OUT_OF_SYNC:
+		/* UGLY magic number here - unfortunately MySQL & MariaDB
+		 * are diverging. MySQL has this as ER_CLIENT_INTERACTION_TIMEOUT
+		 * while MariaDB has it as ER_REFERENCED_TRG_DOES_NOT_EXIST
+		 */
+		case 4031:
 			return -1; /* reconnection error -> <0 */
 		default:
 			LM_CRIT("driver error (%i): %s\n", error, mysql_stmt_error(stmt));
@@ -181,6 +190,15 @@ static inline int wrapper_single_mysql_real_query(const db_con_t *conn,
 		case CR_SERVER_GONE_ERROR:
 		case CR_SERVER_LOST:
 		case CR_COMMANDS_OUT_OF_SYNC:
+		/* UGLY magic number here - unfortunately MySQL & MariaDB
+		 * are diverging. MySQL has this as ER_CLIENT_INTERACTION_TIMEOUT
+		 * while MariaDB has it as ER_REFERENCED_TRG_DOES_NOT_EXIST
+		 */
+		case 4031:
+			return -1; /* reconnection error -> <0 */
+		case ER_LOCK_DEADLOCK:
+			LM_WARN("server error (%i): %s\n", error,
+				mysql_error(CON_CONNECTION(conn)));
 			return -1; /* reconnection error -> <0 */
 		default:
 			LM_CRIT("driver error (%i): %s\n", error,
@@ -207,6 +225,11 @@ static inline int wrapper_single_mysql_send_query(const db_con_t *conn,
 		case CR_SERVER_GONE_ERROR:
 		case CR_SERVER_LOST:
 		case CR_COMMANDS_OUT_OF_SYNC:
+		/* UGLY magic number here - unfortunately MySQL & MariaDB
+		 * are diverging. MySQL has this as ER_CLIENT_INTERACTION_TIMEOUT
+		 * while MariaDB has it as ER_REFERENCED_TRG_DOES_NOT_EXIST
+		 */
+		case 4031:
 			return -1; /* reconnection error -> <0 */
 		default:
 			LM_CRIT("driver error (%i): %s\n", error,
@@ -242,7 +265,7 @@ static inline void reset_all_statements(const db_con_t* conn)
 	struct prep_stmt *pq_ptr;
 	struct my_stmt_ctx *ctx;
 
-	LM_INFO("reseting all statements on connection: (%p) %p\n",
+	LM_INFO("resetting all statements on connection: (%p) %p\n",
 		conn,(void*)conn->tail);
 	for( pq_ptr=CON_PS_LIST(conn); pq_ptr ; pq_ptr=pq_ptr->next ) {
 		for (ctx = pq_ptr->stmts ; ctx ; ctx=ctx->next ) {
@@ -293,7 +316,8 @@ static int db_mysql_submit_query(const db_con_t* _h, const str* _s)
 	for (i=0; i<max_db_queries; i++) {
 		start_expire_timer(start,db_mysql_exec_query_threshold);
 		code = wrapper_single_mysql_real_query(_h, _s);
-		stop_expire_timer(start,db_mysql_exec_query_threshold,"mysql query",_s->s,_s->len,0);
+		_stop_expire_timer(start, db_mysql_exec_query_threshold, "mysql query",
+		            _s->s, _s->len, 0, sql_slow_queries, sql_total_queries);
 		if (code < 0) {
 			/* got disconnected during call */
 			switch_state_to_disconnected(_h);
@@ -599,7 +623,8 @@ static int db_mysql_do_prepared_query(const db_con_t* conn, const str *query,
 		/* get a new context */
 		ctx = get_new_stmt_ctx(conn, query);
 		if (ctx==NULL) {
-			LM_ERR("failed to create new context\n");
+			LM_ERR("failed to create new context for query=|%.*s|\n",
+					query->len, query->s);
 			pkg_free(pq_ptr);
 			return -1;
 		}
@@ -624,7 +649,8 @@ static int db_mysql_do_prepared_query(const db_con_t* conn, const str *query,
 			/* get a new context */
 			ctx = get_new_stmt_ctx(conn, query);
 			if (ctx==NULL) {
-				LM_ERR("failed to create new context\n");
+				LM_ERR("failed to create new context for query=|%.*s|\n",
+						query->len, query->s);
 				return -1;
 			}
 			/* link it */
@@ -648,7 +674,7 @@ static int db_mysql_do_prepared_query(const db_con_t* conn, const str *query,
 			return -1;
 		}
 
-		/* if add succesfull but not yet insertion time
+		/* if add successful but not yet insertion time
 		   return success and wait for number of query rows
 		   to pile up
 		*/
@@ -715,7 +741,8 @@ static int db_mysql_do_prepared_query(const db_con_t* conn, const str *query,
 
 		start_expire_timer(start,db_mysql_exec_query_threshold);
 		code = wrapper_single_mysql_stmt_execute(conn, ctx->stmt);
-		stop_expire_timer(start,db_mysql_exec_query_threshold,"mysql prep stmt",query->s,query->len,0);
+		_stop_expire_timer(start, db_mysql_exec_query_threshold, "mysql prep stmt",
+		        query->s, query->len, 0, sql_slow_queries, sql_total_queries);
 		if (code < 0) {
 			/* got disconnected during call */
 			switch_state_to_disconnected(conn);
@@ -753,9 +780,12 @@ static int db_mysql_do_prepared_query(const db_con_t* conn, const str *query,
 		LM_DBG("prepared statement has %d columns in result\n",cols);
 		/* set the out bind array ? */
 		if (pq_ptr->cols_out==-1) {
+			char *col_bufs;
+
 			pq_ptr->cols_out = cols;
 			pq_ptr->bind_out = (MYSQL_BIND*)pkg_malloc
-				( cols*(sizeof(struct bind_ocontent) + sizeof(MYSQL_BIND)) );
+				( cols*(sizeof(struct bind_ocontent) + sizeof(MYSQL_BIND)
+			            + ps_max_col_size) );
 			if (pq_ptr->bind_out==NULL) {
 				db_mysql_free_pq(pq_ptr);
 				CON_CURR_PS(conn) = NULL;
@@ -763,15 +793,19 @@ static int db_mysql_do_prepared_query(const db_con_t* conn, const str *query,
 				return -1;
 			}
 			memset(pq_ptr->bind_out, 0 ,
-				cols*(sizeof(struct bind_ocontent) + sizeof(MYSQL_BIND)));
+				cols*(sizeof(struct bind_ocontent) + sizeof(MYSQL_BIND)
+			          + ps_max_col_size));
 
 			pq_ptr->out_bufs = (struct bind_ocontent*)(pq_ptr->bind_out+cols);
+			col_bufs = (char*)(pq_ptr->out_bufs+cols);
+
 			mysql_bind = pq_ptr->bind_out;
 			/* prepare the pointers */
-			for( i=0 ; i<pq_ptr->cols_out ; i++ ) {
-				mysql_bind[i].buffer =  pq_ptr->out_bufs[i].buf;
+			for( i=0 ; i<cols ; i++ ) {
+				mysql_bind[i].buffer = pq_ptr->out_bufs[i].buf
+				                     = col_bufs + i*ps_max_col_size;
 				mysql_bind[i].buffer_type = MYSQL_TYPE_STRING;
-				mysql_bind[i].buffer_length = PREP_STMT_VAL_LEN;
+				mysql_bind[i].buffer_length = ps_max_col_size;
 				mysql_bind[i].length = &pq_ptr->out_bufs[i].len;
 				mysql_bind[i].is_null = &pq_ptr->out_bufs[i].null;
 #if (MYSQL_VERSION_ID >= 50030)
@@ -863,16 +897,18 @@ static int db_mysql_store_result(const db_con_t* _h, db_res_t** _r)
 	if (!CON_HAS_PS(_h))
 		CON_RESULT(_h) = mysql_store_result(CON_CONNECTION(_h));
 	if (!CON_RESULT(_h)) {
-		if (mysql_field_count(CON_CONNECTION(_h)) == 0) {
-			(*_r)->col.n = 0;
-			(*_r)->n = 0;
-			goto done;
-		} else {
+		if (mysql_errno(CON_CONNECTION(_h)) > 0) {
 			LM_ERR("driver error: %s\n", mysql_error(CON_CONNECTION(_h)));
 			db_free_result(*_r);
 			*_r = 0;
 			return -3;
 		}
+
+		if (mysql_field_count(CON_CONNECTION(_h)) == 0) {
+			(*_r)->col.n = 0;
+			(*_r)->n = 0;
+			goto done;
+		}
 	}
 
 	if (db_mysql_convert_result(_h, *_r) < 0) {
@@ -1009,17 +1045,20 @@ int db_mysql_fetch_result(const db_con_t* _h, db_res_t** _r, const int nrows)
 		if (!CON_HAS_PS(_h))
 			CON_RESULT(_h) = mysql_store_result(CON_CONNECTION(_h));
 		if (!CON_RESULT(_h)) {
-			if (mysql_field_count(CON_CONNECTION(_h)) == 0) {
-				(*_r)->col.n = 0;
-				(*_r)->n = 0;
-				return 0;
-			} else {
+			if (mysql_errno(CON_CONNECTION(_h)) > 0) {
 				LM_ERR("driver error: %s\n", mysql_error(CON_CONNECTION(_h)));
 				db_free_result(*_r);
 				*_r = 0;
 				return -3;
 			}
+
+			if (mysql_field_count(CON_CONNECTION(_h)) == 0) {
+				(*_r)->col.n = 0;
+				(*_r)->n = 0;
+				return 0;
+			}
 		}
+
 		if (db_mysql_get_columns(_h, *_r) < 0) {
 			LM_ERR("error while getting column names\n");
 			return -4;
@@ -1158,8 +1197,10 @@ int db_mysql_async_raw_query(db_con_t *_h, const str *_s, void **_priv)
 		} else {
 			code = wrapper_single_mysql_real_query(_h, _s);
 		}
-		stop_expire_timer(start, db_mysql_exec_query_threshold,
-						  "mysql async query", _s->s, _s->len, 0);
+		_stop_expire_timer(start, db_mysql_exec_query_threshold,
+				"mysql async query", _s->s, _s->len, 0,
+				sql_slow_queries, sql_total_queries);
+
 		if (code < 0) {
 			/* got disconnected during call */
 			switch_state_to_disconnected(_h);
@@ -1312,7 +1353,10 @@ int db_mysql_delete(const db_con_t* _h, const db_key_t* _k, const db_op_t* _o,
 		if (CON_HAS_UNINIT_PS(_h)||!has_stmt_ctx(_h,&(CON_MYSQL_PS(_h)->ctx))){
 			ret = db_do_delete(_h, _k, _o, _v, _n, db_mysql_val2str,
 				db_mysql_submit_dummy_query);
-			if (ret!=0) {CON_RESET_CURR_PS(_h);return ret;}
+			if (ret != 0) {
+				CON_RESET_CURR_PS(_h);
+				return ret;
+			}
 		}
 		ret = db_mysql_do_prepared_query(_h, &query_holder, _v, _n, NULL, 0);
 		CON_RESET_CURR_PS(_h);
@@ -1345,7 +1389,10 @@ int db_mysql_update(const db_con_t* _h, const db_key_t* _k, const db_op_t* _o,
 		if (CON_HAS_UNINIT_PS(_h)||!has_stmt_ctx(_h,&(CON_MYSQL_PS(_h)->ctx))){
 			ret = db_do_update(_h, _k, _o, _v, _uk, _uv, _n, _un,
 				db_mysql_val2str, db_mysql_submit_dummy_query);
-			if (ret!=0) {CON_RESET_CURR_PS(_h);return ret;}
+			if (ret != 0) {
+				CON_RESET_CURR_PS(_h);
+				return ret;
+			}
 		}
 		ret = db_mysql_do_prepared_query(_h, &query_holder, _uv, _un, _v, _n);
 		CON_RESET_CURR_PS(_h);
@@ -1372,7 +1419,10 @@ int db_mysql_replace(const db_con_t* _h, const db_key_t* _k, const db_val_t* _v,
 		if (CON_HAS_UNINIT_PS(_h)||!has_stmt_ctx(_h,&(CON_MYSQL_PS(_h)->ctx))){
 			ret = db_do_replace(_h, _k, _v, _n, db_mysql_val2str,
 				db_mysql_submit_dummy_query);
-			if (ret!=0) {CON_RESET_CURR_PS(_h);return ret;}
+			if (ret != 0) {
+				CON_RESET_CURR_PS(_h);
+				return ret;
+			}
 		}
 		ret = db_mysql_do_prepared_query(_h, &query_holder, _v, _n, NULL, 0);
 		CON_RESET_CURR_PS(_h);
