@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2009-2014 OpenSIPS Solutions
+ * Copyright (C) 2009-2020 OpenSIPS Solutions
  * Copyright (C) 2006-2009 Voice System SRL
  *
  * This file is part of opensips, a free SIP server.
@@ -16,23 +16,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
- *
- * History:
- * --------
- * 2006-04-14  initial version (bogdan)
- * 2006-11-28  Added num_100s and num_200s to dlg_cell, to aid in adding
- *             statistics tracking of the number of early, and active dialogs.
- *             (Jeffrey Magder - SOMA Networks)
- * 2007-03-06  syncronized state machine added for dialog state. New tranzition
- *             design based on events; removed num_1xx and num_2xx (bogdan)
- * 2007-07-06  added flags, cseq, contact, route_set and bind_addr
- *             to struct dlg_cell in order to store these information into db
- *             (ancuta)
- * 2008-04-17  added new dialog flag to avoid state tranzitions from DELETED to
- *             CONFIRMED_NA due delayed "200 OK" (bogdan)
- * 2009-09-09  support for early dialogs added; proper handling of cseq
- *             while PRACK is used (bogdan)
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 
@@ -42,6 +26,8 @@
 #include "../../locking.h"
 #include "../../context.h"
 #include "../../mi/mi.h"
+#include "../../lib/dbg/struct_hist.h"
+
 #include "dlg_timer.h"
 #include "dlg_cb.h"
 #include "dlg_vals.h"
@@ -62,19 +48,33 @@
 #define DLG_EVENT_REQBYE       7
 #define DLG_EVENT_REQ          8
 
-#define DLG_FLAG_NEW			(1<<0)
-#define DLG_FLAG_CHANGED		(1<<1)
-#define DLG_FLAG_HASBYE			(1<<2)
-#define DLG_FLAG_BYEONTIMEOUT		(1<<3)
-#define DLG_FLAG_ISINIT			(1<<4)
-#define DLG_FLAG_PING_CALLER		(1<<5)
-#define DLG_FLAG_PING_CALLEE		(1<<6)
-#define DLG_FLAG_TOPHIDING		(1<<7)
-#define DLG_FLAG_VP_CHANGED		(1<<8)
-#define DLG_FLAG_DB_DELETED		(1<<9)
-#define DLG_FLAG_CSEQ_ENFORCE		(1<<10)
-#define DLG_FLAG_REINVITE_PING_CALLER	(1<<11)
-#define DLG_FLAG_REINVITE_PING_CALLEE	(1<<12)
+#define DLG_FLAG_NEW				(1<<0)
+#define DLG_FLAG_CHANGED			(1<<1)
+#define DLG_FLAG_HASBYE				(1<<2)
+#define DLG_FLAG_BYEONTIMEOUT			(1<<3)
+#define DLG_FLAG_ISINIT				(1<<4)
+#define DLG_FLAG_PING_CALLER			(1<<5)
+#define DLG_FLAG_PING_CALLEE			(1<<6)
+#define DLG_FLAG_FROM_DB			(1<<7)
+#define DLG_FLAG_VP_CHANGED			(1<<8)
+#define DLG_FLAG_DB_DELETED			(1<<9)
+#define DLG_FLAG_CSEQ_ENFORCE			(1<<10)
+#define DLG_FLAG_REINVITE_PING_CALLER		(1<<11)
+#define DLG_FLAG_REINVITE_PING_CALLEE		(1<<12)
+#define DLG_FLAG_REINVITE_PING_ENGAGED_REQ	(1<<13)
+#define DLG_FLAG_REINVITE_PING_ENGAGED_REPL	(1<<14)
+#define DLG_FLAG_END_ON_RACE_CONDITION		(1<<15)
+#define DLG_FLAG_WAS_CANCELLED			(1<<16)
+#define DLG_FLAG_RACE_CONDITION_OCCURRED	(1<<17)
+#define DLG_FLAG_SELF_EXTENDED_TIMEOUT		(1<<18)
+#define DLG_FLAG_SYNCED                     (1<<19)
+
+#define dlg_has_options_pinging(dlg) \
+	(dlg->flags & DLG_FLAG_PING_CALLER || \
+	 dlg->flags & DLG_FLAG_PING_CALLEE)
+#define dlg_has_reinvite_pinging(dlg) \
+	(dlg->flags & DLG_FLAG_REINVITE_PING_CALLER || \
+	 dlg->flags & DLG_FLAG_REINVITE_PING_CALLEE)
 
 #define DLG_CALLER_LEG         0
 #define DLG_FIRST_CALLEE_LEG   1
@@ -83,6 +83,11 @@
 #define DLG_DIR_DOWNSTREAM     1
 #define DLG_DIR_UPSTREAM       2
 
+struct dlg_leg_cseq_map {
+	struct dlg_cell *dlg;
+	unsigned int msg, gen, leg;
+	struct dlg_leg_cseq_map *next;
+};
 
 struct dlg_leg {
 	int id;
@@ -93,18 +98,22 @@ struct dlg_leg {
 	str from_uri;	/* FROM URI for this leg, in case of FROM URI mangling*/
 	str to_uri;		/* TO URI for this leg, in case of TO URI mangling */
 	str route_set;
-	str contact;
-	str th_sent_contact;	/* topology hiding advertised contact towards this leg - full header */
+	str contact;    /* this leg's Contact URI (most recent version) */
+	str adv_contact;	/* topology hiding advertised contact towards this leg - full header */
+	str in_sdp;			/* latest SDP advertised by the uac ( full body ), after all OpenSIPS changes */
+	str out_sdp;		/* latest SDP advertised towards this leg ( full body ), after all OpenSIPS changes */
+	str tmp_in_sdp;		/* temporarily stored in_sdp until confirmation (200 OK) arrives */
+	str tmp_out_sdp;	/* temporarily stored out_sdp until confirmation (200 OK) arrives */
 	str route_uris[64];
-	str sdp;		/* latest SDP provided by this leg ( full body ), after all OpenSIPS changes */
 	int nr_uris;
 	unsigned int last_gen_cseq; /* FIXME - think this can be atomic_t to avoid locking */
-	unsigned int last_inv_gen_cseq; /* used when translating ACKs */
+	struct dlg_leg_cseq_map *cseq_maps; /* used when translating ACKs */
 	char reply_received;
 	char reinvite_confirmed;
 	struct socket_info *bind_addr;
 };
 
+#define leg_is_answered(dlg_leg) ((dlg_leg)->tag.s)
 
 #define DLG_LEGS_USED      0
 #define DLG_LEGS_ALLOCED   1
@@ -117,6 +126,9 @@ struct dlg_leg {
 #define TOPOH_ONGOING     (1 << 1)
 #define TOPOH_KEEP_USER   (1 << 2)
 #define TOPOH_HIDE_CALLID (1 << 3)
+#define TOPOH_DID_IN_USER (1 << 4)
+#define TOPOH_KEEP_ADV_A  (1 << 5)
+#define TOPOH_KEEP_ADV_B  (1 << 6)
 
 struct dlg_cell
 {
@@ -127,7 +139,11 @@ struct dlg_cell
 	unsigned int         h_entry;
 	unsigned int         state;
 	unsigned int         lifetime;
-	unsigned int         lifetime_dirty; /* 1 if lifetime timer should be updated */
+	unsigned short       lifetime_dirty; /* 1 if lifetime timer should
+	                                      * be updated */
+	unsigned short       locked_by;   /* holds the ID of the process locking
+	                                   * the dialog (if the case) while
+	                                   * calling a callback */
 	unsigned int         start_ts;    /* start time  (absolute UNIX ts)*/
 	unsigned int         flags;
 	unsigned int         from_rr_nb;
@@ -135,7 +151,11 @@ struct dlg_cell
 	unsigned int         mod_flags;
 	unsigned int         initial_t_hash_index;
 	unsigned int         initial_t_label;
+	unsigned int         replicated; /* indicates if the dialog is replicated */
+	unsigned int         del_delay; /* if any custom delay should be done
+	                                 * when deleting this dialog */
 	struct dlg_tl        tl;
+	struct dlg_tl        del_tl;
 	struct dlg_ping_list *pl;
 	struct dlg_ping_list *reinvite_pl;
 	str                  terminate_reason;
@@ -147,6 +167,15 @@ struct dlg_cell
 	struct dlg_head_cbl  cbs;
 	struct dlg_profile_link *profile_links;
 	struct dlg_val       *vals;
+	str                  shtag;
+
+	struct script_route_ref  *rt_on_answer;
+	struct script_route_ref  *rt_on_timeout;
+	struct script_route_ref  *rt_on_hangup;
+
+#ifdef DBG_DIALOG
+	struct struct_hist   *hist;
+#endif
 };
 
 
@@ -169,14 +198,25 @@ struct dlg_table
 	gen_lock_set_t     *locks;
 };
 
-
+extern stat_var *active_dlgs;
+extern stat_var *early_dlgs;
+extern struct struct_hist_list *dlg_hist;
 extern struct dlg_table *d_table;
 extern int ctx_dlg_idx;
+extern int dlg_enable_stats;
 
 #define callee_idx(_dlg) \
 	(((_dlg)->legs_no[DLG_LEG_200OK]==0)? \
 		DLG_FIRST_CALLEE_LEG : (_dlg)->legs_no[DLG_LEG_200OK])
 
+#define other_leg(dlg, l) \
+	(l == DLG_CALLER_LEG? callee_idx(dlg): DLG_CALLER_LEG)
+
+#define dlg_get_out_sdp(_dlg, _dst) \
+	((_dlg)->legs[(_dst)].out_sdp.s? \
+	 (_dlg)->legs[(_dst)].out_sdp:(_dlg)->legs[other_leg(_dlg, _dst)].in_sdp)
+
+
 #define ctx_dialog_get() \
 	((struct dlg_cell*)context_get_ptr(CONTEXT_GLOBAL,current_processing_ctx,ctx_dlg_idx) )
 
@@ -185,7 +225,7 @@ extern int ctx_dlg_idx;
 
 struct dlg_cell *get_current_dialog();
 
-#define dlg_hash(_callid) core_hash(_callid, 0, d_table->size)
+#define dlg_hash(_callid) core_hash(_callid, NULL, d_table->size)
 
 #define dlg_lock(_table, _entry) \
 		lock_set_get( (_table)->locks, (_entry)->lock_idx);
@@ -202,6 +242,28 @@ struct dlg_cell *get_current_dialog();
 #define dlg_unlock_dlg(_dlg) \
 	dlg_unlock( d_table, &(d_table->entries[_dlg->h_entry]))
 
+static inline int ensure_leg_array(int needed_legs, struct dlg_cell *dlg)
+{
+	struct dlg_leg *new_legs;
+
+	while (((int)dlg->legs_no[DLG_LEGS_ALLOCED] - needed_legs) < 0) {
+		new_legs = shm_realloc(dlg->legs,
+			(dlg->legs_no[DLG_LEGS_ALLOCED] + 2) * sizeof *new_legs);
+		if (!new_legs) {
+			LM_ERR("oom\n");
+			return -1;
+		}
+
+		dlg->legs = new_legs;
+		dlg->legs_no[DLG_LEGS_ALLOCED] += 2;
+		memset(dlg->legs + dlg->legs_no[DLG_LEGS_ALLOCED] - 2, 0,
+			2 * sizeof *new_legs);
+	}
+
+	return 0;
+}
+
+
 static inline str* dlg_leg_from_uri(struct dlg_cell *dlg,int leg_no)
 {
 	/* no mangling possible on caller leg */
@@ -233,19 +295,32 @@ static inline str* dlg_leg_to_uri(struct dlg_cell *dlg,int leg_no)
 void unlink_unsafe_dlg(struct dlg_entry *d_entry, struct dlg_cell *dlg);
 void destroy_dlg(struct dlg_cell *dlg);
 
+#ifdef DBG_DIALOG
+#define DBG_REF(dlg, cnt) \
+	sh_log((dlg)->hist, DLG_REF, "h=%d, ref %d with +%d", \
+	       (dlg)->h_entry, (dlg)->ref, (cnt));
+#define DBG_UNREF(dlg, cnt) \
+	sh_log((dlg)->hist, DLG_UNREF, "h=%d, unref %d with -%d", \
+	       (dlg)->h_entry, (dlg)->ref, (cnt));
+#define DBG_FLUSH(dlg) sh_flush((dlg)->hist)
+#else
+#define DBG_REF(dlg, cnt)
+#define DBG_UNREF(dlg, cnt)
+#define DBG_FLUSH(dlg)
+#endif
+
 #define ref_dlg_unsafe(_dlg,_cnt)     \
 	do { \
+		DBG_REF(_dlg, _cnt); \
 		(_dlg)->ref += (_cnt); \
-		LM_DBG("ref dlg %p with %d -> %d\n", \
-			(_dlg),(_cnt),(_dlg)->ref); \
 	}while(0)
 
 #define unref_dlg_unsafe(_dlg,_cnt,_d_entry)   \
 	do { \
+		DBG_UNREF(_dlg, _cnt); \
 		(_dlg)->ref -= (_cnt); \
-		LM_DBG("unref dlg %p with %d -> %d in entry %p\n",\
-			(_dlg),(_cnt),(_dlg)->ref,(_d_entry));\
 		if ((_dlg)->ref<0) {\
+			DBG_FLUSH(_dlg); \
 			LM_CRIT("bogus ref %d with cnt %d for dlg %p [%u:%u] "\
 				"with clid '%.*s' and tags '%.*s' '%.*s'\n",\
 				(_dlg)->ref, _cnt, _dlg,\
@@ -253,11 +328,20 @@ void destroy_dlg(struct dlg_cell *dlg);
 				(_dlg)->callid.len, (_dlg)->callid.s,\
 				dlg_leg_print_info(_dlg, DLG_CALLER_LEG, tag), \
 				dlg_leg_print_info(_dlg, callee_idx(_dlg), tag)); \
+			abort(); \
 		}\
 		if ((_dlg)->ref<=0) { \
-			unlink_unsafe_dlg( _d_entry, _dlg);\
-			LM_DBG("ref <=0 for dialog %p\n",_dlg);\
-			destroy_dlg(_dlg);\
+			/* dlg good to be destried, but be sure it went first
+			 * via the delete timer */ \
+			if ((dlg_del_delay==0 && (_dlg)->del_delay==0) ||    \
+			insert_attempt_dlg_del_timer(&_dlg->del_tl,        \
+			(_dlg)->del_delay?(_dlg)->del_delay:dlg_del_delay)==-2) {\
+				/* no delay on del or not in del timer anymore -> destroy */ \
+				LM_DBG("Destroying dialog %p\n",_dlg); \
+				unlink_unsafe_dlg( _d_entry, _dlg);\
+				destroy_dlg(_dlg);\
+			} /* else, either still in timer (-1), either
+			   * inserted now in del timer (0) -> nothing to do*/ \
 		}\
 	}while(0)
 
@@ -269,7 +353,7 @@ void destroy_dlg(struct dlg_cell *dlg);
 	({ \
 		char *___p; \
 		int ___flags = 0; \
-		for (___p=(input).s; ___p < (input).s + (input).len; ___p++) \
+		for (___p=(input)->s; ___p < (input)->s + (input)->len; ___p++) \
 		{ \
 			switch (*___p) \
 			{ \
@@ -293,6 +377,10 @@ void destroy_dlg(struct dlg_cell *dlg);
 					___flags |= DLG_FLAG_REINVITE_PING_CALLEE; \
 					LM_DBG("re-invite ping callee activated\n"); \
 					break; \
+				case 'E': \
+					___flags |= DLG_FLAG_END_ON_RACE_CONDITION; \
+					LM_DBG("ending call on 200OK race conditions \n"); \
+					break; \
 				default: \
 					LM_DBG("unknown create_dialog flag : [%c] ." \
 						   "Skipping\n", *___p); \
@@ -310,33 +398,98 @@ void destroy_dlg_table();
 struct dlg_cell* build_new_dlg(str *callid, str *from_uri,
 		str *to_uri, str *from_tag);
 
-int dlg_add_leg_info(struct dlg_cell *dlg, str* tag, str *rr,
-		str *contact,str *cseq, struct socket_info *sock,
-		str *mangled_from,str *mangled_to,str *sdp);
+/**
+ * dlg_clone_callee_leg - Clone a callee leg and only fill in shared leg data
+ * @return: index of the new leg or -1 on error
+ */
+int dlg_clone_callee_leg(struct dlg_cell *dlg, int cloned_leg_idx);
+
+int dlg_update_leg_info(int leg_idx, struct dlg_cell *dlg, str* tag, str *rr,
+		str *contact, str *adv_ct, str *cseq, struct socket_info *sock,
+		str *mangled_from,str *mangled_to,str *in_sdp, str *out_sdp);
 
 int dlg_update_cseq(struct dlg_cell *dlg, unsigned int leg, str *cseq,
-						int field_type);
+		int field_type);
 
-int dlg_update_routing(struct dlg_cell *dlg, unsigned int leg,str *rr, str *contact);
+int dlg_update_routing(struct dlg_cell *dlg, unsigned int leg,str *rr,
+		str *contact);
 
-struct dlg_cell* lookup_dlg( unsigned int h_entry, unsigned int h_id);
+struct dlg_cell* lookup_dlg( unsigned int h_entry, unsigned int h_id,
+		int active_only);
 
 struct dlg_cell* get_dlg(str *callid, str *ftag, str *ttag,
 		unsigned int *dir, unsigned int *dst_leg);
 
-struct dlg_cell* get_dlg_by_val(str *attr, str *val);
+struct dlg_cell* get_dlg_by_val(struct sip_msg *msg, str *attr, pv_spec_t *val);
 
-void link_dlg(struct dlg_cell *dlg, int n);
+struct dlg_cell* get_dlg_by_callid(const str *callid, int active_only);
 
-void unref_dlg(struct dlg_cell *dlg, unsigned int cnt);
+struct dlg_cell* get_dlg_by_did(str *did, int active_only);
 
-void ref_dlg(struct dlg_cell *dlg, unsigned int cnt);
+struct dlg_cell* get_dlg_by_ids(unsigned int h_entry, unsigned int h_id, int active_only);
+
+struct dlg_cell *get_dlg_by_dialog_id(str *dialog_id);
+
+int get_dlg_direction(void);
+
+void link_dlg(struct dlg_cell *dlg, int extra_refs);
+
+#define _link_dlg_unsafe(d_entry, dlg) \
+	do { \
+		if (!d_entry->first) { \
+			d_entry->first = d_entry->last = dlg; \
+		} else { \
+			d_entry->last->next = dlg; \
+			dlg->prev = d_entry->last; \
+			d_entry->last = dlg; \
+		} \
+		DBG_REF(dlg, 1); \
+		dlg->ref++; \
+		d_entry->cnt++; \
+	} while (0)
+
+#define link_dlg_unsafe(d_entry, dlg) \
+	do { \
+		dlg->h_id = d_entry->next_id++; \
+		_link_dlg_unsafe(d_entry, dlg); \
+	} while (0)
+
+void _unref_dlg(struct dlg_cell *dlg, unsigned int cnt);
+#define unref_dlg(dlg, cnt) \
+	do { \
+		DBG_UNREF(dlg, cnt); \
+		_unref_dlg(dlg, cnt); \
+	} while (0)
+
+void _ref_dlg(struct dlg_cell *dlg, unsigned int cnt);
+#define ref_dlg(dlg, cnt) \
+	do { \
+		DBG_REF(dlg, cnt); \
+		_ref_dlg(dlg, cnt); \
+	} while (0)
 
 void next_state_dlg(struct dlg_cell *dlg, int event, int dir, int *old_state,
-		int *new_state, int *unref, int last_dst_leg, char is_replicated);
+		int *new_state, int *unref, int last_dst_leg, char replicate_events);
 
-struct mi_root * mi_print_dlgs(struct mi_root *cmd, void *param );
-struct mi_root * mi_print_dlgs_ctx(struct mi_root *cmd, void *param );
+mi_response_t *mi_print_dlgs(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+mi_response_t *mi_print_dlgs_1(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+mi_response_t *mi_print_dlgs_2(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+mi_response_t *mi_print_dlgs_cnt(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+mi_response_t *mi_print_dlgs_ctx(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+mi_response_t *mi_print_dlgs_1_ctx(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+mi_response_t *mi_print_dlgs_2_ctx(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+mi_response_t *mi_print_dlgs_cnt_ctx(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+
+mi_response_t *mi_push_dlg_var(const mi_params_t *params,
+								struct mi_handler *async_hdl);
 
 static inline void unref_dlg_destroy_safe(struct dlg_cell *dlg, unsigned int cnt)
 {
@@ -465,7 +618,44 @@ static inline int match_dialog(struct dlg_cell *dlg, str *callid,
 */
 }
 
-int mi_print_dlg(struct mi_node *rpl, struct dlg_cell *dlg, int with_context);
+/* @return: 0 if found, -1 otherwise */
+static inline int get_dlg_unsafe(struct dlg_entry *d_entry,
+          str *callid, str *from_tag, str *to_tag, struct dlg_cell **out_dlg)
+{
+	struct dlg_cell *it;
+	int callee_leg_idx;
+
+	for (it = d_entry->first; it; it = it->next) {
+		if (it->callid.len == callid->len &&
+			it->legs[DLG_CALLER_LEG].tag.len == from_tag->len &&
+			!memcmp(it->callid.s, callid->s, callid->len) &&
+			!memcmp(it->legs[DLG_CALLER_LEG].tag.s, from_tag->s, from_tag->len)) {
+			/* callid & ftag match */
+			callee_leg_idx = callee_idx(it);
+			if (it->legs[callee_leg_idx].tag.len == to_tag->len &&
+				!memcmp(it->legs[callee_leg_idx].tag.s, to_tag->s, to_tag->len)) {
+				/* full dlg match */
+				*out_dlg = it;
+				return 0;
+			}
+		}
+	}
+
+	*out_dlg = NULL;
+	return -1;
+}
+
+static inline void update_dlg_stats(struct dlg_cell *dlg, int amount)
+{
+	if (dlg->state == DLG_STATE_CONFIRMED_NA ||
+	        dlg->state==DLG_STATE_CONFIRMED) {
+		if_update_stat(dlg_enable_stats, active_dlgs, amount);
+	} else if (dlg->state == DLG_STATE_EARLY) {
+		if_update_stat(dlg_enable_stats, early_dlgs, amount);
+	}
+}
+
+int mi_print_dlg(mi_item_t *dialog_obj, struct dlg_cell *dlg, int with_context);
 
 static inline void init_dlg_term_reason(struct dlg_cell *dlg,char *reason,int reason_len)
 {
@@ -482,7 +672,17 @@ static inline void init_dlg_term_reason(struct dlg_cell *dlg,char *reason,int re
 	}
 }
 
+
 int state_changed_event_init(void);
 void state_changed_event_destroy(void);
 
+#define dlg_get_db_id(_dlg) \
+	(((unsigned long long)(_dlg)->h_entry << 32) | ((_dlg)->h_id))
+
+#define dlg_parse_db_id(_did, _h_entry, _h_id) \
+	do { \
+		(_h_entry) = (unsigned int)((unsigned long long)(_did) >> 32); \
+		(_h_id) = (unsigned int)((unsigned long long)(_did) & 0xFFFFFFFFULL); \
+	} while(0)
+
 #endif
