@@ -54,13 +54,13 @@ static cachedb_con *cdbc = 0;
 static int blacklist_timeout=3600; /* seconds */
 static str cachedb_url = {0,0};
 
-static param_export_t params[]={
+static const param_export_t params[]={
 	{ "cachedb_url",                 STR_PARAM, &cachedb_url.s},
 	{ "blacklist_timeout",           INT_PARAM, &blacklist_timeout},
 	{0,0,0}
 };
 
-static dep_export_t deps = {
+static const dep_export_t deps = {
 	{ /* OpenSIPS module dependencies */
 		{ MOD_TYPE_CACHEDB, NULL, DEP_ABORT },
 		{ MOD_TYPE_NULL, NULL, 0 },
@@ -76,6 +76,7 @@ struct module_exports exports= {
 	MOD_TYPE_DEFAULT,/* class of this module */
 	MODULE_VERSION,
 	DEFAULT_DLFLAGS,			/* dlopen flags */
+	0,							/* load function */
 	&deps,              /* OpenSIPS module dependencies */
 	0,					/* exported functions */
 	0,					/* exported async functions */
@@ -83,11 +84,14 @@ struct module_exports exports= {
 	0,					/* exported statistics */
 	0,					/* exported MI functions */
 	0,					/* exported pseudo-variables */
+	0,			 		/* exported transformations */
 	0,					/* extra processes */
+	0,						/* module pre-initialization function */
 	mod_init,				/* module initialization function */
 	(response_function) 0,      		/* response handling function */
 	(destroy_function)destroy,		/* destroy function */
-	child_init			        /* per-child init function */
+	child_init,			        /* per-child init function */
+	0                           /* reload confirm function */
 };
 
 /**
@@ -102,7 +106,7 @@ static int mod_init(void)
 		return -1;
 	} else {
 		cachedb_url.len = strlen(cachedb_url.s);
-		LM_DBG("using CacheDB URL: %s\n", cachedb_url.s);
+		LM_DBG("using CacheDB URL: %s\n", db_url_escape(&cachedb_url));
 	}
 
 	/* set pointers that resolver will use for caching */
@@ -115,8 +119,8 @@ static int mod_init(void)
 static int child_init(int rank)
 {
 	if (cachedb_bind_mod(&cachedb_url, &cdbf) < 0) {
-		LM_ERR("cannot bind functions for db_url %.*s\n",
-				cachedb_url.len, cachedb_url.s);
+		LM_ERR("cannot bind functions for db_url %s\n",
+				db_url_escape(&cachedb_url));
 		return -1;
 	}
 
@@ -128,7 +132,7 @@ static int child_init(int rank)
 
 	cdbc = cdbf.init(&cachedb_url);
 	if (!cdbc) {
-		LM_ERR("cannot connect to db_url %.*s\n", cachedb_url.len, cachedb_url.s);
+		LM_ERR("cannot connect to db_url %s\n", db_url_escape(&cachedb_url));
 		return -1;
 	}
 
@@ -141,13 +145,13 @@ static int child_init(int rank)
 static void destroy(void)
 {
 	LM_NOTICE("destroy module dns_cache ...\n");
-	if (cdbc)
-		cdbf.destroy(cdbc);
 }
 
 static int rdata_struct_len=sizeof(struct rdata)-sizeof(void *) -
 		sizeof(struct rdata *);
 
+#define MAXALIASES		36
+#define MAXADDRS 		36
 static unsigned char *he_buf=NULL;
 static int he_buf_len=0;
 static char* serialize_he_rdata(struct hostent *he,int *buf_len,int do_encoding)
@@ -163,7 +167,7 @@ static char* serialize_he_rdata(struct hostent *he,int *buf_len,int do_encoding)
 		len+=strlen(he->h_name)+1;
 
 	if (he->h_aliases)
-       		for (i=0;he->h_aliases[i];i++) {
+		for (i=0;he->h_aliases[i]&&alias_no<MAXALIASES-1;i++) {
 			/* integer with len + len bytes of alias */
 			len+=strlen(he->h_aliases[i])+1+sizeof(int);
 			alias_no++;
@@ -172,13 +176,13 @@ static char* serialize_he_rdata(struct hostent *he,int *buf_len,int do_encoding)
 
 	i=0;
 	if (he->h_addr_list)
-       		for (i=0;he->h_addr_list[i];i++) {
+		for (i=0;he->h_addr_list[i]&&addr_no<MAXADDRS-1;i++) {
 			len+=he->h_length;
 			addr_no++;
 		}
 
 	if (do_encoding) {
-		/* backend does not support binary values - allocate continous buffer
+		/* backend does not support binary values - allocate continuous buffer
 		for encoding */
 		base64_len = calc_base64_encode_len(len);
 		needed_len=len+base64_len;
@@ -215,7 +219,7 @@ static char* serialize_he_rdata(struct hostent *he,int *buf_len,int do_encoding)
 
 	/* copy aliases, if any */
 	if (he->h_aliases)
-       		for (i=0;he->h_aliases[i];i++) {
+		for (i=0;i<alias_no;i++) {
 			len=strlen(he->h_aliases[i])+1;
 			/* copy alias length */
 			memcpy(p,&len,sizeof(int));
@@ -231,7 +235,7 @@ static char* serialize_he_rdata(struct hostent *he,int *buf_len,int do_encoding)
 
 	/* copy addresses */
 	if (he->h_addr_list)
-       		for (i=0;he->h_addr_list[i];i++) {
+		for (i=0;i<addr_no;i++) {
 			/* copy addreses. length will be known from the addrtype field */
 			len=he->h_length;
 			memcpy(p,he->h_addr_list[i],len);
@@ -257,8 +261,6 @@ static char* serialize_he_rdata(struct hostent *he,int *buf_len,int do_encoding)
 static unsigned char *dec_he_buf=NULL;
 static int dec_he_buf_len=0;
 static struct hostent dec_global_he;
-#define MAXALIASES		36
-#define MAXADDRS 		36
 static char *h_addr_ptrs[MAXADDRS];
 static char *host_aliases[MAXALIASES];
 static struct hostent* deserialize_he_rdata(char *buff,int buf_len,int do_decoding)
@@ -287,10 +289,8 @@ static struct hostent* deserialize_he_rdata(char *buff,int buf_len,int do_decodi
 
 	/* set pointer in dec_global_he */
 	ap = host_aliases;
-	*ap = NULL;
 	dec_global_he.h_aliases = host_aliases;
 	hap = h_addr_ptrs;
-	*hap = NULL;
 	dec_global_he.h_addr_list = h_addr_ptrs;
 
 	if (do_decoding) {
@@ -327,6 +327,7 @@ static struct hostent* deserialize_he_rdata(char *buff,int buf_len,int do_decodi
 		*ap++ = (char *)p;
 		p+=len;
 	}
+	*ap = NULL;
 
 	/* get number of addresses */
 	memcpy(&addr_no,p,sizeof(int));
@@ -337,6 +338,7 @@ static struct hostent* deserialize_he_rdata(char *buff,int buf_len,int do_decodi
 		*hap++ = (char *)p;
 		p+=dec_global_he.h_length;
 	}
+	*hap = NULL;
 
 	return &dec_global_he;
 }
@@ -459,14 +461,14 @@ static char* serialize_dns_rdata(struct rdata *head,int buf_len,int *len,int do_
 	}
 
 	if (do_encoding) {
-		if (*len)
+		if (len)
 			*len = base64_len;
 
 		/* encode and return beggining of encoding */
 		base64encode(p,rdata_buf,buf_len);
 		return (char *)p;
 	} else {
-		if (*len)
+		if (len)
 			*len = needed_len;
 		return (char *)rdata_buf;
 	}
@@ -805,15 +807,22 @@ void* get_dnscache_value(char *name,int r_type,int name_len)
 }
 
 /* Pushes internal structure ( hostent or rdata ) to a cache backend
- * Params :
- * name - what query to be saved - binary IP for PTR and strings for other queries
- * r_type - type of DNS query
- * record - pointer to hostent or rdata
- * rdata_len - If rdata record, rdata_len holds the actual length of rdata buf,
-		in order to avoid double iterations on the rdata struct. If it's a
-		PTR record, rdata_len is used to differentiate between IP and IPv6
- * failure - should we blacklist or not
- * ttl - seconds the key should be kept in cache */
+ *
+ * Params:
+ *	name - what query to be saved - binary IP for PTR and strings for other queries
+ *	r_type - type of DNS query
+ *	record - pointer to hostent or rdata
+ *	rdata_len - If rdata record, rdata_len holds the actual length of rdata buf,
+ *		in order to avoid double iterations on the rdata struct. If it's a
+ *		PTR record, rdata_len is used to differentiate between IP and IPv6
+ *	failure - should we blacklist or not
+ *	ttl - seconds the key should be kept in cache
+ *
+ * Returns:
+ *	0  - success
+ *	1  - cache not initialized yet
+ *	-1 - internal failure
+ */
 int put_dnscache_value(char *name,int r_type,void *record,int rdata_len,
 				int failure,int ttl)
 {
@@ -822,7 +831,15 @@ int put_dnscache_value(char *name,int r_type,void *record,int rdata_len,
 
 	if (cdbc == NULL) {
 		/* assume dns request before forking - cache is not ready yet */
-		return -1;
+		return 1;
+	}
+
+	/* avoid caching records with TTL=0 */
+	if (!failure && ttl==0) {
+		/* RFC1035 states : "Zero TTL values are interpreted to mean that
+		   the RR can only be used for the transaction in progress, and
+		   should not be cached." */
+		return 1;
 	}
 
 	/* generate key */
@@ -831,6 +848,8 @@ int put_dnscache_value(char *name,int r_type,void *record,int rdata_len,
 		LM_ERR("failed to create key\n");
 		return -1;
 	}
+	value.len = 0;
+	value.s = 0;
 
 	if (failure) {
 		/* just set value as failure marker, and push to back-end
@@ -858,7 +877,8 @@ int put_dnscache_value(char *name,int r_type,void *record,int rdata_len,
 		key_ttl = ttl;
 	}
 
-	LM_DBG("putting value [%.*s] with ttl = %d\n",key.len,key.s,key_ttl);
+	LM_INFO("putting key [%.*s] with value [%.*s] ttl = %d\n",
+		key.len,key.s,value.len,value.s,key_ttl);
 	if (cdbf.set(cdbc,&key,&value,key_ttl) < 0) {
 		LM_ERR("failed to set dns key\n");
 		return -1;
