@@ -1,5 +1,6 @@
 /*
  * Copyright (C) 2005-2008 Voice Sistem SRL
+ * Copyright (C) 2020 OpenSIPS Solutions
  *
  * This file is part of Open SIP Server (OpenSIPS).
  *
@@ -16,15 +17,6 @@
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
- *
- * For any questions about this software and its license, please contact
- * Voice Sistem at following e-mail address:
- *         office@voice-system.ro
- *
- * History:
- * ---------
- *  2005-02-20  first version (cristian)
- *  2005-02-27  ported to 0.9.0 (bogdan)
  */
 
 
@@ -36,6 +28,11 @@
 #include "dr_partitions.h"
 #include "routing.h"
 
-rt_data_t* dr_load_routing_info(struct head_db * ,int persistent_state);
+int dr_set_gw_sock_filter_mode(char *mode);
+
+void dr_update_head_cache(struct head_db *head);
+
+rt_data_t* dr_load_routing_info(struct head_db *current_partition,
+		int persistent_state, str *rules_tables, int rules_no);
 
 #endif
