@@ -50,17 +50,37 @@ static void mod_destroy(void);
 struct dlg_binds dlgb;
 struct rr_binds rr_api;
 
+struct tm_binds eme_tm;
+
+str db_url;
+str *db_table;
+db_func_t db_funcs;
+db_con_t *db_con;
+
+struct esrn_routing **db_esrn_esgwri;
+struct service_provider **db_service_provider;
+
 str callid_aux;
+char* url_vpc;
+
+int emet_size;
+int subst_size;
+
+char *empty;
+char *mandatory_parm;
+
+struct call_htable *call_htable;
+struct subs_htable *subs_htable;
 
 /*
  * Exported functions
  */
-static cmd_export_t cmds[] = {
-	{"emergency_call", (cmd_function) emergency_call, 0, 0, 0,
+static const cmd_export_t cmds[] = {
+	{"emergency_call", (cmd_function) emergency_call, {{0, 0, 0}},
 		REQUEST_ROUTE | BRANCH_ROUTE },
-	{"failure", (cmd_function) failure, 0, 0, 0,
+	{"failure", (cmd_function) failure, {{0, 0, 0}},
 		FAILURE_ROUTE | ONREPLY_ROUTE },
-	{ 0, 0, 0, 0, 0, 0}
+	{ 0, 0, {{0, 0, 0}}, 0}
 };
 
 
@@ -68,13 +88,13 @@ static cmd_export_t cmds[] = {
  * Exported parameters
  */
 
-static param_export_t params[] = {
+static const param_export_t params[] = {
 	{ "emergency_codes", STR_PARAM | USE_FUNC_PARAM, (void *) &set_codes},
 	{ "timer_interval", INT_PARAM, &timer_interval},
 	{ "db_url", STR_PARAM, &db_url.s},
-	{ "db_table_routing", STR_PARAM, &table_name},
-	{ "db_table_report", STR_PARAM, &table_report},
-	{ "db_table_provider", STR_PARAM, &table_provider},
+	{ "db_table_routing", STR_PARAM, &table_name.s},
+	{ "db_table_report", STR_PARAM, &table_report.s},
+	{ "db_table_provider", STR_PARAM, &table_provider.s},
 	{ "url_vpc", STR_PARAM, &url_vpc},
 	{ "contingency_hostname", STR_PARAM, &contingency_hostname},
 	{ "emergency_call_server", STR_PARAM, &call_server_hostname},
@@ -89,7 +109,7 @@ static param_export_t params[] = {
 /*
  * Module parameter variables
  */
-static dep_export_t deps = {
+static const dep_export_t deps = {
 	{ /* OpenSIPS module dependencies */
 		{ MOD_TYPE_SQLDB, NULL, DEP_ABORT },
 		{ MOD_TYPE_NULL, NULL, 0 },
@@ -107,6 +127,7 @@ struct module_exports exports = {
 	MOD_TYPE_DEFAULT,/* class of this module */
 	MODULE_VERSION, /* module version */
 	DEFAULT_DLFLAGS, /* dlopen flags */
+	0,				 /* load function */
 	&deps,           /* OpenSIPS module dependencies */
 	cmds, /* Exported functions */
 	NULL,      /* Exported async functions */
@@ -114,11 +135,14 @@ struct module_exports exports = {
 	0, /* exported statistics */
 	0, /* exported MI functions */
 	0, /* exported pseudo-variables */
+	0, /* exported transformations */
 	0, /* extra processes */
+	0, /* module pre-initialization function */
 	mod_init, /* module initialization function */
 	0, /* response function*/
 	mod_destroy,
-	child_init /* per-child init function */
+	child_init,/* per-child init function */
+	0          /* reload confirm function */
 };
 
 /*
@@ -195,6 +219,10 @@ static int mod_init(void) {
 
 	LM_DBG("Initializing module\n");
 
+	table_name.len = strlen(table_name.s);
+	table_report.len = strlen(table_report.s);
+	table_provider.len = strlen(table_provider.s);
+
 	// checks for mandatory fields
 	mandatory_parm = shm_malloc(2);
 	mandatory_parm[0] = '1';
@@ -273,21 +301,21 @@ static int mod_init(void) {
 
 	db_table = (str *)shm_malloc(sizeof (str));
 	if (!db_table) {
-		LM_ERR("no more memory");
+		LM_ERR("no more memory\n");
 		return -1;
 	}
 	db_table = &table_report;
 
 	db_esrn_esgwri = shm_malloc(sizeof (struct esrn_routing *));
 	if (!db_esrn_esgwri) {
-		LM_ERR("no more memory");
+		LM_ERR("no more memory\n");
 		return -1;
 	}
 	*db_esrn_esgwri = NULL;
 
 	db_service_provider = shm_malloc(sizeof (struct service_provider *));
 	if (!db_service_provider) {
-		LM_ERR("no more memory");
+		LM_ERR("no more memory\n");
 		return -1;
 	}
 	*db_service_provider = NULL;
@@ -341,7 +369,7 @@ static int mod_init(void) {
 static int child_init(int rank) {
 	LM_DBG("Initializing child\n");
 
-	if (db_url.s && rank>PROC_MAIN) {
+	if (db_url.s && rank>=1) {
 		/* open a test connection */
 
 		if ((db_con = db_funcs.init(&db_url)) == 0) {
@@ -368,9 +396,6 @@ static int child_init(int rank) {
 static void mod_destroy(void) {
 	curl_global_cleanup();
 
-	if (db_con != NULL && db_funcs.close != 0)
-		db_funcs.close(db_con);
-
 	if(ref_lock){
 		lock_destroy_rw( ref_lock );
 		ref_lock = NULL;
@@ -410,14 +435,14 @@ void routing_timer(unsigned int ticks, void *attr) {
 
 
 /*
- * - verifying the expiration for packet loss ( timing values are diferent for ACK and BYE)
+ * - verifying the expiration for packet loss ( timing values are different for ACK and BYE)
  * - if there is an expiration the module sends a POST informing the VPC to exclude the number
  *   the key ESQK is retreived from the list calls_eme
  */
 static void libera_esqk(void) {
 
 	time_t rawtime;
-	struct tm * timeinfo;
+	struct tm timeinfo;
 	int resp = 1;
 	char* response;
 	char* esct_callid;
@@ -453,8 +478,8 @@ static void libera_esqk(void) {
 
 						//send esctRequest to the VPC
 						time(&rawtime);
-						timeinfo = localtime(&rawtime);
-						strftime(free_cell->esct->datetimestamp, MAX_TIME_SIZE, "%Y-%m-%dT%H:%M:%S%Z", timeinfo);
+						localtime_r(&rawtime, &timeinfo);
+						strftime(free_cell->esct->datetimestamp, MAX_TIME_SIZE, "%Y-%m-%dT%H:%M:%S%Z", &timeinfo);
 
 						xml = buildXmlFromModel(free_cell->esct);
 						resp = post(url_vpc, xml, &response);
@@ -700,7 +725,7 @@ void reply_in_redirect( struct cell* t, int type, struct tmcb_params *params){
 				}else{
 					if(resp == 0){
 						if(call_cell->lro == empty){
-							LM_ERR("dont exits esgwri/ert or lro to routing\n");
+							LM_ERR("don't exits esgwri/ert or lro to routing\n");
 							goto end;
 						}
 					}
@@ -708,14 +733,14 @@ void reply_in_redirect( struct cell* t, int type, struct tmcb_params *params){
 			}else{
 				LM_DBG("exits lro to routing %d\n", resp);
 				if(call_cell->lro == empty){
-					LM_ERR("dont exits esgwri/ert or lro to routing\n");
+					LM_ERR("don't exits esgwri/ert or lro to routing\n");
 					goto end;
 				}
 			}
 		}
 	}
 
-	hash_code= core_hash(&reply->callid->body, 0, emet_size);
+	hash_code= core_hash(&reply->callid->body, NULL, emet_size);
 	LM_DBG("********************************************HASH_CODE%d\n", hash_code);
 
 	if(insert_ehtable(call_htable,hash_code,call_cell)< 0){
@@ -877,13 +902,13 @@ static int failure(struct sip_msg *msg) {
 	memset(from_tag, 0, pfrom->tag_value.len + 1);
 	strncpy(from_tag, pfrom->tag_value.s, pfrom->tag_value.len);
 
-	hash_code= core_hash(&msg->callid->body, 0, emet_size);
+	hash_code= core_hash(&msg->callid->body, NULL, emet_size);
 	LM_DBG("********************************************HASH_CODE%d\n", hash_code);
 
 	// find the cell with the callid from the list calls_cell
 	s= search_ehtable(call_htable, callidHeader, from_tag, hash_code, 0);
 	if (s == NULL) {
-		LM_ERR(" ---FAILURE treatment did not find the CALLID");
+		LM_ERR(" ---FAILURE treatment did not find the CALLID\n");
 		goto error;
 	}
 
@@ -908,7 +933,7 @@ static int failure(struct sip_msg *msg) {
 			if(strlen(info_call->esgwri) > 1){
 				LM_DBG ("FAILURE REPLY ESGWRI %s \n",info_call->esgwri);
 				if(new_uri_proxy(msg, info_call->esgwri) == -1){
-					LM_ERR(" ERROR IN NEW_URI_PROXY");
+					LM_ERR(" ERROR IN NEW_URI_PROXY\n");
 					pkg_free(cbn_aux);
 					goto lro;
 				}
@@ -952,7 +977,7 @@ lro:
 	// verifica se a chamada tratada teve o numero de contingencia lro fornecido pelo VPC
 	// caso não tenha, não trata failure
 	if (info_call->lro == empty) {
-		LM_ERR(" ---treat FAILURE not found lro");
+		LM_ERR(" ---treat FAILURE not found lro\n");
 		goto error;
 	}
 
@@ -975,11 +1000,11 @@ lro:
 
 		LM_DBG(" ---NEW DESTIN =%s", new_to);
 		if(new_uri_proxy(msg, new_to) == -1){
-			LM_ERR(" ---ERRO EM NEW_URI_PROXY");
+			LM_ERR(" ---ERRO EM NEW_URI_PROXY\n");
 			goto error;
 		}
 	}else{
-		LM_DBG(" ---FAILURE JA TRANSMITIU LRO");
+		LM_DBG(" ---FAILURE JA TRANSMITIU LRO\n");
 		goto error;
 	}
 end:
@@ -1043,7 +1068,7 @@ int is_emergency_call(struct sip_msg *msg) {
 			if (codigo->code.len == msg->parsed_uri.user.len){
 				if (strncmp(codigo->code.s, msg->parsed_uri.user.s , codigo->code.len) == 0) {
 
-					LM_DBG(" ---> CODIGO -- OK %.*s\n\n", codigo->code.len, codigo->code.s);
+					LM_DBG(" ---> CODIGO -- OK %.*s\n", codigo->code.len, codigo->code.s);
 
 					if (check_myself(msg)) {
 						LM_DBG(" --- IT IS ONWER HOST  \n \n");
@@ -1125,10 +1150,10 @@ int send_request_vpc(struct sip_msg *msg) {
 	if (proxy_role == 2) {
 		LM_DBG(" ---role: call server scenario II \n");
 		if (add_hdr_PAI(msg, cbn) == -1) {
-			LM_ERR("FAILURE IN ADD PAI");
+			LM_ERR("FAILURE IN ADD PAI\n");
 		}
 		if (proxy_request(msg,call_server_hostname) == -1) {
-			LM_ERR("ERROR IN ROUTING EMERGENCY REQUEST");
+			LM_ERR("ERROR IN ROUTING EMERGENCY REQUEST\n");
 			return -1;
 		}
 		return 1;
@@ -1138,11 +1163,11 @@ int send_request_vpc(struct sip_msg *msg) {
 		// Call Server SCENARIO III
 		LM_DBG(" ---role: proxy redirect \n");
 		//if (add_hdr_PAI(msg, cbn) == -1) {
-		//    LM_ERR("FAILURE IN ADD PAI");
+		//    LM_ERR("FAILURE IN ADD PAI\n");
 		//}
 		pkg_free(cbn.s);
 		if (proxy_request(msg,call_server_hostname) == -1) {
-			LM_ERR("ERROR IN ROUTING EMERGENCY REQUEST");
+			LM_ERR("ERROR IN ROUTING EMERGENCY REQUEST\n");
 			return -1;
 		}
 		eme_tm.register_tmcb(msg,NULL,TMCB_RESPONSE_IN,reply_in_redirect,0,0);
@@ -1203,7 +1228,7 @@ int send_request_vpc(struct sip_msg *msg) {
 		if(locationHeader && strlen(locationHeader)>1){
 			lie = locationHeader;
 		}else{
-			LM_ERR("INVITE whithout location information\n");
+			LM_ERR("INVITE without location information\n");
 			return -1;
 		}
 	}
@@ -1238,7 +1263,7 @@ int send_request_vpc(struct sip_msg *msg) {
 	}
 
 
-	LM_DBG("END EMERGENCY");
+	LM_DBG("END EMERGENCY\n");
 	resp = 1;
 
 end:
@@ -1270,7 +1295,7 @@ int create_call_cell(PARSED *parsed,struct sip_msg* msg, char* callidHeader, str
 
 	unsigned int hash_code;
 
-	LM_DBG(" ---PARSED ");
+	LM_DBG(" ---PARSED \n");
 	if ((parsed->callid == empty || parsed->result == empty || parsed->vpc->nenaid == empty || parsed->vpc->contact == empty)) {
 		LM_ERR("MANDATORY FIELDS ARE BLANK \n");
 		free_parsed(parsed);
@@ -1342,7 +1367,7 @@ int create_call_cell(PARSED *parsed,struct sip_msg* msg, char* callidHeader, str
 
 		// insert calls_eme in call_htable hash with key source ip address
 
-		hash_code= core_hash(&msg->callid->body, 0, emet_size);
+		hash_code= core_hash(&msg->callid->body, NULL, emet_size);
 		LM_DBG("********************************************HASH_CODE%d\n", hash_code);
 
 		if(insert_ehtable(call_htable, hash_code,call_cell)< 0){
@@ -1404,7 +1429,7 @@ int treat_routing(struct sip_msg* msg, struct esct *call_cell, char* callidHeade
 			}
 
 			if(new_uri_proxy(msg,call_cell->esgwri) == -1){
-				LM_ERR(" ---ERROR IN NEW_URI_PROXY");
+				LM_ERR(" ---ERROR IN NEW_URI_PROXY\n");
 				goto error;
 			}
 
@@ -1522,7 +1547,7 @@ int routing_by_ert( struct sip_msg *msg, ESCT *call_cell, int failure) {
 		}
 
 		if(new_uri_proxy(msg, call_cell->esgwri) == -1){
-			LM_ERR(" ---ERROR IN NEW_URI_PROXY");
+			LM_ERR(" ---ERROR IN NEW_URI_PROXY\n");
 			return -1;
 		}
 
@@ -1565,7 +1590,7 @@ int contingency(struct sip_msg *msg, ESCT *call_cell){
 	sprintf(call_cell->esgwri, "sip:%s@%s;user=phone", lro, contingency_hostname);
 
 	if(new_uri_proxy(msg, call_cell->esgwri) == -1){
-		LM_ERR(" ---ERRO EM NEW_URI_PROXY");
+		LM_ERR(" ---ERRO EM NEW_URI_PROXY\n");
 		return -1;
 	}
 
@@ -1649,7 +1674,7 @@ int routing_ack(struct sip_msg *msg) {
 	strncpy(from_tag, pfrom->tag_value.s, pfrom->tag_value.len);
 	LM_DBG("PFROM_TAGIII: %s \n ", from_tag );
 
-	hash_code= core_hash(&msg->callid->body, 0, emet_size);
+	hash_code= core_hash(&msg->callid->body, NULL, emet_size);
 	LM_DBG("********************************************HASH_CODE%d\n", hash_code);
 
 	LM_DBG(" ---TREATMENT ACK  callid=%s \n", callidHeader);
@@ -1664,9 +1689,9 @@ int routing_ack(struct sip_msg *msg) {
 	info_call = s->esct;
 
 	if (strlen(info_call->esgwri) > 0) {
-		LM_DBG(" ---Routing ACK %s \n\n", info_call->esgwri);
+		LM_DBG(" ---Routing ACK %s \n", info_call->esgwri);
 		if(new_uri_proxy(msg, info_call->esgwri) == -1){
-			LM_ERR(" ---ERROR IN NEW_URI_PROXY");
+			LM_ERR(" ---ERROR IN NEW_URI_PROXY\n");
 			resp = -1;
 			goto end;
 		}
@@ -1699,7 +1724,7 @@ int bye(struct sip_msg *msg, int dir) {
 	char* response;
 	char* esct_callid;
 	time_t rawtime;
-	struct tm * timeinfo;
+	struct tm timeinfo;
 	NODE* info_call;
 	char* xml;
 	struct sm_subscriber*  cell_notif;
@@ -1711,7 +1736,7 @@ int bye(struct sip_msg *msg, int dir) {
 	LM_DBG(" --- BYE \n \n");
 
 	time(&rawtime);
-	timeinfo = localtime(&rawtime);
+	localtime_r(&rawtime, &timeinfo);
 	time_now = (int)rawtime;
 
 	if (proxy_role == 2) {
@@ -1719,7 +1744,7 @@ int bye(struct sip_msg *msg, int dir) {
 		if (dir == 1) {
 			LM_DBG(" ---role: proxy routing \n");
 			if (proxy_request(msg,call_server_hostname) == -1) {
-				LM_ERR("ERROR IN ROUTING EMERGENCY REQUEST");
+				LM_ERR("ERROR IN ROUTING EMERGENCY REQUEST\n");
 				return -1;
 			}
 			return 1;
@@ -1817,7 +1842,7 @@ int bye(struct sip_msg *msg, int dir) {
 
 	}
 
-	hash_code= core_hash(&msg->callid->body, 0, emet_size);
+	hash_code= core_hash(&msg->callid->body, NULL, emet_size);
 	LM_DBG("********************************************HASH_CODE%d\n", hash_code);
 
 	// search call hash with hash_code, callidHeader and from/to_tag params
@@ -1845,9 +1870,9 @@ int bye(struct sip_msg *msg, int dir) {
 		// routing BYE to same direction that INVITE
 
 		if (strlen(info_call->esct->esgwri) > 0) {
-			LM_DBG(" ---Routing BYE %s \n\n", info_call->esct->esgwri);
+			LM_DBG(" ---Routing BYE %s \n", info_call->esct->esgwri);
 			if(new_uri_proxy(msg, info_call->esct->esgwri) == -1){
-				LM_ERR(" ---ERROR IN NEW_URI_PROXY");
+				LM_ERR(" ---ERROR IN NEW_URI_PROXY\n");
 				shm_free(info_call->esct->esgwri);
 				shm_free(info_call);
 				resp = -1;
@@ -1864,7 +1889,7 @@ int bye(struct sip_msg *msg, int dir) {
 
 			LM_DBG(" --- SEND ESQK =%s\n \n",info_call->esct->esqk);
 
-			strftime(info_call->esct->datetimestamp, MAX_TIME_SIZE, "%Y-%m-%dT%H:%M:%S%Z", timeinfo);
+			strftime(info_call->esct->datetimestamp, MAX_TIME_SIZE, "%Y-%m-%dT%H:%M:%S%Z", &timeinfo);
 
 			xml = buildXmlFromModel(info_call->esct);
 			LM_DBG(" --- TREAT BYE - XML ESCT %s \n \n", xml);
@@ -1970,8 +1995,10 @@ int check_myself(struct sip_msg *msg) {
 		LM_ERR("cannot parse msg URI\n");
 		return 0;
 	}
-	LM_DBG(" --- opensips host %.*s \n \n", msg->parsed_uri.host.len, msg->parsed_uri.host.s);
-	ret = check_self_op(EQUAL_OP, &msg->parsed_uri.host, 0);
+	LM_DBG(" --- opensips host %.*s \n \n",
+		msg->parsed_uri.host.len, msg->parsed_uri.host.s);
+
+	ret=check_self(&msg->parsed_uri.host, 0, 0);
 	return ret;
 }
 
@@ -2004,7 +2031,7 @@ char* formatted_xml(struct sip_msg *msg, char* lie, char* callidHeader, char* cb
 	char* xml;
 	char formated_time[80];
 	time_t rawtime;
-	struct tm * timeinfo;
+	struct tm timeinfo;
 	struct service_provider* source_provider;
 	struct service_provider* vpc_provider;
 	struct service_provider* vsp_provider;
@@ -2026,8 +2053,8 @@ char* formatted_xml(struct sip_msg *msg, char* lie, char* callidHeader, char* cb
 	vsp_cert_uri = empty;
 
 	time(&rawtime);
-	timeinfo = localtime(&rawtime);
-	strftime(formated_time, 80, "%Y-%m-%dT%H:%M:%S%Z", timeinfo);
+	localtime_r(&rawtime, &timeinfo);
+	strftime(formated_time, 80, "%Y-%m-%dT%H:%M:%S%Z", &timeinfo);
 	LM_DBG(" --- INIT  send_request_vpc\n \n");
 	LM_DBG(" --- FORMAT XML \n \n");
 
