@@ -45,22 +45,28 @@
 #include "../../sr_module.h"
 #include "../../str.h"
 #include "../../ut.h"
-#include "../../sliblist.h"
+#include "../../lib/sliblist.h"
+#include "../../reactor_proc.h"
 #include "httpd_load.h"
+#include "httpd_proc.h"
 
 
 extern int port;
 extern str ip;
 extern str buffer;
 extern int post_buf_size;
+extern str tls_cert_file;
+extern str tls_key_file;
+extern str tls_ciphers;
 extern struct httpd_cb *httpd_cb_list;
+static union sockaddr_union httpd_server_info;
 
 static const str MI_HTTP_U_URL = str_init("<html><body>"
 "Unable to parse URL!</body></html>");
 static const str MI_HTTP_U_METHOD = str_init("<html><body>"
 "Unsupported HTTP request!</body></html>");
-static const str MI_HTTP_U_CNT_TYPE = str_init("<html><body>"
-"Unsupported Content-Type!</body></html>");
+
+static char * load_file(char * );
 
 /**
  * Data structure to store inside elents of slinkedl_list list.
@@ -192,7 +198,7 @@ skip:
  * Handle regular POST data.
  *
  */
-static int post_iterator (void *cls,
+static MHD_RET post_iterator (void *cls,
 		enum MHD_ValueKind kind,
 		const char *key,
 		const char *filename,
@@ -283,7 +289,7 @@ static int post_iterator (void *cls,
  * @return MHD_YES to continue iterating,
  *         MHD_NO to abort the iteration.
  */
-int getConnectionHeader(void *cls, enum MHD_ValueKind kind,
+MHD_RET getConnectionHeader(void *cls, enum MHD_ValueKind kind,
 					const char *key, const char *value)
 {
 	struct post_request *pr = (struct post_request*)cls;
@@ -322,8 +328,14 @@ int getConnectionHeader(void *cls, enum MHD_ValueKind kind,
 			pr->content_type = HTTPD_TEXT_XML_CNT_TYPE;
 		else if (strncasecmp("application/json", value, 16) == 0)
 			pr->content_type = HTTPD_APPLICATION_JSON_CNT_TYPE;
-		else
+		else if (strncasecmp("text/html", value, 9) == 0)
+			pr->content_type = HTTPD_TEXT_HTML_TYPE;
+		else if (strncasecmp("text/plain", value, 10) == 0)
+			pr->content_type = HTTPD_TEXT_PLAIN_TYPE;
+		else {
 			pr->content_type = HTTPD_UNKNOWN_CNT_TYPE;
+			LM_ERR("Unexpected Content-Type=[%s]\n", value);
+		}
 		if (p) *p = bk;
 		goto done;
 	}
@@ -339,6 +351,7 @@ int getConnectionHeader(void *cls, enum MHD_ValueKind kind,
 		goto done;
 	}
 
+	LM_DBG("key=[%s] value=[%s]\n", key, value);
 	return MHD_YES;
 
 done:
@@ -389,8 +402,13 @@ void httpd_lookup_arg(void *connection, const char *key,
 	return;
 }
 
+union sockaddr_union* httpd_get_server_info(void)
+{
+	return &httpd_server_info;
+}
 
-int answer_to_connection (void *cls, struct MHD_Connection *connection,
+/* 0x00097001 changed the returned result */
+MHD_RET answer_to_connection (void *cls, struct MHD_Connection *connection,
 		const char *url, const char *method,
 		const char *version, const char *upload_data,
 		size_t *upload_data_size, void **con_cls)
@@ -399,14 +417,19 @@ int answer_to_connection (void *cls, struct MHD_Connection *connection,
 	struct MHD_Response *response;
 	int ret;
 	void *async_data = NULL;
-	struct httpd_cb *cb;
+	struct httpd_cb *cb = NULL;
 	const char *normalised_url;
 	struct post_request *pr;
 	str_str_t *kv;
 	char *p;
-	int cnt_type = HTTPD_STD_CNT_TYPE;
-	int accept_type = HTTPD_STD_CNT_TYPE;
 	int ret_code = MHD_HTTP_OK;
+	str saved_body = STR_NULL;
+
+#if ( MHD_VERSION >= 0x000092800 )
+	int sv_sockfd;
+	socklen_t addrlen=sizeof(httpd_server_info);
+#endif
+	union sockaddr_union* cl_socket;
 
 	LM_DBG("START *** cls=%p, connection=%p, url=%s, method=%s, "
 			"versio=%s, upload_data[%zu]=%p, *con_cls=%p\n",
@@ -425,6 +448,26 @@ int answer_to_connection (void *cls, struct MHD_Connection *connection,
 		pr = NULL;
 	}
 
+	/* we're safe here since this returns a struct sockaddr* and
+	 * sockaddr_union contains sockaddr* inside */
+	cl_socket = (union sockaddr_union *)MHD_get_connection_info(connection,
+			MHD_CONNECTION_INFO_CLIENT_ADDRESS)->client_addr;
+
+#if ( MHD_VERSION >= 0x000092800 )
+	sv_sockfd = MHD_get_connection_info(connection, MHD_CONNECTION_INFO_CONNECTION_FD)->connect_fd;
+	if (getsockname( sv_sockfd, &httpd_server_info.s, &addrlen) < 0) {
+		LM_ERR("cannot resolve server's IP: %s:%d\n", strerror(errno), errno);
+		return MHD_NO;
+	}
+
+	/* we could do
+	 * httpd_server_info.sin.sin_port = ntohs(httpd_server_info.sin.sin_port);
+	 * but it has no sense since we already know the port since we initialised
+	 * httpd server
+	 */
+	httpd_server_info.sin.sin_port = port;
+#endif
+
 	if(strncmp(method, "POST", 4)==0) {
 		if(pr == NULL){
 			pr = *con_cls;
@@ -438,74 +481,27 @@ int answer_to_connection (void *cls, struct MHD_Connection *connection,
 											post_buf_size,
 											&post_iterator,
 											pr);
-			if(pr->pp==NULL) {
-				if (*upload_data_size == 0) {
-					/* We need to wait for morte data before
-					 * handling the POST request */
-					return MHD_YES;
-				}
-				LM_DBG("NOT a regular POST :o)\n");
-				if (pr->content_type==0 && pr->content_len==0)
-					MHD_get_connection_values(connection, MHD_HEADER_KIND,
-											&getConnectionHeader, pr);
-				if (pr->content_type<=0 || pr->content_len<=0) {
-					LM_ERR("got a bogus request\n");
-					return MHD_NO;
-				}
-				if (*upload_data_size != pr->content_len) {
-					/* For now, we don't support large POST with truncated data */
-					LM_ERR("got a truncated POST request\n");
-					return MHD_NO;
-				}
-				LM_DBG("got ContentType [%d] with len [%d]: %.*s\\n",
-					pr->content_type, pr->content_len,
-					(int)*upload_data_size, upload_data);
-				/* Here we save data. */
-				switch (pr->content_type) {
-				case HTTPD_TEXT_XML_CNT_TYPE:
-				case HTTPD_APPLICATION_JSON_CNT_TYPE:
-					/* Save the entire body as 'body' */
-					kv = (str_str_t*)slinkedl_append(pr->p_list,
-							sizeof(str_str_t) + 1 +
-							*upload_data_size);
-					p = (char*)(kv + 1);
-					kv->key.len = 1; kv->key.s = p;
-					memcpy(p, "1", 1);
-					p += 1;
-					kv->val.len = *upload_data_size;
-					kv->val.s = p;
-					memcpy(p, upload_data, *upload_data_size);
-					break;
-				default:
-					LM_ERR("Unhandled data for ContentType [%d]\n",
-							pr->content_type);
-					return MHD_NO;
-				}
-				/* Mark the fact that we consumed all data */
-				*upload_data_size = 0;
-				return MHD_YES;
-			}
-
 			LM_DBG("pr=[%p] pp=[%p] p_list=[%p]\n",
 					pr, pr->pp, pr->p_list);
+
+			/* We need to wait for the actual data in the POST request */
 			return MHD_YES;
 		} else {
 			if (pr->pp==NULL) {
 				if (*upload_data_size == 0) {
-					if (pr->content_type==HTTPD_TEXT_XML_CNT_TYPE)
-						cnt_type = HTTPD_TEXT_XML_CNT_TYPE;
-					if (pr->content_type==HTTPD_APPLICATION_JSON_CNT_TYPE)
-						cnt_type = HTTPD_APPLICATION_JSON_CNT_TYPE;
 					*con_cls = pr->p_list;
 					cb = get_httpd_cb(url);
 					if (cb) {
 						normalised_url = &url[cb->http_root->len+1];
 						LM_DBG("normalised_url=[%s]\n", normalised_url);
+						kv = slinkedl_peek(pr->p_list);
+						if (kv)
+							saved_body = ((str_str_t *)kv)->val;
 						ret_code = cb->callback(cls, (void*)connection,
 								normalised_url,
 								method, version,
-								upload_data, upload_data_size, con_cls,
-								&buffer, &page);
+								saved_body.s, saved_body.len,
+								con_cls, &buffer, &page, cl_socket);
 					} else {
 						page = MI_HTTP_U_URL;
 						ret_code = MHD_HTTP_BAD_REQUEST;
@@ -520,8 +516,13 @@ int answer_to_connection (void *cls, struct MHD_Connection *connection,
 				if (pr->content_type==0 && pr->content_len==0)
 					MHD_get_connection_values(connection, MHD_HEADER_KIND,
 											&getConnectionHeader, pr);
-				if (pr->content_type<=0 || pr->content_len<=0) {
-					LM_ERR("got a bogus request\n");
+				if (pr->content_type==0) {
+					LM_ERR("missing Content-Type header\n");
+					return MHD_NO;
+				}
+				if (pr->content_type<0) {
+					/* Unexpected Content-Type header:
+					err log printed in getConnectionHeader() */
 					return MHD_NO;
 				}
 				if (*upload_data_size != pr->content_len) {
@@ -529,14 +530,14 @@ int answer_to_connection (void *cls, struct MHD_Connection *connection,
 					LM_ERR("got a truncated POST request\n");
 					return MHD_NO;
 				}
-				LM_DBG("got ContentType [%d] with len [%d]: %.*s\\n",
+				LM_DBG("got ContentType [%d] with len [%d]: %.*s\n",
 					pr->content_type, pr->content_len,
 					(int)*upload_data_size, upload_data);
 				/* Here we save data. */
 				switch (pr->content_type) {
 				case HTTPD_TEXT_XML_CNT_TYPE:
 				case HTTPD_APPLICATION_JSON_CNT_TYPE:
-					/* Save the entire body as 'body' */
+					/* Save the entire body as the '1' key */
 					kv = (str_str_t*)slinkedl_append(pr->p_list,
 							sizeof(str_str_t) + 1 +
 							*upload_data_size);
@@ -590,8 +591,8 @@ int answer_to_connection (void *cls, struct MHD_Connection *connection,
 				ret_code = cb->callback(cls, (void*)connection,
 						normalised_url,
 						method, version,
-						upload_data, upload_data_size, con_cls,
-						&buffer, &page);
+						upload_data, *upload_data_size, con_cls,
+						&buffer, &page, cl_socket);
 			} else {
 				page = MI_HTTP_U_URL;
 				ret_code = MHD_HTTP_BAD_REQUEST;
@@ -604,9 +605,7 @@ int answer_to_connection (void *cls, struct MHD_Connection *connection,
 		pr = *con_cls;
 		MHD_get_connection_values(connection, MHD_HEADER_KIND,
 								&getConnectionHeader, pr);
-		accept_type = pr->accept_type;
 		pkg_free(pr); *con_cls = pr = NULL;
-		LM_DBG("accept_type=[%d]\n", accept_type);
 		cb = get_httpd_cb(url);
 		if (cb) {
 			normalised_url = &url[cb->http_root->len+1];
@@ -614,40 +613,69 @@ int answer_to_connection (void *cls, struct MHD_Connection *connection,
 			ret_code = cb->callback(cls, (void*)connection,
 					normalised_url,
 					method, version,
-					upload_data, upload_data_size, con_cls,
-					&buffer, &page);
+					upload_data, *upload_data_size, con_cls,
+					&buffer, &page, cl_socket);
 		} else {
 			page = MI_HTTP_U_URL;
 			ret_code = MHD_HTTP_BAD_REQUEST;
 		}
 	}else{
 		page = MI_HTTP_U_METHOD;
+#ifdef MHD_HTTP_NOT_ACCEPTABLE
+		ret_code = MHD_HTTP_NOT_ACCEPTABLE;
+#else
 		ret_code = MHD_HTTP_METHOD_NOT_ACCEPTABLE;
+#endif
 	}
 
 send_response:
 	if (page.s) {
-		LM_DBG("MHD_create_response_from_data [%p:%d]\n",
-			page.s, page.len);
+#if defined MHD_VERSION && MHD_VERSION >= 0x00090000
+		response = MHD_create_response_from_buffer(page.len,
+							(void*)page.s,
+							MHD_RESPMEM_MUST_COPY);
+#else
+		/* use old constructor */
 		response = MHD_create_response_from_data(page.len,
 							(void*)page.s,
 							0, 1);
-	} else {
+#endif
+		LM_DBG("MHD_create_response_from_data [%p:%d]\n",
+			page.s, page.len);
+	} else if (cb) {
 		LM_DBG("MHD_create_response_from_callback\n");
 		response = MHD_create_response_from_callback (MHD_SIZE_UNKNOWN,
 							buffer.len,
 							(MHD_ContentReaderCallback)cb->flush_data_callback,
 							(void*)async_data,
 							NULL);
+	} else {
+		return MHD_NO;
+	}
+
+	if (cb && cb->type>0) {
+		if (cb->type==HTTPD_TEXT_XML_CNT_TYPE)
+			MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_TYPE,
+				"text/xml; charset=utf-8");
+		else if (cb->type==HTTPD_APPLICATION_JSON_CNT_TYPE)
+			MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_TYPE,
+				"application/json");
+		else if (cb->type==HTTPD_TEXT_HTML_TYPE)
+			MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_TYPE,
+				"text/html");
+		else if (cb->type==HTTPD_TEXT_PLAIN_TYPE)
+			MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_TYPE,
+				"text/plain");
+		else if (cb->type==HTTPD_TEXT_PLAIN_PROMETHEUS_TYPE)
+			MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_TYPE,
+				"text/plain; version=0.0.4");
+		else
+			LM_BUG("unhandled content type %d\n",cb->type);
+	} else {
+		/* 'page' for sure contains some HTML error we pushed */
+		MHD_add_response_header(response, MHD_HTTP_HEADER_CONTENT_TYPE,
+				"text/html");
 	}
-	if (cnt_type==HTTPD_TEXT_XML_CNT_TYPE || accept_type==HTTPD_TEXT_XML_CNT_TYPE)
-		MHD_add_response_header(response,
-								MHD_HTTP_HEADER_CONTENT_TYPE,
-								"text/xml; charset=utf-8");
-	if (cnt_type==HTTPD_APPLICATION_JSON_CNT_TYPE || accept_type==HTTPD_APPLICATION_JSON_CNT_TYPE)
-		MHD_add_response_header(response,
-								MHD_HTTP_HEADER_CONTENT_TYPE,
-								"application/json");
 	ret = MHD_queue_response (connection, ret_code, response);
 	MHD_destroy_response (response);
 
@@ -655,15 +683,28 @@ send_response:
 }
 #endif
 
+
+#ifdef LIBMICROHTTPD
+static fd_set mhd_rs;
+static fd_set mhd_ws;
+static fd_set mhd_es;
+
+int httpd_callback(int fd, void *dmn, int was_timeout)
+{
+	int status;
+	status = MHD_run_from_select((struct MHD_Daemon *)dmn,
+		&mhd_rs, &mhd_ws, &mhd_es);
+	if (status == MHD_NO) {
+		LM_ERR("failed to run http daemon\n");
+		return -1;
+	}
+	return 0;
+}
+#endif
+
+
 void httpd_proc(int rank)
 {
-#ifdef LIBMICROHTTPD
-	int status;
-	fd_set rs;
-	fd_set ws;
-	fd_set es;
-	int max;
-#endif
 	struct httpd_cb *cb = httpd_cb_list;
 
 	/*child's initial settings*/
@@ -673,7 +714,7 @@ void httpd_proc(int rank)
 	}
 
 	/* Allocating http response buffer */
-	buffer.s = (char*)pkg_malloc(sizeof(char)*buffer.len);
+	buffer.s = (char*)malloc(sizeof(char)*buffer.len);
 	if (buffer.s==NULL) {
 		LM_ERR("oom\n");
 		return;
@@ -686,23 +727,127 @@ void httpd_proc(int rank)
 	}
 
 #ifdef LIBMICROHTTPD
-	struct timeval tv;
-	struct sockaddr_in saddr_in;
+	unsigned int mhd_flags = MHD_USE_DEBUG;
+	unsigned int mhdi_flags;
+	int mhd_opt_n = 0;
+	char *key_pem, *cert_pem, *ip_repr, ip6buf[1+IP_ADDR_MAX_STR_SIZE+1];
+	void *saddr;
+	struct sockaddr_in saddr4;
+	struct MHD_OptionItem mhd_opts[4];
+	const union MHD_DaemonInfo *dmni;
+	int fd;
 
-	memset(&saddr_in, 0, sizeof(saddr_in));
-	if (ip.s)
-		saddr_in.sin_addr.s_addr = inet_addr(ip.s);
-	else
-		saddr_in.sin_addr.s_addr = INADDR_ANY;
-	saddr_in.sin_family = AF_INET;
-	saddr_in.sin_port = htons(port);
+	if (tls_key_file.s && tls_cert_file.s) {
+
+		key_pem = load_file(tls_key_file.s);
+		if (NULL == key_pem) {
+			LM_ERR("unable to load tls key\n");
+			return;
+		}
+
+  		cert_pem = load_file(tls_cert_file.s);
+		if (NULL == cert_pem) {
+			LM_ERR("unable to load tls certificate\n");
+			return;
+		}
+
+		mhd_flags = mhd_flags | MHD_USE_SSL;
+
+		mhd_opts[mhd_opt_n].option = MHD_OPTION_HTTPS_MEM_KEY;
+		mhd_opts[mhd_opt_n].value = 0;
+		mhd_opts[mhd_opt_n].ptr_value = key_pem;
+		mhd_opt_n++;
+		mhd_opts[mhd_opt_n].option = MHD_OPTION_HTTPS_MEM_CERT;
+		mhd_opts[mhd_opt_n].value = 0;
+		mhd_opts[mhd_opt_n].ptr_value = cert_pem;
+		mhd_opt_n++;
+		mhd_opts[mhd_opt_n].option = MHD_OPTION_HTTPS_PRIORITIES;
+		mhd_opts[mhd_opt_n].ptr_value = tls_ciphers.s;
+		mhd_opts[mhd_opt_n].value = 0;
+		mhd_opt_n++;
+
+	} else
+		mhd_flags = mhd_flags | MHD_NO_FLAG;
+
+	mhd_opts[mhd_opt_n].option = MHD_OPTION_END;
+	mhd_opts[mhd_opt_n].value = 0;
+	mhd_opts[mhd_opt_n].ptr_value = NULL;
+
+	#if (MHD_VERSION <= 0x00095000)
+	mhd_flags = mhd_flags | MHD_USE_EPOLL_LINUX_ONLY;
+	#else
+	mhd_flags = mhd_flags | MHD_USE_EPOLL;
+	#endif
+
+#if ( MHD_VERSION >= 0x000092800 )
+	struct sockaddr_in6 saddr6;
+	if (ip.s && strcmp(ip.s, "*")) {
+		if (q_memchr(ip.s, ':', ip.len)) {
+			LM_DBG("preparing to listen on IPv6 interface '%s'\n", ip.s);
+			memset(&saddr6, 0, sizeof saddr6);
+
+			if (inet_pton(AF_INET6, ip.s, &saddr6.sin6_addr) <= 0) {
+				LM_ERR("failed to parse 'ip' modparam: %s\n", ip.s);
+				return;
+			}
+
+			saddr6.sin6_family = AF_INET6;
+			saddr6.sin6_port = htons(port);
+			saddr = &saddr6;
+			sprintf(ip6buf, "[%s]", !strcmp(ip.s, "::0") ? "::" : ip.s);
+			ip_repr = ip6buf;
+			mhd_flags |= MHD_USE_IPv6;
+		} else {
+			LM_DBG("preparing to listen on IPv4 interface '%s'\n", ip.s);
+			memset(&saddr4, 0, sizeof saddr4);
+
+			if (inet_pton(AF_INET, ip.s, &saddr4.sin_addr) <= 0) {
+				LM_ERR("failed to parse 'ip' modparam: %s\n", ip.s);
+				return;
+			}
+
+			saddr4.sin_family = AF_INET;
+			saddr4.sin_port = htons(port);
+			saddr = &saddr4;
+			ip_repr = ip.s;
+		}
+	} else {
+		LM_DBG("preparing to listen on all IPv6 + IPv4 interfaces\n");
+		memset(&saddr6, 0, sizeof saddr6);
+
+		saddr6.sin6_addr = in6addr_any;
+		saddr6.sin6_family = AF_INET6;
+		saddr6.sin6_port = htons(port);
+		saddr = &saddr6;
+		ip_repr = "*";
+
+		mhd_flags |= MHD_USE_DUAL_STACK;
+	}
+#else
+	memset(&saddr4, 0, sizeof saddr4);
+	saddr4.sin_family = AF_INET;
+	saddr4.sin_port = htons(port);
+	saddr = &saddr4;
+
+	if (ip.s) {
+		saddr4.sin_addr.s_addr = inet_addr(ip.s);
+		ip_repr = ip.s;
+	} else {
+		saddr4.sin_addr.s_addr = INADDR_ANY;
+		ip_repr = "0.0.0.0";
+	}
+
+	memcpy( &httpd_server_info, &saddr, sizeof(struct sockaddr_in) );
+	httpd_server_info.sin.sin_port = port;
+#endif
 
 	LM_DBG("init_child [%d] - [%d] HTTP Server init [%s:%d]\n",
-		rank, getpid(), (ip.s?ip.s:"INADDR_ANY"), port);
-	set_proc_attrs("HTTPD %s:%d", (ip.s?ip.s:"INADDR_ANY"), port);
-	dmn = MHD_start_daemon(MHD_NO_FLAG|MHD_USE_DEBUG, port, NULL, NULL,
+		rank, getpid(), ip_repr, port);
+	set_proc_attrs("HTTPD %s:%d", ip_repr, port);
+	dmn = MHD_start_daemon(mhd_flags, port, NULL, NULL,
 			&(answer_to_connection), NULL,
-			MHD_OPTION_SOCK_ADDR, &saddr_in,
+			MHD_OPTION_SOCK_ADDR, saddr,
+			MHD_OPTION_ARRAY, mhd_opts,
 			MHD_OPTION_END);
 
 	if (NULL == dmn) {
@@ -710,40 +855,49 @@ void httpd_proc(int rank)
 		return;
 	}
 
-	while(1) {
-		max = 0;
-		FD_ZERO (&rs);
-		FD_ZERO (&ws);
-		FD_ZERO (&es);
-		if (MHD_YES != MHD_get_fdset (dmn, &rs, &ws, &es, &max)) {
-			LM_ERR("unable to get file descriptors\n");
-			return;
-		}
-		tv.tv_sec = 1;
-		tv.tv_usec = 0;
-		//LM_DBG("select(%d,%p,%p,%p,%p)\n",max+1, &rs, &ws, &es, &tv);
-		status = select(max+1, &rs, &ws, &es, &tv);
-		if (status < 0) {
-			switch(errno){
-				case EINTR:
-					LM_DBG("error returned by select:"
-							" [%d] [%d][%s]\n",
-							status, errno, strerror(errno));
-					break;
-				default:
-					LM_WARN("error returned by select:"
-							" [%d] [%d][%s]\n",
-							status, errno, strerror(errno));
-					return;
-			}
-		}
-		//LM_DBG("select returned %d\n", status);
-		status = MHD_run(dmn);
-		if (status == MHD_NO) {
-			LM_ERR("unable to run http daemon\n");
-			return;
-		}
+	/* as we use the MHD_USE_EPOLL (so libmicrohttp is internally using
+	 * epoll), the MHD_get_fdset() will return only the epoll controll fd.
+	 * This fd never changes and is set only for reading.
+	 * So, we "learn" that fd right now and bypass MHD_get_fdset() on
+	 * further iterations.*/
+
+	FD_ZERO (&mhd_rs);
+	FD_ZERO (&mhd_ws);
+	FD_ZERO (&mhd_es);
+	MHD_get_fdset (dmn, &mhd_rs, &mhd_ws, &mhd_es, NULL);
+
+#if (MHD_VERSION <= 0x00095100)
+	mhdi_flags = MHD_DAEMON_INFO_EPOLL_FD_LINUX_ONLY;
+#else
+	mhdi_flags = MHD_DAEMON_INFO_EPOLL_FD;
+#endif
+
+	dmni = MHD_get_daemon_info(dmn, mhdi_flags);
+	if (!dmni) {
+		LM_ERR("unable to get file descriptors\n");
+		return;
 	}
+#if (MHD_VERSION <= 0x00095200)
+	fd = dmni->listen_fd;
+#else
+	fd = dmni->epoll_fd;
+#endif
+
+	LM_DBG("found [%d] epoll ctl fd\n",fd);
+
+	if (reactor_proc_init( "HTTPD" )<0) {
+		LM_ERR("failed to init the HTTPD reactor\n");
+		return;
+	}
+
+	if (reactor_proc_add_fd( fd, httpd_callback, (void*)dmn)<0) {
+		LM_CRIT("failed to add Datagram listen socket to reactor\n");
+		return;
+	}
+
+	reactor_proc_loop();
+
+	/* we get here only if the "loop"-ing failed to start*/
 #endif
 	LM_DBG("HTTP Server stopped!\n");
 }
@@ -756,3 +910,33 @@ void httpd_proc_destroy(void)
 #endif
 	return;
 }
+
+static char * load_file(char * filename)
+{
+	FILE *f = fopen(filename, "rb");
+	if(!f)
+		return NULL;
+
+	fseek(f, 0, SEEK_END);
+	long fsize = ftell(f);
+
+	if(fsize == 0) {
+		fclose(f);
+		return NULL;
+	}
+
+	fseek(f, 0, SEEK_SET);
+
+	char *string = malloc(fsize + 1);
+	size_t bread;
+
+	bread = fread(string, 1, fsize, f);
+	if (bread == 0 || ferror(f))
+		LM_ERR("error while reading from %s (bytes read: %lu)\n",
+				filename, (unsigned long)bread);
+	fclose(f);
+
+	string[fsize] = 0;
+
+	return string;
+}
