@@ -57,31 +57,12 @@
  * Some functions are based on or copied from the source of the Textops module.
  */
 
-
-/**
- * make strptime available
- * use 600 for 'Single UNIX Specification, Version 3'
- */
-
-
-
 #include <fnmatch.h>
 
-#define _XOPEN_SOURCE 600          /* glibc2 on linux, bsd */
-#define _XOPEN_SOURCE_EXTENDED 1   /* solaris */
-
-/**
- * _XOPEN_SOURCE creates conflict in swab definition in Solaris
- */
-#ifdef __OS_solaris
-	#undef _XOPEN_SOURCE
-#endif
-
+/* make strptime available */
+#define _GNU_SOURCE
 #include <time.h>
 
-#undef _XOPEN_SOURCE
-#undef _XOPEN_SOURCE_EXTENDED
-
 #include <stdlib.h>
 #include <locale.h>
 #include <openssl/x509.h>
@@ -107,6 +88,9 @@
 #include "identity.h"
 
 
+#if OPENSSL_VERSION_NUMBER < 0x10100000L
+#define EVP_MD_CTX_free EVP_MD_CTX_cleanup
+#endif
 
 /* parameters */
 
@@ -141,17 +125,15 @@ static X509_STORE * store = NULL;
 /* needed for certificate verification */
 static X509_STORE_CTX * verify_ctx = NULL;
 
-
-/** module functions */
-static cmd_export_t cmds[]={
-	{"authservice",  (cmd_function)authservice_,  0, 0, 0,
+static const cmd_export_t cmds[]={
+	{"authservice",(cmd_function)authservice_, {{0,0,0}},
 		REQUEST_ROUTE | BRANCH_ROUTE | LOCAL_ROUTE},
-	{"verifier",     (cmd_function)verifier_,     0, 0, 0,
+	{"verifier",(cmd_function)verifier_, {{0,0,0}},
 		REQUEST_ROUTE},
-	{0,0,0,0,0,0}
+	{0,0,{{0,0,0}},0}
 };
 
-static param_export_t params[]={
+static const param_export_t params[]={
 	{"authCert", STR_PARAM, &authCert},
 	{"privKey", STR_PARAM, &privKey},
 	{"certUri", STR_PARAM, &certUri},
@@ -168,6 +150,7 @@ struct module_exports exports= {
 	MOD_TYPE_DEFAULT,/* class of this module */
 	MODULE_VERSION,
 	DEFAULT_DLFLAGS, /* dlopen flags */
+	0,				 /* load function */
 	NULL,            /* OpenSIPS module dependencies */
 	cmds, /* exported functions */
 	0,    /* exported async functions */
@@ -175,11 +158,14 @@ struct module_exports exports= {
 	0,          /* exported statistics */
 	0,          /* exported MI functions */
 	0,          /* exported pseudo-variables */
+	0,			/* exported transformations */
 	0,          /* local processes */
+	0,          /* module pre-initialization function */
 	mod_init,   /* module initialization function */
 	(response_function) 0, /* response function */
 	mod_destroy, /* destroy function */
-	0            /* per-child init function */
+	0,           /* per-child init function */
+	0            /* reload confirm function */
 };
 
 
@@ -519,6 +505,7 @@ static int addDate(char * dateHF, time_t * dateHFValue, struct sip_msg * msg)
 	char* buf;
 	size_t len = 0;
 	struct tm * bd_time = NULL;
+	struct tm bd_time_buff;
 
 	if(!dateHF || !dateHFValue || !msg)
 	{
@@ -528,7 +515,7 @@ static int addDate(char * dateHF, time_t * dateHFValue, struct sip_msg * msg)
 
 	*dateHFValue = time(0);
 
-	bd_time = gmtime(dateHFValue);
+	bd_time = gmtime_r(dateHFValue, &bd_time_buff);
 	if(!bd_time)
 	{
 		LM_ERR("gmtime failed\n");
@@ -831,7 +818,11 @@ static int addIdentity(char * dateHF, struct sip_msg * msg)
 {
 	#define IDENTITY_HDR_S  "Identity: \""
 	#define IDENTITY_HDR_L  (sizeof(IDENTITY_HDR_S)-1)
-	EVP_MD_CTX ctx;
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+	EVP_MD_CTX *pctx;
+#else
+	EVP_MD_CTX ctx, *pctx = &ctx;
+#endif
 	unsigned int siglen = 0;
 	int b64len = 0;
 	unsigned char * sig = NULL;
@@ -843,27 +834,30 @@ static int addIdentity(char * dateHF, struct sip_msg * msg)
 		LM_ERR("error making digest string\n");
 		return 0;
 	}
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+	pctx = EVP_MD_CTX_new();
+#endif
 
-	EVP_SignInit(&ctx, EVP_sha1());
+	EVP_SignInit(pctx, EVP_sha1());
 
-	EVP_SignUpdate(&ctx, digestString, strlen(digestString));
+	EVP_SignUpdate(pctx, digestString, strlen(digestString));
 
 	sig = pkg_malloc(EVP_PKEY_size(privKey_evp));
 	if(!sig)
 	{
-		EVP_MD_CTX_cleanup(&ctx);
+		EVP_MD_CTX_free(pctx);
 		LM_ERR("failed allocating memory\n");
 		return 0;
 	}
 
-	if(!EVP_SignFinal(&ctx, sig, &siglen, privKey_evp))
+	if(!EVP_SignFinal(pctx, sig, &siglen, privKey_evp))
 	{
-		EVP_MD_CTX_cleanup(&ctx);
+		EVP_MD_CTX_free(pctx);
 		pkg_free(sig);
 		LM_ERR("error calculating signature\n");
 		return 0;
 	}
-	EVP_MD_CTX_cleanup(&ctx);
+	EVP_MD_CTX_free(pctx);
 
 	/* ###Base64-encoding### */
 	/* annotation: The next few lines are based on example 7-11 of [VIE-02] */
@@ -1138,6 +1132,10 @@ static int checkAuthority(X509 * cert, struct sip_msg * msg)
 	const unsigned char * data;
 	STACK_OF(CONF_VALUE) * val;
 	CONF_VALUE * nval;
+	int len;
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+	ASN1_OCTET_STRING *adata;
+#endif
 
 	if(!cert || !msg)
 	{
@@ -1190,15 +1188,22 @@ static int checkAuthority(X509 * cert, struct sip_msg * msg)
 				LM_ERR("X509V3_EXT_get failed\n");
 				return 0;
 			}
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+			adata = X509_EXTENSION_get_data(cext);
+			data = ASN1_STRING_get0_data(adata);
+			len = ASN1_STRING_length(adata);
+#else
 			data = cext->value->data;
+			len = cext->value->length;
+#endif
 			if(meth->it)
 			{
 				ext_str = ASN1_item_d2i(NULL, &data,
-					cext->value->length, ASN1_ITEM_ptr(meth->it));
+					len, ASN1_ITEM_ptr(meth->it));
 			}
 			else
 			{
-				 ext_str = meth->d2i(NULL, &data, cext->value->length);
+				 ext_str = meth->d2i(NULL, &data, len);
 			}
 
 			val = meth->i2v(meth, ext_str, NULL);
@@ -1251,7 +1256,11 @@ static int checkSign(X509 * cert, char * identityHF, struct sip_msg * msg)
 	int siglen = -1;
 	unsigned char * sigbuf = NULL;
 	int b64len = 0;
-	EVP_MD_CTX ctx;
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+	EVP_MD_CTX *pctx;
+#else
+	EVP_MD_CTX ctx, *pctx = &ctx;
+#endif
 	int result = 0;
 	char *p;
 	unsigned long err;
@@ -1295,22 +1304,25 @@ static int checkSign(X509 * cert, char * identityHF, struct sip_msg * msg)
 	p=strstr(identityHF , "=");
 	siglen-=strspn(p , "=");
 
-	EVP_VerifyInit(&ctx, EVP_sha1());
-	EVP_VerifyUpdate(&ctx, digestString, strlen(digestString));
+#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+	pctx = EVP_MD_CTX_new();
+#endif
+	EVP_VerifyInit(pctx, EVP_sha1());
+	EVP_VerifyUpdate(pctx, digestString, strlen(digestString));
 
 	pubkey = X509_get_pubkey(cert);
 	if(!pubkey)
 	{
-		EVP_MD_CTX_cleanup(&ctx);
+		EVP_MD_CTX_free(pctx);
 		pkg_free(sigbuf);
 		LM_ERR("error reading pubkey from cert\n");
 		return 0;
 	}
 
-	result = EVP_VerifyFinal(&ctx, sigbuf, siglen, pubkey);
+	result = EVP_VerifyFinal(pctx, sigbuf, siglen, pubkey);
 
 	EVP_PKEY_free(pubkey);
-	EVP_MD_CTX_cleanup(&ctx);
+	EVP_MD_CTX_free(pctx);
 	pkg_free(sigbuf);
 
 	switch(result)
@@ -1715,8 +1727,9 @@ static int verify_callback(int ok, X509_STORE_CTX * stor)
 {
 	if (!ok)
 	{
+		int err = X509_STORE_CTX_get_error(stor);
 		LM_INFO("certificate validation failed: %s\n",
-			X509_verify_cert_error_string(stor->error));
+			X509_verify_cert_error_string(err));
 	}
 
 	return ok;
