@@ -17,11 +17,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
- *
- * History:
- * --------
- *  2009-02-01 initial version (bogdan)
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 
@@ -33,13 +29,21 @@
 #include "../../parser/parse_uri.h"
 #include "../../mem/shm_mem.h"
 #include "../../evi/evi.h"
+#include "../../rw_locking.h"
 #include "lb_parser.h"
 #include "lb_data.h"
+#include "lb_clustering.h"
 #include "lb_db.h"
 
 /* dialog stuff */
 extern struct dlg_binds lb_dlg_binds;
 
+extern int fetch_freeswitch_stats;
+extern int initial_fs_load;
+extern struct fs_binds fs_api;
+
+/* reader-writers lock for data reloading */
+rw_lock_t *ref_lock = NULL;
 
 
 struct lb_data* load_lb_data(void)
@@ -215,7 +219,7 @@ static int lb_set_resource_bitmask(struct lb_resource *res, unsigned int bit)
 
 
 int add_lb_dsturi( struct lb_data *data, int id, int group, char *uri,
-											char* resource, unsigned int flags)
+						char* resource, char* attrs, unsigned int flags)
 {
 	struct lb_res_str_list *lb_rl;
 	struct lb_res_str *r;
@@ -224,15 +228,17 @@ int add_lb_dsturi( struct lb_data *data, int id, int group, char *uri,
 	struct sip_uri puri;
 	struct proxy_l *proxy;
 	union sockaddr_union sau;
-	int len;
+	int uri_len, attrs_len;
 	int i;
+	str fs_url = { NULL, 0 };
+	str lb_str = { MI_SSTR("load_balancer") };
 
 	LM_DBG("uri=<%s>, grp=%d, res=<%s>\n",uri, group, resource);
 
 	/* check uri */
-	len = strlen(uri);
-	if(parse_uri(uri, len, &puri)!=0 ) {
-		LM_ERR("bad uri [%.*s] for destination\n", len, uri);
+	uri_len = strlen(uri);
+	if(parse_uri(uri, uri_len, &puri)!=0 ) {
+		LM_ERR("bad uri [%.*s] for destination\n", uri_len, uri);
 		return -1;
 	}
 
@@ -243,25 +249,31 @@ int add_lb_dsturi( struct lb_data *data, int id, int group, char *uri,
 		return -1;
 	}
 
+	attrs_len = attrs ? strlen(attrs) : 0;
+
 	/*add new destination */
 	dst = (struct lb_dst*)shm_malloc( sizeof(struct lb_dst)
-		+ lb_rl->n*sizeof(struct lb_resource_map) + len +
+		+ lb_rl->n*sizeof(struct lb_resource_map) + uri_len + attrs_len +
 		(3+2*sizeof(struct lb_dst*)));
 	if (dst==NULL) {
 		LM_ERR("failed to get shmem\n");
 		goto error;
 	}
 	memset( dst, 0, sizeof(struct lb_dst)+
-		lb_rl->n*sizeof(struct lb_resource_map) + len +
+		lb_rl->n*sizeof(struct lb_resource_map) + uri_len + attrs_len +
 		(3+2*sizeof(struct lb_dst*)) );
 
 	dst->rmap = (struct lb_resource_map*)(dst+1);
 
 	dst->uri.s = (char*)(dst->rmap + lb_rl->n);
-	dst->uri.len = len;
-	memcpy( dst->uri.s , uri, len);
+	dst->uri.len = uri_len;
+	memcpy( dst->uri.s , uri, uri_len);
 
-	dst->profile_id.s = dst->uri.s + len;
+	dst->attrs.s = dst->uri.s + uri_len;
+	dst->attrs.len = attrs_len;
+	memcpy(dst->attrs.s, attrs, attrs_len);
+
+	dst->profile_id.s = dst->attrs.s + attrs_len;
 	dst->profile_id.len = snprintf(dst->profile_id.s,
 		2+2*sizeof(struct lb_dst*), "%X", id);
 
@@ -273,8 +285,9 @@ int add_lb_dsturi( struct lb_data *data, int id, int group, char *uri,
 	/* add or update resource list */
 	for( i=0 ; i<lb_rl->n ; i++) {
 		r = lb_rl->resources + i;
-		LM_DBG(" setting for uri=<%s> (%d) resource=<%.*s>, val=%d\n",
-			uri, data->dst_no+1, r->name.len, r->name.s, r->val);
+		LM_DBG(" setting for uri=<%s> (%d) resource=<%.*s>, val=%d, fs=%.*s\n",
+			uri, data->dst_no+1, r->name.len, r->name.s, r->val,
+		    r->fs_url.len, r->fs_url.s);
 		res = get_resource_by_name( data, &r->name);
 		if (res==NULL) {
 			/* add new resource */
@@ -291,7 +304,13 @@ int add_lb_dsturi( struct lb_data *data, int id, int group, char *uri,
 		}
 		/* set the pointer and the max load */
 		dst->rmap[i].resource = res;
-		dst->rmap[i].max_load = r->val;
+		if (fetch_freeswitch_stats && r->fs_url.s) {
+			fs_url = r->fs_url;
+			dst->rmap[i].max_load = initial_fs_load;
+			dst->rmap[i].fs_enabled = 1;
+		} else {
+			dst->rmap[i].max_load = r->val;
+		}
 	}
 
 	/* Do a SIP wise DNS-Lookup for the domain part */
@@ -322,6 +341,13 @@ int add_lb_dsturi( struct lb_data *data, int id, int group, char *uri,
 	free_proxy(proxy);
 	pkg_free(proxy);
 
+	if (fetch_freeswitch_stats && fs_url.s && fs_url.len > 0) {
+		dst->fs_sock = fs_api.get_stats_evs(&fs_url, &lb_str);
+		if (!dst->fs_sock) {
+			LM_ERR("failed to create FreeSWITCH stats socket!\n");
+		}
+	}
+
 	/* link at the end */
 	if (data->last_dst==NULL) {
 		data->dsts = data->last_dst = dst;
@@ -344,6 +370,7 @@ void free_lb_data(struct lb_data *data)
 {
 	struct lb_resource *lbr1, *lbr2;
 	struct lb_dst *lbd1, *lbd2;
+	str lb_str = { MI_SSTR("load_balancer") };
 
 	if (data==NULL)
 		return;
@@ -365,6 +392,9 @@ void free_lb_data(struct lb_data *data)
 	for( lbd1=data->dsts ; lbd1 ; ) {
 		lbd2 = lbd1;
 		lbd1 = lbd1->next;
+		if (lbd2->fs_sock) {
+			fs_api.put_stats_evs(lbd2->fs_sock, &lb_str);
+		}
 		shm_free(lbd2);
 	}
 
@@ -410,8 +440,15 @@ static int get_dst_load(struct lb_resource **res, unsigned int res_no,
 }
 
 
+/* Performce the LB logic. It may return:
+ *   0 - success
+ *  -1 - generic error
+ *  -2 - no capacity (destinations may exist)
+ *  -3 - no destination at all
+ *  -4 - bad resources
+ */
 int lb_route(struct sip_msg *req, int group, struct lb_res_str_list *rl,
-						unsigned int flags, struct lb_data *data, int reuse)
+					unsigned int flags, struct lb_data *data, int reuse, str *attrs)
 {
 	/* resources for previous iteration */
 	static struct lb_resource **res_prev = NULL;
@@ -452,7 +489,7 @@ int lb_route(struct sip_msg *req, int group, struct lb_res_str_list *rl,
 	struct lb_dst *it_d;
 	struct lb_resource *it_r;
 	int load, it_l;
-	int i, j, cond;
+	int i, j, cond, cnt_aval_dst;
 
 
 	/* init control vars state */
@@ -489,7 +526,7 @@ int lb_route(struct sip_msg *req, int group, struct lb_res_str_list *rl,
 		if( i != res_new_n ) {
 			LM_ERR("initial call of LB - unknown resource found in "
 				"input string\n");
-			return -1;
+			return -4;
 		}
 
 		/* set 'res_new' as current iteration buffer */
@@ -529,7 +566,7 @@ int lb_route(struct sip_msg *req, int group, struct lb_res_str_list *rl,
 	if(
 		/* we should reuse previous resources list */
 		reuse ||
-		/* we have 'last_dst', i.e. previous iteration was successfull and
+		/* we have 'last_dst', i.e. previous iteration was successful and
 		 * we need to clean it up */
 		(last_dst != NULL)
 	) {
@@ -716,11 +753,13 @@ int lb_route(struct sip_msg *req, int group, struct lb_res_str_list *rl,
 	cond = 0; /* use it here as a 'first iteration' flag */
 	load = it_l = 0;
 	dsts_size_cur = 0;
+	cnt_aval_dst = 0;
 	for( it_d=data->dsts,i=0,j=0 ; it_d ; it_d=it_d->next ) {
 		if( it_d->group == group ) {
 			if( (dst_bitmap_cur[i] & (1 << j)) &&
 			((it_d->flags & LB_DST_STAT_DSBL_FLAG) == 0) ) {
 				/* valid destination (group & resources & status) */
+				cnt_aval_dst++;
 				if( get_dst_load(res_cur, res_cur_n, it_d, flags, &it_l) ) {
 					/* only valid load here */
 					if( (it_l > 0) || (flags & LB_FLAGS_NEGATIVE) ) {
@@ -731,6 +770,7 @@ int lb_route(struct sip_msg *req, int group, struct lb_res_str_list *rl,
 							dsts_size_cur = 0;
 						} else if( it_l < load ) {
 							/* lower availability -> new iteration */
+							if( ++j == (8 * sizeof(unsigned int)) ) { i++; j=0; }
 							continue;
 						}
 
@@ -863,30 +903,33 @@ int lb_route(struct sip_msg *req, int group, struct lb_res_str_list *rl,
 	/* add new resources */
 	for( i=0 ; i<res_cur_n ; i++ ) {
 		res_val.s = res_cur[i]->name;
-		if( add_avp(AVP_VAL_STR, res_avp_name, res_val) != 0 )
+		if( add_avp_last(AVP_VAL_STR, res_avp_name, res_val) != 0 )
 			LM_ERR("failed to add RES AVP\n");
 	}
 
 	/* outcome: set dst uri */
 	if( (dst != NULL) && (set_dst_uri(req, &dst->uri) != 0) ) {
 		LM_ERR("failed to set duri\n");
-		return -2;
+		return -1;
 	}
 
-	return dst ? 0 : -2;
+	if (dst && attrs)
+		*attrs = dst->attrs;
+
+	return dst ? 0 : (cnt_aval_dst? -2 /*no capacity*/ : -3 /* no dests*/ );
 }
 
 
 int do_lb_start(struct sip_msg *req, int group, struct lb_res_str_list *rl,
-									unsigned int flags, struct lb_data *data)
+						unsigned int flags, struct lb_data *data, str *attrs)
 {
-	return lb_route(req, group, rl, flags, data, 0/*no data reusage*/);
+	return lb_route(req, group, rl, flags, data, 0/*no data reusage*/, attrs);
 }
 
 
-int do_lb_next(struct sip_msg *req, struct lb_data *data)
+int do_lb_next(struct sip_msg *req, struct lb_data *data, str *attrs)
 {
-	return lb_route(req, -1, NULL, 0, data, 1/*reuse previous data*/);
+	return lb_route(req, -1, NULL, 0, data, 1/*reuse previous data*/, attrs);
 }
 
 
@@ -989,7 +1032,7 @@ int do_lb_disable_dst(struct sip_msg *req, struct lb_data *data, unsigned int ve
 				dst->flags |= LB_DST_STAT_DSBL_FLAG;
 
 				if( dst->flags != old_flags ) {
-					lb_raise_event(dst);
+					lb_status_changed(dst);
 					if( verbose )
 						LM_INFO("manually disable destination %d <%.*s> "
 							"from script\n",dst->id, dst->uri.len, dst->uri.s);
@@ -1007,37 +1050,16 @@ int do_lb_disable_dst(struct sip_msg *req, struct lb_data *data, unsigned int ve
 /* Checks, if the IP PORT is a LB destination
  */
 int lb_is_dst(struct lb_data *data, struct sip_msg *_m,
-				pv_spec_t *pv_ip, gparam_t *pv_port, int group, int active)
+				str *ip_str, int port, int group, int active, str *attrs)
 {
-	pv_value_t val;
 	struct ip_addr *ip;
-	int port;
 	struct lb_dst *dst;
 	int k;
 
-	/* get the address to test */
-	if (pv_get_spec_value( _m, pv_ip, &val)!=0) {
-		LM_ERR("failed to get IP value from PV\n");
+	if ( (ip=str2ip(ip_str))==NULL  && (ip=str2ip6(ip_str))==NULL) {
+		LM_ERR("IP val is not IP <%.*s>\n",ip_str->len,ip_str->s);
 		return -1;
 	}
-	if ( (val.flags&PV_VAL_STR)==0 ) {
-		LM_ERR("IP PV val is not string\n");
-		return -1;
-	}
-	if ( (ip=str2ip( &val.rs ))==NULL ) {
-		LM_ERR("IP val is not IP <%.*s>\n",val.rs.len,val.rs.s);
-		return -1;
-	}
-
-	/* get the port to test */
-	if (pv_port) {
-		if (fixup_get_ivalue(_m, (gparam_p)pv_port, &port) != 0) {
-			LM_ERR("failed to get PORT value from PV\n");
-			return -1;
-		}
-	} else {
-		port = 0;
-	}
 
 	/* and now search !*/
 	for( dst=data->dsts ; dst ; dst=dst->next) {
@@ -1048,6 +1070,8 @@ int lb_is_dst(struct lb_data *data, struct sip_msg *_m,
 			for(k=0 ; k<dst->ips_cnt ; k++ ) {
 				if ( (dst->ports[k]==0 || port==0 || port==dst->ports[k]) &&
 				ip_addr_cmp( ip, &dst->ips[k]) ) {
+					if (attrs)
+						*attrs = dst->attrs;
 					/* found */
 					return 1;
 				}
@@ -1207,3 +1231,15 @@ void lb_raise_event(struct lb_dst *dst)
 error:
 	evi_free_params(list);
 }
+
+
+void lb_status_changed(struct lb_dst *dst)
+{
+	/* do BIN replication if configured */
+	replicate_lb_status( dst );
+
+	/* raise the event */
+	lb_raise_event(dst);
+}
+
+
