@@ -40,10 +40,9 @@
 #include "../../mi/mi.h"
 #include "../../mem/mem.h"
 #include "../../mem/shm_mem.h"
+#include "../../reactor_proc.h"
 #include "mi_fifo.h"
 #include "fifo_fnc.h"
-#include "mi_parser.h"
-#include "mi_writer.h"
 
 static char *mi_buf = 0;
 static char *reply_fifo_s = 0;
@@ -52,9 +51,14 @@ static char *mi_fifo_name = NULL;
 static int mi_fifo_mode;
 static int mi_fifo_uid;
 static int mi_fifo_gid;
+static str backend = str_init("fifo");
 
 static int volatile mi_reload_fifo = 0;
 
+str correlation_value;
+extern int mi_trace_mod_id;
+int mi_fifo_pp;
+
 FILE* mi_create_fifo(void)
 {
 	static int  mi_fifo_read = 0;
@@ -64,19 +68,21 @@ FILE* mi_create_fifo(void)
 
 	/* create FIFO ... */
 	if ((mkfifo(mi_fifo_name, mi_fifo_mode)<0)) {
-		LM_ERR("can't create FIFO: %s (mode=%d)\n", strerror(errno), mi_fifo_mode);
+		LM_ERR("can't create FIFO: %s (mode=%o)\n", strerror(errno), mi_fifo_mode);
 		return 0;
 	}
 
 	LM_DBG("FIFO created @ %s\n", mi_fifo_name );
 
 	if ((chmod(mi_fifo_name, mi_fifo_mode)<0)) {
-		LM_ERR("can't chmod FIFO: %s (mode=%d)\n", strerror(errno), mi_fifo_mode);
+		LM_ERR("can't chmod FIFO: %s (mode=%o)\n", strerror(errno), mi_fifo_mode);
 		return 0;
 	}
 
 	if ((mi_fifo_uid!=-1) || (mi_fifo_gid!=-1)){
 		if (chown(mi_fifo_name, mi_fifo_uid, mi_fifo_gid)<0){
+
+#include "../../mi/mi_trace.h"
 			LM_ERR("failed to change the owner/group for %s to %d.%d; %s[%d]\n",
 				mi_fifo_name, mi_fifo_uid, mi_fifo_gid, strerror(errno), errno);
 			return 0;
@@ -137,7 +143,7 @@ FILE* mi_init_fifo_server(char *fifo_name, int fifo_mode,
 
 
 	/* allocate all static buffers */
-	mi_buf = pkg_malloc(MAX_MI_FIFO_BUFFER);
+	mi_buf = pkg_malloc(MAX_MI_FIFO_BUFFER + 1);
 	reply_fifo_s = pkg_malloc(MAX_MI_FILENAME);
 	if ( mi_buf==NULL|| reply_fifo_s==NULL) {
 		LM_ERR("no more private memory\n");
@@ -160,6 +166,7 @@ FILE* mi_init_fifo_server(char *fifo_name, int fifo_mode,
 
 	if (signal(SIGHUP, mi_sig_hup) == SIG_ERR ) {
 		LM_ERR("cannot install SIGHUP signal\n");
+		fclose(fifo_stream);
 		pkg_free(reply_fifo_s);
 		return 0;
 	}
@@ -269,7 +276,7 @@ tryagain:
 		if (errno==ENXIO) {
 			/* give up on the client - we can't afford server blocking */
 			if (retries==0) {
-				LM_ERR("no client at %s\n",pipe_name );
+				LM_NOTICE("no client at %s\n", pipe_name );
 				return 0;
 			}
 			/* don't be noisy on the very first try */
@@ -323,6 +330,59 @@ static FILE *mi_init_read(FILE *stream, int *fd, fd_set *fds)
 }
 
 
+static int mi_read_fifo(char *b, int max, FILE **stream, int *read_len)
+{
+	int ret = 0;
+	int done, i, fd;
+	struct timeval tv;
+	fd_set fds, init_fds;
+	FILE *new_stream;
+
+	/* first check if we need to update our fifo file */
+	if (!(new_stream = mi_init_read(*stream, &fd, &init_fds)))
+		return -1;
+
+	done = 0;
+	for (i = 0; !done && i < max; i++) {
+		fds = init_fds;
+		tv.tv_sec = FIFO_CHECK_WAIT;
+		tv.tv_usec = 0;
+retry:
+		ret = select(fd + 1, &fds, NULL, NULL, &tv);
+		if (ret < 0)  {
+			if (errno == EAGAIN)
+				goto retry;
+			/* interrupted by signal or ... */
+			if (errno == EINTR) {
+				if (!(new_stream = mi_init_read(new_stream, &fd, &init_fds)))
+					return -1;
+			} else {
+				kill(0, SIGTERM);
+			}
+		} else if (ret == 0) {
+			if (!(new_stream = mi_init_read(new_stream, &fd, &init_fds)))
+				return -1;
+			--i;
+			continue;
+		}
+		ret = read(fd, b, max);
+		if (ret < 0)
+			return ret;
+		else
+			done = 1;
+	}
+
+	if (!done) {
+		LM_ERR("request line too long\n");
+		fclose(new_stream);
+		return -1;
+	}
+	*read_len = ret;
+	*stream = new_stream;
+
+	return 0;
+}
+
 int mi_read_line( char *b, int max, FILE **stream, int *read_len)
 {
 	int ret = 0;
@@ -367,6 +427,7 @@ retry:
 
 	if (!done) {
 		LM_ERR("request line too long\n");
+		fclose(new_stream);
 		return -1;
 	}
 	*read_len = i;
@@ -379,7 +440,7 @@ retry:
 
 static inline char *get_reply_filename( char * file, int len )
 {
-	if ( strchr(file,'.') || strchr(file,'/') || strchr(file, '\\') ) {
+	if (memchr(file,'.',len) || memchr(file,'/',len) || memchr(file,'\\', len)) {
 		LM_ERR("forbidden filename: %s\n", file);
 		return 0;
 	}
@@ -392,234 +453,381 @@ static inline char *get_reply_filename( char * file, int len )
 	memcpy( reply_fifo_s+reply_fifo_len, file, len );
 	reply_fifo_s[reply_fifo_len+len]=0;
 
-
 	return reply_fifo_s;
 }
 
 
-static inline void free_async_handler( struct mi_handler *hdl )
+#define mi_open_reply(_name,_file,_err) \
+	do { \
+		if (!(_file) && (_name)) { \
+			_file = mi_open_reply_pipe( _name ); \
+			if (!(_file)) { \
+				LM_NOTICE("cannot open reply pipe %s\n", _name); \
+				goto _err; \
+			} \
+		} \
+	} while(0)
+
+#define mi_write_err2buf( _buf, _max_size, _err, ...) \
+	do { \
+		_buf.len = snprintf( _buf.s, _max_size, __VA_ARGS__); \
+		if ( _buf.len >= _max_size ) { \
+			LM_ERR("can't fit message in reply buffer!\n"); \
+			goto _err; \
+		} \
+	} while(0);
+
+#define mi_trace_fifo_request(method, params) \
+	do { \
+		if ((!cmd || is_mi_cmd_traced(mi_trace_mod_id, cmd))) { \
+			mi_trace_request(0, 0, method, strlen(method), \
+						params, &backend, t_dst); \
+		} \
+	} while(0)
+
+#define mi_trace_fifo_reply(message) \
+	do { \
+		if ((!cmd || is_mi_cmd_traced(mi_trace_mod_id, cmd))) { \
+			mi_trace_reply(0, 0, message, t_dst); \
+		} \
+	} while(0)
+
+#define mi_throw_error(_stream, _file, _err, _msg) \
+	do { \
+		if (_file) { \
+			str _s = str_init(_msg); \
+			mi_open_reply( _file, _stream, _err); \
+			if (mi_fifo_write(_file, _stream, &_s, cmd) < 0) { \
+				LM_ERR("cannot reply %s error\n", _msg); \
+				goto _err; \
+			} \
+			mi_trace_fifo_reply(&_s); \
+		} \
+	} while(0);
+
+struct mi_fifo_flush_params {
+	FILE *stream;
+	char *file;
+	struct mi_cmd *cmd;
+};
+
+static int mi_fifo_write(char *file, FILE *stream, str *msg, struct mi_cmd *cmd)
 {
-	if (hdl)
-		shm_free(hdl);
+	int ret, written;
+	FILE *old_stream = stream;
+
+	mi_open_reply(file, stream, error);
+
+	written = 0;
+	do {
+		ret = fwrite(msg->s + written, 1, msg->len - written, stream);
+		if (ret <= 0) {
+			if (errno!=EINTR && errno!=EAGAIN && errno!=EWOULDBLOCK)
+				goto error;
+		} else
+			written += ret;
+	} while (msg->len > written);
+
+	mi_trace_fifo_reply(msg);
+
+	if (!old_stream && stream)
+		fclose(stream);
+	return written;
+
+error:
+	if (!old_stream && stream)
+		fclose(stream);
+	return -1;
 }
 
-
-static void fifo_close_async( struct mi_root *mi_rpl, struct mi_handler *hdl,
-																	int done)
+static int mi_fifo_flush(unsigned char *buf, int len, void *param)
 {
-	FILE *reply_stream;
-	char *name;
+	str msg;
+	struct mi_fifo_flush_params *params =
+		(struct mi_fifo_flush_params *)param;
 
-	name = (char*)hdl->param;
+	if (!params)
+		return len;
 
-	if ( mi_rpl!=0 || done ) {
-		/*open fifo reply*/
-		reply_stream = mi_open_reply_pipe( name );
-		if (reply_stream==NULL) {
-			LM_ERR("cannot open reply pipe %s\n", name );
-			return;
-		}
+	msg.s = (char *)buf;
+	msg.len = len;
 
-		if (mi_rpl!=0) {
-			mi_write_tree( reply_stream, mi_rpl);
-			free_mi_tree( mi_rpl );
+	return mi_fifo_write(params->file, params->stream, &msg, params->cmd);
+}
+
+struct mi_async_param {
+	mi_item_t *id;
+	char *file;
+};
+
+static inline void free_async_handler(struct mi_handler *hdl)
+{
+	if (hdl) {
+		free_shm_mi_item(((struct mi_async_param *)hdl->param)->id);
+		shm_free(hdl);
+	}
+}
+
+static int mi_fifo_reply(FILE *reply_stream, char *file, str *buf,
+		mi_response_t *response, mi_item_t *id, struct mi_cmd *cmd)
+{
+	struct mi_fifo_flush_params params;
+
+	params.cmd = cmd;
+	params.file = file;
+	params.stream = reply_stream;
+
+	return print_mi_response_flush(response, id,
+			mi_fifo_flush, &params, buf, mi_fifo_pp);
+}
+
+static void fifo_close_async(mi_response_t *resp, struct mi_handler *hdl, int done)
+{
+	FILE *reply_stream = NULL;
+	struct mi_async_param *p;
+	int rc;
+	char buffer[MAX_MI_FIFO_BUFFER];
+	str buf;
+	buf.s = buffer;
+	buf.len = MAX_MI_FIFO_BUFFER;
+	struct mi_cmd *cmd = NULL; /* used by mi_throw_error */
+
+	p = (struct mi_async_param *)hdl->param;
+
+	if (resp || done) {
+		if (resp!=0) {
+			rc = mi_fifo_reply(NULL, p->file, &buf, resp, p->id, NULL);
+			if (rc == MI_NO_RPL) {
+				LM_DBG("No reply for jsonrpc notification\n");
+			} else if (rc < 0) {
+				LM_ERR("failed to print json response\n");
+				mi_throw_error(reply_stream, p->file, free_request,
+						"failed to print response");
+			} else
+				free_mi_response(resp);
 		} else {
-			mi_fifo_reply( reply_stream, "500 command failed\n");
+			mi_throw_error(reply_stream, p->file, free_request,
+					"failed to build response");
 		}
 
-		fclose(reply_stream);
+		if (reply_stream)
+			fclose(reply_stream);
 	}
 
+free_request:
 	if (done)
-		free_async_handler( hdl );
+		free_async_handler(hdl);
 	return;
 }
 
 
-static inline struct mi_handler* build_async_handler( char *name, int len)
+static inline struct mi_handler* build_async_handler(char *name, int len, mi_item_t *id)
 {
-	struct mi_handler *hdl;
-	char *p;
+	struct {
+		struct mi_handler hdl;
+		struct mi_async_param p;
+		char file[0];
+	} *buf;
 
-	hdl = (struct mi_handler*)shm_malloc( sizeof(struct mi_handler) + len + 1);
-	if (hdl==0) {
+	buf = shm_malloc(sizeof(*buf) + len + 1);
+	if (buf == NULL) {
 		LM_ERR("no more shared memory\n");
 		return 0;
 	}
+	buf->p.file = buf->file;
+	buf->p.id = shm_clone_mi_item(id);
 
-	p = (char*)(hdl) + sizeof(struct mi_handler);
-	memcpy( p, name, len+1 );
+	memcpy(buf->file, name, len+1 );
 
-	hdl->handler_f = fifo_close_async;
-	hdl->param = (void*)p;
+	buf->hdl.handler_f = fifo_close_async;
+	buf->hdl.param = &buf->p;
 
-	return hdl;
+	return &buf->hdl;
 }
 
 
-#define mi_do_consume() \
-	do { \
-		LM_DBG("entered consume\n"); \
-		/* consume the rest of the fifo request */ \
-		do { \
-			mi_read_line(mi_buf,MAX_MI_FIFO_BUFFER,&fifo_stream,&line_len); \
-		} while(line_len>1); \
-		LM_DBG("**** done consume\n"); \
-	} while(0)
+int mi_fifo_callback(int fd, void *fs, int was_timeout)
+{
+	static int remain_len = 0;
+	FILE *fifo_stream = (FILE*)fs;
+	const char *parse_end = NULL;
+	mi_request_t request;
+	int read_len, parse_len;
+	char *req_method = NULL;
+	char *file_sep, *file, *p, *f;
+	struct mi_cmd *cmd = NULL;
+	FILE *reply_stream;
+	struct mi_handler *hdl = NULL;
+	mi_response_t *response = NULL;
+	int rc;
+	str buf;
+
+	/* commands must look this way ':[filename]:' */
+	if (mi_read_fifo(mi_buf + remain_len,
+			MAX_MI_FIFO_BUFFER - remain_len,
+			&fifo_stream, &read_len)) {
+		LM_ERR("failed to read command\n");
+		goto skip_unparsed;
+	}
+	parse_len = remain_len + read_len;
+
+retry:
+	p = mi_buf;
+	reply_stream = NULL;
 
 
-#define mi_open_reply(_name,_file,_err) \
-	do { \
-		_file = mi_open_reply_pipe( _name ); \
-		if (_file==NULL) { \
-			LM_ERR("cannot open reply pipe %s\n", _name); \
-			goto _err; \
-		} \
-	} while(0)
+	while (parse_len && is_ws(*p)) {
+		p++;
+		parse_len--;
+	}
 
+	if (parse_len==0) {
+		LM_DBG("command file is empty\n");
+		goto skip_unparsed;
+	}
+	if (parse_len<3) {
+		LM_DBG("command must have at least 3 chars (has %d)\n", parse_len);
+		return -1;
+	}
+	if (*p!=MI_CMD_SEPARATOR) {
+		LM_ERR("command must begin with '%c': %.*s\n",
+			MI_CMD_SEPARATOR, parse_len, p);
+		goto skip_unparsed;
+	}
+	p++;
+	parse_len--;
+	file = p;
+	file_sep=memchr(p, MI_CMD_SEPARATOR , parse_len);
+	if (file_sep==NULL) {
+		LM_DBG("file separator missing: %.*s\n", read_len, mi_buf);
+		return 0;
+	}
+	if (file_sep - file + 1 >= parse_len) {
+		LM_DBG("no command specified yet: %.*s\n", read_len, mi_buf);
+		return -1;
+	}
+	p = file_sep + 1;
+	parse_len -= file_sep - file + 1;
+	if (file_sep==file) {
+		file = NULL; /* no reply expected */
+	} else {
+		f = get_reply_filename(file, file_sep - file);
+		if (f==NULL) {
+			LM_ERR("error trimming filename: %.*s\n",
+				(int)(file_sep - file), file);
+			goto skip_unparsed;
+		}
+		file = f;
+	}
+
+	/* make the command null terminated */
+	p[parse_len] = '\0';
+	memset(&request, 0, sizeof request);
+	if (parse_mi_request(p, &parse_end, &request) < 0) {
+		LM_ERR("cannot parse command: %.*s\n", parse_len, p);
+		return -1;
+	}
+
+	if (parse_end)
+		LM_DBG("running command [%.*s]\n", (int)(parse_end - p), p);
+	else
+		LM_DBG("running command [%s]\n", p);
+
+	if (parse_end && parse_len != parse_end - p) {
+		parse_len -= parse_end - p;
+		p = (char *)parse_end;
+		memmove(mi_buf, p, parse_len);
+	} else
+		parse_len = 0;
+	remain_len = parse_len;
+
+	req_method = mi_get_req_method(&request);
+	if (req_method)
+		cmd = lookup_mi_cmd(req_method, strlen(req_method));
+	/* if asyncron cmd, build the async handler */
+	if (cmd && cmd->flags&MI_ASYNC_RPL_FLAG) {
+		hdl = build_async_handler(file, strlen(file), request.id);
+		if (hdl==0) {
+			LM_ERR("failed to build async handler\n");
+
+			mi_throw_error(reply_stream, file, free_request,
+					"failed to build async handler");
+
+			goto free_request;
+		}
+	} else {
+		hdl = 0;
+		mi_open_reply( file, reply_stream, free_request);
+		if (!cmd)
+			LM_INFO("command %s not found!\n", req_method);
+	}
+
+	mi_trace_fifo_request(req_method, request.params);
+	response = handle_mi_request(&request, cmd, hdl);
+	LM_DBG("got mi response = [%p]\n", response);
+
+	if (response == NULL) {
+		LM_ERR("failed to build response!\n");
+		mi_throw_error(reply_stream, file, free_request,
+				"failed to build response");
+	} else if (response != MI_ASYNC_RPL) {
+		buf.s = mi_buf + remain_len;
+		buf.len = MAX_MI_FIFO_BUFFER - remain_len;
+		if (file) {
+			rc = mi_fifo_reply(reply_stream, file, &buf,
+					response, request.id, cmd);
+			if (rc == MI_NO_RPL) {
+				LM_DBG("No reply for jsonrpc notification\n");
+			} else if (rc < 0) {
+				LM_ERR("failed to print json response\n");
+				mi_throw_error(reply_stream, file, free_request,
+						"failed to print response");
+			} else
+				free_mi_response(response);
+		}
+		/* if there is no file specified, there is nothing to reply */
+	} else
+		goto end;
+
+free_request:
+	free_async_handler(hdl);
+	free_mi_request_parsed(&request);
+	if (reply_stream)
+		fclose(reply_stream);
+end:
+	if (parse_len)
+		goto retry;
+	return 0;
+skip_unparsed:
+	remain_len = 0;
+	return 0;
+}
 
 
 void mi_fifo_server(FILE *fifo_stream)
 {
-	struct mi_root *mi_cmd;
-	struct mi_root *mi_rpl;
-	struct mi_handler *hdl;
-	int line_len;
-	char *file_sep, *command, *file;
-	struct mi_cmd *f;
-	FILE *reply_stream;
+	int fd;
 
-	while(1) {
-		reply_stream = NULL;
-
-		/* commands must look this way ':<command>:[filename]' */
-		if (mi_read_line(mi_buf,MAX_MI_FIFO_BUFFER,&fifo_stream, &line_len)) {
-			LM_ERR("failed to read command\n");
-			continue;
-		}
-
-		/* trim from right */
-		while(line_len) {
-			if(mi_buf[line_len-1]=='\n' || mi_buf[line_len-1]=='\r'
-				|| mi_buf[line_len-1]==' ' || mi_buf[line_len-1]=='\t' ) {
-				line_len--;
-				mi_buf[line_len]=0;
-			} else break;
-		}
-
-		if (line_len==0) {
-			LM_DBG("command empty\n");
-			continue;
-		}
-		if (line_len<3) {
-			LM_ERR("command must have at least 3 chars\n");
-			continue;
-		}
-		if (*mi_buf!=MI_CMD_SEPARATOR) {
-			LM_ERR("command must begin with %c: %.*s\n", MI_CMD_SEPARATOR, line_len, mi_buf );
-			goto consume1;
-		}
-		command = mi_buf+1;
-		file_sep=strchr(command, MI_CMD_SEPARATOR );
-		if (file_sep==NULL) {
-			LM_ERR("file separator missing\n");
-			goto consume1;
-		}
-		if (file_sep==command) {
-			LM_ERR("empty command\n");
-			goto consume1;
-		}
-		if (*(file_sep+1)==0) {
-			file = NULL;
-		} else {
-			file = file_sep+1;
-			file = get_reply_filename(file, mi_buf+line_len-file);
-			if (file==NULL) {
-				LM_ERR("trimming filename\n");
-				goto consume1;
-			}
-		}
-		/* make command zero-terminated */
-		*file_sep=0;
-
-		f=lookup_mi_cmd( command, strlen(command) );
-		if (f==0) {
-			LM_ERR("command %s is not available\n", command);
-			mi_open_reply( file, reply_stream, consume1);
-			mi_fifo_reply( reply_stream, "500 command '%s' not available\n",
-				command);
-			goto consume2;
-		}
-
-		/* if asyncron cmd, build the async handler */
-		if (f->flags&MI_ASYNC_RPL_FLAG) {
-			hdl = build_async_handler( file, strlen(file) );
-			if (hdl==0) {
-				LM_ERR("failed to build async handler\n");
-				mi_open_reply( file, reply_stream, consume1);
-				mi_fifo_reply( reply_stream, "500 Internal server error\n");
-				goto consume2;
-			}
-		} else {
-			hdl = 0;
-			mi_open_reply( file, reply_stream, consume1);
-		}
-
-		if (f->flags&MI_NO_INPUT_FLAG) {
-			mi_cmd = 0;
-			mi_do_consume();
-		} else {
-			mi_cmd = mi_parse_tree(fifo_stream);
-			if (mi_cmd==NULL){
-				LM_ERR("error parsing MI tree\n");
-				if (!reply_stream)
-					mi_open_reply( file, reply_stream, consume3);
-				mi_fifo_reply( reply_stream, "400 parse error in "
-					"command '%s'\n", command);
-				goto consume3;
-			}
-			mi_cmd->async_hdl = hdl;
-		}
-
-		LM_DBG("done parsing the mi tree\n");
-
-		if ( (mi_rpl=run_mi_cmd(f, mi_cmd,
-		(mi_flush_f *)mi_flush_tree, reply_stream))==0 ) {
-			if (!reply_stream)
-				mi_open_reply( file, reply_stream, failure);
-			mi_fifo_reply(reply_stream, "500 command '%s' failed\n", command);
-			LM_ERR("command (%s) processing failed\n", command );
-		} else if (mi_rpl!=MI_ROOT_ASYNC_RPL) {
-			if (!reply_stream)
-				mi_open_reply( file, reply_stream, failure);
-			mi_write_tree( reply_stream, mi_rpl);
-			free_mi_tree( mi_rpl );
-		} else {
-			if (mi_cmd) free_mi_tree( mi_cmd );
-			continue;
-		}
-
-		free_async_handler(hdl);
-		/* close reply fifo */
-		fclose(reply_stream);
-		/* destroy request tree */
-		if (mi_cmd) free_mi_tree( mi_cmd );
-		continue;
-
-failure:
-		free_async_handler(hdl);
-		/* destroy request tree */
-		if (mi_cmd) free_mi_tree( mi_cmd );
-		/* destroy the reply tree */
-		if (mi_rpl) free_mi_tree(mi_rpl);
-		continue;
-
-consume3:
-		free_async_handler(hdl);
-		if (reply_stream)
-consume2:
-		fclose(reply_stream);
-consume1:
-		mi_do_consume();
+	if (reactor_proc_init( "MI FIFO" )<0) {
+		LM_ERR("failed to init the MI FIFO reactor\n");
+		return;
 	}
+
+	fd = fileno(fifo_stream);
+	if (fd<0) {
+		LM_ERR("failed to retriev fd from stream\n");
+		return;
+	}
+
+	if (reactor_proc_add_fd( fd, mi_fifo_callback, fifo_stream)<0) {
+		LM_CRIT("failed to add FIFO listen socket to reactor\n");
+		return;
+	}
+
+	reactor_proc_loop();
+
+	/* we get here only if the "loop"-ing failed to start*/
+
+	return;
 }
+
