@@ -125,14 +125,14 @@ static str ms_db_table = str_init("silo");
 str  ms_reminder = {NULL, 0};
 str  ms_outbound_proxy = {NULL, 0};
 
-char*  ms_from = NULL; /*"sip:registrar@example.org";*/
-char*  ms_contact = NULL; /*"Contact: <sip:registrar@example.org>\r\n";*/
-char*  ms_content_type = NULL; /*"Content-Type: text/plain\r\n";*/
-char*  ms_offline_message = NULL; /*"<em>I'm offline.</em>"*/
-void**  ms_from_sp = NULL;
-void**  ms_contact_sp = NULL;
-void**  ms_content_type_sp = NULL;
-void**  ms_offline_message_sp = NULL;
+str  ms_from = {0,0}; /*"sip:registrar@example.org";*/
+str  ms_contact = {0,0}; /*"Contact: <sip:registrar@example.org>\r\n";*/
+str  ms_content_type = {0,0}; /*"Content-Type: text/plain\r\n";*/
+str  ms_offline_message = {0,0}; /*"<em>I'm offline.</em>"*/
+pv_elem_t *ms_from_sp = NULL;
+pv_elem_t *ms_contact_sp = NULL;
+pv_elem_t *ms_content_type_sp = NULL;
+pv_elem_t *ms_offline_message_sp = NULL;
 
 int  ms_expire_time = 259200;
 int  ms_check_time = 60;
@@ -152,8 +152,8 @@ str msg_type = str_init("MESSAGE");
 static int mod_init(void);
 static int child_init(int);
 
-static int m_store(struct sip_msg*, char*, char*);
-static int m_dump(struct sip_msg*, char*, char*);
+static int m_store(struct sip_msg*, str*);
+static int m_dump(struct sip_msg* msg, str* owner, int* maxmsg);
 
 void destroy(void);
 
@@ -168,29 +168,25 @@ int check_message_support(struct sip_msg* msg);
 static void m_tm_callback( struct cell *t, int type, struct tmcb_params *ps);
 
 /* commands wrappers and fixups */
-static int fixup_m_dump(void** param, int param_no);
 
-static cmd_export_t cmds[]={
-	{"m_store",  (cmd_function)m_store, 0, 0, 0,
+static const cmd_export_t cmds[]={
+	{"m_store",  (cmd_function)m_store, {
+		{CMD_PARAM_STR|CMD_PARAM_OPT,0,0}, {0,0,0}},
 		REQUEST_ROUTE | FAILURE_ROUTE},
-	{"m_store",  (cmd_function)m_store, 1, fixup_spve_null, 0,
-		REQUEST_ROUTE | FAILURE_ROUTE},
-	{"m_dump",   (cmd_function)m_dump,  0, 0, 0,
+	{"m_dump",   (cmd_function)m_dump, {
+		{CMD_PARAM_STR|CMD_PARAM_OPT,0,0},
+		{CMD_PARAM_INT|CMD_PARAM_OPT,0,0},
+		{0,0,0}},
 		REQUEST_ROUTE | STARTUP_ROUTE | TIMER_ROUTE | EVENT_ROUTE},
-	{"m_dump",   (cmd_function)m_dump,  1, fixup_spve_null, 0,
-		REQUEST_ROUTE | STARTUP_ROUTE | TIMER_ROUTE | EVENT_ROUTE},
-	{"m_dump",   (cmd_function)m_dump,  2, fixup_m_dump, 0,
-		REQUEST_ROUTE | STARTUP_ROUTE | TIMER_ROUTE | EVENT_ROUTE},
-	{0,0,0,0,0,0}
+	{0,0,{{0,0,0}},0}
 };
 
-
-static param_export_t params[]={
+static const param_export_t params[]={
 	{ "db_url",           STR_PARAM, &ms_db_url.s             },
 	{ "db_table",         STR_PARAM, &ms_db_table.s           },
-	{ "from_address",     STR_PARAM, &ms_from                 },
-	{ "contact_hdr",      STR_PARAM, &ms_contact              },
-	{ "content_type_hdr", STR_PARAM, &ms_content_type         },
+	{ "from_address",     STR_PARAM, &ms_from.s                 },
+	{ "contact_hdr",      STR_PARAM, &ms_contact.s              },
+	{ "content_type_hdr", STR_PARAM, &ms_content_type.s         },
 	{ "offline_message",  STR_PARAM, &ms_offline_message      },
 	{ "reminder",         STR_PARAM, &ms_reminder.s           },
 	{ "outbound_proxy",   STR_PARAM, &ms_outbound_proxy.s     },
@@ -224,7 +220,7 @@ stat_var* ms_failed_msgs;
 stat_var* ms_dumped_rmds;
 stat_var* ms_failed_rmds;
 
-static stat_export_t msilo_stats[] = {
+static const stat_export_t msilo_stats[] = {
 	{"stored_messages" ,  0,  &ms_stored_msgs  },
 	{"dumped_messages" ,  0,  &ms_dumped_msgs  },
 	{"failed_messages" ,  0,  &ms_failed_msgs  },
@@ -234,7 +230,7 @@ static stat_export_t msilo_stats[] = {
 };
 #endif
 
-static dep_export_t deps = {
+static const dep_export_t deps = {
 	{ /* OpenSIPS module dependencies */
 		{ MOD_TYPE_DEFAULT, "tm", DEP_ABORT },
 		{ MOD_TYPE_SQLDB,   NULL, DEP_ABORT },
@@ -251,6 +247,7 @@ struct module_exports exports= {
 	MOD_TYPE_DEFAULT,/* class of this module */
 	MODULE_VERSION,
 	DEFAULT_DLFLAGS, /* dlopen flags */
+	0,				 /* load function */
 	&deps,           /* OpenSIPS module dependencies */
 	cmds,       /* module's exported functions */
 	0,          /* module's exported async functions */
@@ -262,11 +259,14 @@ struct module_exports exports= {
 #endif
 	0,          /* exported MI functions */
 	0,          /* exported pseudo-variables */
+	0,			/* exported transformations */
 	0,          /* extra processes */
+	0,          /* module pre-initialization function */
 	mod_init,   /* module initialization function */
 	(response_function) 0,       /* response handler */
 	(destroy_function) destroy,  /* module destroy function */
-	child_init  /* per-child init function */
+	child_init,  /* per-child init function */
+	0            /* reload confirm function */
 };
 
 /**
@@ -344,67 +344,39 @@ static int mod_init(void)
 		return -1;
 	}
 
-	if(ms_from!=NULL)
+	if(ms_from.s!=NULL)
 	{
-		ms_from_sp = (void**)pkg_malloc(sizeof(void*));
-		if(ms_from_sp==NULL)
-		{
-			LM_ERR("no more pkg\n");
-			return -1;
-		}
-		*ms_from_sp = (void*)ms_from;
-		if(fixup_spve_null(ms_from_sp, 1)!=0)
-		{
-			LM_ERR("bad contact parameter\n");
+		ms_from.len = strlen (ms_from.s);
+		if (pv_parse_format(&ms_from, &ms_from_sp) < 0) {
+			LM_ERR("Failed to parse formatted string in parameter\n");
 			return -1;
 		}
 	}
-	if(ms_contact!=NULL)
+	if(ms_contact.s!=NULL)
 	{
-		ms_contact_sp = (void**)pkg_malloc(sizeof(void*));
-		if(ms_contact_sp==NULL)
-		{
-			LM_ERR("no more pkg\n");
+		ms_contact.len = strlen (ms_contact.s);
+		if (pv_parse_format(&ms_contact, &ms_contact_sp) < 0) {
+			LM_ERR("Failed to parse formatted string in parameter\n");
 			return -1;
-		}
-		*ms_contact_sp = (void*)ms_contact;
-		if(fixup_spve_null(ms_contact_sp, 1)!=0)
-		{
-			LM_ERR("bad contact parameter\n");
+		}	
+	}
+	if(ms_content_type.s!=NULL)
+	{
+		ms_content_type.len = strlen (ms_content_type.s);
+		if (pv_parse_format(&ms_content_type, &ms_content_type_sp) < 0) {
+			LM_ERR("Failed to parse formatted string in parameter\n");
 			return -1;
 		}
 	}
-	if(ms_content_type!=NULL)
+	if(ms_offline_message.s!=NULL)
 	{
-		ms_content_type_sp = (void**)pkg_malloc(sizeof(void*));
-		if(ms_content_type_sp==NULL)
-		{
-			LM_ERR("no more pkg\n");
-			return -1;
-		}
-		*ms_content_type_sp = (void*)ms_content_type;
-		if(fixup_spve_null(ms_content_type_sp, 1)!=0)
-		{
-			LM_ERR("bad content_type parameter\n");
+		ms_offline_message.len = strlen (ms_offline_message.s);
+		if (pv_parse_format(&ms_offline_message, &ms_offline_message_sp) < 0) {
+			LM_ERR("Failed to parse formatted string in parameter\n");
 			return -1;
 		}
 	}
-	if(ms_offline_message!=NULL)
-	{
-		ms_offline_message_sp = (void**)pkg_malloc(sizeof(void*));
-		if(ms_offline_message_sp==NULL)
-		{
-			LM_ERR("no more pkg\n");
-			return -1;
-		}
-		*ms_offline_message_sp = (void*)ms_offline_message;
-		if(fixup_spve_null(ms_offline_message_sp, 1)!=0)
-		{
-			LM_ERR("bad offline_message parameter\n");
-			return -1;
-		}
-	}
-	if(ms_offline_message!=NULL && ms_content_type==NULL)
+	if(ms_offline_message.s!=NULL && ms_content_type.s==NULL)
 	{
 		LM_ERR("content_type parameter must be set\n");
 		return -1;
@@ -471,12 +443,12 @@ static int child_init(int rank)
  * 		= "2" -- look for outgoing URI only at to header
  */
 
-static int m_store(struct sip_msg* msg, char* owner, char* s2)
+static int m_store(struct sip_msg* msg, str* owner)
 {
 	str body, str_hdr, ctaddr;
 	struct to_body *pto, *pfrom;
 	struct sip_uri puri;
-	str duri, owner_s;
+	str duri;
 	db_key_t db_keys[NR_KEYS-1];
 	db_val_t db_vals[NR_KEYS-1];
 	db_key_t db_cols[1];
@@ -518,17 +490,12 @@ static int m_store(struct sip_msg* msg, char* owner, char* s2)
 	memset(&puri, 0, sizeof(struct sip_uri));
 	if(owner)
 	{
-		if(fixup_get_svalue(msg, (gparam_p)owner, &owner_s)!=0)
-		{
-			LM_ERR("invalid owner uri parameter");
-			return -1;
-		}
-		if(parse_uri(owner_s.s, owner_s.len, &puri)!=0)
+		if(parse_uri(owner->s, owner->len, &puri)!=0)
 		{
 			LM_ERR("bad owner SIP address!\n");
 			goto error;
 		} else {
-			LM_DBG("using user id [%.*s]\n", owner_s.len, owner_s.s);
+			LM_DBG("using user id [%.*s]\n", owner->len, owner->s);
 		}
 	} else { /* get it from R-URI */
 		if(msg->new_uri.len <= 0)
@@ -713,31 +680,27 @@ static int m_store(struct sip_msg* msg, char* owner, char* s2)
 	update_stat(ms_stored_msgs, 1);
 #endif
 
-	if(ms_from==NULL || ms_offline_message == NULL)
+	if(ms_from.s==NULL || ms_offline_message.s == NULL)
 		goto done;
 
 	LM_DBG("sending info message.\n");
-	if(fixup_get_svalue(msg, (gparam_p)*ms_from_sp, &notify_from)!=0
-			|| notify_from.len<=0)
+	if (pv_printf_s(msg, ms_from_sp, &notify_from) != 0)
 	{
 		LM_WARN("cannot get notification From address\n");
 		goto done;
 	}
-	if(fixup_get_svalue(msg, (gparam_p)*ms_offline_message_sp, &notify_body)!=0
-			|| notify_body.len<=0)
+	if (pv_printf_s(msg, ms_offline_message_sp, &notify_body) != 0)
 	{
 		LM_WARN("cannot get notification body\n");
 		goto done;
 	}
-	if(fixup_get_svalue(msg, (gparam_p)*ms_content_type_sp, &notify_ctype)!=0
-			|| notify_ctype.len<=0)
+	if (pv_printf_s(msg, ms_content_type_sp, &notify_ctype) != 0)
 	{
 		LM_WARN("cannot get notification content type\n");
 		goto done;
 	}
 
-	if(ms_contact!=NULL && fixup_get_svalue(msg, (gparam_p)*ms_contact_sp,
-				&notify_contact)==0 && notify_contact.len>0)
+	if (ms_contact.s!=NULL && pv_printf_s(msg, ms_contact_sp, &notify_contact) != 0)
 	{
 		if(notify_contact.len+notify_ctype.len>=MS_BUF1_SIZE)
 		{
@@ -799,7 +762,7 @@ error:
 /**
  * dump message
  */
-static int m_dump(struct sip_msg* msg, char* owner, char* maxmsg)
+static int m_dump(struct sip_msg* msg, str* owner, int* maxmsg)
 {
 	struct to_body *pto = NULL;
 	db_key_t db_keys[3];
@@ -809,12 +772,11 @@ static int m_dump(struct sip_msg* msg, char* owner, char* maxmsg)
 	db_key_t db_cols[6];
 	db_res_t* db_res = NULL;
 	int i, db_no_cols = 6, db_no_keys = 3, mid, n;
-	int sent_cnt = 0;
-	int maxmsg_i;
+	unsigned int sent_cnt = 0;
+	unsigned int maxmsg_i = 0;
 	static char hdr_buf[1024];
 	static char body_buf[1024];
 	struct sip_uri puri;
-	str owner_s;
 
 	str str_vals[4], hdr_str , body_str;
 	time_t rtime;
@@ -836,29 +798,24 @@ static int m_dump(struct sip_msg* msg, char* owner, char* maxmsg)
 	db_cols[4]=&sc_ctype;
 	db_cols[5]=&sc_inc_time;
 
-
 	hdr_str.s=hdr_buf;
 	hdr_str.len=1024;
 	body_str.s=body_buf;
 	body_str.len=1024;
 
-	maxmsg_i = (int)(long)maxmsg;
+	if (maxmsg)
+		maxmsg_i = *(unsigned int *)maxmsg;
 
 	/* get the owner */
 	memset(&puri, 0, sizeof(struct sip_uri));
 	if(owner)
 	{
-		if(fixup_get_svalue(msg, (gparam_p)owner, &owner_s)!=0)
-		{
-			LM_ERR("invalid owner uri parameter");
-			return -1;
-		}
-		if(parse_uri(owner_s.s, owner_s.len, &puri)!=0)
+		if(parse_uri(owner->s, owner->len, &puri)!=0)
 		{
 			LM_ERR("bad owner SIP address!\n");
 			goto error;
 		} else {
-			LM_DBG("using user id [%.*s]\n", owner_s.len, owner_s.s);
+			LM_DBG("using user id [%.*s]\n", owner->len, owner->s);
 		}
 	} else { /* get it from  To URI */
 		/* check for TO header */
@@ -982,30 +939,30 @@ static int m_dump(struct sip_msg* msg, char* owner, char* maxmsg)
 
 		/** sending using TM function: t_uac */
 		body_str.len = 1024;
-		n = m_build_body(&body_str, rtime, str_vals[2/*body*/], 0);
+		n = m_build_body(&body_str, rtime, str_vals[2/*body*/],
+			rtime /*Date*/, 0 /*not a reminder*/);
 		if(n<0)
 			LM_DBG("sending simple body\n");
 		else
 			LM_DBG("sending composed body\n");
 
-			tmb.t_request(&msg_type,  /* Type of the message */
-					&str_vals[1],     /* Request-URI (To) */
-					&str_vals[1],     /* To */
-					&str_vals[0],     /* From */
-					&hdr_str,         /* Optional headers including CRLF */
-					(n<0)?(RES_ROWS(db_res)[i].values[3].nul?NULL:&str_vals[2]):&body_str, /* Message body */
-					(ms_outbound_proxy.s)?&ms_outbound_proxy:0,
-									/* outbound uri */
-					m_tm_callback,    /* Callback function */
-					(void*)(long)mid, /* Callback parameter */
-					NULL
-				);
-			if (maxmsg_i > 0) {
-				LM_DBG("Maximum number of dumped messages: %d\n", maxmsg_i);
-				sent_cnt++;
-				if (sent_cnt >= maxmsg_i)
-					break;
-			}
+		tmb.t_request(&msg_type,  /* Type of the message */
+				&str_vals[1],     /* Request-URI (To) */
+				&str_vals[1],     /* To */
+				&str_vals[0],     /* From */
+				&hdr_str,         /* Optional headers including CRLF */
+				(n<0)?(RES_ROWS(db_res)[i].values[3].nul?NULL:&str_vals[2]):&body_str, /* Message body */
+				(ms_outbound_proxy.s)?&ms_outbound_proxy:0, /* outbound uri */
+				m_tm_callback,    /* Callback function */
+				(void*)(long)mid, /* Callback parameter */
+				NULL
+		);
+		if (maxmsg_i > 0) {
+			LM_DBG("Maximum number of dumped messages: %d\n", maxmsg_i);
+			sent_cnt++;
+			if (sent_cnt >= maxmsg_i)
+				break;
+		}
 	}
 
 done:
@@ -1106,9 +1063,6 @@ void destroy(void)
 {
 	LM_DBG("msilo destroy module ...\n");
 	msg_list_free(ml);
-
-	if(db_con && msilo_dbf.close)
-		msilo_dbf.close(db_con);
 }
 
 /**
@@ -1154,6 +1108,7 @@ void m_send_ontimer(unsigned int ticks, void *param)
 	static char hdr_buf[1024];
 	static char uri_buf[1024];
 	static char body_buf[1024];
+	char cbuf[26];
 	str puri;
 	time_t ttime;
 
@@ -1204,12 +1159,13 @@ void m_send_ontimer(unsigned int ticks, void *param)
 	if((msilo_dbf.query(db_con,db_keys,db_ops,db_vals,db_cols,db_no_keys,
 				db_no_cols, NULL,&db_res)!=0) || (RES_ROW_N(db_res) <= 0))
 	{
-		LM_DBG("no message for <%.*s>!\n", 24, ctime((const time_t*)&ttime));
+		ctime_r((const time_t*)&ttime, cbuf);
+		LM_DBG("no message for <%.*s>!\n", 24, cbuf);
 		goto done;
 	}
 
-	LM_DBG("dumping [%d] messages for <%.*s>!!!\n", RES_ROW_N(db_res), 24,
-			ctime((const time_t*)&ttime));
+	ctime_r((const time_t*)&ttime, cbuf);
+	LM_DBG("dumping [%d] messages for <%.*s>!!!\n", RES_ROW_N(db_res), 24, cbuf);
 
 	for(i = 0; i < RES_ROW_N(db_res); i++)
 	{
@@ -1250,7 +1206,8 @@ void m_send_ontimer(unsigned int ticks, void *param)
 		body_str.len = 1024;
 		stime =
 			(time_t)RES_ROWS(db_res)[i].values[5/*snd time*/].val.int_val;
-		n = m_build_body(&body_str, 0, str_vals[2/*body*/], stime);
+		n = m_build_body(&body_str, 0, str_vals[2/*body*/], stime,
+			1 /*is a reminder*/);
 		if(n<0)
 			LM_DBG("sending simple body\n");
 		else
@@ -1319,16 +1276,6 @@ int ms_reset_stime(int mid)
 	return 0;
 }
 
-static int fixup_m_dump(void** param, int param_no)
-{
-	if (param_no==1) {
-		return fixup_spve(param);
-	} else if (param_no==2) {
-		return fixup_uint(param);
-	}
-	return 0;
-}
-
 /*
  * Check if REGISTER request has contacts that support MESSAGE method or
  * if MESSAGE method is listed in Allow header and contact does not have
