@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2007-2009 Dan Pascu
+ * Copyright (C) 2007-2019 Dan Pascu
  *
  * This file is part of opensips, a free SIP server.
  *
@@ -40,7 +40,6 @@
 #include "../../timer.h"
 #include "../../resolve.h"
 #include "../../data_lump.h"
-#include "../../mod_fix.h"
 #include "../../script_cb.h"
 #include "../../sl_cb.h"
 #include "../../parser/msg_parser.h"
@@ -50,7 +49,7 @@
 #include "../../parser/contact/parse_contact.h"
 #include "../dialog/dlg_load.h"
 #include "../tm/tm_load.h"
-
+#include "clustering.h"
 
 
 
@@ -173,22 +172,20 @@ typedef struct Keepalive_Params {
 //
 static int NAT_Keepalive(struct sip_msg *msg);
 static int FixContact(struct sip_msg *msg);
-static int ClientNatTest(struct sip_msg *msg, unsigned int tests);
+static int ClientNatTest(struct sip_msg *msg, int *tests);
 
 static Bool test_private_contact(struct sip_msg *msg);
 static Bool test_source_address(struct sip_msg *msg);
 static Bool test_private_via(struct sip_msg *msg);
 static Bool test_source_contact(struct sip_msg *msg);
 
-static INLINE char* shm_strdup(char *source);
-
 static int  mod_init(void);
 static int  child_init(int rank);
 static void mod_destroy(void);
 static int  preprocess_request(struct sip_msg *msg, void *param);
 static int  reply_filter(struct sip_msg *reply);
 
-static int pv_parse_nat_contact_name(pv_spec_p sp, str *in);
+static int pv_parse_nat_contact_name(pv_spec_p sp, const str *in);
 static int pv_get_keepalive_socket(struct sip_msg *msg, pv_param_t *param, pv_value_t *res);
 static int pv_get_source_uri(struct sip_msg *msg, pv_param_t *param, pv_value_t *res);
 static int pv_get_track_dialog(struct sip_msg *msg, pv_param_t *param, pv_value_t *res);
@@ -218,13 +215,6 @@ stat_var *registered_endpoints = 0;
 stat_var *subscribed_endpoints = 0;
 stat_var *dialog_endpoints = 0;
 
-static NetInfo rfc1918nets[] = {
-    {"10.0.0.0",    0x0a000000UL, 0xff000000UL},
-    {"172.16.0.0",  0xac100000UL, 0xfff00000UL},
-    {"192.168.0.0", 0xc0a80000UL, 0xffff0000UL},
-    {NULL,          0UL,          0UL}
-};
-
 static NatTest NAT_Tests[] = {
     {NTPrivateContact, test_private_contact},
     {NTSourceAddress,  test_source_address},
@@ -233,24 +223,25 @@ static NatTest NAT_Tests[] = {
     {NTNone,           NULL}
 };
 
-
-static cmd_export_t commands[] = {
-    {"nat_keepalive",   (cmd_function)NAT_Keepalive, 0, NULL, 0, REQUEST_ROUTE},
-    {"fix_contact",     (cmd_function)FixContact,    0, NULL, 0, REQUEST_ROUTE | ONREPLY_ROUTE | BRANCH_ROUTE |LOCAL_ROUTE},
-    {"client_nat_test", (cmd_function)ClientNatTest, 1, fixup_uint_null, 0, REQUEST_ROUTE | ONREPLY_ROUTE | FAILURE_ROUTE | BRANCH_ROUTE|LOCAL_ROUTE},
-    {0, 0, 0, 0, 0, 0}
+static const cmd_export_t commands[] = {
+    {"nat_keepalive",   (cmd_function)NAT_Keepalive, {{0, 0, 0}}, REQUEST_ROUTE},
+    {"fix_contact",     (cmd_function)FixContact,    {{0, 0, 0}}, REQUEST_ROUTE | ONREPLY_ROUTE | BRANCH_ROUTE | LOCAL_ROUTE},
+    {"client_nat_test", (cmd_function)ClientNatTest, {{CMD_PARAM_INT, 0, 0}, {0, 0, 0}}, REQUEST_ROUTE | ONREPLY_ROUTE | FAILURE_ROUTE | BRANCH_ROUTE | LOCAL_ROUTE},
+    {0, 0, {{0, 0, 0}}, 0}
 };
 
-static param_export_t parameters[] = {
+static const param_export_t parameters[] = {
     {"keepalive_interval",       INT_PARAM, &keepalive_interval},
     {"keepalive_method",         STR_PARAM, &keepalive_params.method},
     {"keepalive_from",           STR_PARAM, &keepalive_params.from},
     {"keepalive_extra_headers",  STR_PARAM, &keepalive_params.extra_headers},
     {"keepalive_state_file",     STR_PARAM, &keepalive_state_file},
+    {"cluster_id",               INT_PARAM, &nt_cluster_id},
+    {"cluster_sharing_tag",      STR_PARAM, &nt_cluster_shtag},
     {0, 0, 0}
 };
 
-static pv_export_t pvars[] = {
+static const pv_export_t pvars[] = {
     {str_init("keepalive.socket"), 1000, pv_get_keepalive_socket, NULL, pv_parse_nat_contact_name, NULL, NULL, 0},
     {str_init("source_uri"), 1000, pv_get_source_uri, NULL, NULL, NULL, NULL, 0},
     {str_init("nat_traversal.track_dialog"), 1000, pv_get_track_dialog, pv_set_track_dialog, NULL, NULL, NULL, 0},
@@ -258,7 +249,7 @@ static pv_export_t pvars[] = {
 };
 
 #ifdef STATISTICS
-static stat_export_t statistics[] = {
+static const stat_export_t statistics[] = {
     {"keepalive_endpoints",  STAT_NO_RESET, &keepalive_endpoints},
     {"registered_endpoints", STAT_NO_RESET, &registered_endpoints},
     {"subscribed_endpoints", STAT_NO_RESET, &subscribed_endpoints},
@@ -267,35 +258,41 @@ static stat_export_t statistics[] = {
 };
 #endif
 
-static dep_export_t deps = {
-	{ /* OpenSIPS module dependencies */
-		{ MOD_TYPE_DEFAULT, "sl",     DEP_ABORT  },
-		{ MOD_TYPE_DEFAULT, "tm",     DEP_ABORT  },
-		{ MOD_TYPE_DEFAULT, "dialog", DEP_SILENT },
-		{ MOD_TYPE_NULL, NULL, 0 },
-	},
-	{ /* modparam dependencies */
-		{ NULL, NULL },
-	},
+static const dep_export_t deps = {
+    // OpenSIPS module dependencies
+    {
+        {MOD_TYPE_DEFAULT, "sl",     DEP_ABORT},
+        {MOD_TYPE_DEFAULT, "tm",     DEP_ABORT},
+        {MOD_TYPE_DEFAULT, "dialog", DEP_SILENT},
+        {MOD_TYPE_NULL, NULL, 0},
+    },
+    // modparam dependencies
+    {
+        {NULL, NULL}
+    }
 };
 
 struct module_exports exports = {
-    "nat_traversal", // module name
-    MOD_TYPE_DEFAULT,// class of this module
-    MODULE_VERSION,  // module version
-    DEFAULT_DLFLAGS, // dlopen flags
-    &deps,           // OpenSIPS module dependencies
-    commands,        // exported functions
-    0,               // exported async functions
-    parameters,      // exported parameters
-    NULL,            // exported statistics (initialized early in mod_init)
-    NULL,            // exported MI functions
-    pvars,           // exported pseudo-variables
-    NULL,            // extra processes
-    mod_init,        // module init function (before fork. kids will inherit)
-    reply_filter,    // reply processing function
-    mod_destroy,     // destroy function
-    child_init       // child init function
+    "nat_traversal",  // module name
+    MOD_TYPE_DEFAULT, // class of this module
+    MODULE_VERSION,   // module version
+    DEFAULT_DLFLAGS,  // dlopen flags
+    NULL,             // load function
+    &deps,            // module dependencies
+    commands,         // exported functions
+    NULL,             // exported async functions
+    parameters,       // exported parameters
+    NULL,             // exported statistics (initialized early in mod_init)
+    NULL,             // exported MI functions
+    pvars,            // exported pseudo-variables
+    NULL,             // exported transformations
+    NULL,             // extra processes
+    NULL,             // pre-init function
+    mod_init,         // module init function (before fork. kids will inherit)
+    reply_filter,     // reply processing function
+    mod_destroy,      // destroy function
+    child_init,       // child init function
+    NULL              // reload confirm function
 };
 
 
@@ -723,22 +720,6 @@ Dialog_Param_add_candidate(Dialog_Param *param, char *candidate)
 // Miscellaneous helper functions
 //
 
-static INLINE char*
-shm_strdup(char *source)
-{
-    char *copy;
-
-    if (!source)
-        return NULL;
-
-    copy = (char*)shm_malloc(strlen(source) + 1);
-    if (!copy)
-        return NULL;
-    strcpy(copy, source);
-
-    return copy;
-}
-
 
 static Bool
 get_contact_uri(struct sip_msg* msg, struct sip_uri *uri, contact_t **_c)
@@ -766,33 +747,6 @@ get_contact_uri(struct sip_msg* msg, struct sip_uri *uri, contact_t **_c)
     return True;
 }
 
-
-#define is_private_address(x) (rfc1918address(x)==1 ? 1 : 0)
-
-// Test if IP in `address' belongs to a RFC1918 network
-static INLINE int
-rfc1918address(str *address)
-{
-    struct ip_addr *ip;
-    uint32_t netaddr;
-    int i;
-
-    ip = str2ip(address);
-    if (ip == NULL)
-        return -1; // invalid address to test
-
-    netaddr = ntohl(ip->u.addr32[0]);
-
-    for (i=0; rfc1918nets[i].name!=NULL; i++) {
-        if ((netaddr & rfc1918nets[i].mask)==rfc1918nets[i].address) {
-            return 1;
-        }
-    }
-
-    return 0;
-}
-
-
 // Test if address of signaling is different from address in 1st Via field
 static Bool
 test_source_address(struct sip_msg *msg)
@@ -808,7 +762,7 @@ test_source_address(struct sip_msg *msg)
 }
 
 
-// Test if Contact field contains a private IP address as defined in RFC1918
+// Test if Contact field contains a private IP address
 static Bool
 test_private_contact(struct sip_msg *msg)
 {
@@ -818,15 +772,15 @@ test_private_contact(struct sip_msg *msg)
     if (!get_contact_uri(msg, &uri, &contact))
         return False;
 
-    return is_private_address(&(uri.host));
+    return ip_addr_is_1918(&(uri.host), 0);
 }
 
 
-// Test if top Via field contains a private IP address as defined in RFC1918
+// Test if top Via field contains a private IP address
 static Bool
 test_private_via(struct sip_msg *msg)
 {
-    return is_private_address(&(msg->via1->host));
+    return ip_addr_is_1918(&(msg->via1->host), 0);
 }
 
 
@@ -1567,12 +1521,12 @@ FixContact(struct sip_msg *msg)
 
 
 static int
-ClientNatTest(struct sip_msg *msg, unsigned int tests)
+ClientNatTest(struct sip_msg *msg, int *tests)
 {
     int i;
 
     for (i=0; NAT_Tests[i].test!=NTNone; i++) {
-        if ((tests & NAT_Tests[i].test)!=0 && NAT_Tests[i].proc(msg)) {
+        if ((*tests & NAT_Tests[i].test)!=0 && NAT_Tests[i].proc(msg)) {
             return 1;
         }
     }
@@ -1631,7 +1585,7 @@ send_keepalive(NAT_Contact *contact)
                    keepalive_params.extra_headers);
 
     if (len >= sizeof(buffer)) {
-        LM_ERR("keepalive message is longer than %lu bytes\n", (unsigned long)sizeof(buffer));
+        LM_ERR("keepalive message is longer than %zu bytes\n", sizeof(buffer));
         return;
     }
 
@@ -1642,13 +1596,13 @@ send_keepalive(NAT_Contact *contact)
     nat_port = strtol(ptr+1, NULL, 10);
     hostent = sip_resolvehost(&nat_ip, NULL, NULL, False, NULL);
     hostent2su(&to, hostent, 0, nat_port);
+    tolen=sockaddru_len(to);
 
-	tolen=sockaddru_len(to);
 again:
-	if (sendto(contact->socket->socket, buffer, len, 0, &to.s, tolen)==-1) {
-		if (errno==EINTR) goto again;
-		LM_ERR("sendto() failed with %s(%d)\n", strerror(errno),errno);
-	}
+    if (sendto(contact->socket->socket, buffer, len, 0, &to.s, tolen)==-1) {
+        if (errno==EINTR) goto again;
+        LM_ERR("sendto() failed with %s(%d)\n", strerror(errno),errno);
+    }
 }
 
 
@@ -1666,6 +1620,9 @@ keepalive_timer(unsigned int ticks, void *counter)
     time_t now;
     int i;
 
+    if ( !nt_cluster_shtag_is_active() )
+        return;
+
     now = time(NULL);
 
     for (i=0; i<nat_table->size; i++) {
@@ -1762,7 +1719,8 @@ restore_keepalive_state(void)
     res = fscanf(f, STATE_FILE_HEADER); // skip header
 
     while (True) {
-        res = fscanf(f, "%63s %63s %ld %ld", uri, socket, &rtime, &stime);
+        res = fscanf(f, "%63s %63s %lld %lld", uri, socket,
+				(long long *)&rtime, (long long *)&stime);
         if (res == EOF) {
             if (ferror(f))
                 LM_ERR("error while reading keepalive state file: %s\n", strerror(errno));
@@ -1806,6 +1764,7 @@ restore_keepalive_state(void)
 static int
 mod_init(void)
 {
+#define MIN_KEEPALIVE_INTERVAL 2
     int *param;
 
     if (keepalive_interval <= 0) {
@@ -1879,25 +1838,32 @@ mod_init(void)
     restore_keepalive_state();
 
     // check keepalive interval and add keepalive timer process
-    if (keepalive_interval < 10) {
-        LM_WARN("keepalive_interval should be at least 10 seconds\n");
-        LM_NOTICE("using 10 seconds for keepalive_interval\n");
-        keepalive_interval = 10;
+    if (keepalive_interval < MIN_KEEPALIVE_INTERVAL) {
+        LM_WARN("keepalive_interval should be at least %d seconds\n", MIN_KEEPALIVE_INTERVAL);
+        LM_NOTICE("using %d seconds for keepalive_interval\n", MIN_KEEPALIVE_INTERVAL);
+        keepalive_interval = MIN_KEEPALIVE_INTERVAL;
     }
+
     // allocate a shm variable to keep the counter used by the keepalive
     // timer routine - it must be shared as the routine get executed
     // in different processes
-    if (NULL==(param=(int*) shm_malloc(sizeof(int)))) {
+    param = (int*)shm_malloc(sizeof(int));
+    if (param == NULL) {
         LM_ERR("cannot allocate shm memory for keepalive counter\n");
         return -1;
     }
     *param = 0;
-    if (register_timer( "nt-pinger", keepalive_timer, (void*)(long)param, 1,
-    TIMER_FLAG_DELAY_ON_DELAY)<0) {
+
+    if (register_timer( "nt-pinger", keepalive_timer, (void*)(long)param, 1, TIMER_FLAG_DELAY_ON_DELAY) < 0) {
         LM_ERR("failed to register keepalive timer\n");
         return -1;
     }
 
+    if (nt_cluster_id>0 && nt_init_cluster()<0) {
+        LM_ERR("failed to initialized the clustering support\n");
+        return -1;
+    }
+
     return 0;
 }
 
@@ -1952,7 +1918,7 @@ preprocess_request(struct sip_msg *msg, void *_param)
         msg->msg_flags |= FL_NAT_TRACK_DIALOG;
     }
 
-	return SCB_RUN_ALL;
+    return SCB_RUN_ALL;
 }
 
 
@@ -2007,7 +1973,7 @@ reply_filter(struct sip_msg *reply)
 //
 
 static int
-pv_parse_nat_contact_name(pv_spec_p sp, str *in)
+pv_parse_nat_contact_name(pv_spec_p sp, const str *in)
 {
     char *p;
     char *s;
@@ -2091,7 +2057,10 @@ pv_get_source_uri(struct sip_msg *msg, pv_param_t *param, pv_value_t *res)
     if (msg==NULL || res==NULL)
         return -1;
 
-    snprintf(uri, 64, "sip:%s:%d", ip_addr2a(&msg->rcv.src_ip), msg->rcv.src_port);
+    if (msg->rcv.src_ip.af == AF_INET6)
+        snprintf(uri, 64, "sip:[%s]:%d", ip_addr2a(&msg->rcv.src_ip), msg->rcv.src_port);
+    else
+        snprintf(uri, 64, "sip:%s:%d", ip_addr2a(&msg->rcv.src_ip), msg->rcv.src_port);
 
     switch (msg->rcv.proto) {
     case PROTO_TCP:
