@@ -43,6 +43,7 @@
 #include "presence.h"
 #include "notify.h"
 #include "utils_func.h"
+#include "clustering.h"
 
 #define MAX_FORWARD 70
 
@@ -75,6 +76,7 @@ str str_status_col = str_init("status");
 str str_reason_col = str_init("reason");
 str str_socket_info_col = str_init("socket_info");
 str str_local_contact_col = str_init("local_contact");
+str str_sharing_tag_col = str_init("sharing_tag");
 str str_version_col = str_init("version");
 str str_presentity_uri_col = str_init("presentity_uri");
 str str_inserted_time_col = str_init("inserted_time");
@@ -116,7 +118,8 @@ inline void printf_subs(subs_t* subs)
 		subs->contact.s,subs->record_route.len,subs->record_route.s);
 }
 
-int build_str_hdr(subs_t* subs, int is_body, str* hdr, str* extra_hdrs)
+int build_str_hdr(subs_t* subs, int is_body, str* hdr, str *ct_body,
+															str* extra_hdrs)
 {
 	int len = 0;
 	int extra_len = 0;
@@ -146,13 +149,14 @@ int build_str_hdr(subs_t* subs, int is_body, str* hdr, str* extra_hdrs)
 	}
 	status.len = strlen(status.s);
 
-	len = 7 /*Event: */ + subs->event->name.len +4 /*;id=*/+ subs->event_id.len+
-		CRLF_LEN + 10 /*Contact: <*/ + subs->local_contact.len + 1/*>*/ +
-		((subs->sockinfo && subs->sockinfo->proto!=PROTO_UDP)?
-		 15/*";transport=xxxx"*/:0) + CRLF_LEN + 20 /*Subscription-State: */ +
+	len = 7 /*Event: */ + subs->event->name.len + 4 /*;id=*/ +
+		subs->event_id.len + CRLF_LEN +
+		10 /*Contact: <*/ + subs->local_contact.len + 1/*>*/ + CRLF_LEN +
+		20 /*Subscription-State: */ +
 		status.len + ((subs->status== TERMINATED_STATUS)?(10/*;reason=*/+
-		subs->reason.len):9/*expires=*/ + lexpire_len) + CRLF_LEN + (is_body?
-		(14 /*Content-Type: */+subs->event->content_type.len + CRLF_LEN):0);
+		subs->reason.len):9/*expires=*/ + lexpire_len) + CRLF_LEN +
+		(is_body ? (14 /*Content-Type: */+
+			(ct_body->len?ct_body->len:subs->event->content_type.len)+CRLF_LEN):0);
 
 	if(extra_hdrs && extra_hdrs->s && extra_hdrs->len)
 		extra_len = extra_hdrs->len;
@@ -183,33 +187,14 @@ int build_str_hdr(subs_t* subs, int is_body, str* hdr, str* extra_hdrs)
  		memcpy(p, subs->event_id.s, subs->event_id.len);
  		p += subs->event_id.len;
  	}
-	memcpy(p, CRLF, CRLF_LEN);
-	p += CRLF_LEN;
 
-	memcpy(p ,"Contact: <", 10);
-	p += 10;
+	memcpy(p ,CRLF "Contact: <", CRLF_LEN+10);
+	p += CRLF_LEN+10;
 	memcpy(p, subs->local_contact.s, subs->local_contact.len);
 	p +=  subs->local_contact.len;
 
-	if (subs->sockinfo && subs->sockinfo->proto!=PROTO_UDP)
-	{
-		memcpy(p,";transport=",11);
-		p += 11;
-		p = proto2str(subs->sockinfo->proto, p);
-		if (p == NULL)
-		{
-			LM_ERR("invalid proto\n");
-			pkg_free(hdr->s);
-			return -1;
-		}
-	}
-	*(p++) = '>';
-
-	memcpy(p, CRLF, CRLF_LEN);
-	p += CRLF_LEN;
-
-	memcpy(p, "Subscription-State: ", 20);
-	p+= 20;
+	memcpy(p, ">" CRLF "Subscription-State: " , 1+CRLF_LEN+20);
+	p += 1+CRLF_LEN+20;
 
 	memcpy(p, status.s, status.len);
 	p += status.len;
@@ -234,12 +219,16 @@ int build_str_hdr(subs_t* subs, int is_body, str* hdr, str* extra_hdrs)
 		p += CRLF_LEN;
 	}
 
-	if(is_body && subs->event->content_type.s && subs->event->content_type.len)
+	if (!ct_body->len &&
+		subs->event->content_type.s && subs->event->content_type.len)
+		ct_body = &subs->event->content_type;
+
+	if(is_body && ct_body->len)
 	{
 		memcpy(p,"Content-Type: ", 14);
 		p += 14;
-		memcpy(p, subs->event->content_type.s , subs->event->content_type.len);
-		p += subs->event->content_type.len;
+		memcpy(p, ct_body->s , ct_body->len);
+		p += ct_body->len;
 		memcpy(p, CRLF, CRLF_LEN);
 		p += CRLF_LEN;
 	}
@@ -291,7 +280,7 @@ int get_wi_subs_db(subs_t* subs, watcher_t* watchers)
 		goto error;
 	}
 
-//	CON_PS_REFERENCE(pa_db) = &my_ps;
+//	CON_SET_CURR_PS(pa_db, &my_ps);
 	if (pa_dbf.query (pa_db, query_cols, query_ops, query_vals,
 		 result_cols, n_query_cols, n_result_cols, 0,  &result) < 0)
 	{
@@ -359,7 +348,7 @@ str* get_wi_notify_body(subs_t* subs, subs_t* watcher_subs)
 	version_str = int2str(subs->version, &len);
 	if(version_str ==NULL)
 	{
-		LM_ERR("converting int to str\n ");
+		LM_ERR("converting int to str\n");
 		goto error;
 	}
 
@@ -429,6 +418,7 @@ str* get_wi_notify_body(subs_t* subs, subs_t* watcher_subs)
 				subs->event->wipeer->name)< 0 )
 	{
 		LM_ERR("failed to add waiting watchers\n");
+		lock_release(&subs_htable[hash_code].lock);
 		goto error;
 	}
 
@@ -448,12 +438,6 @@ done:
 	return notify_body;
 
 error:
-	if(notify_body)
-	{
-		if(notify_body->s)
-			xmlFree(notify_body->s);
-		pkg_free(notify_body);
-	}
 	free_watcher_list(watchers);
 	return NULL;
 }
@@ -580,8 +564,7 @@ int add_waiting_watchers(watcher_t* watchers, str pres_uri, str event)
 		return -1;
 	}
 
-//	CON_PS_REFERENCE(pa_db) = &my_ps;
-
+//	CON_SET_CURR_PS(pa_db, &my_ps);
 	if (pa_dbf.query (pa_db, query_cols, 0, query_vals,
 		 result_cols, n_query_cols, n_result_cols, 0, &result) < 0)
 	{
@@ -728,7 +711,7 @@ str* build_empty_bla_body(str pres_uri)
 	body = (str*) pkg_malloc(sizeof(str));
 	if(body== NULL)
 	{
-		LM_ERR("no more private memory");
+		LM_ERR("no more private memory\n");
 		pkg_free(entity);
 		goto error;
 	}
@@ -737,7 +720,7 @@ str* build_empty_bla_body(str pres_uri)
 	body->s = (char*) pkg_malloc(len);
 	if(body->s == NULL)
 	{
-		LM_ERR("no more private memory");
+		LM_ERR("no more private memory\n");
 		pkg_free(body);
 		pkg_free(entity);
 		goto error;
@@ -759,7 +742,7 @@ error:
 }
 
 
-static inline db_res_t* pres_search_db(struct sip_uri* uri,str* ev_name, int* body_col,
+db_res_t* pres_search_db(struct sip_uri* uri,str* ev_name, int* body_col,
 		int* extra_hdrs_col, int* expires_col, int* etag_col)
 {
 /*	static db_ps_t my_ps = NULL; */
@@ -812,7 +795,7 @@ static inline db_res_t* pres_search_db(struct sip_uri* uri,str* ev_name, int* bo
 				query_vals[i].val.str_val.s);
 	}
 
-/*	CON_PS_REFERENCE(pa_db) = &my_ps; */
+/*	CON_SET_CURR_PS(pa_db, &my_ps); */
 	if (pa_dbf.query (pa_db, query_cols, 0, query_vals,
 		 result_cols, n_query_cols, n_result_cols, &query_str, &result) < 0)
 	{
@@ -919,8 +902,9 @@ error:
 	return NULL;
 }
 
-str* get_p_notify_body(str pres_uri, pres_ev_t* event, str* etag, str* publ_body,
-		str* contact, str* dbody, str* extra_hdrs, free_body_t** free_fct, int from_publish)
+static str* get_p_notify_body(str pres_uri, pres_ev_t* event, str* etag,
+		str* publ_body, str* contact, str* dbody, str* extra_hdrs,
+		free_body_t** free_fct, int from_publish, int query_cluster)
 {
 	int body_col, extra_hdrs_col, expires_col, etag_col= 0;
 	db_res_t *result = NULL;
@@ -989,6 +973,10 @@ str* get_p_notify_body(str pres_uri, pres_ev_t* event, str* etag, str* publ_body
 		LM_DBG("No record exists in hash_table\n");
 		if(!fallback2db)
 		{
+			/* we do not have the presentity */
+			if (query_cluster && is_federation_full_sharing())
+				query_cluster_for_presentity(&pres_uri, event->evp);
+
 			/* for pidf manipulation && dialog-presence mixing */
 			if(event->agg_nbody)
 			{
@@ -1037,6 +1025,11 @@ str* get_p_notify_body(str pres_uri, pres_ev_t* event, str* etag, str* publ_body
 		pa_dbf.free_result(pa_db, result);
 		result= NULL;
 
+		/* we do not have the presentity */
+		if (query_cluster && is_federation_full_sharing() &&
+		is_event_clustered(event->evp->parsed))
+			query_cluster_for_presentity(&pres_uri, event->evp);
+
 		if(event->agg_nbody)
 		{
 			/* broken - it will not work with pidf manipulation */
@@ -1412,12 +1405,12 @@ error:
 }
 
 int get_subs_db(str* pres_uri, pres_ev_t* event, str* sender,
-		subs_t** s_array, int* n)
+		subs_t** s_array, int* n, str **sh_tags)
 {
 //	static db_ps_t my_ps = NULL;
-	db_key_t query_cols[7];
-	db_op_t  query_ops[7];
-	db_val_t query_vals[7];
+	db_key_t query_cols[8];
+	db_op_t  query_ops[8];
+	db_val_t query_vals[8];
 	db_key_t result_cols[19];
 	int n_result_cols = 0, n_query_cols = 0;
 	db_row_t *row ;
@@ -1433,6 +1426,7 @@ int get_subs_db(str* pres_uri, pres_ev_t* event, str* sender,
 	str sockinfo_str;
 	int port, proto;
 	str host;
+	unsigned int tag_no = 0;
 
 	if (pa_dbf.use_table(pa_db, &active_watchers_table) < 0)
 	{
@@ -1466,7 +1460,6 @@ int get_subs_db(str* pres_uri, pres_ev_t* event, str* sender,
 	query_ops[n_query_cols] = OP_NEQ;
 	query_vals[n_query_cols].type = DB_STR;
 	query_vals[n_query_cols].nul = 0;
-
 	if(sender)
 	{
 		LM_DBG("Do not send Notify to:[uri]= %.*s\n",sender->len,sender->s);
@@ -1479,6 +1472,17 @@ int get_subs_db(str* pres_uri, pres_ev_t* event, str* sender,
 	}
 	n_query_cols++;
 
+	/* this must be the last select key! */
+	if (sh_tags) {
+		query_cols[n_query_cols] = &str_sharing_tag_col;
+		query_ops[n_query_cols] = OP_EQ;
+		query_vals[n_query_cols].type = DB_STR;
+		query_vals[n_query_cols].nul = 0;
+		n_query_cols++;
+		/* the tag value is filled during the 'while' loop ; we rely on the
+		 * fact that there is at least one tag in the list !! */
+	}
+
 	result_cols[to_user_col=n_result_cols++]      =   &str_to_user_col;
 	result_cols[to_domain_col=n_result_cols++]    =   &str_to_domain_col;
 	result_cols[from_user_col=n_result_cols++]    =   &str_watcher_username_col;
@@ -1496,119 +1500,125 @@ int get_subs_db(str* pres_uri, pres_ev_t* event, str* sender,
 	result_cols[local_contact_col=n_result_cols++]=   &str_local_contact_col;
 	result_cols[version_col=n_result_cols++]      =   &str_version_col;
 
-	//CON_PS_REFERENCE(pa_db) = &my_ps;
-	if (pa_dbf.query(pa_db, query_cols, query_ops, query_vals,result_cols,
+	do {
+
+		if (sh_tags)
+			query_vals[n_query_cols-1].val.str_val = *sh_tags[tag_no];
+
+		//CON_SET_CURR_PS(pa_db, &my_ps);
+		if (pa_dbf.query(pa_db, query_cols, query_ops, query_vals,result_cols,
 				n_query_cols, n_result_cols, 0, &result) < 0)
-	{
-		LM_ERR("while querying database\n");
-		if(result)
 		{
-			pa_dbf.free_result(pa_db, result);
+			LM_ERR("while querying database\n");
+			if(result)
+				pa_dbf.free_result(pa_db, result);
+			return -1;
 		}
-		return -1;
-	}
 
-	if(result== NULL)
-		return -1;
+		if(result== NULL)
+			return -1;
 
-	if(result->n <=0 )
-	{
-		LM_DBG("The query for subscribtion for [uri]= %.*s for [event]= %.*s"
-			" returned no result\n",pres_uri->len, pres_uri->s,
-			event->name.len, event->name.s);
-		pa_dbf.free_result(pa_db, result);
-		return 0;
-	}
-	LM_DBG("found %d dialogs\n", result->n);
-
-	for(i=0; i<result->n; i++)
-	{
-		row = &result->rows[i];
-		row_vals = ROW_VALUES(row);
-
-	//	if(row_vals[expires_col].val.int_val< (int)time(NULL))
-	//		continue;
-
-        if(row_vals[reason_col].val.string_val)
-        {
-            if(strlen(row_vals[reason_col].val.string_val) != 0)
-                continue;
-        }
-		//	s.reason.len= strlen(s.reason.s);
-
-		memset(&s, 0, sizeof(subs_t));
-		s.status= ACTIVE_STATUS;
-
-		s.pres_uri= *pres_uri;
-		s.to_user.s= (char*)row_vals[to_user_col].val.string_val;
-		s.to_user.len= strlen(s.to_user.s);
-
-		s.to_domain.s= (char*)row_vals[to_domain_col].val.string_val;
-		s.to_domain.len= strlen(s.to_domain.s);
-
-		s.from_user.s= (char*)row_vals[from_user_col].val.string_val;
-		s.from_user.len= strlen(s.from_user.s);
-
-		s.from_domain.s= (char*)row_vals[from_domain_col].val.string_val;
-		s.from_domain.len= strlen(s.from_domain.s);
-
-		s.event_id.s=(char*)row_vals[event_id_col].val.string_val;
-		s.event_id.len= (s.event_id.s)?strlen(s.event_id.s):0;
-
-		s.to_tag.s= (char*)row_vals[to_tag_col].val.string_val;
-		s.to_tag.len= strlen(s.to_tag.s);
-
-		s.from_tag.s= (char*)row_vals[from_tag_col].val.string_val;
-		s.from_tag.len= strlen(s.from_tag.s);
-
-		s.callid.s= (char*)row_vals[callid_col].val.string_val;
-		s.callid.len= strlen(s.callid.s);
-
-		s.record_route.s=  (char*)row_vals[record_route_col].val.string_val;
-		s.record_route.len= (s.record_route.s)?strlen(s.record_route.s):0;
-
-		s.contact.s= (char*)row_vals[contact_col].val.string_val;
-		s.contact.len= strlen(s.contact.s);
-
-		sockinfo_str.s = (char*)row_vals[sockinfo_col].val.string_val;
-		if (sockinfo_str.s)
+		if(result->n <=0 )
 		{
-			sockinfo_str.len = strlen(sockinfo_str.s);
-			if (parse_phostport (sockinfo_str.s, sockinfo_str.len,&host.s,
-					&host.len, &port, &proto )< 0)
+			LM_DBG("The query for subscribtion for [uri]= %.*s for [event]= %.*s"
+				" returned no result\n",pres_uri->len, pres_uri->s,
+				event->name.len, event->name.s);
+			pa_dbf.free_result(pa_db, result);
+			return 0;
+		}
+		LM_DBG("found %d dialogs\n", result->n);
+
+		for(i=0; i<result->n; i++)
+		{
+			row = &result->rows[i];
+			row_vals = ROW_VALUES(row);
+
+			//	if(row_vals[expires_col].val.int_val< (int)time(NULL))
+			//		continue;
+
+			if(row_vals[reason_col].val.string_val)
 			{
-				LM_ERR("bad format for stored sockinfo string\n");
+				if(strlen(row_vals[reason_col].val.string_val) != 0)
+					continue;
+			}
+			//	s.reason.len= strlen(s.reason.s);
+
+			memset(&s, 0, sizeof(subs_t));
+			s.status= ACTIVE_STATUS;
+
+			s.pres_uri= *pres_uri;
+			s.to_user.s= (char*)row_vals[to_user_col].val.string_val;
+			s.to_user.len= strlen(s.to_user.s);
+
+			s.to_domain.s= (char*)row_vals[to_domain_col].val.string_val;
+			s.to_domain.len= strlen(s.to_domain.s);
+
+			s.from_user.s= (char*)row_vals[from_user_col].val.string_val;
+			s.from_user.len= strlen(s.from_user.s);
+
+			s.from_domain.s= (char*)row_vals[from_domain_col].val.string_val;
+			s.from_domain.len= strlen(s.from_domain.s);
+
+			s.event_id.s=(char*)row_vals[event_id_col].val.string_val;
+			s.event_id.len= (s.event_id.s)?strlen(s.event_id.s):0;
+
+			s.to_tag.s= (char*)row_vals[to_tag_col].val.string_val;
+			s.to_tag.len= strlen(s.to_tag.s);
+
+			s.from_tag.s= (char*)row_vals[from_tag_col].val.string_val;
+			s.from_tag.len= strlen(s.from_tag.s);
+
+			s.callid.s= (char*)row_vals[callid_col].val.string_val;
+			s.callid.len= strlen(s.callid.s);
+
+			s.record_route.s=  (char*)row_vals[record_route_col].val.string_val;
+			s.record_route.len= (s.record_route.s)?strlen(s.record_route.s):0;
+
+			s.contact.s= (char*)row_vals[contact_col].val.string_val;
+			s.contact.len= strlen(s.contact.s);
+
+			sockinfo_str.s = (char*)row_vals[sockinfo_col].val.string_val;
+			if (sockinfo_str.s)
+			{
+				sockinfo_str.len = strlen(sockinfo_str.s);
+				if (parse_phostport (sockinfo_str.s, sockinfo_str.len,&host.s,
+						&host.len, &port, &proto )< 0)
+				{
+					LM_ERR("bad format for stored sockinfo string\n");
+					goto error;
+				}
+				s.sockinfo = grep_sock_info(&host, (unsigned short) port,
+						(unsigned short) proto);
+			}
+
+			s.local_contact.s = (char*)row_vals[local_contact_col].val.string_val;
+			s.local_contact.len = s.local_contact.s?strlen(s.local_contact.s):0;
+
+			s.event= event;
+			s.local_cseq = row_vals[cseq_col].val.int_val;
+
+			if(row_vals[expires_col].val.int_val < (int)time(NULL))
+				s.expires = 0;
+			else
+				s.expires = row_vals[expires_col].val.int_val -(int)time(NULL);
+			s.version = row_vals[version_col].val.int_val;
+
+			s_new= mem_copy_subs(&s, PKG_MEM_TYPE);
+			if(s_new== NULL)
+			{
+				LM_ERR("while copying subs_t structure\n");
 				goto error;
 			}
-			s.sockinfo = grep_sock_info(&host, (unsigned short) port,
-					(unsigned short) proto);
+			s_new->next= (*s_array);
+			(*s_array)= s_new;
+
+			printf_subs(s_new);
+			inc++;
 		}
+		pa_dbf.free_result(pa_db, result);
 
-		s.local_contact.s = (char*)row_vals[local_contact_col].val.string_val;
-		s.local_contact.len = s.local_contact.s?strlen(s.local_contact.s):0;
+	} while(sh_tags && sh_tags[++tag_no]);
 
-		s.event= event;
-		s.local_cseq = row_vals[cseq_col].val.int_val;
-
-		if(row_vals[expires_col].val.int_val < (int)time(NULL))
-			s.expires = 0;
-		else
-			s.expires = row_vals[expires_col].val.int_val -(int)time(NULL);
-		s.version = row_vals[version_col].val.int_val;
-
-		s_new= mem_copy_subs(&s, PKG_MEM_TYPE);
-		if(s_new== NULL)
-		{
-			LM_ERR("while copying subs_t structure\n");
-			goto error;
-		}
-		s_new->next= (*s_array);
-		(*s_array)= s_new;
-
-		printf_subs(s_new);
-		inc++;
-	}
-	pa_dbf.free_result(pa_db, result);
 	*n= inc;
 
 	return 0;
@@ -1660,19 +1670,123 @@ int update_in_list(subs_t* s, subs_t* s_array, int new_rec_no, int n)
 	return -1;
 }
 
-subs_t* get_subs_dialog(str* pres_uri, pres_ev_t* event, str* sender)
+int presentity_has_subscribers(str* pres_uri, pres_ev_t* event)
+{
+	static db_ps_t ps = NULL;
+	unsigned int hash_code;
+	subs_t* s;
+	time_t now;
+	db_key_t keys[3];
+	db_val_t vals[3];
+	db_key_t cols[1];
+	db_res_t *res;
+
+	/* first check the in-memory hash, maybe we are lucky */
+	hash_code= core_hash(pres_uri, &event->name, shtable_size);
+
+	lock_get(&subs_htable[hash_code].lock);
+	s = subs_htable[hash_code].entries;
+	now = time(NULL);
+
+	while (s->next) {
+		s= s->next;
+
+		/* expired & active ? */
+		if ( (s->expires<(int)now) || (s->status!=ACTIVE_STATUS) ||
+		(s->reason.len!=0) )
+			continue;
+
+		if (s->event==event && s->pres_uri.len== pres_uri->len &&
+		strncmp(s->pres_uri.s, pres_uri->s, pres_uri->len)==0) {
+			/* found a subscriber for our presentity*/
+			lock_release(&subs_htable[hash_code].lock);
+			return 1;
+		}
+
+	}
+	lock_release(&subs_htable[hash_code].lock);
+
+	/* presentity has no subscriber in hash. Check db ?? */
+	if(fallback2db==0)
+		return 0;
+
+	keys[0] = &str_presentity_uri_col;
+	vals[0].type = DB_STR;
+	vals[0].nul = 0;
+	vals[0].val.str_val = *pres_uri;
+
+	keys[1] = &str_event_col;
+	vals[1].type = DB_STR;
+	vals[1].nul = 0;
+	vals[1].val.str_val = event->name;
+
+	keys[2] = &str_status_col;
+	vals[2].type = DB_INT;
+	vals[2].nul = 0;
+	vals[2].val.int_val = ACTIVE_STATUS;
+
+	cols[0] = &str_watcher_username_col;
+
+	if (pa_dbf.use_table(pa_db, &active_watchers_table) < -1) {
+		LM_ERR("in use_table\n");
+		goto error;
+	}
+
+	CON_SET_CURR_PS(pa_db, ps);
+	if (pa_dbf.query(pa_db, keys, 0, vals, cols, 3, 1, 0, &res) < 0) {
+		LM_ERR("DB query failed\n");
+		goto error;
+	}
+
+	if ( RES_ROW_N(res)>0 ) {
+		pa_dbf.free_result(pa_db, res);
+		return 1;
+	}
+
+	pa_dbf.free_result(pa_db, res);
+
+	/* no subscriber for presentity + event */
+	return 0;
+error:
+	return 0;
+}
+
+
+static int is_in_shtag_list(str *tag, str **list)
+{
+	unsigned int n = 0;
+
+	while( list[n] ) {
+		if (list[n]->len==tag->len &&
+		strncasecmp(list[n]->s,tag->s,tag->len)==0)
+			return 1;
+		n++;
+	}
+
+	return 0;
+}
+
+
+subs_t* get_subs_dialog(str* pres_uri, pres_ev_t* event, str* sender,
+															str **sh_tags)
 {
 	unsigned int hash_code;
 	subs_t* s= NULL, *s_new;
 	subs_t* s_array= NULL;
 	int n= 0, i= 0;
 
+	/* if tag filtering is enabled but not active tag is present
+	 * in the list, simply return a 0 len list */
+	if (sh_tags && sh_tags[0]==NULL) {
+		return 0;
+	}
+
 	/* if fallback2db -> should take all dialogs from db
 	 * and the only those dialogs from cache with db_flag= INSERTDB_FLAG */
 
 	if(fallback2db)
 	{
-		if(get_subs_db(pres_uri, event, sender, &s_array, &n)< 0)
+		if(get_subs_db(pres_uri, event, sender, &s_array, &n, sh_tags)< 0)
 		{
 			LM_ERR("getting dialogs from database\n");
 			goto error;
@@ -1704,7 +1818,8 @@ subs_t* get_subs_dialog(str* pres_uri, pres_ev_t* event, str* sender)
 				s->event== event && s->pres_uri.len== pres_uri->len &&
 				strncmp(s->pres_uri.s, pres_uri->s, pres_uri->len)== 0)) ||
 				(sender && sender->len== s->contact.len &&
-				strncmp(sender->s, s->contact.s, sender->len)== 0))
+				strncmp(sender->s, s->contact.s, sender->len)== 0) ||
+				(sh_tags && !is_in_shtag_list(&s->sh_tag, sh_tags) ) )
 				continue;
 
 			s_new= mem_copy_subs(s, PKG_MEM_TYPE);
@@ -1733,7 +1848,7 @@ error:
 }
 
 int publ_notify(presentity_t* p, str pres_uri, str* body, str* offline_etag,
-		str* rules_doc, str* dialog_body, int from_publish)
+		str* rules_doc, str* dialog_body, int from_publish, str **sh_tags)
 {
 	str *notify_body = NULL;
 	str notify_extra_hdrs = {NULL, 0};
@@ -1741,7 +1856,7 @@ int publ_notify(presentity_t* p, str pres_uri, str* body, str* offline_etag,
 	int ret_code= -1;
 	free_body_t* free_fct = 0;
 
-	subs_array= get_subs_dialog(&pres_uri, p->event , p->sender);
+	subs_array= get_subs_dialog(&pres_uri, p->event , p->sender, sh_tags);
 	if(subs_array == NULL)
 	{
 		LM_DBG("Could not find subs_dialog\n");
@@ -1755,7 +1870,7 @@ int publ_notify(presentity_t* p, str pres_uri, str* body, str* offline_etag,
 		notify_body = get_p_notify_body(pres_uri, p->event , offline_etag, body,
 				NULL, dialog_body,
 				p->extra_hdrs?p->extra_hdrs:&notify_extra_hdrs, &free_fct,
-				from_publish);
+				from_publish, 0);
 	}
 
 	s= subs_array;
@@ -1793,6 +1908,23 @@ done:
 	return ret_code;
 }
 
+
+int virtual_notify(str *pres_uri, pres_ev_t *ev, str *body)
+{
+	presentity_t pres;
+	str l_pres_uri;
+
+	memset( &pres, 0, sizeof(presentity_t) );
+	pres.event = ev;
+
+	l_pres_uri = *pres_uri;
+
+	return publ_notify( &pres, l_pres_uri, body, NULL /*offline_etag*/,
+		NULL/*rules_doc*/, NULL/*dialog_body*/,
+		1 /*from_publish*/, NULL/*sh_tag*/);
+}
+
+
 int query_db_notify(str* pres_uri, pres_ev_t* event, subs_t* watcher_subs)
 {
 	subs_t* subs_array = NULL, *s= NULL;
@@ -1801,7 +1933,7 @@ int query_db_notify(str* pres_uri, pres_ev_t* event, subs_t* watcher_subs)
 	int ret_code= -1;
 	free_body_t* free_fct = 0;
 
-	subs_array= get_subs_dialog(pres_uri, event , NULL);
+	subs_array= get_subs_dialog(pres_uri, event , NULL, NULL);
 	if(subs_array == NULL)
 	{
 		LM_DBG("Could not get subscription dialog\n");
@@ -1812,7 +1944,7 @@ int query_db_notify(str* pres_uri, pres_ev_t* event, subs_t* watcher_subs)
 	if(event->type & PUBL_TYPE)
 	{
 		notify_body = get_p_notify_body(*pres_uri, event, 0, 0, 0, 0,
-				&notify_extra_hdrs, &free_fct, 0);
+				&notify_extra_hdrs, &free_fct, 0, 1);
 	}
 
 	s= subs_array;
@@ -1867,6 +1999,8 @@ int send_notify_request(subs_t* subs, subs_t * watcher_subs,
 	str* final_body= NULL;
 	str* aux_body = 0;
 	free_body_t* free_fct = 0;
+	str ct_body = {NULL,0};
+	int suppress_notify = 0;
 
 	LM_DBG("enter: have_body=%d force_null=%d dialog info:\n",
 	  (n_body!=0&&n_body->s!=0)?1:0, force_null_body);
@@ -1918,13 +2052,20 @@ int send_notify_request(subs_t* subs, subs_t * watcher_subs,
 			}
 			else
 			{
-				if (from_publish && n_body!= 0 && n_body->s!= 0)
+				if (from_publish && n_body!= 0 && n_body->s!= 0) {
 					notify_body = n_body;
-				else
+				} else if (subs->event->build_notify_body) {
+					notify_body = subs->event->build_notify_body(&subs->pres_uri,
+						&subs->subs_body, &ct_body, &suppress_notify);
+					if (suppress_notify)
+						return 0;
+					free_fct = subs->event->free_body;
+				} else {
 					notify_body = get_p_notify_body(subs->pres_uri,
 							subs->event, 0, 0, (subs->contact.s)?&subs->contact:NULL,
 							NULL, extra_hdrs?extra_hdrs:&notify_extra_hdrs,
-							&free_fct, from_publish);
+							&free_fct, from_publish, 1);
+				}
 				if(notify_body == NULL || notify_body->s== NULL)
 				{
 					LM_DBG("Could not get the notify_body\n");
@@ -1965,7 +2106,7 @@ jump_over_body:
 
 	/* build extra headers */
 	if( build_str_hdr( subs, notify_body?1:0, &str_hdr,
-				extra_hdrs?extra_hdrs:&notify_extra_hdrs)< 0 )
+				&ct_body, extra_hdrs?extra_hdrs:&notify_extra_hdrs)< 0 )
 	{
 		LM_ERR("while building headers\n");
 		goto error;
@@ -2055,6 +2196,8 @@ error:
 		pkg_free(str_hdr.s);
 	if (notify_extra_hdrs.s)
 		pkg_free(notify_extra_hdrs.s);
+	if (ct_body.s)
+		pkg_free(ct_body.s);
 
 	if((int)(long)n_body!= (int)(long)notify_body)
 	{
@@ -2065,8 +2208,8 @@ error:
 				if(subs->event->type& WINFO_TYPE)
 					xmlFree(notify_body->s);
 				else
-				if(subs->event->apply_auth_nbody== NULL && subs->event->agg_nbody== NULL)
-					pkg_free(notify_body->s);
+				if(free_fct)
+					free_fct(notify_body->s);
 				else
 				subs->event->free_body(notify_body->s);
 			}
@@ -2173,7 +2316,7 @@ c_back_param* shm_dup_cbparam(subs_t* subs)
 	cb_param= (c_back_param*)shm_malloc(size);
 	if(cb_param== NULL)
 	{
-		LM_ERR("no more shared memory");
+		LM_ERR("no more shared memory\n");
 		return 0;
 	}
 	memset(cb_param, 0, size);
