@@ -33,6 +33,7 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <time.h>
+#include <fnmatch.h>
 
 #include "../../sr_module.h"
 #include "../../db/db.h"
@@ -46,37 +47,40 @@
 #include "../../mem/mem.h"
 #include "../../mem/shm_mem.h"
 #include "../../usr_avp.h"
-#include "../tm/tm_load.h"
-#include "../signaling/signaling.h"
 #include "../../pt.h"
 #include "../../mi/mi.h"
+#include "../../evi/evi_modules.h"
+#include "../tm/tm_load.h"
+#include "../signaling/signaling.h"
 #include "../pua/hash.h"
 #include "publish.h"
 #include "subscribe.h"
 #include "event_list.h"
 #include "bind_presence.h"
 #include "notify.h"
-#include "../../evi/evi_modules.h"
+#include "utils_func.h"
+#include "clustering.h"
 
 
 #define S_TABLE_VERSION  4
 #define P_TABLE_VERSION  5
-#define ACTWATCH_TABLE_VERSION 11
+#define ACTWATCH_TABLE_VERSION 12
 
 char *log_buf = NULL;
-static int clean_period=100;
-static int watchers_clean_period=3600;
-static int db_update_period=100;
+static int clean_period = 100;
+static int watchers_clean_period = 3600;
+static int db_update_period = 100;
 
 /* database connection */
 db_con_t *pa_db = NULL;
 db_func_t pa_dbf;
-str presentity_table= str_init("presentity");
+str presentity_table = str_init("presentity");
 str active_watchers_table = str_init("active_watchers");
-str watchers_table= str_init("watchers");
+str watchers_table = str_init("watchers");
+
+int library_mode = 0;
+str contact_user = str_init("presence");
 
-int library_mode= 0;
-str server_address= {0, 0};
 evlist_t* EvList= NULL;
 
 /* TM bind */
@@ -90,12 +94,22 @@ static int mod_init(void);
 static int child_init(int);
 static void destroy(void);
 int stored_pres_info(struct sip_msg* msg, char* pres_uri, char* s);
-static int fixup_presence(void** param, int param_no);
-static int fixup_subscribe(void** param, int param_no);
-static struct mi_root* mi_refreshWatchers(struct mi_root* cmd, void* param);
-static struct mi_root* mi_cleanup(struct mi_root* cmd, void* param);
-static struct mi_root* mi_list_phtable(struct mi_root* cmd, void* param);
-static struct mi_root* mi_list_shtable(struct mi_root* cmd, void* param);
+static int fixup_presence(void** param);
+static int fixup_subscribe(void** param);
+static mi_response_t *mi_refresh_watchers(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_cleanup(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_list_phtable(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_list_shtable_1(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_list_shtable_2(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_pres_expose_1(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_pres_expose_2(const mi_params_t *params,
+								struct mi_handler *async_hdl);
 static int update_pw_dialogs(subs_t* subs, unsigned int hash_code, subs_t** subs_array);
 int update_watchers_status(str pres_uri, pres_ev_t* ev, str* rules_doc);
 int refresh_send_winfo_notify(watcher_t* watcher, str pres_uri,
@@ -120,6 +134,9 @@ int end_sub_on_timeout= 1;
 pres_ev_t** pres_event_p= NULL;
 pres_ev_t** dialog_event_p= NULL;
 
+char *federation_mode_str = NULL;
+char *cluster_active_shtag_str = NULL;
+
 int phtable_size= 9;
 phtable_t* pres_htable = NULL;
 unsigned int waiting_subs_daysno = 0;
@@ -130,19 +147,24 @@ int fix_remote_target=1;
 
 /* event id */
 static str presence_publish_event = str_init("E_PRESENCE_PUBLISH");
+static str presence_exposed_event = str_init("E_PRESENCE_EXPOSED");
 event_id_t presence_event_id = EVI_ERROR;
+event_id_t exposed_event_id = EVI_ERROR;
 
-static cmd_export_t cmds[]=
-{
-	{"handle_publish",  (cmd_function)handle_publish,  0,fixup_presence,0, REQUEST_ROUTE},
-	{"handle_publish",  (cmd_function)handle_publish,  1,fixup_presence, 0, REQUEST_ROUTE},
-	{"handle_subscribe",(cmd_function)handle_subscribe,0,fixup_subscribe,0, REQUEST_ROUTE},
-	{"handle_subscribe",(cmd_function)handle_subscribe,1,fixup_subscribe,0, REQUEST_ROUTE},
-	{"bind_presence",   (cmd_function)bind_presence,   1,     0,         0,  0},
-	{ 0,                    0,                         0,     0,         0,  0}
+static const cmd_export_t cmds[]={
+	{"handle_publish",  (cmd_function)handle_publish, {
+		{CMD_PARAM_STR|CMD_PARAM_OPT,fixup_presence,0}, {0,0,0}},
+		REQUEST_ROUTE},
+	{"handle_subscribe",(cmd_function)handle_subscribe, {
+		{CMD_PARAM_INT|CMD_PARAM_OPT,fixup_subscribe,0},
+		{CMD_PARAM_STR|CMD_PARAM_OPT,fixup_subscribe,0},
+		{0,0,0}},
+		REQUEST_ROUTE},
+	{"bind_presence",(cmd_function)bind_presence,{{0,0,0}},0},
+	{0,0,{{0,0,0}},0}
 };
 
-static param_export_t params[]={
+static const param_export_t params[]={
 	{ "db_url",                 STR_PARAM, &db_url.s},
 	{ "presentity_table",       STR_PARAM, &presentity_table.s},
 	{ "active_watchers_table",  STR_PARAM, &active_watchers_table.s},
@@ -152,29 +174,56 @@ static param_export_t params[]={
 	{ "expires_offset",         INT_PARAM, &expires_offset },
 	{ "max_expires_subscribe",  INT_PARAM, &max_expires_subscribe },
 	{ "max_expires_publish",    INT_PARAM, &max_expires_publish },
-	{ "server_address",         STR_PARAM, &server_address.s},
+	{ "contact_user",           STR_PARAM, &contact_user.s},
 	{ "subs_htable_size",       INT_PARAM, &shtable_size},
 	{ "pres_htable_size",       INT_PARAM, &phtable_size},
 	{ "fallback2db",            INT_PARAM, &fallback2db},
 	{ "enable_sphere_check",    INT_PARAM, &sphere_enable},
 	{ "waiting_subs_daysno",    INT_PARAM, &waiting_subs_daysno},
 	{ "mix_dialog_presence",    INT_PARAM, &mix_dialog_presence},
-	{ "bla_presentity_spec",    STR_PARAM, &bla_presentity_spec_param},
+	{ "bla_presentity_spec",    STR_PARAM, &bla_presentity_spec_param.s},
 	{ "bla_fix_remote_target",  INT_PARAM, &fix_remote_target},
 	{ "notify_offline_body",    INT_PARAM, &notify_offline_body},
 	{ "end_sub_on_timeout",     INT_PARAM, &end_sub_on_timeout},
+	{ "cluster_id",             INT_PARAM, &pres_cluster_id},
+	{ "cluster_federation_mode",STR_PARAM, &federation_mode_str},
+	{ "cluster_be_active_shtag",STR_PARAM, &cluster_active_shtag_str},
+	{ "cluster_pres_events",    STR_PARAM, &clustering_events.s},
 	{0,0,0}
 };
 
-static mi_export_t mi_cmds[] = {
-	{ "refreshWatchers",   0, mi_refreshWatchers,    0,  0,  0},
-	{ "cleanup",           0, mi_cleanup,            0,  0,  0},
-	{ "pres_phtable_list", 0, mi_list_phtable,       0,  0,  0},
-	{ "subs_phtable_list", 0, mi_list_shtable,       0,  0,  0},
-	{  0,                  0, 0,                     0,  0,  0}
+static const mi_export_t mi_cmds[] = {
+	// refreshWatchers is a deprecated alias for refresh_watchers. To be removed later.
+	{ "refreshWatchers", 0,0,0, {
+		{mi_refresh_watchers, {"presentity_uri", "event", "refresh_type", 0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{ "refresh_watchers", 0,0,0, {
+		{mi_refresh_watchers, {"presentity_uri", "event", "refresh_type", 0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{ "cleanup", 0,0,0, {
+		{mi_cleanup, {0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{ "pres_expose", 0,0,0, {
+		{mi_pres_expose_1, {"event", 0}},
+		{mi_pres_expose_2, {"event", "filter", 0}},
+		{EMPTY_MI_RECIPE}} 
+	},
+	{ "pres_phtable_list", 0,0,0, {
+		{mi_list_phtable, {0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{ "subs_phtable_list", 0,0,0, {
+		{mi_list_shtable_1, {0}},
+		{mi_list_shtable_2, {"from", "to", 0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{EMPTY_MI_EXPORT}
 };
 
-static dep_export_t deps = {
+static const dep_export_t deps = {
 	{ /* OpenSIPS module dependencies */
 		{ MOD_TYPE_DEFAULT, "tm",        DEP_ABORT },
 		{ MOD_TYPE_DEFAULT, "signaling", DEP_ABORT },
@@ -182,16 +231,19 @@ static dep_export_t deps = {
 	},
 	{ /* modparam dependencies */
 		{ "db_url", get_deps_sqldb_url },
+		{ "cluster_id", get_deps_clusterer },
 		{ NULL, NULL },
 	},
 };
 
+
 /** module exports */
 struct module_exports exports= {
 	"presence",					/* module name */
 	MOD_TYPE_DEFAULT,           /* class of this module */
 	MODULE_VERSION,
 	DEFAULT_DLFLAGS,			/* dlopen flags */
+	0,							/* load function */
 	&deps,                      /* OpenSIPS module dependencies */
 	cmds,						/* exported functions */
 	0,							/* exported async functions */
@@ -199,11 +251,14 @@ struct module_exports exports= {
 	0,							/* exported statistics */
 	mi_cmds,					/* exported MI functions */
 	0,							/* exported pseudo-variables */
+	0,			 				/* exported transformations */
 	0,							/* extra processes */
+	0,							/* module pre-initialization function */
 	mod_init,					/* module initialization function */
 	(response_function) 0,      /* response handling function */
 	(destroy_function) destroy, /* destroy function */
-	child_init                  /* per-child init function */
+	child_init,                 /* per-child init function */
+	0                           /* reload confirm function */
 };
 
 /**
@@ -211,18 +266,21 @@ struct module_exports exports= {
  */
 static int mod_init(void)
 {
-	/* register event E_PRESENCE_NOTIFY */ 
-	if( (presence_event_id=evi_publish_event(presence_publish_event)) == EVI_ERROR )
-		LM_ERR("Cannot register E_PRESENCE_PUBLISH event\n");
-	else
-		LM_NOTICE("E_PRESENCE_PUBLISH event registered\n");
 	db_url.len = db_url.s ? strlen(db_url.s) : 0;
-	LM_DBG("db_url=%s/%d/%p\n", ZSW(db_url.s), db_url.len,db_url.s);
+	LM_DBG("db_url=%s\n", db_url_escape(&db_url));
+
 	presentity_table.len = strlen(presentity_table.s);
 	active_watchers_table.len = strlen(active_watchers_table.s);
 	watchers_table.len = strlen(watchers_table.s);
 
-	LM_NOTICE("initializing module ...\n");
+	/* register event E_PRESENCE_NOTIFY */ 
+	presence_event_id = evi_publish_event(presence_publish_event);
+	if ( presence_event_id == EVI_ERROR )
+		LM_ERR("Cannot register E_PRESENCE_PUBLISH event\n");
+
+	exposed_event_id=evi_publish_event(presence_exposed_event);
+	if ( exposed_event_id == EVI_ERROR )
+		LM_ERR("Cannot register E_PRESENCE_EXPOSED event\n");
 
 	EvList= init_evlist();
 	if(!EvList)
@@ -257,13 +315,7 @@ static int mod_init(void)
 	if(max_expires_publish<= 0)
 		max_expires_publish = 3600;
 
-	if(server_address.s== NULL)
-		LM_DBG("server_address parameter not set in configuration file\n");
-
-	if(server_address.s)
-		server_address.len= strlen(server_address.s);
-	else
-		server_address.len= 0;
+	contact_user.len = strlen(contact_user.s);
 
 	/* load SIGNALING API */
 	if(load_sig_api(&sigb)< 0)
@@ -279,9 +331,8 @@ static int mod_init(void)
 		return -1;
 	}
 
-	if(db_url.s== NULL)
-	{
-		LM_ERR("database url not set!\n");
+	if (init_pres_clustering(federation_mode_str, cluster_active_shtag_str)<0) {
+		LM_ERR("failed to init clustering support\n");
 		return -1;
 	}
 
@@ -307,9 +358,12 @@ static int mod_init(void)
 	}
 
 	/*verify table versions */
-	if((db_check_table_version(&pa_dbf, pa_db, &presentity_table, P_TABLE_VERSION) < 0) ||
-		(db_check_table_version(&pa_dbf, pa_db, &active_watchers_table, ACTWATCH_TABLE_VERSION) < 0) ||
-		(db_check_table_version(&pa_dbf, pa_db, &watchers_table, S_TABLE_VERSION) < 0)) {
+	if ( (db_check_table_version(
+		&pa_dbf, pa_db, &presentity_table, P_TABLE_VERSION) < 0) ||
+	(db_check_table_version(
+		&pa_dbf, pa_db, &active_watchers_table, ACTWATCH_TABLE_VERSION) < 0) ||
+	(db_check_table_version(
+		&pa_dbf, pa_db, &watchers_table, S_TABLE_VERSION) < 0) ) {
 			LM_ERR("error during table version check\n");
 			return -1;
 	}
@@ -352,10 +406,10 @@ static int mod_init(void)
 
 	if(clean_period>0)
 	{
-		register_timer("presence-pclean", msg_presentity_clean, 0,
-			clean_period, TIMER_FLAG_DELAY_ON_DELAY);
-		register_timer("presence-wclean", msg_watchers_clean, 0,
-			watchers_clean_period, TIMER_FLAG_DELAY_ON_DELAY);
+		register_timer("presence-pclean", msg_presentity_clean,
+			(void*)(long)clean_period, clean_period,TIMER_FLAG_DELAY_ON_DELAY);
+		register_timer("presence-wclean", msg_watchers_clean,
+			0, watchers_clean_period, TIMER_FLAG_DELAY_ON_DELAY);
 	}
 
 	if(db_update_period>0)
@@ -415,8 +469,6 @@ static int mod_init(void)
  */
 static int child_init(int rank)
 {
-	LM_NOTICE("init_child [%d]  pid [%d]\n", rank, getpid());
-
 	pid = my_pid();
 
 	if(library_mode)
@@ -447,7 +499,7 @@ static void destroy(void)
 {
 	LM_NOTICE("destroy module ...\n");
 
-	if(subs_htable && pa_db)
+	if(subs_htable && !library_mode && child_init(process_no)==0)
 		timer_db_update(0, 0);
 
 	if(subs_htable)
@@ -467,37 +519,19 @@ static void destroy(void)
 	destroy_evlist();
 }
 
-static int fixup_presence(void** param, int param_no)
+static int fixup_presence(void** param)
 {
- 	pv_elem_t *model;
-	str s;
-
 	if(library_mode)
 	{
 		LM_ERR("Bad config - you can not call 'handle_publish' function"
 				" (db_url not set)\n");
 		return -1;
 	}
-	if(param_no== 0)
-		return 0;
-
-	if(*param)
- 	{
-		s.s = (char*)(*param); s.len = strlen(s.s);
- 		if(pv_parse_format(&s, &model)<0)
- 		{
- 			LM_ERR( "wrong format[%s]\n",(char*)(*param));
- 			return E_UNSPEC;
- 		}
-
- 		*param = (void*)model;
- 		return 0;
- 	}
- 	LM_ERR( "null format\n");
- 	return E_UNSPEC;
+	
+	return 0;
 }
 
-static int fixup_subscribe(void** param, int param_no)
+static int fixup_subscribe(void** param)
 {
 	if(library_mode)
 	{
@@ -505,71 +539,49 @@ static int fixup_subscribe(void** param, int param_no)
 				" (db_url not set)\n");
 		return -1;
 	}
+
 	return 0;
 }
 
 /*
- *  mi cmd: refreshWatchers
+ *  mi cmd: refresh_watchers
  *			<presentity_uri>
  *			<event>
  *          <refresh_type> // can be:  = 0 -> watchers autentification type or
  *									  != 0 -> publish type //
  *		* */
 
-static struct mi_root* mi_refreshWatchers(struct mi_root* cmd, void* param)
+static mi_response_t *mi_refresh_watchers(const mi_params_t *params,
+								struct mi_handler *async_hdl)
 {
-	struct mi_node* node= NULL;
 	str pres_uri, event;
 	struct sip_uri uri;
 	pres_ev_t* ev;
 	str* rules_doc= NULL;
 	int result;
-	unsigned int refresh_type;
+	int refresh_type;
 
 	LM_DBG("start\n");
 
-	node = cmd->node.kids;
-	if(node == NULL)
-		return init_mi_tree(404, "No parameters", 13);
-
-	/* Get presentity URI */
-	pres_uri = node->value;
+	if (get_mi_string_param(params, "presentity_uri", &pres_uri.s, &pres_uri.len) < 0)
+		return init_mi_param_error();
 	if(pres_uri.s == NULL || pres_uri.len== 0)
 	{
 		LM_ERR( "empty uri\n");
-		return init_mi_tree(404, "Empty presentity URI", 20);
+		return init_mi_error(404, MI_SSTR("Empty presentity URI"));
 	}
 
-	node = node->next;
-	if(node == NULL)
-		return 0;
-	event= node->value;
+	if (get_mi_string_param(params, "event", &event.s, &event.len) < 0)
+		return init_mi_param_error();
 	if(event.s== NULL || event.len== 0)
 	{
 		LM_ERR( "empty event parameter\n");
-		return init_mi_tree(400, "Empty event parameter", 21);
+		return init_mi_error(400, MI_SSTR("Empty event parameter"));
 	}
 	LM_DBG("event '%.*s'\n",  event.len, event.s);
 
-	node = node->next;
-	if(node == NULL)
-		return 0;
-	if(node->value.s== NULL || node->value.len== 0)
-	{
-		LM_ERR( "empty event parameter\n");
-		return init_mi_tree(400, "Empty event parameter", 21);
-	}
-	if(str2int(&node->value, &refresh_type)< 0)
-	{
-		LM_ERR("converting string to int\n");
-		goto error;
-	}
-
-	if(node->next!= NULL)
-	{
-		LM_ERR( "Too many parameters\n");
-		return init_mi_tree(400, "Too many parameters", 19);
-	}
+	if (get_mi_int_param(params, "refresh_type", &refresh_type) < 0)
+		return init_mi_param_error();
 
 	ev= contains_event(&event, NULL);
 	if(ev== NULL)
@@ -621,7 +633,7 @@ static struct mi_root* mi_refreshWatchers(struct mi_root* cmd, void* param)
 
 	}
 
-	return init_mi_tree( 200, MI_OK_S, MI_OK_LEN);
+	return init_mi_result_ok();
 
 error:
 	if(rules_doc)
@@ -636,41 +648,44 @@ error:
 /*
  *  mi cmd: cleanup
  *		* */
-
-static struct mi_root* mi_cleanup(struct mi_root* cmd, void* param)
+static mi_response_t *mi_cleanup(const mi_params_t *params,
+								struct mi_handler *async_hdl)
 {
 	LM_DBG("mi_cleanup:start\n");
 
 	(void)msg_watchers_clean(0,0);
 	(void)msg_presentity_clean(0,0);
 
-	return init_mi_tree(200, MI_OK_S, MI_OK_LEN);
+	return init_mi_result_ok();
 }
 
 
-static inline int mi_print_phtable_record(struct mi_node *rpl, pres_entry_t* pres)
+static inline int mi_print_phtable_record(mi_item_t *p_item,pres_entry_t* pres)
 {
-	struct mi_node* node;
-	struct mi_attr* attr;
-	char* p;
-	int len;
+	mi_item_t *item;
+
+	item = add_mi_object(p_item, NULL, 0);
+	if (!item)
+		goto error;
+
+	if (add_mi_string(item, MI_SSTR("pres_uri"),
+		pres->pres_uri.s, pres->pres_uri.len) < 0)
+		goto error;
+	
+	if (add_mi_number(item, MI_SSTR("event"), pres->event) < 0)
+		goto error;
+	
+	if (add_mi_number(item, MI_SSTR("etag_count"), pres->etag_count) < 0)
+		goto error;
 
-	node = add_mi_node_child(rpl, 0, "pres_uri", 8, pres->pres_uri.s, pres->pres_uri.len);
-	if(node == NULL) goto error;
-	p= int2str((unsigned long)pres->event, &len);
-	attr = add_mi_attr(node, MI_DUP_VALUE, "event", 5, p, len);
-	if(attr == NULL) goto error;
-	p= int2str((unsigned long)pres->etag_count, &len);
-	attr = add_mi_attr(node, MI_DUP_VALUE, "etag_count", 10, p, len);
-	if(attr == NULL) goto error;
 	if (pres->sphere)
-	{
-		attr = add_mi_attr(node, MI_DUP_VALUE, "sphere", 6,
-						pres->sphere, strlen(pres->sphere));
-		if(attr == NULL) goto error;
-	}
-	attr = add_mi_attr(node, MI_DUP_VALUE, "etag", 4, pres->etag, pres->etag_len);
-	if(attr == NULL) goto error;
+		if (add_mi_string(item, MI_SSTR("sphere"),
+			pres->sphere, strlen(pres->sphere)) < 0)
+			goto error;
+
+	if (add_mi_string(item, MI_SSTR("etag"),
+		pres->etag, pres->etag_len) < 0)
+		goto error;
 
 	return 0;
 error:
@@ -678,18 +693,17 @@ error:
 	return -1;
 }
 
-static struct mi_root* mi_list_phtable(struct mi_root* cmd, void* param)
+static mi_response_t *mi_list_phtable(const mi_params_t *params,
+								struct mi_handler *async_hdl)
 {
-
-	struct mi_root *rpl_tree;
-	struct mi_node* rpl;
+	mi_response_t *resp;
+	mi_item_t *resp_arr;
 	pres_entry_t* p;
 	unsigned int i;
 
-	rpl_tree = init_mi_tree( 200, MI_OK_S, MI_OK_LEN);
-	if (rpl_tree==NULL) return NULL;
-	rpl = &rpl_tree->node;
-	rpl->flags |= MI_IS_ARRAY;
+	resp = init_mi_result_array(&resp_arr);
+	if (!resp)
+		return 0;
 
 	for(i= 0; i<phtable_size; i++)
 	{
@@ -697,79 +711,110 @@ static struct mi_root* mi_list_phtable(struct mi_root* cmd, void* param)
 		p = pres_htable[i].entries->next;
 		while(p)
 		{
-			if(mi_print_phtable_record(rpl, p)<0) goto error;
+			if(mi_print_phtable_record(resp_arr, p)<0) goto error;
 			p= p->next;;
 		}
 		lock_release(&pres_htable[i].lock);
 	}
-	return rpl_tree;
+	return resp;
 error:
 	lock_release(&pres_htable[i].lock);
 	LM_ERR("Unable to create reply\n");
-	free_mi_tree(rpl_tree);
+	free_mi_response(resp);
 	return NULL;
 }
 
 
-static inline int mi_print_shtable_record(struct mi_node *rpl, subs_t* s)
+static inline int mi_print_shtable_record(mi_item_t *p_item, subs_t* s)
 {
-	struct mi_node *node, *node1;
-	struct mi_attr *attr;
 	time_t _ts;
 	char date_buf[MI_DATE_BUF_LEN];
 	int date_buf_len;
-	char *p;
-	int len;
+	rr_t *rr_head = NULL;
+	mi_item_t *item;
+	struct tm t;
+
+	item = add_mi_object(p_item, NULL, 0);
+	if (!item)
+		return 0;
+
+	if (add_mi_string(item, MI_SSTR("pres_uri"),
+		s->pres_uri.s, s->pres_uri.len) < 0)
+		goto error;
+	
+	if (add_mi_string(item, MI_SSTR("event"),
+		s->event->name.s, s->event->name.len) < 0)
+		goto error;
 
-	node = add_mi_node_child(rpl, 0, "pres_uri", 8, s->pres_uri.s, s->pres_uri.len);
-	if (node==NULL) goto error;
-	attr = add_mi_attr(node, MI_DUP_VALUE, "event", 5, s->event->name.s, s->event->name.len);
-	if (attr==NULL) goto error;
 	/*
 	attr = add_mi_attr(node, MI_DUP_VALUE, "event_id", 8, s->event_id.s, s->event_id.len);
 	if (attr==NULL) goto error;
 	*/
-	if (attr==NULL) goto error;
 	_ts = (time_t)s->expires;
+	localtime_r(&_ts, &t);
 	date_buf_len = strftime(date_buf, MI_DATE_BUF_LEN - 1,
-						"%Y-%m-%d %H:%M:%S", localtime(&_ts));
+						"%Y-%m-%d %H:%M:%S", &t);
 	if (date_buf_len != 0) {
-		attr = add_mi_attr(node, MI_DUP_VALUE, "expires", 7, date_buf, date_buf_len);
+		if (add_mi_string(item, MI_SSTR("expires"),
+			date_buf, date_buf_len) < 0)
+			goto error;
 	} else {
-		p= int2str((unsigned long)s->expires, &len);
-		attr = add_mi_attr(node, MI_DUP_VALUE, "expires", 7, p, len);
+		if (add_mi_number(item, MI_SSTR("expires"), s->expires) < 0)
+			goto error;
 	}
-	if (attr==NULL) goto error;
-	p= int2str((unsigned long)s->db_flag, &len);
-	attr = add_mi_attr(node, MI_DUP_VALUE, "db_flag", 7, p, len);
-	if (attr==NULL) goto error;
-	p= int2str((unsigned long)s->version, &len);
-	attr = add_mi_attr(node, MI_DUP_VALUE, "version", 7, p, len);
-	if (attr==NULL) goto error;
 
-	node1 = add_mi_node_child(node, 0, "to_user", 7, s->to_user.s, s->to_user.len);
-	if (node1==NULL) goto error;
-	attr = add_mi_attr(node1, MI_DUP_VALUE, "to_domain", 9, s->to_domain.s, s->to_domain.len);
-	if (attr==NULL) goto error;
-	attr = add_mi_attr(node1, MI_DUP_VALUE, "to_tag", 6, s->to_tag.s, s->to_tag.len);
-	if (attr==NULL) goto error;
+	if (add_mi_number(item, MI_SSTR("db_flag"), s->db_flag) < 0)
+		goto error;
 
-	node1 = add_mi_node_child(node, 0, "from_user", 9, s->from_user.s, s->from_user.len);
-	if (node1==NULL) goto error;
-	attr = add_mi_attr(node1, MI_DUP_VALUE, "from_domain", 11,
-				s->from_domain.s, s->from_domain.len);
-	if (attr==NULL) goto error;
-	attr = add_mi_attr(node1, MI_DUP_VALUE, "from_tag", 8, s->from_tag.s, s->from_tag.len);
-	if (attr==NULL) goto error;
+	if (add_mi_number(item, MI_SSTR("version"), s->version) < 0)
+		goto error;
 
-	node1 = add_mi_node_child(node, 0, "callid", 6, s->callid.s, s->callid.len);
-	if (node1==NULL) goto error;
-	p= int2str((unsigned long)s->local_cseq, &len);
-	attr = add_mi_attr(node1, MI_DUP_VALUE, "local_cseq", 10, p, len);
-	if (attr==NULL) goto error;
-	p= int2str((unsigned long)s->remote_cseq, &len);
-	attr = add_mi_attr(node1, MI_DUP_VALUE, "remote_cseq", 11, p, len);
-	if (attr==NULL) goto error;
+	if (s->sh_tag.len && add_mi_string(item, MI_SSTR("sharing_tag"),
+	s->sh_tag.s, s->sh_tag.len) < 0)
+		goto error;
+
+	if (add_mi_string(item, MI_SSTR("to_user"),
+		s->to_user.s, s->to_user.len) < 0)
+		goto error;
+	if (add_mi_string(item, MI_SSTR("to_domain"),
+		s->to_domain.s, s->to_domain.len) < 0)
+		goto error;
+	if (add_mi_string(item, MI_SSTR("to_tag"),
+		s->to_tag.s, s->to_tag.len) < 0)
+		goto error;
+
+	if (add_mi_string(item, MI_SSTR("from_user"),
+		s->from_user.s, s->from_user.len) < 0)
+		goto error;
+	if (add_mi_string(item, MI_SSTR("from_domain"),
+		s->from_domain.s, s->from_domain.len) < 0)
+		goto error;
+	if (add_mi_string(item, MI_SSTR("from_tag"),
+		s->from_tag.s, s->from_tag.len) < 0)
+		goto error;
+
+	if (add_mi_string(item, MI_SSTR("contact"),
+		s->contact.s, s->contact.len) < 0)
+		goto error;
+
+	if (s->record_route.s && s->record_route.len &&
+		parse_rr_body(s->record_route.s, s->record_route.len, &rr_head) < 0)
+		goto error;
+	if (rr_head) {
+		if (add_mi_string(item, MI_SSTR("next_hop"),
+			rr_head->nameaddr.uri.s, rr_head->nameaddr.uri.len) < 0)
+			goto error;
+	}
+
+	if (add_mi_string(item, MI_SSTR("callid"),
+		s->callid.s, s->callid.len) < 0)
+		goto error;
+
+	if (add_mi_number(item, MI_SSTR("local_cseq"), s->local_cseq) < 0)
+		goto error;
+
+	if (add_mi_number(item, MI_SSTR("remote_cseq"), s->remote_cseq) < 0)
+		goto error;
 
 	return 0;
 error:
@@ -778,43 +823,110 @@ error:
 }
 
 
-static struct mi_root* mi_list_shtable(struct mi_root* cmd, void* param)
+static inline int from_to_match_subs(subs_t *s, str *match_from, str *match_to,
+									char *from_w, char *to_w)
 {
-	struct mi_root *rpl_tree;
-	struct mi_node* rpl;
-	subs_t *s;
-	unsigned int i,j;
+	if (match_from->s)
+		pkg_free(match_from->s);
+	if (match_to->s)
+		pkg_free(match_to->s);
 
-	rpl_tree = init_mi_tree( 200, MI_OK_S, MI_OK_LEN);
-	if (rpl_tree==NULL) return NULL;
-	rpl = &rpl_tree->node;
-	rpl->flags |= MI_IS_ARRAY;
+	if (uandd_to_uri(s->from_user, s->from_domain, match_from) < 0)
+		return -1;
+	if (uandd_to_uri(s->to_user, s->to_domain, match_to) < 0)
+		return -1;
 
-	for(i=0,j=0; i< shtable_size; i++)
-	{
-		lock_get(&subs_htable[i].lock);
-		s = subs_htable[i].entries->next;
-		while(s)
-		{
-			if(mi_print_shtable_record(rpl, s)<0) goto error;
-			s= s->next;
-			j++;
-		}
-		lock_release(&subs_htable[i].lock);
+	if (fnmatch(from_w, match_from->s, 0) || fnmatch(to_w, match_to->s, 0))
+		return 1;
 
-		if ((j % 50) == 0)
-			flush_mi_tree(rpl_tree);
-	}
-	return rpl_tree;
-error:
-	lock_release(&subs_htable[i].lock);
-	LM_ERR("Unable to create reply\n");
-	free_mi_tree(rpl_tree);
-	return NULL;
+	return 0;
 }
 
 
-int pres_update_status(subs_t subs, str reason, db_key_t* query_cols,
+static mi_response_t *mi_list_shtable(const mi_params_t *params, str *from, str *to)
+{
+	mi_response_t *resp;
+	mi_item_t *resp_arr;
+	subs_t *s;
+	unsigned int i,j;
+	char from_w[256], to_w[256];
+	str match_from = {0,0}, match_to = {0,0};
+	int rc;
+
+	resp = init_mi_result_array(&resp_arr);
+	if (!resp)
+		return 0;
+
+	/* from wildcard */
+	if (from) {
+		memcpy(from_w, from->s, from->len);
+		from_w[from->len] = 0;
+	}
+
+	/* to wildcard */
+	if (to) {
+		memcpy(to_w, to->s, to->len);
+		to_w[to->len] = 0;
+	}
+
+	for (i = 0, j = 0; i < shtable_size; i++) {
+
+		lock_get(&subs_htable[i].lock);
+		for (s = subs_htable[i].entries->next; s; s = s->next) {
+			if (from) {
+				/* print subscribtion if "from" and "to" match with given wildcard */
+				rc = from_to_match_subs(s, &match_from, &match_to, from_w, to_w);
+				if (rc < 0)
+					goto error;
+				else if (rc == 1)
+					continue;
+			}
+
+			if (mi_print_shtable_record(resp_arr, s) < 0)
+				goto error;
+			j++;
+		}
+		lock_release(&subs_htable[i].lock);
+	}
+
+	if (match_from.s)
+		pkg_free(match_from.s);
+	if (match_to.s)
+		pkg_free(match_to.s);
+
+	return resp;
+
+error:
+	if (match_from.s)
+		pkg_free(match_from.s);
+	if (match_to.s)
+		pkg_free(match_to.s);
+	lock_release(&subs_htable[i].lock);
+	LM_ERR("Unable to create reply\n");
+	free_mi_response(resp);
+	return NULL;
+}
+
+static mi_response_t *mi_list_shtable_1(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	return mi_list_shtable(params, NULL, NULL);
+}
+
+static mi_response_t *mi_list_shtable_2(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	str from, to;
+
+	if (get_mi_string_param(params, "from", &from.s, &from.len) < 0)
+		return init_mi_param_error();
+	if (get_mi_string_param(params, "to", &to.s, &to.len) < 0)
+		return init_mi_param_error();
+
+	return mi_list_shtable(params, &from, &to);
+}
+
+int pres_update_status(subs_t *subs, str reason, db_key_t* query_cols,
         db_val_t* query_vals, int n_query_cols, subs_t** subs_array)
 {
 	static db_ps_t my_del_ps = NULL;
@@ -844,29 +956,29 @@ int pres_update_status(subs_t subs, str reason, db_key_t* query_cols,
 	update_vals[u_reason_col].type= DB_STR;
 	n_update_cols++;
 
-	status= subs.status;
-	if(subs.event->get_auth_status(&subs)< 0)
+	status= subs->status;
+	if(subs->event->get_auth_status(subs)< 0)
 	{
 		LM_ERR( "getting status from rules document\n");
 		return -1;
 	}
-	LM_DBG("subs.status= %d\n", subs.status);
-	if(get_status_str(subs.status)== NULL)
+	LM_DBG("subs.status= %d\n", subs->status);
+	if(get_status_str(subs->status)== NULL)
 	{
-		LM_ERR("wrong status: %d\n", subs.status);
+		LM_ERR("wrong status: %d\n", subs->status);
 		return -1;
 	}
 
-	if(subs.status!= status || reason.len!= subs.reason.len ||
-		(reason.s && subs.reason.s && strncmp(reason.s, subs.reason.s,
+	if(subs->status!= status || reason.len!= subs->reason.len ||
+		(reason.s && subs->reason.s && strncmp(reason.s, subs->reason.s,
 											reason.len)))
 	{
 		/* update in watchers_table */
-		query_vals[q_wuser_col].val.str_val= subs.from_user;
-		query_vals[q_wdomain_col].val.str_val= subs.from_domain;
+		query_vals[q_wuser_col].val.str_val= subs->from_user;
+		query_vals[q_wdomain_col].val.str_val= subs->from_domain;
 
-		update_vals[u_status_col].val.int_val= subs.status;
-		update_vals[u_reason_col].val.str_val= subs.reason;
+		update_vals[u_status_col].val.int_val= subs->status;
+		update_vals[u_reason_col].val.str_val= subs->reason;
 
 		if (pa_dbf.use_table(pa_db, &watchers_table) < 0)
 		{
@@ -876,10 +988,10 @@ int pres_update_status(subs_t subs, str reason, db_key_t* query_cols,
 
 		/* if status is terminated and reason="deactivated",
 		 * delete the record from table */
-		if(subs.status == TERMINATED_STATUS && subs.reason.len==11 &&
-				strncmp(subs.reason.s, "deactivated", 11)==0)
+		if(subs->status == TERMINATED_STATUS && subs->reason.len==11 &&
+				strncmp(subs->reason.s, "deactivated", 11)==0)
 		{
-			CON_PS_REFERENCE(pa_db) = &my_del_ps;
+			CON_SET_CURR_PS(pa_db, &my_del_ps);
 			if(pa_dbf.delete(pa_db, query_cols, 0, query_vals, n_query_cols)<0)
 			{
 				LM_ERR( "in sql delete\n");
@@ -888,7 +1000,7 @@ int pres_update_status(subs_t subs, str reason, db_key_t* query_cols,
 		}
 		else
 		{
-			CON_PS_REFERENCE(pa_db) = &my_upd_ps;
+			CON_SET_CURR_PS(pa_db, &my_upd_ps);
 			if(pa_dbf.update(pa_db, query_cols, 0, query_vals, update_cols,
 						update_vals, n_query_cols, n_update_cols)< 0)
 			{
@@ -899,12 +1011,12 @@ int pres_update_status(subs_t subs, str reason, db_key_t* query_cols,
 
 		/* save in the list all affected dialogs */
 		/* if status switches to terminated -> delete dialog */
-		if(update_pw_dialogs(&subs, subs.db_flag, subs_array)< 0)
+		if(update_pw_dialogs(subs, subs->db_flag, subs_array)< 0)
 		{
 			LM_ERR( "extracting dialogs from [watcher]=%.*s@%.*s to"
-				" [presentity]=%.*s\n",	subs.from_user.len, subs.from_user.s,
-				subs.from_domain.len, subs.from_domain.s, subs.pres_uri.len,
-				subs.pres_uri.s);
+				" [presentity]=%.*s\n",	subs->from_user.len, subs->from_user.s,
+				subs->from_domain.len, subs->from_domain.s, subs->pres_uri.len,
+				subs->pres_uri.s);
 			return -1;
 		}
 	}
@@ -948,8 +1060,7 @@ int pres_db_delete_status(subs_t* s)
 	query_vals[n_query_cols].val.str_val= s->from_domain;
 	n_query_cols++;
 
-	CON_PS_REFERENCE(pa_db) = &my_ps;
-
+	CON_SET_CURR_PS(pa_db, &my_ps);
 	if(pa_dbf.delete(pa_db, query_cols, 0, query_vals, n_query_cols)< 0)
 	{
 		LM_ERR("sql delete failed\n");
@@ -966,7 +1077,7 @@ int terminate_watchers(str *pres_uri, pres_ev_t* ev)
 	subs_t *tmp;
 
 	/* get all watchers for the presentity */
-	all_s = get_subs_dialog( pres_uri, ev, NULL);
+	all_s = get_subs_dialog( pres_uri, ev, NULL, NULL);
 	if ( all_s==NULL ) {
 		LM_DBG("No subscription dialogs found for <%.*s>\n",
 			pres_uri->len, pres_uri->s);
@@ -1056,7 +1167,7 @@ int update_watchers_status(str pres_uri, pres_ev_t* ev, str* rules_doc)
 		goto done;
 	}
 
-//	CON_PS_REFERENCE(pa_db) = &my_ps;
+//	CON_SET_CURR_PS(pa_db, &my_ps);
 	if(pa_dbf.query(pa_db, query_cols, 0, query_vals, result_cols,n_query_cols,
 				n_result_cols, 0, &result)< 0)
 	{
@@ -1150,7 +1261,7 @@ int update_watchers_status(str pres_uri, pres_ev_t* ev, str* rules_doc)
 			subs.status = ws_list[i].status;
 			memset(&subs.reason, 0, sizeof(str));
 
-			if( pres_update_status(subs, reason, query_cols, query_vals,
+			if( pres_update_status(&subs, reason, query_cols, query_vals,
 					n_query_cols, &subs_array)< 0)
 			{
 				LM_ERR("failed to update watcher status\n");
@@ -1193,7 +1304,7 @@ int update_watchers_status(str pres_uri, pres_ev_t* ev, str* rules_doc)
 		subs.status= status;
 		memset(&subs.reason, 0, sizeof(str));
 
- 		if( pres_update_status(subs,reason, query_cols, query_vals,
+		if( pres_update_status(&subs,reason, query_cols, query_vals,
 					n_query_cols, &subs_array)< 0)
 		{
 			LM_ERR("failed to update watcher status\n");
@@ -1289,7 +1400,7 @@ int refresh_send_winfo_notify(watcher_t* watchers, str pres_uri,
 	if(watchers->next== NULL)
 		return 0;
 
-	subs_array= get_subs_dialog(&pres_uri, ev, NULL);
+	subs_array= get_subs_dialog(&pres_uri, ev, NULL, NULL);
 	if(subs_array == NULL)
 	{
 		LM_DBG("Could not get subscription dialog\n");
@@ -1402,3 +1513,48 @@ static int update_pw_dialogs(subs_t* subs, unsigned int hash_code, subs_t** subs
 
     return 0;
 }
+
+static mi_response_t *mi_pres_expose_1(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	pres_ev_t* ev;
+	str event;
+
+	if (get_mi_string_param(params, "event", &event.s, &event.len) < 0)
+		return init_mi_param_error();
+	if (event.s == NULL || event.len == 0)
+		return init_mi_error(404, MI_SSTR("Invalid event"));
+
+	ev = contains_event(&event, NULL);
+	if (!ev)
+		return init_mi_error(404, MI_SSTR("unknown event"));
+
+	if (pres_expose_evi(ev, NULL) < 0)
+		return NULL;
+
+	return init_mi_result_ok();
+}
+
+static mi_response_t *mi_pres_expose_2(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	str event, pres_fn;
+	pres_ev_t* ev;
+
+	if (get_mi_string_param(params, "event", &event.s, &event.len) < 0)
+		return init_mi_param_error();
+	if (event.s == NULL || event.len == 0)
+		return init_mi_error(404, MI_SSTR("Invalid event"));
+
+	ev = contains_event(&event, NULL);
+	if (!ev)
+		return init_mi_error(404, MI_SSTR("unknown event"));
+
+	if (get_mi_string_param(params, "filter", &pres_fn.s, &pres_fn.len) < 0)
+		return init_mi_param_error();
+
+	if (pres_expose_evi(ev, &pres_fn) < 0)
+		return NULL;
+
+	return init_mi_result_ok();
+}
