@@ -73,7 +73,7 @@ int generate_ETag(int publ_count, str* etag)
 			prefix, (int)startup_time, pid, counter, publ_count);
 	if( etag->len <0 )
 	{
-		LM_ERR("unsuccessfull sprintf\n ");
+		LM_ERR("unsuccessful sprintf\n");
 		return -1;
 	}
 	if(etag->len > ETAG_LEN)
@@ -127,7 +127,7 @@ int publ_send200ok(struct sip_msg *msg, int lexpire, str etag)
 	hdr_append2.len = sprintf(hdr_append2.s, "SIP-ETag: %.*s\r\n", etag.len, etag.s);
 	if(hdr_append2.len < 0)
 	{
-		LM_ERR("unsuccessful sprintf\n ");
+		LM_ERR("unsuccessful sprintf\n");
 		goto error;
 	}
 	if(hdr_append2.len+1 > size)
@@ -392,7 +392,8 @@ int check_if_dialog(str body, int *is_dialog)
 }
 
 
-int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_reply)
+int update_presentity(struct sip_msg* msg, presentity_t* presentity,
+															int* sent_reply)
 {
 //	static db_ps_t my_ps_insert = NULL, my_ps_update_no_body = NULL,
 //		   my_ps_update_body = NULL;
@@ -402,9 +403,7 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 	db_val_t query_vals[13], update_vals[8];
 	int n_query_cols = 0;
 	int n_update_cols = 0;
-	str etag= {NULL, 0};
 	str notify_body = {NULL, 0};
-	str cur_etag= {NULL, 0};
 	str* rules_doc= NULL;
 	str pres_uri= {NULL, 0};
 	pres_entry_t* p= NULL;
@@ -453,19 +452,13 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 	query_vals[n_query_cols].val.str_val = presentity->event->name;
 	n_query_cols++;
 
-	query_cols[n_query_cols] = &str_etag_col;
-	query_ops[n_query_cols] = OP_EQ;
-	query_vals[n_query_cols].type = DB_STR;
-	query_vals[n_query_cols].nul = 0;
-	query_vals[n_query_cols].val.str_val = presentity->etag;
-	n_query_cols++;
-
 	result_cols[0] = &str_etag_col;
 	hash_code= core_hash(&pres_uri, NULL, phtable_size);
 
 	if(presentity->etag_new)
 	{
-		if (msg && publ_send200ok(msg,presentity->expires,presentity->etag)<0)
+		if (msg && publ_send200ok(msg, presentity->expires,
+		presentity->new_etag)<0)
 		{
 			LM_ERR("sending 200OK\n");
 			goto error;
@@ -474,7 +467,8 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 
 		/* insert new record in hash_table */
 		p = insert_phtable(&pres_uri, presentity->event->evp->parsed,
-					&presentity->etag, presentity->sphere, 1);
+				&presentity->new_etag, presentity->sphere,
+				presentity->flags, 1);
 		if (p==NULL)
 		{
 			LM_ERR("inserting record in hash table\n");
@@ -482,6 +476,13 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 		}
 
 		/* insert new record into database */
+		query_cols[n_query_cols] = &str_etag_col;
+		query_ops[n_query_cols] = OP_EQ;
+		query_vals[n_query_cols].type = DB_STR;
+		query_vals[n_query_cols].nul = 0;
+		query_vals[n_query_cols].val.str_val = presentity->new_etag;
+		n_query_cols++;
+
 		query_cols[n_query_cols] = &str_expires_col;
 		query_vals[n_query_cols].type = DB_INT;
 		query_vals[n_query_cols].nul = 0;
@@ -533,7 +534,7 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 
 		LM_DBG("inserting %d cols into table\n",n_query_cols);
 
-		//CON_PS_REFERENCE(pa_db) = &my_ps_insert;
+		//CON_SET_CURR_PS(pa_db, &my_ps_insert);
 		if (pa_dbf.insert(pa_db, query_cols, query_vals, n_query_cols) < 0)
 		{
 			LM_ERR("inserting new record in database\n");
@@ -543,24 +544,32 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 	}
 	else
 	{
+		query_cols[n_query_cols] = &str_etag_col;
+		query_ops[n_query_cols] = OP_EQ;
+		query_vals[n_query_cols].type = DB_STR;
+		query_vals[n_query_cols].nul = 0;
+		query_vals[n_query_cols].val.str_val = presentity->old_etag;
+		n_query_cols++;
+
 		lock_get(&pres_htable[hash_code].lock);
 		p = search_phtable_etag(&pres_uri, presentity->event->evp->parsed,
-				&presentity->etag, hash_code);
+				&presentity->old_etag, hash_code);
 		if (p) {
 
 			turn = p->last_turn++;
 			LM_DBG("pres <%.*s> my turn is %d, current turn is %d\n",pres_uri.len,
 				pres_uri.s, turn, p->current_turn);
 
-			/* wait to get our turn as order of handling pubishs
+			/* wait to get our turn as order of handling publish
 			   (need to wait the ongoing published to terminate
 			   before starting */
 			while (p && turn!=p->current_turn) {
 				lock_release(&pres_htable[hash_code].lock);
 				sleep_us(100);
 				lock_get(&pres_htable[hash_code].lock);
-				p = search_phtable_etag(&pres_uri, presentity->event->evp->parsed,
-					&presentity->etag, hash_code);
+				p = search_phtable_etag(&pres_uri,
+					presentity->event->evp->parsed, &presentity->old_etag,
+					hash_code);
 			}
 
 		} else {
@@ -572,6 +581,7 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 					LM_ERR("unsuccessful sql use table\n");
 					goto error;
 			}
+
 			if (pa_dbf.query (pa_db, query_cols, query_ops, query_vals,
 					 result_cols, n_query_cols, 1, 0, &result) < 0)
 			{
@@ -584,8 +594,8 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 			if (result->n <= 0)
 			{
 					pa_dbf.free_result(pa_db, result);
-					LM_ERR("No E_Tag match [%.*s]\n", presentity->etag.len,
-							presentity->etag.s);
+					LM_ERR("No E_Tag match [%.*s]\n", presentity->old_etag.len,
+							presentity->old_etag.s);
 					if (msg && sigb.reply(msg, 412, &pu_412_rpl, 0)==-1 )
 					{
 						LM_ERR("sending '412 Conditional request failed' reply\n");
@@ -597,7 +607,7 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 
 			pa_dbf.free_result(pa_db, result);
 			LM_INFO("*** found in db but not in htable [%.*s]\n",
-					presentity->etag.len, presentity->etag.s);
+					presentity->old_etag.len, presentity->old_etag.s);
 		}
 
 		/* record found */
@@ -612,7 +622,8 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 			p = NULL;
 
 			lock_release(&pres_htable[hash_code].lock);
-			if(msg && publ_send200ok(msg,presentity->expires,presentity->etag)<0)
+			if(msg && publ_send200ok(msg, presentity->expires,
+			presentity->old_etag)<0)
 			{
 				LM_ERR("sending 200OK reply\n");
 				goto error;
@@ -620,7 +631,7 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 			*sent_reply= 1;
 
 			if(publ_notify(presentity, pres_uri, body.s ? &body : 0,
-			&presentity->etag, rules_doc, NULL, 1) < 0)
+			&presentity->old_etag, rules_doc, NULL, 1, NULL) < 0)
 			{
 				LM_ERR("while sending notify\n");
 				goto error;
@@ -631,7 +642,7 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 				LM_ERR("unsuccessful sql use table\n");
 				goto error;
 			}
-			//CON_PS_REFERENCE(pa_db) = &my_ps_delete;
+			//CON_SET_CURR_PS(pa_db, &my_ps_delete);
 			if(pa_dbf.delete(pa_db,query_cols,0,query_vals,n_query_cols)<0)
 			{
 				LM_ERR("unsuccessful sql delete operation");
@@ -642,7 +653,8 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 			/* Send another NOTIFY, this time rely on whatever is on the DB,
 			 *  so in case there are no documents an empty
 			 * NOTIFY will be sent to the watchers */
-			if(publ_notify(presentity, pres_uri, NULL, NULL, rules_doc, NULL, 1) < 0)
+			if (publ_notify(presentity, pres_uri, NULL, NULL, rules_doc, NULL,
+			1, NULL) < 0)
 			{
 				LM_ERR("while sending notify\n");
 				goto error;
@@ -653,23 +665,29 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 
 		if(presentity->event->etag_not_new== 0)
 		{
-			if(generate_ETag(p?p->etag_count:0, &etag) < 0)
+			/* if we already have a new etag (probably received via cluster
+			 * replication) do not generate a new local one */
+			if (presentity->new_etag.s==NULL)
 			{
-				LM_ERR("while generating etag\n");
-				lock_release(&pres_htable[hash_code].lock);
-				goto error;
+				if(generate_ETag(p?p->etag_count:0, &presentity->new_etag) < 0)
+				{
+					LM_ERR("while generating etag\n");
+					lock_release(&pres_htable[hash_code].lock);
+					goto error;
+				}
 			}
-			cur_etag= etag;
 			if(p)
 			{
-				update_pres_etag(p, &etag);
+				update_pres_etag(p, &presentity->new_etag);
+				p->flags = presentity->flags;
 				lock_release(&pres_htable[hash_code].lock);
 			}
 			else
 			{
 				lock_release(&pres_htable[hash_code].lock);
 				p = insert_phtable(&pres_uri, presentity->event->evp->parsed,
-							&presentity->etag, presentity->sphere, 1);
+						&presentity->new_etag, presentity->sphere,
+						presentity->flags, 1);
 				if ( p==NULL )
 				{
 					LM_ERR("inserting record in hash table\n");
@@ -680,14 +698,14 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 		else
 		{
 			lock_release(&pres_htable[hash_code].lock);
-			cur_etag= presentity->etag;
+			presentity->new_etag = presentity->old_etag;
 		}
 
 		n_update_cols= 0;
 		update_keys[n_update_cols] = &str_etag_col;
 		update_vals[n_update_cols].type = DB_STR;
 		update_vals[n_update_cols].nul = 0;
-		update_vals[n_update_cols].val.str_val = cur_etag;
+		update_vals[n_update_cols].val.str_val = presentity->new_etag;
 		n_update_cols++;
 
 		update_keys[n_update_cols] = &str_expires_col;
@@ -757,11 +775,11 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 					goto error;
 				}
 			}
-			//CON_PS_REFERENCE(pa_db) = &my_ps_update_body;
+			//CON_SET_CURR_PS(pa_db, &my_ps_update_body);
 		}
 		else
 		{
-			//CON_PS_REFERENCE(pa_db) = &my_ps_update_no_body;
+			//CON_SET_CURR_PS(pa_db, &my_ps_update_no_body);
 		}
 
 		if (pa_dbf.use_table(pa_db, &presentity_table) < 0)
@@ -778,7 +796,8 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 		}
 
 		/* send 200OK */
-		if (msg && publ_send200ok(msg, presentity->expires, cur_etag)<0 )
+		if (msg && publ_send200ok(msg, presentity->expires,
+		presentity->new_etag)<0 )
 		{
 			LM_ERR("sending 200OK reply\n");
 			goto error;
@@ -792,7 +811,7 @@ int update_presentity(struct sip_msg* msg, presentity_t* presentity, int* sent_r
 send_notify:
 
 	if (publ_notify(presentity, pres_uri, body.s?&body:0,
-				NULL, rules_doc, NULL, 1)<0)
+				NULL, rules_doc, NULL, 1, NULL)<0)
 	{
 		LM_ERR("while sending Notify requests to watchers\n");
 		goto error;
@@ -803,7 +822,8 @@ send_mxd_notify:
 	if(mix_dialog_presence && *pres_event_p &&
 	presentity->event->evp->parsed == EVENT_DIALOG)
 	{
-		str* dialog_body= NULL;
+		str* dialog_body = NULL;
+		pres_ev_t *ev_bk;
 
 		LM_DBG("Publish for event dialog - try to send Notify for presence\n");
 
@@ -811,10 +831,13 @@ send_mxd_notify:
 		if(dialog_body)
 		{
 			/* send Notify for presence */
+			ev_bk = presentity->event;
 			presentity->event = *pres_event_p;
-			if (publ_notify(presentity, pres_uri, 0, NULL, 0, dialog_body, 1)<0)
+			if (publ_notify(presentity, pres_uri, 0, NULL, 0, dialog_body,
+			1, NULL)<0)
 			{
 				LM_ERR("while sending Notify requests to watchers\n");
+				presentity->event = ev_bk;
 				if(dialog_body && dialog_body!=FAKED_BODY)
 				{
 					xmlFree(dialog_body->s);
@@ -822,6 +845,7 @@ send_mxd_notify:
 				}
 				goto error;
 			}
+			presentity->event = ev_bk;
 			if(dialog_body && dialog_body!=FAKED_BODY)
 			{
 				xmlFree(dialog_body->s);
@@ -853,9 +877,6 @@ error:
 	if (p)
 		next_turn_phtable( p, hash_code);
 
-	if(result)
-		pa_dbf.free_result(pa_db, result);
-
 	if (notify_body.s)
 		xmlFree(notify_body.s);
 
@@ -871,7 +892,7 @@ error:
 }
 
 
-/*  This is just a wrappter over "update_presentity" to be able to export the
+/*  This is just a wrapper over "update_presentity" to be able to export the
  *  function via the internal API - basically provides the "publishing" support
  *  without actually having the PUBLISH SIP request, but directly a presentity
  */
@@ -915,7 +936,7 @@ int pres_htable_restore(void)
 		goto error;
 	}
 
-	/* select the whole tabel and all the columns */
+	/* select the whole table and all the columns */
 	if (DB_CAPABILITY(pa_dbf, DB_CAP_FETCH))
 	{
 		if(pa_dbf.query(pa_db,0,0,0,result_cols, 0,
@@ -1005,9 +1026,9 @@ int pres_htable_restore(void)
 				sphere= extract_sphere(body);
 			}
 
-			if(insert_phtable(&uri, event, &etag, sphere, 0)== NULL)
+			if(insert_phtable(&uri, event, &etag, sphere, 0, 0)== NULL)
 			{
-				LM_ERR("inserting record in presentity hash table");
+				LM_ERR("inserting record in presentity hash table\n");
 				pkg_free(uri.s);
 				if(sphere)
 					pkg_free(sphere);
@@ -1042,6 +1063,145 @@ error:
 	return -1;
 }
 
+int pres_expose_evi(pres_ev_t *ev, str *filter)
+{
+	int user_col, domain_col, expires_col, body_col, etag_col;
+	int no_rows = 0, nr_rows, i;
+	int n_result_cols= 0;
+	presentity_t presentity;
+	db_key_t result_cols[5];
+	db_key_t query_cols[3];
+	db_val_t query_vals[3];
+	db_op_t query_ops[3];
+	db_res_t* result = NULL;
+	db_row_t *rows = NULL;
+	db_val_t *row_vals;
+	int nr_vals = 0;
+
+	if (pa_dbf.use_table(pa_db, &presentity_table) < 0) {
+		LM_ERR("cannot select table \"%.*s\"\n",
+				presentity_table.len, presentity_table.s);
+		goto error;
+	}
+
+	query_cols[nr_vals] = &str_expires_col;
+	query_ops[nr_vals] = OP_GT;
+	query_vals[nr_vals].nul = 0;
+	query_vals[nr_vals].type = DB_INT;
+	query_vals[nr_vals].val.int_val = (int)time(NULL);
+	nr_vals++;
+	query_cols[nr_vals] = &str_event_col;
+	query_ops[nr_vals] = OP_EQ;
+	query_vals[nr_vals].nul = 0;
+	query_vals[nr_vals].type = DB_STR;
+	query_vals[nr_vals].val.str_val = ev->name;
+	nr_vals++;
+
+	if (filter) {
+		query_cols[nr_vals] = &str_username_col;
+		query_ops[nr_vals] = " REGEXP ";
+		query_vals[nr_vals].nul = 0;
+		query_vals[nr_vals].type = DB_STR;
+		query_vals[nr_vals].val.str_val = *filter;
+		nr_vals++;
+	}
+
+	result_cols[user_col = n_result_cols++]= &str_username_col;
+	result_cols[domain_col = n_result_cols++]= &str_domain_col;
+	result_cols[expires_col = n_result_cols++]= &str_expires_col;
+	result_cols[etag_col = n_result_cols++]= &str_etag_col;
+	result_cols[body_col = n_result_cols++]= &str_body_col;
+
+	if (DB_CAPABILITY(pa_dbf, DB_CAP_FETCH)) {
+		if (pa_dbf.query(pa_db, query_cols, query_ops, query_vals, result_cols,
+				nr_vals, n_result_cols, 0, 0) < 0) {
+			LM_ERR("cannot query table!\n");
+			goto error;
+		}
+		no_rows = estimate_available_rows(64+64+4+64+512 /* body estimation */,
+				n_result_cols);
+		if (no_rows==0) no_rows = 10;
+		if(pa_dbf.fetch_result(pa_db, &result, no_rows)<0) {
+			LM_ERR("Error fetching rows\n");
+			goto error;
+		}
+	} else {
+		if (pa_dbf.query(pa_db, query_cols, query_ops, query_vals, result_cols,
+				nr_vals, n_result_cols, 0, &result) < 0) {
+			LM_ERR("cannot query table!\n");
+			goto error;
+		}
+	}
+
+	LM_DBG("%d records found in %.*s%s%.*s\n",
+			RES_ROW_N(result), presentity_table.len, presentity_table.s,
+			(filter ? " with filter ": ""), (filter ? filter->len: 0),
+			(filter ? filter->s : ""));
+
+	nr_rows = RES_ROW_N(result);
+
+	do {
+		LM_DBG("loading information from database for %i records\n", nr_rows);
+
+		rows = RES_ROWS(result);
+
+		/* for every row */
+		for(i = 0; i < nr_rows; i++)
+		{
+			row_vals = ROW_VALUES(rows + i);
+
+			if (VAL_NULL(row_vals + user_col) || VAL_NULL(row_vals + domain_col))
+			{
+				LM_ERR("columns %.*s or/and %.*s cannot be null -> skipping\n",
+					str_username_col.len, str_username_col.s,
+					str_domain_col.len, str_domain_col.s);
+				continue;
+			}
+
+			/* not sure if this is needed */
+			if(row_vals[expires_col].val.int_val< (int)time(NULL))
+				continue;
+
+			memset(&presentity, 0, sizeof(presentity_t));
+			presentity.domain.s = (char*)row_vals[domain_col].val.string_val;
+			presentity.domain.len = strlen(presentity.domain.s);
+			presentity.user.s = (char*)row_vals[user_col].val.string_val;
+			presentity.user.len = strlen(presentity.user.s);
+			presentity.new_etag.s = (char*)row_vals[etag_col].val.string_val;
+			presentity.new_etag.len = strlen(presentity.new_etag.s);
+			presentity.event = ev;
+			presentity.expires = row_vals[expires_col].val.int_val;
+			presentity.body.s = (char*)row_vals[body_col].val.string_val;
+			presentity.body.len = strlen(presentity.body.s);
+
+			/* send event E_PRESENCE_EXPOSED */
+			presence_raise_event(exposed_event_id, &presentity);
+		}
+
+		/* any more data to be fetched ?*/
+		if (DB_CAPABILITY(pa_dbf, DB_CAP_FETCH))
+		{
+			if (pa_dbf.fetch_result(pa_db, &result, no_rows) < 0)
+			{
+				LM_ERR("fetching more rows failed\n");
+				goto error;
+			}
+			nr_rows = RES_ROW_N(result);
+		} else
+			nr_rows = 0;
+
+	} while(nr_rows > 0);
+
+	pa_dbf.free_result(pa_db, result);
+
+	return 0;
+
+error:
+	if (result)
+		pa_dbf.free_result(pa_db, result);
+	return -1;
+}
+
 char* extract_sphere(str body)
 {
 
@@ -1194,7 +1354,7 @@ char* get_sphere(str* pres_uri)
 		return NULL;
 	}
 
-	// CON_PS_REFERENCE(pa_db) = &my_ps;
+	// CON_SET_CURR_PS(pa_db, &my_ps);
 	if (pa_dbf.query (pa_db, query_cols, 0, query_vals,
 		 result_cols, n_query_cols, n_result_cols, &query_str ,  &result) < 0)
 	{
