@@ -89,7 +89,7 @@ int dialog_offline_body(str* body, str** offline_body)
 
 	if (!offline_body)
 	{
-		LM_ERR("invalid parameter");
+		LM_ERR("invalid parameter\n");
 		return OFFB_STATUS_ERROR;
 	}
 	*offline_body = NULL;
@@ -104,7 +104,7 @@ int dialog_offline_body(str* body, str** offline_body)
 	node= xmlDocGetNodeByName(doc, "dialog", NULL);
 	if(node== NULL)
 	{
-		LM_DBG("no dialog nodes found");
+		LM_DBG("no dialog nodes found\n");
 		rc = OFFB_STATUS_NO_DIALOG;
 		goto done;
 	}
@@ -119,7 +119,7 @@ int dialog_offline_body(str* body, str** offline_body)
 	new_body = (str*)pkg_malloc(sizeof(str));
 	if(new_body == NULL)
 	{
-		LM_ERR("No more pkg memory");
+		LM_ERR("No more pkg memory\n");
 		goto done;
 	}
 	memset(new_body, 0, sizeof(str));
@@ -177,7 +177,7 @@ int presence_offline_body(str* body, str** offline_body)
 	}
 
 	pres_node= xmlDocGetNodeByName(doc, "presence", NULL);
-	if(node== NULL)
+	if(pres_node== NULL)
 	{
 		LM_ERR("while extracting presence node\n");
 		goto done;
@@ -223,7 +223,7 @@ int presence_offline_body(str* body, str** offline_body)
 	new_body = (str*)pkg_malloc(sizeof(str));
 	if(new_body == NULL)
 	{
-		LM_ERR("No more pkg memory");
+		LM_ERR("No more pkg memory\n");
 		goto done;
 	}
 	memset(new_body, 0, sizeof(str));
@@ -268,7 +268,7 @@ str* agregate_dialog_xmls(str* pres_user, str* pres_domain, str** body_array, in
 	if(xml_array== NULL)
 	{
 
-		LM_ERR("while alocating memory");
+		LM_ERR("while allocating memory\n");
 		return NULL;
 	}
 	memset(xml_array, 0, (n+2)*sizeof(xmlDocPtr)) ;
@@ -301,7 +301,7 @@ str* agregate_dialog_xmls(str* pres_user, str* pres_domain, str** body_array, in
 	p_root = xmlDocGetNodeByName( xml_array[j], root_name, NULL);
 	if(p_root ==NULL)
 	{
-		LM_ERR("while geting the xml_tree root\n");
+		LM_ERR("while getting the xml_tree root\n");
 		goto error;
 	}
 
@@ -312,7 +312,7 @@ str* agregate_dialog_xmls(str* pres_user, str* pres_domain, str** body_array, in
 		new_p_root= xmlDocGetNodeByName( xml_array[i], root_name, NULL);
 		if(new_p_root ==NULL)
 		{
-			LM_ERR("while geting the xml_tree root\n");
+			LM_ERR("while getting the xml_tree root\n");
 			goto error;
 		}
 
@@ -485,7 +485,7 @@ str* agregate_presence_xmls(str* pres_user, str* pres_domain, str** body_array,
     xml_array = (xmlDocPtr*)pkg_malloc( (n+2)*sizeof(xmlDocPtr));
     if(xml_array == NULL)
     {
-        LM_ERR("while alocating memory");
+        LM_ERR("while allocating memory\n");
         return NULL;
     }
     memset(xml_array, 0, (n+2)*sizeof(xmlDocPtr)) ;
@@ -574,7 +574,7 @@ str* agregate_presence_xmls(str* pres_user, str* pres_domain, str** body_array,
         current_doc_root = xmlDocGetRootElement(xml_array[i]);
         if(current_doc_root == NULL)
         {
-            LM_ERR("while geting the xml_tree root\n");
+            LM_ERR("while getting the xml_tree root\n");
             continue;
         }
 
@@ -941,6 +941,7 @@ str* get_final_notify_body( subs_t *subs, str* notify_body, xmlNodePtr rule_node
 	char* deviceID = NULL;
 	char* content = NULL;
 	char all_name[20];
+	int namelen;
 
 	strcpy(all_name, "all-");
 
@@ -980,7 +981,12 @@ str* get_final_notify_body( subs_t *subs, str* notify_body, xmlNodePtr rule_node
 			continue;
 
 		LM_DBG("transf_node->name:%s\n",node->name);
+		namelen = strlen((char *)node->name + 8);
 
+		if (namelen >= 15) {
+			LM_ERR("node name too large (%d) [%s]\n", namelen, (char *)node->name + 8);
+			continue;
+		}
 		strcpy((char*)name ,(char*)(node->name + 8));
 		strcpy(all_name+4, name);
 
@@ -1218,6 +1224,8 @@ error:
 			xmlFree(new_body->s);
 		pkg_free(new_body);
 	}
+	/* none of these are used before error!
+	 * Detected by Coverity 40606-40609
 	if(class_cont)
 		xmlFree(class_cont);
 	if(occurence_ID)
@@ -1226,6 +1234,7 @@ error:
 		xmlFree(deviceID);
 	if(service_uri)
 		xmlFree(service_uri);
+	*/
 
 	return NULL;
 }
