@@ -1,6 +1,5 @@
 /*
- * Copyright (C) 2015 - OpenSIPS Foundation
- * Copyright (C) 2001-2003 FhG Fokus
+ * Copyright (C) 2015 - OpenSIPS Solutions
  *
  * This file is part of opensips, a free SIP server.
  *
@@ -34,11 +33,11 @@
 #include "../../sr_module.h"
 #include "../../net/api_proto.h"
 #include "../../net/api_proto_net.h"
+#include "../../net/tcp_common.h"
 #include "../../net/net_tcp.h"
 #include "../../net/net_udp.h"
 #include "../../socket_info.h"
 #include "../../receive.h"
-#include "../../tsend.h"
 #include "../../net/proto_tcp/tcp_common_defs.h"
 #include "../../net/proto_tcp/tcp_common_defs.h"
 #include "../../pt.h"
@@ -47,8 +46,6 @@
 #include "hep.h"
 #include "hep_cb.h"
 
-#define HEP_FIRST 1
-#define HEP_LAST  3
 
 
 static int mod_init(void);
@@ -56,20 +53,19 @@ static void destroy(void);                          /*!< Module destroy function
 static int proto_hep_init_udp(struct proto_info *pi);
 static int proto_hep_init_tcp(struct proto_info *pi);
 static int proto_hep_init_udp_listener(struct socket_info *si);
-static int hep_conn_init(struct tcp_connection* c);
-static void hep_conn_clean(struct tcp_connection* c);
-static int hep_write_async_req(struct tcp_connection* con,int fd);
 static int hep_tcp_read_req(struct tcp_connection* con, int* bytes_read);
 static int hep_udp_read_req(struct socket_info *si, int* bytes_read);
 static int hep_udp_send (struct socket_info* send_sock,
-		char *buf, unsigned int len, union sockaddr_union *to, int id);
+		char *buf, unsigned int len, union sockaddr_union *to,
+		unsigned int id);
 static int hep_tcp_send (struct socket_info* send_sock,
-		char *buf, unsigned int len, union sockaddr_union *to, int id);
+		char *buf, unsigned int len, union sockaddr_union *to,
+		unsigned int id);
 static void update_recv_info(struct receive_info *ri, struct hep_desc *h);
 
 void free_hep_context(void* ptr);
 
-static int hep_port = 5656;
+static int hep_port = HEP_PORT;
 static int hep_async = 1;
 static int hep_send_timeout = 100;
 static int hep_async_max_postponed_chunks = 32;
@@ -82,6 +78,9 @@ int hep_ctx_idx=0;
 int hep_capture_id = 1;
 int payload_compression=0;
 
+int homer5_on=1;
+str homer5_delim = {":", 0};
+
 compression_api_t compression_api;
 load_compression_f load_compression;
 
@@ -90,33 +89,24 @@ static struct tcp_req hep_current_req;
  * so we need to know what is the current version of hep */
 static int hep_current_proto;
 
-struct hep_send_chunk {
-	char *buf; /* buffer that needs to be sent out */
-	char *pos; /* the position that we should be writing next */
-	int len;   /* length of the buffer */
-	int ticks; /* time at which this chunk was initially
-				  attempted to be written */
+union sockaddr_union local_su;
+
+static const cmd_export_t cmds[] = {
+	{"proto_init", (cmd_function)proto_hep_init_udp, {{0,0,0}}, 0},
+	{"proto_init", (cmd_function)proto_hep_init_tcp, {{0,0,0}}, 0},
+	{"load_hep", (cmd_function)bind_proto_hep, {{0,0,0}}, 0},
+	{"trace_bind_api", (cmd_function)hep_bind_trace_api, {{0,0,0}}, 0},
+	{"correlate", (cmd_function)correlate_w, {
+		{CMD_PARAM_STR,0,0},
+		{CMD_PARAM_STR,0,0},
+		{CMD_PARAM_STR,0,0},
+		{CMD_PARAM_STR,0,0},
+		{CMD_PARAM_STR,0,0}, {0,0,0}},
+		REQUEST_ROUTE|FAILURE_ROUTE|ONREPLY_ROUTE|BRANCH_ROUTE|LOCAL_ROUTE},
+	{0,0,{{0,0,0}},0}
 };
 
-struct hep_data {
-	/* the chunks that need to be written on this
-	 * connection when it will become writable */
-	struct hep_send_chunk **async_chunks;
-	/* the total number of chunks pending to be written */
-	int async_chunks_no;
-	/* the oldest chunk in our write list */
-	int oldest_chunk;
-};
-
-
-static cmd_export_t cmds[] = {
-	{"proto_init",            (cmd_function)proto_hep_init_udp,        0, 0, 0, 0},
-	{"proto_init",            (cmd_function)proto_hep_init_tcp,        0, 0, 0, 0},
-	{"load_hep",			  (cmd_function)bind_proto_hep,        1, 0, 0, 0},
-	{0,0,0,0,0,0}
-};
-
-static param_export_t params[] = {
+static const param_export_t params[] = {
 	{ "hep_port",                        INT_PARAM, &hep_port				},
 	{ "hep_send_timeout",                INT_PARAM, &hep_send_timeout		},
 	{ "hep_max_msg_chunks",              INT_PARAM, &hep_max_msg_chunks     },
@@ -130,11 +120,14 @@ static param_export_t params[] = {
 	{ "hep_async_local_write_timeout",   INT_PARAM,
 											&hep_async_local_write_timeout  },
 	{ "compressed_payload",				 INT_PARAM, &payload_compression},
+	{ "hep_id",						 STR_PARAM|USE_FUNC_PARAM, parse_hep_id },
+	{ "homer5_on",						 INT_PARAM, &homer5_on              },
+	{ "homer5_delim",					 STR_PARAM, &homer5_delim.s },
 	{0, 0, 0}
 };
 
 
-static module_dependency_t *get_deps_compression(param_export_t *param)
+static module_dependency_t *get_deps_compression(const param_export_t *param)
 {
 	int do_compression= *(int *)param->param_pointer;
 
@@ -145,7 +138,7 @@ static module_dependency_t *get_deps_compression(param_export_t *param)
 
 }
 
-static dep_export_t deps = {
+static const dep_export_t deps = {
 	{ /* OpenSIPS module dependencies */
 		{ MOD_TYPE_NULL, NULL, 0 },
 	},
@@ -162,6 +155,7 @@ struct module_exports exports = {
 	MOD_TYPE_DEFAULT,/* class of this module */
 	MODULE_VERSION,
 	DEFAULT_DLFLAGS, /* dlopen flags */
+	0,				 /* load function */
 	&deps,            /* OpenSIPS module dependencies */
 	cmds,       /* exported functions */
 	0,          /* exported async functions */
@@ -169,18 +163,32 @@ struct module_exports exports = {
 	0,          /* exported statistics */
 	0,          /* exported MI functions */
 	0,          /* exported pseudo-variables */
+	0,			/* exported transformations */
 	0,          /* extra processes */
+	0,          /* module pre-initialization function */
 	mod_init,   /* module initialization function */
 	0,          /* response function */
 	destroy,	/* destroy function */
 	0,          /* per-child init function */
+	0           /* reload confirm function */
 };
 
 static int mod_init(void)
 {
+	/* check if any listeners defined for this proto */
+	if ( !protos[PROTO_HEP_UDP].listeners && !protos[PROTO_HEP_TCP].listeners ) {
+		LM_ERR("No HEP listener defined, neither TCP nor UDP!\n");
+		return -1;
+	}
+
+	if (init_hep_id() < 0) {
+		LM_ERR("could not initialize HEP id list!\n");
+		return -1;
+	}
+
 	if (payload_compression) {
 		load_compression =
-			(load_compression_f)find_export("load_compression", 1, 0);
+			(load_compression_f)find_export("load_compression", 0);
 		if (!load_compression) {
 			LM_ERR("can't bind compression module!\n");
 			return -1;
@@ -193,6 +201,11 @@ static int mod_init(void)
 	}
 
 	hep_ctx_idx = context_register_ptr(CONTEXT_GLOBAL, 0);
+	homer5_delim.len = strlen(homer5_delim.s);
+
+	local_su.sin.sin_addr.s_addr = TRACE_INADDR_LOOPBACK;
+	local_su.sin.sin_port = 0;
+	local_su.sin.sin_family = AF_INET;
 
 	return 0;
 }
@@ -200,6 +213,7 @@ static int mod_init(void)
 static void destroy(void)
 {
 	free_hep_cbs();
+	destroy_hep_id();
 }
 
 void free_hep_context(void *ptr)
@@ -265,373 +279,27 @@ static int proto_hep_init_tcp(struct proto_info *pi)
 	pi->net.flags			= PROTO_NET_USE_TCP;
 
 	pi->net.read			= (proto_net_read_f)hep_tcp_read_req;
-	pi->net.write			= (proto_net_write_f)hep_write_async_req;
+	pi->net.write			= (proto_net_write_f)tcp_async_write;
 
 	pi->tran.send			= hep_tcp_send;
 
 
-	if (hep_async) {
-		pi->net.conn_init	= hep_conn_init;
-		pi->net.conn_clean	= hep_conn_clean;
-	}
+	if (hep_async)
+		pi->net.async_chunks= hep_async_max_postponed_chunks;
 
 	return 0;
 }
 
 
 
-static int hep_conn_init(struct tcp_connection* c)
-{
-	struct hep_data *d;
-
-	/* allocate the tcp_data and the array of chunks as a single mem chunk */
-	d = (struct hep_data*)shm_malloc( sizeof(struct hep_data) +
-		sizeof(struct hep_send_chunk *) * hep_async_max_postponed_chunks );
-	if (d == NULL) {
-		LM_ERR("failed to create tcp chunks in shm mem\n");
-		return -1;
-	}
-
-	d->async_chunks = (struct hep_send_chunk **)(d+1);
-	d->async_chunks_no = 0;
-	d->oldest_chunk = 0;
-
-	c->proto_data = (void*)d;
-	return 0;
-}
-
-static void hep_conn_clean(struct tcp_connection* c)
-{
-	struct hep_data *d = (struct hep_data*)c->proto_data;
-	int r;
-
-	for (r = 0; r < d->async_chunks_no; r++) {
-		shm_free(d->async_chunks[r]);
-	}
-
-	shm_free(d);
-
-	c->proto_data = NULL;
-}
-
-
 static int proto_hep_init_udp_listener(struct socket_info *si)
 {
 	return udp_init_listener(si, hep_async?O_NONBLOCK:0);
 }
 
-static int add_write_chunk(struct tcp_connection *con,char *buf,int len,
-					int lock)
-{
-	struct hep_send_chunk *c;
-	struct hep_data *d = (struct hep_data*)con->proto_data;
-
-	c = shm_malloc(sizeof(struct hep_send_chunk) + len);
-	if (!c) {
-		LM_ERR("No more SHM\n");
-		return -1;
-	}
-
-	c->len = len;
-	c->ticks = get_ticks();
-	c->buf = (char *)(c+1);
-	memcpy(c->buf,buf,len);
-	c->pos = c->buf;
-
-	if (lock)
-		lock_get(&con->write_lock);
-
-	if (d->async_chunks_no == hep_async_max_postponed_chunks) {
-		LM_ERR("We have reached the limit of max async postponed chunks\n");
-		if (lock)
-			lock_release(&con->write_lock);
-		shm_free(c);
-		return -2;
-	}
-
-	d->async_chunks[d->async_chunks_no++] = c;
-	if (d->async_chunks_no == 1)
-		d->oldest_chunk = c->ticks;
-
-	if (lock)
-		lock_release(&con->write_lock);
-
-	return 0;
-}
-
-static int async_tsend_stream(struct tcp_connection *c,
-		int fd, char* buf, unsigned int len, int timeout)
-{
-	int written;
-	int n;
-	struct pollfd pf;
-
-	pf.fd=fd;
-	pf.events=POLLOUT;
-	written=0;
-
-again:
-	n=send(fd, buf, len,0);
-	if (n<0){
-		if (errno==EINTR) goto again;
-		else if (errno!=EAGAIN && errno!=EWOULDBLOCK) {
-			LM_ERR("Failed first TCP async send : (%d) %s\n",
-					errno, strerror(errno));
-			return -1;
-		} else
-			goto poll_loop;
-	}
-
-	written+=n;
-	if (n < len) {
-		/* partial write */
-		buf += n;
-		len -= n;
-	} else {
-		/* succesful write from the first try */
-		LM_DBG("Async succesful write from first try on %p\n",c);
-		return len;
-	}
-
-poll_loop:
-	n = poll(&pf,1,timeout);
-	if (n<0) {
-		if (errno==EINTR)
-			goto poll_loop;
-		LM_ERR("Polling while trying to async send failed %s [%d]\n",
-				strerror(errno), errno);
-		return -1;
-	} else if (n == 0) {
-		LM_DBG("timeout -> do an async write (add it to conn)\n");
-		/* timeout - let's just pass to main */
-		if (add_write_chunk(c,buf,len,0) < 0) {
-			LM_ERR("Failed to add write chunk to connection \n");
-			return -1;
-		} else {
-			/* we have succesfully added async write chunk
-			 * tell MAIN to poll out for us */
-			LM_DBG("Data still pending for write on conn %p\n",c);
-			return 0;
-		}
-	}
-
-	if (pf.events&POLLOUT)
-		goto again;
-
-	/* some other events triggered by poll - treat as errors */
-	return -1;
-}
-
-
-
-static struct tcp_connection* hep_sync_connect(struct socket_info* send_sock,
-		union sockaddr_union* server, int *fd)
-{
-	int s;
-	union sockaddr_union my_name;
-	socklen_t my_name_len;
-	struct tcp_connection* con;
-
-	s=socket(AF2PF(server->s.sa_family), SOCK_STREAM, 0);
-	if (s==-1){
-		LM_ERR("socket: (%d) %s\n", errno, strerror(errno));
-		goto error;
-	}
-	if (tcp_init_sock_opt(s)<0){
-		LM_ERR("tcp_init_sock_opt failed\n");
-		goto error;
-	}
-	my_name_len = sockaddru_len(send_sock->su);
-	memcpy( &my_name, &send_sock->su, my_name_len);
-	su_setport( &my_name, 0);
-	if (bind(s, &my_name.s, my_name_len )!=0) {
-		LM_ERR("bind failed (%d) %s\n", errno,strerror(errno));
-		goto error;
-	}
-
-	if (tcp_connect_blocking(s, &server->s, sockaddru_len(*server))<0){
-		LM_ERR("tcp_blocking_connect failed\n");
-		goto error;
-	}
-	con = tcp_conn_create(s, server, send_sock, S_CONN_OK);
-	if (con==NULL){
-		LM_ERR("tcp_conn_create failed, closing the socket\n");
-		goto error;
-	}
-	*fd = s;
-	return con;
-	/*FIXME: set sock idx! */
-error:
-	/* close the opened socket */
-	if (s!=-1) close(s);
-	return 0;
-}
-
-
-
-static int tcpconn_async_connect(struct socket_info* send_sock,
-					union sockaddr_union* server, char *buf, unsigned len,
-					struct tcp_connection** c, int *ret_fd)
-{
-	int fd, n;
-	union sockaddr_union my_name;
-	socklen_t my_name_len;
-	struct tcp_connection* con;
-
-	struct pollfd pf;
-
-	unsigned int elapsed,to;
-	int err;
-	unsigned int err_len;
-	int poll_err;
-	char *ip;
-	unsigned short port;
-	struct timeval begin;
-
-	/* create the socket */
-	fd = socket(AF2PF(server->s.sa_family), SOCK_STREAM, 0);
-	if (fd == -1){
-		LM_ERR("socket: (%d) %s\n", errno, strerror(errno));
-		return -1;
-	}
-	if (tcp_init_sock_opt(fd)<0){
-		LM_ERR("tcp_init_sock_opt failed\n");
-		goto error;
-	}
-	my_name_len = sockaddru_len(send_sock->su);
-	memcpy( &my_name, &send_sock->su, my_name_len);
-	su_setport( &my_name, 0);
-	if (bind(fd, &my_name.s, my_name_len )!=0) {
-		LM_ERR("bind failed (%d) %s\n", errno,strerror(errno));
-		goto error;
-	}
-
-	/* attempt to do connect and see if we do block or not */
-	poll_err = 0;
-	elapsed = 0;
-	to = hep_async_local_connect_timeout*1000;
-
-	if (gettimeofday(&(begin), NULL)) {
-		LM_ERR("Failed to get TCP connect start time\n");
-		goto error;
-	}
-
-again:
-	n = connect(fd, &server->s, sockaddru_len(*server));
-	if (n == -1) {
-		if (errno == EINTR){
-			elapsed=get_time_diff(&begin);
-			if (elapsed < to) goto again;
-			else {
-				LM_DBG("Local connect attempt failed \n");
-				goto async_connect;
-			}
-		}
-		if (errno != EINPROGRESS && errno!=EALREADY) {
-			get_su_info(&server->s, ip, port);
-			LM_ERR("[server=%s:%d] (%d) %s\n",ip, port, errno,strerror(errno));
-			goto error;
-		}
-	} else goto local_connect;
-
-	/* let's poll for a little */
-
-	pf.fd = fd;
-	pf.events = POLLOUT;
-
-	while(1){
-		elapsed = get_time_diff(&begin);
-		if (elapsed < to)
-			to -= elapsed;
-		else {
-			LM_DBG("Polling is overdue \n");
-			goto async_connect;
-		}
-
-		n = poll(&pf, 1, to/1000);
-
-		if (n < 0){
-			if (errno == EINTR) continue;
-			get_su_info(&server->s, ip, port);
-			LM_ERR("poll/select failed:[server=%s:%d] (%d) %s\n",
-				ip, port, errno, strerror(errno));
-			goto error;
-		} else if (n==0) /* timeout */ continue;
-
-		if (pf.revents & (POLLERR|POLLHUP|POLLNVAL)){
-			LM_ERR("poll error: flags %x\n", pf.revents);
-			poll_err=1;
-		}
-
-
-		err_len=sizeof(err);
-		getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, &err_len);
-		if ((err==0) && (poll_err==0)) goto local_connect;
-		if (err!=EINPROGRESS && err!=EALREADY){
-			get_su_info(&server->s, ip, port);
-			LM_ERR("failed to retrieve SO_ERROR [server=%s:%d] (%d) %s\n",
-				ip, port, err, strerror(err));
-			goto error;
-		}
-	}
-
-async_connect:
-	LM_DBG("Create connection for async connect\n");
-	/* create a new dummy connection */
-	con = tcp_conn_create(fd, server, send_sock, S_CONN_CONNECTING);
-	if (con==NULL) {
-		LM_ERR("tcp_conn_create failed\n");
-		goto error;
-	}
-
-	/* attach the write buffer to it */
-	lock_get(&con->write_lock);
-
-	if (add_write_chunk(con,buf,len,0) < 0) {
-		LM_ERR("Failed to add the initial write chunk\n");
-		/* FIXME - seems no more SHM now ...
-		 * continue the async connect process ? */
-	}
-
-	lock_release(&con->write_lock);
-	/* report an async, in progress connect */
-	*c = con;
-	return 0;
-
-local_connect:
-	con = tcp_conn_create(fd, server, send_sock, S_CONN_OK);
-	if (con==NULL) {
-		LM_ERR("tcp_conn_create failed, closing the socket\n");
-		goto error;
-	}
-	*c = con;
-	*ret_fd = fd;
-	/* report a local connect */
-	return 1;
-
-error:
-	close(fd);
-	*c = NULL;
-	return -1;
-}
-
-inline static int _hep_write_on_socket(struct tcp_connection *c, int fd,
-												char *buf, int len){
-	int n;
-
-	lock_get(&c->write_lock);
-	if (hep_async) {
-		n=async_tsend_stream(c,fd,buf,len, hep_async_local_write_timeout);
-	} else {
-		n = tsend_stream(fd, buf, len, hep_send_timeout);
-	}
-	lock_release(&c->write_lock);
-
-	return n;
-}
-
 static int hep_udp_send (struct socket_info* send_sock,
-		char *buf, unsigned int len, union sockaddr_union *to, int id)
+		char *buf, unsigned int len, union sockaddr_union *to,
+		unsigned int id)
 {
 	int n, tolen;
 
@@ -653,7 +321,8 @@ again:
 }
 
 static int hep_tcp_send (struct socket_info* send_sock,
-		char *buf, unsigned int len, union sockaddr_union *to, int id)
+		char *buf, unsigned int len, union sockaddr_union *to,
+		unsigned int id)
 {
 	struct tcp_connection *c;
 	int port=0;
@@ -664,9 +333,9 @@ static int hep_tcp_send (struct socket_info* send_sock,
 	if (to) {
 		su2ip_addr(&ip, to);
 		port=su_getport(to);
-		n = tcp_conn_get(id,&ip, port, PROTO_HEP_TCP, &c, &fd);
+		n = tcp_conn_get(id,&ip, port, PROTO_HEP_TCP, NULL, &c, &fd, send_sock);
 	} else if (id) {
-		n = tcp_conn_get(id, 0, 0, PROTO_NONE, &c, &fd);
+		n = tcp_conn_get(id, 0, 0, PROTO_NONE, NULL, &c, &fd, NULL);
 	} else {
 		LM_CRIT("tcp_send called with null id & to\n");
 		return -1;
@@ -674,15 +343,18 @@ static int hep_tcp_send (struct socket_info* send_sock,
 
 	if (n < 0) {
 		/* error during conn get, return with error too */
-		LM_ERR("failed to aquire connection\n");
+		LM_ERR("failed to acquire connection\n");
 		return -1;
 	}
 
 	/* was connection found ?? */
 	if (c==0) {
-		if (tcp_no_new_conn) {
+		struct tcp_conn_profile prof;
+		int matched = tcp_con_get_profile(to, &send_sock->su, send_sock->proto, &prof);
+
+		if ((matched && prof.no_new_conn) || (!matched && tcp_no_new_conn))
 			return -1;
-		}
+
 		if (!to) {
 			LM_ERR("Unknown destination - cannot open new tcp connection\n");
 			return -1;
@@ -690,22 +362,33 @@ static int hep_tcp_send (struct socket_info* send_sock,
 		LM_DBG("no open tcp connection found, opening new one, async = %d\n",hep_async);
 		/* create tcp connection */
 		if (hep_async) {
-			n = tcpconn_async_connect(send_sock, to, buf, len, &c, &fd);
+			n = tcp_async_connect(send_sock, to, &prof,
+					hep_async_local_connect_timeout, &c, &fd, 1);
 			if ( n<0 ) {
 				LM_ERR("async TCP connect failed\n");
 				return -1;
 			}
 			/* connect succeeded, we have a connection */
 			if (n==0) {
+				/* attach the write buffer to it */
+				if (tcp_async_add_chunk(c, buf, len, 1) < 0) {
+					LM_ERR("Failed to add the initial write chunk\n");
+					len = -1; /* report an error - let the caller decide what to do */
+				}
+
+				/* mark the ID of the used connection (tracing purposes) */
+				last_outgoing_tcp_id = c->id;
+				send_sock->last_local_real_port = c->rcv.dst_port;
+				send_sock->last_remote_real_port = c->rcv.src_port;
 				/* connect is still in progress, break the sending
 				 * flow now (the actual write will be done when
 				 * connect will be completed */
-				LM_DBG("Succesfully started async connection \n");
+				LM_DBG("Successfully started async connection \n");
 				tcp_conn_release(c, 0);
 				return len;
 			}
 			/* our first connect attempt succeeded - go ahead as normal */
-		} else if ((c=hep_sync_connect(send_sock, to, &fd))==0) {
+		} else if ((c=tcp_sync_connect(send_sock, to, &prof, &fd, 1))==0) {
 			LM_ERR("connect failed\n");
 			return -1;
 		}
@@ -725,7 +408,7 @@ static int hep_tcp_send (struct socket_info* send_sock,
 			 * case we ever manage to get through */
 			LM_DBG("We have acquired a TCP connection which is still "
 				"pending to connect - delaying write \n");
-			n = add_write_chunk(c,buf,len,1);
+			n = tcp_async_add_chunk(c,buf,len,1);
 			if (n < 0) {
 				LM_ERR("Failed to add another write chunk to %p\n",c);
 				/* we failed due to internal errors - put the
@@ -734,7 +417,12 @@ static int hep_tcp_send (struct socket_info* send_sock,
 				return -1;
 			}
 
-			/* we succesfully added our write chunk - success */
+			/* mark the ID of the used connection (tracing purposes) */
+			last_outgoing_tcp_id = c->id;
+			send_sock->last_local_real_port = c->rcv.dst_port;
+			send_sock->last_remote_real_port = c->rcv.src_port;
+
+			/* we successfully added our write chunk - success */
 			tcp_conn_release(c, 0);
 			return len;
 		} else {
@@ -748,11 +436,12 @@ static int hep_tcp_send (struct socket_info* send_sock,
 send_it:
 	LM_DBG("sending via fd %d...\n",fd);
 
-	n = _hep_write_on_socket(c, fd, buf, len);
+	n = tcp_write_on_socket(c, fd, buf, len,
+			hep_send_timeout, hep_async_local_write_timeout);
 
-	tcp_conn_set_lifetime( c, tcp_con_lifetime);
+	tcp_conn_reset_lifetime(c);
 
-	LM_DBG("after write: c= %p n=%d fd=%d\n",c, n, fd);
+	LM_DBG("after write: c= %p n/len=%d/%d fd=%d\n",c, n, len, fd);
 	/* LM_DBG("buf=\n%.*s\n", (int)len, buf); */
 	if (n<0){
 		LM_ERR("failed to send\n");
@@ -768,6 +457,11 @@ send_it:
 	if (c->proc_id != process_no)
 		close(fd);
 
+	/* mark the ID of the used connection (tracing purposes) */
+	last_outgoing_tcp_id = c->id;
+	send_sock->last_local_real_port = c->rcv.dst_port;
+	send_sock->last_remote_real_port = c->rcv.src_port;
+
 	tcp_conn_release(c, (n<len)?1:0/*pending data in async mode?*/ );
 
 	return n;
@@ -796,11 +490,11 @@ static void hep_parse_headers(struct tcp_req *req){
 	length = ntohs(ctrl->length);
 	req->content_len = (unsigned short)length;
 
-	if(req->pos - req->parsed == req->content_len){
+	if(req->pos - req->buf == req->content_len){
 		LM_DBG("received a COMPLETE message\n");
 		req->complete = 1;
 		req->parsed = req->buf + req->content_len;
-	} else if(req->pos - req->parsed > req->content_len){
+	} else if(req->pos - req->buf > req->content_len){
 		LM_DBG("received MORE than a message\n");
 		req->complete = 1;
 		req->parsed = req->buf + req->content_len;
@@ -809,11 +503,11 @@ static void hep_parse_headers(struct tcp_req *req){
 		/* FIXME should we update parsed? we didn't receive the
 		 * full message; we wait for the full mesage and only
 		 * after that we update parsed */
-		//req->parsed = req->pos;
+		req->parsed = req->pos;
 	}
 }
 
-int tcp_read(struct tcp_connection *c,struct tcp_req *r) {
+static int _tcp_read(struct tcp_connection *c,struct tcp_req *r) {
 	int bytes_free, bytes_read;
 	int fd;
 
@@ -831,8 +525,13 @@ again:
 	if(bytes_read==-1){
 		if (errno == EWOULDBLOCK || errno == EAGAIN){
 			return 0; /* nothing has been read */
-		}else if (errno == EINTR) goto again;
-		else{
+		} else if (errno == EINTR) {
+			goto again;
+		} else if (errno == ECONNRESET) {
+			c->state=S_CONN_EOF;
+			LM_DBG("EOF on %p, FD %d\n", c, fd);
+			bytes_read = 0;
+		} else {
 			LM_ERR("error reading: %s\n",strerror(errno));
 			r->error=TCP_READ_ERROR;
 			return -1;
@@ -859,12 +558,12 @@ static inline int hep_handle_req(struct tcp_req *req,
 
 	int ret=0;
 
-	struct hep_context *hep_ctx;
+	struct hep_context *hep_ctx=NULL;
 	context_p ctx=NULL;
 
 	if (req->complete){
 		/* update the timeout - we successfully read the request */
-		tcp_conn_set_lifetime( con, tcp_con_lifetime);
+		tcp_conn_reset_lifetime(con);
 		con->timeout=con->lifetime;
 
 		/* just for debugging use sendipv4 as receiving socket  FIXME*/
@@ -873,7 +572,7 @@ static inline int hep_handle_req(struct tcp_req *req,
 		/* prepare for next request */
 		size=req->pos-req->parsed;
 
-		msg_buf = req->start;
+		msg_buf = req->buf;
 		msg_len = req->parsed-req->start;
 		local_rcv = con->rcv;
 
@@ -922,13 +621,13 @@ static inline int hep_handle_req(struct tcp_req *req,
 			/* run hep callbacks; set the current processing context
 			 * to hep context; this way callbacks will have all the data
 			 * needed */
-			current_processing_ctx = ctx;
+			set_global_context(ctx);
 			ret=run_hep_cbs();
 			if (ret < 0) {
 				LM_ERR("failed to run hep callbacks\n");
 				goto error_free_hep;
 			}
-			current_processing_ctx = NULL;
+			set_global_context(NULL);
 
 			msg_len = hep_ctx->h.u.hepv3.payload_chunk.chunk.length-
 											sizeof(hep_chunk_t);
@@ -937,27 +636,34 @@ static inline int hep_handle_req(struct tcp_req *req,
 		}
 
 		/* skip receive msg if we were told so from at least one callback */
-		if (ret != HEP_SCRIPT_SKIP &&
-				receive_msg(msg_buf, msg_len, &local_rcv, ctx) <0)
+		if ( ret != HEP_SCRIPT_SKIP ) {
+			if ( receive_msg(msg_buf, msg_len, &local_rcv, ctx, 0) <0 ) {
 				LM_ERR("receive_msg failed \n");
+			}
+		} else {
+			if ( ctx ) {
+				context_free( ctx );
+			}
+		}
 
-		free_hep_context(hep_ctx);
+		if (hep_ctx)
+			free_hep_context(hep_ctx);
 
 		if (!size && req != &hep_current_req) {
 			/* if we no longer need this tcp_req
 			 * we can free it now */
-			pkg_free(req);
+			shm_free(req);
+			con->con_req = NULL;
 		}
 
+		con->msg_attempts = 0;
+
 		if (size) {
 			memmove(req->buf, req->parsed, size);
-			req->pos = req->buf + size;
-			/* anything that was parsed was already processed */
-			req->parsed = req->buf;
-		}
+			init_tcp_req( req, size);
 
-		/* if we still have some unparsed bytes, try to  parse them too*/
-		if (size) return 1;
+			return 1;
+		}
 	} else {
 		/* request not complete - check the if the thresholds are exceeded */
 		if (con->msg_attempts==0)
@@ -976,7 +682,7 @@ static inline int hep_handle_req(struct tcp_req *req,
 			/* let's duplicate this - most likely another conn will come in */
 
 			LM_DBG("We didn't manage to read a full request\n");
-			con->con_req = pkg_malloc(sizeof(struct tcp_req));
+			con->con_req = shm_malloc(sizeof(struct tcp_req));
 			if (con->con_req == NULL) {
 				LM_ERR("No more mem for dynamic con request buffer\n");
 				goto error;
@@ -1043,7 +749,7 @@ static int hep_tcp_read_req(struct tcp_connection* con, int* bytes_read)
 		if (req->parsed < req->pos){
 			bytes=0;
 		} else {
-			bytes=tcp_read(con,req);
+			bytes=_tcp_read(con,req);
 			if (bytes < 0) {
 				LM_ERR("failed to read \n");
 				goto error;
@@ -1074,7 +780,10 @@ static int hep_tcp_read_req(struct tcp_connection* con, int* bytes_read)
 		goto error;
 	}
 
-	switch (hep_handle_req(req, con, hep_max_msg_chunks) ) {
+	int max_chunks = tcp_attr_isset(con, TCP_ATTR_MAX_MSG_CHUNKS) ?
+			con->profile.attrs[TCP_ATTR_MAX_MSG_CHUNKS] : hep_max_msg_chunks;
+
+	switch (hep_handle_req(req, con, max_chunks) ) {
 		case 1:
 			goto again;
 		case -1:
@@ -1093,75 +802,11 @@ error:
 
 
 
-static int hep_write_async_req(struct tcp_connection* con,int fd)
-{
-	int n,left;
-	struct hep_send_chunk *chunk;
-	struct hep_data *d = (struct hep_data*)con->proto_data;
-
-	if (d->async_chunks_no == 0) {
-		LM_DBG("The connection has been triggered "
-		" for a write event - but we have no pending write chunks\n");
-		return 0;
-	}
-
-next_chunk:
-	chunk=d->async_chunks[0];
-again:
-	left = (int)((chunk->buf+chunk->len)-chunk->pos);
-	LM_DBG("Trying to send %d bytes from chunk %p in conn %p - %d %d \n",
-		   left,chunk,con,chunk->ticks,get_ticks());
-	n = send(fd, chunk->pos, left, 0);
-	if (n<0) {
-		if (errno == EINTR)
-			goto again;
-		else if (errno == EAGAIN || errno == EWOULDBLOCK) {
-			LM_DBG("Can't finish to write chunk %p on conn %p\n",
-				   chunk,con);
-			/* report back we have more writting to be done */
-			return 1;
-		} else {
-			LM_ERR("Error occured while sending async chunk %d (%s)\n",
-				   errno,strerror(errno));
-			/* report the conn as broken */
-			return -1;
-		}
-	}
-
-	if (n < left) {
-		/* partial write */
-		chunk->pos+=n;
-		goto again;
-	} else {
-		/* written a full chunk - move to the next one, if any */
-		shm_free(chunk);
-		d->async_chunks_no--;
-		if (d->async_chunks_no == 0) {
-			LM_DBG("We have finished writing all our async chunks in %p\n",con);
-			d->oldest_chunk=0;
-			/*  report back everything ok */
-			return 0;
-		} else {
-			LM_DBG("We still have %d chunks pending on %p\n",
-					d->async_chunks_no,con);
-			memmove(&d->async_chunks[0],&d->async_chunks[1],
-					d->async_chunks_no * sizeof(struct hep_send_chunk*));
-			d->oldest_chunk = d->async_chunks[0]->ticks;
-			goto next_chunk;
-		}
-	}
-}
-
-
 static int hep_udp_read_req(struct socket_info *si, int* bytes_read)
 {
 	struct receive_info ri;
 	int len;
-#ifdef DYN_BUF
-	char* buf;
-#else
 	static char buf [BUF_SIZE+1];
-#endif
 	unsigned int fromlen;
 	str msg;
 
@@ -1171,14 +816,6 @@ static int hep_udp_read_req(struct socket_info *si, int* bytes_read)
 
 	context_p ctx=NULL;
 
-#ifdef DYN_BUF
-	buf=pkg_malloc(BUF_SIZE+1);
-	if (buf==0){
-		LM_ERR("could not allocate receive buffer\n");
-		goto error;
-	}
-#endif
-
 	fromlen=sockaddru_len(si->su);
 	len=recvfrom(bind_address->socket, buf, BUF_SIZE,0,&ri.src_su.s,&fromlen);
 	if (len==-1){
@@ -1226,12 +863,14 @@ static int hep_udp_read_req(struct socket_info *si, int* bytes_read)
 
 	if (!memcmp(buf, HEP_HEADER_ID, HEP_HEADER_ID_LEN)) {
 		/* HEPv3 */
+		/* coverity[tainted_data] */
 		if (unpack_hepv3(buf, len, &hep_ctx->h)) {
 			LM_ERR("hepv3 unpacking failed\n");
 			return -1;
 		}
 	} else {
 		/* HEPv2 */
+		/* coverity[tainted_data] */
 		if (unpack_hepv12(buf, len, &hep_ctx->h)) {
 			LM_ERR("hepv12 unpacking failed\n");
 			return -1;
@@ -1253,13 +892,13 @@ static int hep_udp_read_req(struct socket_info *si, int* bytes_read)
 	/* run hep callbacks; set the current processing context
 	 * to hep context; this way callbacks will have all the data
 	 * needed */
-	current_processing_ctx = ctx;
+	set_global_context(ctx);
 	ret=run_hep_cbs();
+	set_global_context(NULL);
 	if (ret < 0) {
 		LM_ERR("failed to run hep callbacks\n");
 		return -1;
 	}
-	current_processing_ctx = NULL;
 
 	if (hep_ctx->h.version == 3) {
 		/* HEPv3 */
@@ -1279,7 +918,11 @@ static int hep_udp_read_req(struct socket_info *si, int* bytes_read)
 
 	if (ret != HEP_SCRIPT_SKIP) {
 		/* receive_msg must free buf too!*/
-		receive_msg( msg.s, msg.len, &ri, ctx);
+		receive_msg( msg.s, msg.len, &ri, ctx, 0);
+	} else {
+		if ( ctx ) {
+			context_free( ctx );
+		}
 	}
 
 	free_hep_context(hep_ctx);
