@@ -26,6 +26,8 @@
 #ifndef _WS_HANDSHAKE_H_
 #define _WS_HANDSHAKE_H_
 
+#include "../../ip_addr.h"
+
 #define HTTP_SEP			"\r\n"
 #define HTTP_SEP_LEN		(sizeof(HTTP_SEP) - 1)
 #define HTTP_END HTTP_SEP HTTP_SEP
@@ -48,8 +50,8 @@
 #define WS_HOST_F		(1 << 0)
 #define WS_UPGRADE_F	(1 << 1)
 #define WS_CONN_F		(1 << 2)
-#define WS_ORIGIN_F		(1 << 4)
 #define WS_KEY_F		(1 << 3)
+#define WS_ORIGIN_F		(1 << 4)
 #define WS_VER_F		(1 << 5)
 /* for SIP connections, RFC7118 requires sip protocol */
 #define WS_PROTO_F		(1 << 6)
@@ -64,21 +66,6 @@
 #define WS_UPGRADE_HDR "Upgrade"
 #define WS_UPGRADE_HDR_LEN (sizeof(WS_UPGRADE_HDR) - 1)
 
-/* all flags for req */
-#define WS_ALL_REQ_F (WS_HOST_F | \
-					WS_UPGRADE_F | \
-					WS_CONN_F | \
-					WS_ORIGIN_F | \
-					WS_KEY_F | \
-					WS_VER_F | \
-					WS_PROTO_F)
-
-/* all flags for reply */
-#define WS_ALL_RPL_F (WS_UPGRADE_F | \
-					WS_CONN_F | \
-					WS_ACCEPT_F | \
-					WS_PROTO_F)
-
 #define GET_LOWER(_p) \
 	((*(_p)) | 0x20)
 #define GET_LOWER_DWORD(_p) \
@@ -116,7 +103,7 @@
 	"Sec-WebSocket-Protocol: " WS_PROTO_SIP HTTP_END
 #define HTTP_HANDSHAKE_END_LEN (sizeof(HTTP_HANDSHAKE_END) - 1)
 
-#define MAX_HOST_LEN IP_ADDR_MAX_STR_SIZE /*IP*/ + 1 /*':'*/ + 5 /*65535*/
+#define MAX_HOST_LEN IP_ADDR_MAX_STR_SIZE /*IP*/ + 2 /* '[' & ']' for ipv6 */ + 1 /*':'*/ + 5 /*65535*/
 
 #include "../../sha1.h"
 
@@ -126,6 +113,11 @@ static int ws_parse_rpl_handshake(struct tcp_connection *c, char *msg, int len);
 static int ws_complete_handshake(struct tcp_connection *c);
 static int ws_start_handshake(struct tcp_connection *c);
 static int ws_bad_handshake(struct tcp_connection *c);
+static int trace_ws( struct tcp_connection* conn, trans_trace_event event, str* req);
+static int complete_ws_trace( struct tcp_connection* conn, trans_trace_status status, str* rpl, str* message);
+
+#define WS_TRACE_MAX 1024
+static char ws_trace_buf[WS_TRACE_MAX];
 
 /* safety checks */
 #ifndef _ws_common_module
@@ -149,13 +141,31 @@ static int ws_bad_handshake(struct tcp_connection *c);
 #ifndef _ws_common_write_tout
 #error "_ws_common_write_tout not defined!"
 #endif
+#ifndef _ws_common_require_origin
+#error "_ws_common_require_origin not defined!"
+#endif
+
+/* all flags for req */
+#define WS_ALL_REQ_F (WS_HOST_F | \
+					WS_UPGRADE_F | \
+					WS_CONN_F | \
+					(_ws_common_require_origin?WS_ORIGIN_F:0) | \
+					WS_KEY_F | \
+					WS_VER_F | \
+					WS_PROTO_F)
+
+/* all flags for reply */
+#define WS_ALL_RPL_F (WS_UPGRADE_F | \
+					WS_CONN_F | \
+					WS_ACCEPT_F | \
+					WS_PROTO_F)
 
 
 #define WS_KEY_LEN 24
 static char ws_key[WS_KEY_LEN];
 static const char base64alphabet[] =
 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
-#define BASE64ALPHABET_LEN (sizeof base64alphabet)
+#define BASE64ALPHABET_LEN (sizeof(base64alphabet)-1)
 
 
 /* we're using a completely random key - no reason yet for something else */
@@ -164,8 +174,12 @@ static str ws_rand_key(void)
 	static str key = { ws_key, WS_KEY_LEN };
 	int i;
 
-	for (i = 0; i < WS_KEY_LEN; i++)
+	/* randomly selected 16-byte base64 encoded value requires
+	 * 22 characters and 2 paddings at the end */
+	for (i = 0; i < WS_KEY_LEN - 2; i++)
 		ws_key[i] = base64alphabet[rand() % BASE64ALPHABET_LEN];
+	ws_key[i++] = '=';
+	ws_key[i++] = '=';
 
 	return key;
 }
@@ -183,6 +197,7 @@ static inline int ws_client_handshake(struct tcp_connection *con)
 	struct timeval begin;
 	unsigned int err_len, poll_err = 0, err;
 	int n;
+	str trace_str;
 #if defined(HAVE_SELECT) && defined(BLOCKING_USE_SELECT)
 	fd_set sel_set;
 	fd_set orig_set;
@@ -281,8 +296,11 @@ static inline int ws_client_handshake(struct tcp_connection *con)
 			goto error;
 		}
 
+		int max_chunks = tcp_attr_isset(con, TCP_ATTR_MAX_MSG_CHUNKS) ?
+			con->profile.attrs[TCP_ATTR_MAX_MSG_CHUNKS] : _ws_common_max_msg_chunks;
+
 		con->msg_attempts++;
-		if (con->msg_attempts == _ws_common_max_msg_chunks) {
+		if (con->msg_attempts == max_chunks) {
 			LM_ERR("Made %u read attempts but message is not complete yet - "
 				   "closing connection \n",con->msg_attempts);
 			goto error;
@@ -310,8 +328,8 @@ static inline int ws_client_handshake(struct tcp_connection *con)
 #endif
 	}
 
-	/* update the timeout - we succesfully read the request */
-	tcp_conn_set_lifetime(con, tcp_con_lifetime);
+	/* update the timeout - we successfully read the request */
+	tcp_conn_reset_lifetime(con);
 	con->timeout=con->lifetime;
 
 	con->rcv.proto_reserved1=con->id; /* copy the id */
@@ -345,6 +363,16 @@ static inline int ws_client_handshake(struct tcp_connection *con)
 		goto error;
 	}
 
+	if ( ((struct ws_data *) con->proto_data)->dest ) {
+		trace_str.len = msg_len;
+		trace_str.s = msg_buf;
+
+		if ( complete_ws_trace( con, TRANS_TRACE_SUCCESS,
+							&trace_str, &CONNECT_OK) < 0 ) {
+			LM_ERR("failed to complete web socket trace!\n");
+		}
+	}
+
 	init_tcp_req(req, 0);
 	con->msg_attempts = 0;
 
@@ -356,16 +384,26 @@ done:
 	/* connection will be released */
 	return size;
 error:
+	/* ws_start_handshake must be completed so trace_ws was called
+	 * in order to get here so we're safe */
+	if ( ((struct ws_data *) con->proto_data)->dest ) {
+		if ( complete_ws_trace( con, TRANS_TRACE_FAILURE,
+							0, &CONNECT_FAIL) < 0 ) {
+			LM_ERR("failed to trace ws connect failure!\n");
+		}
+	}
+
 	return -1;
 }
 
 static int ws_server_handshake(struct tcp_connection *con)
 {
-	int bytes, total_bytes = 0;
+	int bytes;
 	long size = 0;
 	int msg_len;
 	char *msg_buf;
 	struct tcp_req *req;
+	str trace_str;
 
 	if (con->con_req) {
 		if (WS_TYPE(con) != WS_SERVER) {
@@ -394,7 +432,6 @@ static int ws_server_handshake(struct tcp_connection *con)
 			goto error;
 		}
 
-		total_bytes+=bytes;
 		/* eof check:
 		 * is EOF if eof on fd and r.  not complete yet,
 		 * if r. is complete we might have a second unparsed
@@ -430,7 +467,7 @@ static int ws_server_handshake(struct tcp_connection *con)
 		}
 
 		/* update the timeout - we successfully read the request */
-		tcp_conn_set_lifetime( con, tcp_con_lifetime);
+		tcp_conn_reset_lifetime(con);
 		con->timeout=con->lifetime;
 
 		con->rcv.proto_reserved1=con->id; /* copy the id */
@@ -470,14 +507,26 @@ static int ws_server_handshake(struct tcp_connection *con)
 			goto error;
 		}
 
+		if ( ((struct ws_data *) con->proto_data)->dest ) {
+			trace_str.len = msg_len;
+			trace_str.s = msg_buf;
+
+			if ( trace_ws( con, TRANS_TRACE_ACCEPTED, &trace_str) < 0 ) {
+				LM_ERR("failed to trace WS request!\n");
+			}
+		}
+
+
 		if (ws_complete_handshake(con) < 0) {
 			LM_DBG("cannot complete handshake\n");
 			goto error;
 		}
 
 		con->msg_attempts = 0;
-		if (req != &_ws_common_tcp_current_req)
-			pkg_free(req);
+		if (req != &_ws_common_tcp_current_req) {
+			shm_free(req);
+			con->con_req = NULL;
+		}
 
 		/* handshake now completed, destroy the handshake data */
 		WS_STATE(con) = WS_CON_HANDSHAKE_DONE;
@@ -486,10 +535,13 @@ static int ws_server_handshake(struct tcp_connection *con)
 		goto done;
 
 	} else {
+		int max_chunks = tcp_attr_isset(con, TCP_ATTR_MAX_MSG_CHUNKS) ?
+			con->profile.attrs[TCP_ATTR_MAX_MSG_CHUNKS] : _ws_common_max_msg_chunks;
+
 		/* request not complete - check the if the thresholds are exceeded */
 
 		con->msg_attempts++;
-		if (con->msg_attempts == _ws_common_max_msg_chunks) {
+		if (con->msg_attempts == max_chunks) {
 			LM_ERR("Made %u read attempts but message is not complete yet - "
 				   "closing connection \n",con->msg_attempts);
 			goto error;
@@ -499,7 +551,7 @@ static int ws_server_handshake(struct tcp_connection *con)
 	if (!req->complete && (req == &_ws_common_tcp_current_req)) {
 		/* let's duplicate this - most likely another conn will come in */
 
-		con->con_req = pkg_malloc(sizeof(struct tcp_req));
+		con->con_req = shm_malloc(sizeof(struct tcp_req));
 		if (con->con_req == NULL) {
 			LM_ERR("No more mem for dynamic con request buffer\n");
 			goto error;
@@ -546,7 +598,7 @@ error:
 	if (WS_STATE(con) == WS_CON_BAD_REQ)
 		ws_bad_handshake(con);
 	if (req != &_ws_common_tcp_current_req) {
-		pkg_free(req);
+		shm_free(req);
 		con->con_req = NULL;
 	}
 	return -1;
@@ -618,7 +670,7 @@ static inline int ws_parse_req_http_fl(struct tcp_connection *c,
 	tmp.s = spe + 5; /* start of the major */
 	sp = q_memchr(tmp.s, '.', end - tmp.s);
 	if (!sp || sp == tmp.s) {
-		LM_ERR("cannot find DOT");
+		LM_ERR("cannot find DOT\n");
 		goto version_error;
 	}
 	tmp.len = sp - tmp.s;
@@ -664,6 +716,9 @@ static inline int ws_parse_rpl_http_fl(struct tcp_connection *c,
 	end = cr;
 	len = end - p;
 
+	*msg_buf = cr + 2;
+	*msg_len -= len + 2;
+
 	if (len < HTTP_VERSION_LEN) {
 		LM_ERR("invalid first line: version too small <%.*s>\n", len, p);
 		goto error;
@@ -676,7 +731,7 @@ static inline int ws_parse_rpl_http_fl(struct tcp_connection *c,
 	tmp.s = p;
 	sp = q_memchr(tmp.s, '.', end - tmp.s);
 	if (!sp || sp == tmp.s) {
-		LM_ERR("cannot find version DOT");
+		LM_ERR("cannot find version DOT\n");
 		goto error;
 	}
 	tmp.len = sp - tmp.s;
@@ -688,7 +743,7 @@ static inline int ws_parse_rpl_http_fl(struct tcp_connection *c,
 	tmp.s = p;
 	sp = q_memchr(tmp.s, ' ', end - tmp.s);
 	if (!sp || sp == tmp.s) {
-		LM_ERR("cannot find version separator");
+		LM_ERR("cannot find version separator\n");
 		goto error;
 	}
 	tmp.len = sp - tmp.s;
@@ -770,7 +825,7 @@ static int ws_parse_req_handshake(struct tcp_connection *c, char *msg, int len)
 	struct sip_msg tmp_msg;
 	struct hdr_field *hf;
 	unsigned version;
-	char flags = 0;
+	unsigned char flags = 0;
 	unsigned ver_min, ver_maj;
 
 	if (ws_parse_req_http_fl(c, &msg, &len, &ver_maj, &ver_min) != 0) {
@@ -791,7 +846,7 @@ static int ws_parse_req_handshake(struct tcp_connection *c, char *msg, int len)
 	tmp_msg.buf = tmp_msg.unparsed = msg;
 	if (parse_headers(&tmp_msg, HDR_EOH_F, 0) < 0) {
 		LM_ERR("cannot parse headers\n%.*s\n", len, msg);
-		goto error;
+		goto ws_error;
 	}
 	/* verify headers according to RFC6455 */
 	for (hf = tmp_msg.headers; hf; hf = hf->next) {
@@ -842,13 +897,13 @@ static int ws_parse_req_handshake(struct tcp_connection *c, char *msg, int len)
 			flags |= WS_CONN_F;
 			break;
 		case GET_DWORD('o', 'r', 'i', 'g'): /* Origin */
-			/* TODO: always check for origin? */
 			if (hf->name.len !=  HDR_LEN("Origin") ||
 					GET_LOWER(hf->name.s + 4) != 'i' ||
 					GET_LOWER(hf->name.s + 5) != 'n')
 				break;
 
-			flags |= WS_ORIGIN_F;
+			if (_ws_common_require_origin)
+				flags |= WS_ORIGIN_F;
 			break;
 		case GET_DWORD('s', 'e', 'c', '-'): /* Sec-* */
 			if (hf->name.len < HDR_LEN("Sec-Websocket-*") ||
@@ -871,10 +926,17 @@ static int ws_parse_req_handshake(struct tcp_connection *c, char *msg, int len)
 
 				str_trim_spaces_lr(hf->body);
 
-				/* the key is already in the buffer, so we can just copy it */
-				WS_KEY(c) = hf->body;
-
-				flags |= WS_KEY_F;
+				/* RFC-6455 4.1: Opening Handshake: Client Requirements
+				 * 7.  The request MUST include a header field with the name
+				 *     |Sec-WebSocket-Key|.  The value of this header field MUST be a
+				 *     nonce consisting of a randomly selected 16-byte value that has
+				 *     been base64-encoded (see Section 4 of [RFC4648]).  The nonce
+				 *     MUST be selected randomly for each connection. */
+				if (hf->body.len == WS_KEY_LEN) {
+					/* the key is already in the buffer, so we can just copy it */
+					WS_KEY(c) = hf->body;
+					flags |= WS_KEY_F;
+				}
 			} else if (hf->name.len == HDR_LEN("Sec-WebSocket-Version") &&
 					GET_LOWER(hf->name.s + 14) == 'v' &&
 					GET_LOWER(hf->name.s + 15) == 'e' &&
@@ -914,7 +976,7 @@ static int ws_parse_req_handshake(struct tcp_connection *c, char *msg, int len)
 
 	}
 
-	if (flags != WS_ALL_REQ_F) {
+	if (flags != (unsigned char)WS_ALL_REQ_F) {
 		/* negate so we can easily compare them */
 		flags = ~flags;
 		if (flags & WS_HOST_F)
@@ -923,10 +985,11 @@ static int ws_parse_req_handshake(struct tcp_connection *c, char *msg, int len)
 			LM_ERR("Upgrade header not present!\n");
 		if (flags & WS_CONN_F)
 			LM_ERR("Connection header not present!\n");
-		if (flags & WS_ORIGIN_F)
+		if (_ws_common_require_origin && (flags & WS_ORIGIN_F))
 			LM_ERR("Origin header not present!\n");
 		if (flags & WS_KEY_F)
-			LM_ERR("Sec-WebSocket-Key header not present!\n");
+			LM_ERR("Sec-WebSocket-Key header not present or does not "
+					"have the desired length (%d)!\n", WS_KEY_LEN);
 		if (flags & WS_VER_F)
 			LM_ERR("Sec-WebSocket-Version header not present!\n");
 		if (flags & WS_PROTO_F)
@@ -964,7 +1027,7 @@ static void ws_compute_key(str *key)
 static int ws_is_valid_key(str *key, str *accept)
 {
 	ws_compute_key(key);
-	return strncasecmp((char *)ws_accept_buf, accept->s, accept->len);
+	return strncasecmp((char *)ws_accept_buf, accept->s, accept->len) == 0;
 }
 
 static int ws_complete_handshake(struct tcp_connection *c)
@@ -977,16 +1040,36 @@ static int ws_complete_handshake(struct tcp_connection *c)
 		{ (void *)HTTP_END, HTTP_END_LEN }/* message end */
 	};
 
-	reset_tcp_vars(tcpthreshold);
-	start_expire_timer(get, tcpthreshold);
+	str trace_str = { ws_trace_buf, 0 };
+	int iov_len = sizeof(iov) / sizeof(struct iovec), i;
+
+	reset_tcp_vars(c->profile.send_threshold);
+	start_expire_timer(get, c->profile.send_threshold);
 
 	/* compute the ws_key in ws_accept_buf */
 	ws_compute_key(&WS_KEY(c));
 
 	n = _ws_common_writev(c, c->fd, iov, 3, _ws_common_write_tout);
-	stop_expire_timer(get, tcpthreshold,
+	stop_expire_timer(get, c->profile.send_threshold,
 			_ws_common_module " handshake", "", 0, 1);
 
+	if ( ((struct ws_data *) c->proto_data)->dest ) {
+		for ( i=0; i < iov_len; i++ ) {
+			/* avoid overflow */
+			if ( trace_str.len + iov[i].iov_len > WS_TRACE_MAX )
+				break;
+
+			memcpy( trace_str.s + trace_str.len,
+					iov[i].iov_base, iov[i].iov_len );
+			trace_str.len += iov[i].iov_len;
+		}
+
+		if ( complete_ws_trace( c, TRANS_TRACE_SUCCESS,
+									&trace_str, &ACCEPT_OK ) < 0 ) {
+			LM_ERR("failed to complete webSocket handshake!\n");
+		}
+	}
+
 	return n;
 }
 
@@ -998,12 +1081,21 @@ static int ws_bad_handshake(struct tcp_connection *c)
 		{ (void*)WS_HTTP_BAD_REQ, WS_HTTP_BAD_REQ_LEN },
 	};
 
-	reset_tcp_vars(tcpthreshold);
-	start_expire_timer(get, tcpthreshold);
+	str trace_str = { WS_HTTP_BAD_REQ, WS_HTTP_BAD_REQ_LEN };
+
+	reset_tcp_vars(c->profile.send_threshold);
+	start_expire_timer(get, c->profile.send_threshold);
 	n = _ws_common_writev(c, c->fd, iov, 1, _ws_common_write_tout);
-	stop_expire_timer(get, tcpthreshold,
+	stop_expire_timer(get, c->profile.send_threshold,
 			_ws_common_module " handshake", "", 0, 1);
 
+	if ( ((struct ws_data *) c->proto_data)->dest ) {
+		if ( complete_ws_trace( c, TRANS_TRACE_FAILURE,
+							&trace_str, &ACCEPT_FAIL) < 0 ) {
+			LM_ERR("failed to complete web socket trace!\n");
+		}
+	}
+
 	return n;
 }
 
@@ -1011,7 +1103,7 @@ static int ws_parse_rpl_handshake(struct tcp_connection *c, char *msg, int len)
 {
 	struct sip_msg tmp_msg;
 	struct hdr_field *hf;
-	char flags = 0;
+	unsigned char flags = 0;
 	unsigned ver_min, ver_maj;
 
 	if (ws_parse_rpl_http_fl(c, &msg, &len, &ver_maj, &ver_min) != 0) {
@@ -1032,7 +1124,7 @@ static int ws_parse_rpl_handshake(struct tcp_connection *c, char *msg, int len)
 	tmp_msg.buf = tmp_msg.unparsed = msg;
 	if (parse_headers(&tmp_msg, HDR_EOH_F, 0) < 0) {
 		LM_ERR("cannot parse headers\n%.*s\n", len, msg);
-		goto error;
+		goto ws_error;
 	}
 	/* verify headers according to RFC6455 */
 	for (hf = tmp_msg.headers; hf; hf = hf->next) {
@@ -1146,21 +1238,15 @@ static int ws_parse_rpl_handshake(struct tcp_connection *c, char *msg, int len)
 
 	}
 
-	if (flags != WS_ALL_REQ_F) {
+	if (flags != (unsigned char)WS_ALL_RPL_F) {
 		/* negate so we can easily compare them */
 		flags = ~flags;
-		if (flags & WS_HOST_F)
-			LM_ERR("Host header not present!\n");
 		if (flags & WS_UPGRADE_F)
 			LM_ERR("Upgrade header not present!\n");
 		if (flags & WS_CONN_F)
 			LM_ERR("Connection header not present!\n");
-		if (flags & WS_ORIGIN_F)
-			LM_ERR("Origin header not present!\n");
-		if (flags & WS_KEY_F)
-			LM_ERR("Sec-WebSocket-Key header not present!\n");
-		if (flags & WS_VER_F)
-			LM_ERR("Sec-WebSocket-Version header not present!\n");
+		if (flags & WS_ACCEPT_F)
+			LM_ERR("Sec-WebSocket-Protocol header not present!\n");
 		if (flags & WS_PROTO_F)
 			LM_ERR("Sec-WebSocket-Protocol header not present!\n");
 		goto ws_error;
@@ -1181,11 +1267,16 @@ error:
 static int ws_start_handshake(struct tcp_connection *c)
 {
 	int n;
+	int i;
 	struct timeval get;
 	char *ip;
 	char *port;
 	int port_len;
 	static char host_orig_buf[MAX_HOST_LEN];
+	char *h;
+
+	str trace_str = { ws_trace_buf, 0 };
+
 	static struct iovec iov[] = {
 		{ (void*)HTTP_GET_METHOD, HTTP_GET_METHOD_LEN },	/* GET method */
 		{ (void*)" ", 1 },
@@ -1205,29 +1296,53 @@ static int ws_start_handshake(struct tcp_connection *c)
 		{ (void*)HTTP_HANDSHAKE_END, HTTP_HANDSHAKE_END_LEN }, /* constant part */
 	};
 
-	reset_tcp_vars(tcpthreshold);
-	start_expire_timer(get, tcpthreshold);
+	int iov_len = sizeof(iov) / sizeof(struct iovec);
 
-	ip = ip_addr2a(&c->rcv.dst_ip);
-	port = int2str(c->rcv.dst_port, &port_len);
+	reset_tcp_vars(c->profile.send_threshold);
+	start_expire_timer(get, c->profile.send_threshold);
+
+	ip = ip_addr2a(&c->rcv.src_ip);
+	port = int2str(c->rcv.src_port, &port_len);
 	n = strlen(ip);
-	memcpy(host_orig_buf, ip, n);
-	host_orig_buf[n] = ':';
-	memcpy(host_orig_buf + n + 1, port, port_len);
+	h = host_orig_buf;
+	if (c->rcv.src_ip.af == AF_INET6)
+		*h++ = '[';
+	memcpy(h, ip, n);
+	h += n;
+	if (c->rcv.src_ip.af == AF_INET6)
+		*h++ = ']';
+	*h++ = ':';
+	memcpy(h, port, port_len);
+	h += port_len;
 
 	iov[2].iov_base = _ws_common_resource.s;
 	iov[2].iov_len = _ws_common_resource.len;
 
-	iov[7].iov_len = n + port_len + 1;
+	iov[7].iov_len = h - host_orig_buf;
 	iov[10].iov_len = iov[7].iov_len;
 
 	iov[13].iov_base = WS_KEY(c).s;
 	iov[13].iov_len = WS_KEY(c).len;
 
 	n = _ws_common_writev(c, c->fd, iov, 16, _ws_common_write_tout);
-	stop_expire_timer(get, tcpthreshold,
+	stop_expire_timer(get, c->profile.send_threshold,
 			_ws_common_module " start handshake", "", 0, 1);
 
+	if ( ((struct ws_data *) c->proto_data)->dest ) {
+		for ( i=0; i < iov_len; i++ ) {
+			/* avoid overflow */
+			if ( trace_str.len + iov[i].iov_len > WS_TRACE_MAX )
+				break;
+
+			memcpy( trace_str.s + trace_str.len,
+					iov[i].iov_base, iov[i].iov_len );
+			trace_str.len += iov[i].iov_len;
+		}
+		if ( trace_ws( c, TRANS_TRACE_CONNECTED, &trace_str ) < 0 ) {
+			LM_ERR("WS trace failed!\n");
+		}
+	}
+
 	return n;
 }
 
@@ -1532,4 +1647,72 @@ skip:
 	return bytes;
 }
 
+static int trace_ws( struct tcp_connection* conn, trans_trace_event event, str* req)
+{
+#define WS_TRACE_IS_ON( CONN ) (CONN->proto_data && \
+		((struct ws_data*)CONN->proto_data)->tprot && \
+			((struct ws_data*)CONN->proto_data)->dest && \
+			*((struct ws_data*)CONN->proto_data)->trace_is_on)
+
+
+	struct ws_data* d;
+	union sockaddr_union src, dst;
+
+	if ( !conn || !req || !req->s || !req->len ||
+			!WS_TRACE_IS_ON(conn) || ! (d = conn->proto_data) )
+		return 0;
+
+	if ( ref_script_route_is_valid(d->trace_route_ref) ) {
+		check_trace_route( d->trace_route_ref, conn );
+		/* avoid doing this multiple times */
+		d->trace_route_ref = NULL;
+	}
+
+	/* check if tracing is deactivated from the route for this connection */
+	if ( conn->flags & F_CONN_TRACE_DROPPED )
+		return 0;
+
+	if ( !d->message  ) {
+		if ( tcpconn2su( conn, &src, &dst ) < 0 ) {
+			LM_ERR("can't fetch network info!\n");
+			return -1;
+		}
+
+		if ( (d->message = create_trace_message( conn->cid, &src, &dst,
+						conn->type, d->dest )) == 0 )  {
+			LM_ERR(" can't init trace_message!\n");
+			return -1;
+		}
+	}
+
+	add_trace_data( d->message, "Event", &trans_trace_str_event[event]);
+	add_trace_data( d->message, "Ws-Request", req);
+
+	return 0;
+#undef WS_TRACE_IS_ON
+}
+
+static int complete_ws_trace( struct tcp_connection* conn, trans_trace_status status, str* rpl, str* message)
+{
+	struct ws_data* d;
+
+	if ( !conn || !rpl || !rpl->s || !rpl->len || !(d = conn->proto_data) || !d->message )
+		return 0;
+
+	if ( !(*d->trace_is_on) || conn->flags & F_CONN_TRACE_DROPPED )
+		return 0;
+
+	/* most probably tracing was activated after the request was processed */
+	if ( !d->message )
+		return 0;
+
+	add_trace_data( d->message, "Status", &trans_trace_str_status[status]);
+	add_trace_data( d->message, "Ws-Reply", rpl);
+	if ( message && message->s && message->len )
+		add_trace_data( d->message, "Message", message);
+
+	return 0;
+}
+
+#undef WS_TRACE_MAX
 #endif /* _WS_HANDSHAKE_COMMON_H_ */
