@@ -58,17 +58,19 @@
 #include "../../dset.h"
 #include "../../mod_fix.h"
 #include "../../data_lump.h"
+#include "../../lib/reg/common.h"
+
 #include "../usrloc/usrloc.h"
-#include "common.h"
+
 #include "sip_msg.h"
-#include "rerrno.h"
 #include "reply.h"
 #include "reg_mod.h"
-#include "regtime.h"
-#include "path.h"
 #include "save.h"
 #include "lookup.h"
 
+int filter_contacts(urecord_t *r, struct sip_msg *by_msg);
+void restore_contacts(urecord_t *r);
+
 /*! \brief
  * Process request that contained a star, in that case,
  * we will remove all bindings with the given username
@@ -113,48 +115,6 @@ static inline int star(udomain_t* _d, struct save_ctx *_sctx,
 }
 
 
-/*! \brief
- */
-static struct socket_info *get_sock_hdr(struct sip_msg *msg)
-{
-	struct socket_info *sock;
-	struct hdr_field *hf;
-	str socks;
-	str hosts;
-	int port;
-	int proto;
-
-	if (parse_headers( msg, HDR_EOH_F, 0) == -1) {
-		LM_ERR("failed to parse message\n");
-		return 0;
-	}
-
-	hf = get_header_by_name( msg, sock_hdr_name.s, sock_hdr_name.len);
-	if (hf==0)
-		return 0;
-
-	trim_len( socks.len, socks.s, hf->body );
-	if (socks.len==0)
-		return 0;
-
-	if (parse_phostport( socks.s, socks.len, &hosts.s, &hosts.len,
-	&port, &proto)!=0) {
-		LM_ERR("bad socket <%.*s> in \n",
-			socks.len, socks.s);
-		return 0;
-	}
-	set_sip_defaults( port, proto);
-	sock = grep_sock_info(&hosts,(unsigned short)port,(unsigned short)proto);
-	if (sock==0) {
-		LM_ERR("non-local socket <%.*s>\n",	socks.len, socks.s);
-		return 0;
-	}
-
-	LM_DBG("%d:<%.*s>:%d -> p=%p\n", proto,socks.len,socks.s,port,sock );
-
-	return sock;
-}
-
 
 
 /*! \brief
@@ -163,193 +123,78 @@ static struct socket_info *get_sock_hdr(struct sip_msg *msg)
  * containing a list of all existing bindings for the
  * given username (in To HF)
  */
-static inline int no_contacts(udomain_t* _d, str* _a,struct sip_msg *_m)
+static inline int no_contacts(udomain_t* _d, struct save_ctx *_sctx,
+                              struct sip_msg *_m)
 {
 	urecord_t* r;
 	int res;
 
-	ul.lock_udomain(_d, _a);
-	res = ul.get_urecord(_d, _a, &r);
+	ul.lock_udomain(_d, &_sctx->aor);
+	res = ul.get_urecord(_d, &_sctx->aor, &r);
 	if (res < 0) {
 		rerrno = R_UL_GET_R;
 		LM_ERR("failed to retrieve record from usrloc\n");
-		ul.unlock_udomain(_d, _a);
+		ul.unlock_udomain(_d, &_sctx->aor);
 		return -1;
 	}
 
 	if (res == 0) {  /* Contacts found */
+		if (_sctx->flags & REG_SAVE_REQ_CT_ONLY_FLAG)
+			filter_contacts(r, _m);
+
 		build_contact(r->contacts,_m);
+
+		if (_sctx->flags & REG_SAVE_REQ_CT_ONLY_FLAG)
+			restore_contacts(r);
 	}
-	ul.unlock_udomain(_d, _a);
+	ul.unlock_udomain(_d, &_sctx->aor);
 	return 0;
 }
 
-
-
 /*! \brief
- * Fills the common part (for all contacts) of the info structure
  */
-static inline ucontact_info_t* pack_ci( struct sip_msg* _m, contact_t* _c,
-						unsigned int _e, unsigned int _f, unsigned int _flags)
+static int set_sock_hdr(struct sip_msg *msg, ucontact_info_t *ci,
+                        unsigned int reg_flags)
 {
-	static ucontact_info_t ci;
-	static str no_ua = str_init("n/a");
-	static str callid;
-	static str path_received = {0,0};
-	static str path;
-	static str received = {0,0};
-	static int received_found;
-	static unsigned int allowed, allow_parsed;
-	static struct sip_msg *m = 0;
-	static int_str attr_avp_value;
-	static struct usr_avp *avp_attr;
-	int_str val;
+	struct socket_info *sock;
+	struct hdr_field *hf;
+	str socks;
+	str hosts;
+	int port;
+	int proto;
 
-	if (_m!=0) {
-		memset( &ci, 0, sizeof(ucontact_info_t));
+	if (!msg || !(reg_flags & REG_SAVE_SOCKET_FLAG))
+		return 1;
 
-		/* Get callid of the message */
-		callid = _m->callid->body;
-		trim_trailing(&callid);
-		if (callid.len > CALLID_MAX_SIZE) {
-			rerrno = R_CALLID_LEN;
-			LM_ERR("callid too long\n");
-			goto error;
-		}
-		ci.callid = &callid;
-
-		/* Get CSeq number of the message */
-		if (str2int(&get_cseq(_m)->number, (unsigned int*)&ci.cseq) < 0) {
-			rerrno = R_INV_CSEQ;
-			LM_ERR("failed to convert cseq number\n");
-			goto error;
-		}
-
-		/* set received socket */
-		if ( _flags&REG_SAVE_SOCKET_FLAG) {
-			ci.sock = get_sock_hdr(_m);
-			if (ci.sock==0)
-				ci.sock = _m->rcv.bind_address;
-		} else {
-			ci.sock = _m->rcv.bind_address;
-		}
-
-		/* additional info from message */
-		if (parse_headers(_m, HDR_USERAGENT_F, 0) != -1 && _m->user_agent &&
-		_m->user_agent->body.len>0 && _m->user_agent->body.len<UA_MAX_SIZE) {
-			ci.user_agent = &_m->user_agent->body;
-		} else {
-			ci.user_agent = &no_ua;
-		}
-
-		/* extract Path headers */
-		if ( _flags&REG_SAVE_PATH_FLAG ) {
-			if (build_path_vector(_m, &path, &path_received, _flags) < 0) {
-				rerrno = R_PARSE_PATH;
-				goto error;
-			}
-			if (path.len && path.s) {
-				ci.path = &path;
-				/* save in msg too for reply */
-				if (set_path_vector(_m, &path) < 0) {
-					rerrno = R_PARSE_PATH;
-					goto error;
-				}
-			}
-		}
-
-		ci.last_modified = act_time;
-
-		/* set flags */
-		ci.flags  = _f;
-		ci.cflags =  getb0flags(_m);
-
-		/* get received */
-		if (path_received.len && path_received.s) {
-			ci.cflags |= ul.nat_flag;
-			ci.received = path_received;
-		}
-
-		allow_parsed = 0; /* not parsed yet */
-		received_found = 0; /* not found yet */
-		m = _m; /* remember the message */
+	if (parse_headers( msg, HDR_EOH_F, 0) == -1) {
+		LM_ERR("failed to parse message\n");
+		return 1;
 	}
 
-	if(_c!=0) {
-		/* Calculate q value of the contact */
-		if (calc_contact_q(_c->q, &ci.q) < 0) {
-			rerrno = R_INV_Q;
-			LM_ERR("failed to calculate q\n");
-			goto error;
-		}
+	hf = get_header_by_name( msg, sock_hdr_name.s, sock_hdr_name.len);
+	if (hf==0)
+		return 1;
 
-		/* set expire time */
-		ci.expires = _e;
+	trim_len( socks.len, socks.s, hf->body );
+	if (socks.len==0)
+		return 1;
 
-		/* Get methods of contact */
-		if (_c->methods) {
-			if (parse_methods(&(_c->methods->body), &ci.methods) < 0) {
-				rerrno = R_PARSE;
-				LM_ERR("failed to parse contact methods\n");
-				goto error;
-			}
-		} else {
-			/* check on Allow hdr */
-			if (allow_parsed == 0) {
-				if (m && parse_allow( m ) != -1) {
-					allowed = get_allow_methods(m);
-				} else {
-					allowed = ALL_METHODS;
-				}
-				allow_parsed = 1;
-			}
-			ci.methods = allowed;
-		}
-
-		if (_c->instance) {
-			ci.instance = _c->instance->body;
-		}
-
-		/* get received */
-		if (ci.received.len==0) {
-			if (_c->received) {
-				ci.received = _c->received->body;
-			} else {
-				if (received_found==0) {
-					memset(&val, 0, sizeof(int_str));
-					if (rcv_avp_name>=0
-								&& search_first_avp(rcv_avp_type, rcv_avp_name, &val, 0)
-								&& val.s.len > 0) {
-						if (val.s.len>RECEIVED_MAX_SIZE) {
-							rerrno = R_CONTACT_LEN;
-							LM_ERR("received too long\n");
-							goto error;
-						}
-						received = val.s;
-					} else {
-						received.s = 0;
-						received.len = 0;
-					}
-					received_found = 1;
-				}
-				ci.received = received;
-			}
-		}
-
-		/* additional information (script pvar) */
-		if (attr_avp_name != -1) {
-			avp_attr = search_first_avp(attr_avp_type, attr_avp_name,
-										&attr_avp_value, NULL);
-			if (avp_attr) {
-				ci.attr = &attr_avp_value.s;
-
-				LM_DBG("Attributes: %.*s\n", ci.attr->len, ci.attr->s);
-			}
-		}
+	if (parse_phostport( socks.s, socks.len, &hosts.s, &hosts.len,
+	&port, &proto)!=0) {
+		LM_ERR("bad socket <%.*s> in \n",
+			socks.len, socks.s);
+		return 1;
+	}
+	set_sip_defaults( port, proto);
+	sock = grep_sock_info(&hosts,(unsigned short)port,(unsigned short)proto);
+	if (sock==0) {
+		LM_ERR("non-local socket <%.*s>\n",	socks.len, socks.s);
+		return 1;
 	}
 
-	return &ci;
-error:
+	LM_DBG("%d:<%.*s>:%d -> p=%p\n", proto,socks.len,socks.s,port,sock );
+
+	ci->sock = sock;
 	return 0;
 }
 
@@ -396,10 +241,12 @@ static inline int insert_contacts(struct sip_msg* _m, contact_t* _c,
 				if (r==NULL || r->contacts==NULL) {
 					LM_CRIT("BUG - overflow detected with r=%p and "
 						"contacts=%p\n",r,r->contacts);
+					rerrno = R_INTERNAL;
 					goto error;
 				}
-				if (ul.delete_ucontact( r, r->contacts, 0)!=0) {
+				if (ul.delete_ucontact( r, r->contacts, &_sctx->cmatch, 0)!=0) {
 					LM_ERR("failed to remove contact\n");
+					rerrno = R_INTERNAL;
 					goto error;
 				}
 			} else {
@@ -421,26 +268,35 @@ static inline int insert_contacts(struct sip_msg* _m, contact_t* _c,
 		}
 
 		/* pack the contact_info */
-		if ( (ci=pack_ci( (ci==0)?_m:0, _c, e, cflags, _sctx->flags))==0 ) {
+		if ( (ci=pack_ci( (ci==0)?_m:0, _c, e, cflags, ul.nat_flag,
+		_sctx->flags, &_sctx->ownership_tag, &_sctx->cmatch))==0 ) {
 			LM_ERR("failed to extract contact info\n");
 			goto error;
 		}
 
+		set_sock_hdr(_m, ci, _sctx->flags);
+
 		if ( r->contacts==0 ||
-		ul.get_ucontact(r, &_c->uri, ci->callid, ci->cseq+1, &c)!=0 ) {
-			if (ul.insert_ucontact( r, &_c->uri, ci, &c, 0) < 0) {
+		ul.get_ucontact(r, &_c->uri, ci->callid, ci->cseq+1, &_sctx->cmatch,
+		&c)!=0 ){
+			if (ul.insert_ucontact( r, &_c->uri, ci, &_sctx->cmatch,
+				    0, &c) < 0) {
 				rerrno = R_UL_INS_C;
 				LM_ERR("failed to insert contact\n");
 				goto error;
 			}
 		} else {
-			if (ul.update_ucontact( r, c, ci, 0) < 0) {
+			if (ul.update_ucontact( r, c, ci, &_sctx->cmatch, 0) < 0) {
 				rerrno = R_UL_UPD_C;
 				LM_ERR("failed to update contact\n");
 				goto error;
 			}
 		}
 
+		if (pn_enable && pn_add_reply_purr(c) != 0)
+			LM_ERR("failed to add +sip.pnspurr for Contact: '%.*s'\n",
+			       _c->uri.len, _c->uri.s);
+
 		if (tcp_check) {
 			/* parse contact uri to see if transport is TCP */
 			if (parse_uri( _c->uri.s, _c->uri.len, &uri)<0) {
@@ -459,13 +315,19 @@ static inline int insert_contacts(struct sip_msg* _m, contact_t* _c,
 
 	if (r) {
 		if (r->contacts) {
+			if (_sctx->flags & REG_SAVE_REQ_CT_ONLY_FLAG)
+				filter_contacts(r, _m);
+
 			build_contact(r->contacts,_m);
+
+			if (_sctx->flags & REG_SAVE_REQ_CT_ONLY_FLAG)
+				restore_contacts(r);
 		}
 		ul.release_urecord(r, 0);
 	}
 
 	if ( tcp_check && e_max>0 ) {
-		e_max -= act_time;
+		e_max -= get_act_time();
 		trans_set_dst_attr( &_m->rcv, DST_FCNTL_SET_LIFETIME,
 			(void*)(long)(e_max + 10) );
 	}
@@ -507,17 +369,20 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 	cflags = (_sctx->flags&REG_SAVE_MEMORY_FLAG)?FL_MEM:FL_NONE;
 
 	/* pack the contact_info */
-	if ( (ci=pack_ci( _m, 0, 0, cflags, _sctx->flags))==0 ) {
+	if ( (ci=pack_ci( _m, 0, 0, cflags, ul.nat_flag, _sctx->flags,
+					&_sctx->ownership_tag, &_sctx->cmatch))==0 ) {
 		LM_ERR("failed to initial pack contact info\n");
 		goto error;
 	}
 
+	set_sock_hdr(_m, ci, _sctx->flags);
+
 	/* count how many contacts we have right now */
 	num = 0;
 	if (_sctx->max_contacts) {
 		c = _r->contacts;
 		while(c) {
-			if (VALID_CONTACT(c, act_time)) num++;
+			if (VALID_CONTACT(c, get_act_time())) num++;
 			c = c->next;
 		}
 	}
@@ -534,7 +399,8 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 		calc_contact_expires(_m, _c->expires, &e, _sctx);
 
 		/* search for the contact*/
-		ret = ul.get_ucontact( _r, &_c->uri, ci->callid, ci->cseq, &c);
+		ret = ul.get_ucontact( _r, &_c->uri, ci->callid, ci->cseq,
+			&_sctx->cmatch, &c);
 		if (ret==-1) {
 			LM_ERR("invalid cseq for aor <%.*s>\n",_r->aor.len,_r->aor.s);
 			rerrno = R_INV_CSEQ;
@@ -555,7 +421,7 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 					   so remove the oldest valid one to prevent this */
 					for( c_it=_r->contacts,c_last=NULL ; c_it ;
 					c_it=c_it->next )
-						if (VALID_CONTACT(c_it, act_time))
+						if (VALID_CONTACT(c_it, get_act_time()))
 							c_last=c_it;
 					if (c_last==NULL) {
 						LM_CRIT("BUG - overflow detected but no valid "
@@ -564,7 +430,7 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 					}
 					LM_DBG("overflow on inserting new contact -> removing "
 						"<%.*s>\n", c_last->c.len, c_last->c.s);
-					if (ul.delete_ucontact( _r, c_last, 0)!=0) {
+					if (ul.delete_ucontact( _r, c_last, &_sctx->cmatch, 0)!=0) {
 						LM_ERR("failed to remove contact\n");
 						goto error;
 					}
@@ -578,16 +444,20 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 			}
 
 			/* pack the contact_info */
-			if ( (ci=pack_ci( 0, _c, e, 0, _sctx->flags))==0 ) {
+			if ( (ci=pack_ci( 0, _c, e, 0, ul.nat_flag, _sctx->flags,
+							&_sctx->ownership_tag, &_sctx->cmatch))==0 ) {
 				LM_ERR("failed to extract contact info\n");
 				goto error;
 			}
 
-			if (ul.insert_ucontact( _r, &_c->uri, ci, &c, 0) < 0) {
+			if (ul.insert_ucontact( _r, &_c->uri, ci, &_sctx->cmatch,
+				    0, &c) < 0) {
 				rerrno = R_UL_INS_C;
 				LM_ERR("failed to insert contact\n");
 				goto error;
 			}
+
+			num++;
 		} else {
 			/* Contact found */
 			if (e == 0) {
@@ -598,16 +468,19 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 					c->flags &= ~FL_MEM;
 				}
 
-				if (ul.delete_ucontact(_r, c, 0) < 0) {
+				if (ul.delete_ucontact(_r, c, &_sctx->cmatch, 0) < 0) {
 					rerrno = R_UL_DEL_C;
 					LM_ERR("failed to delete contact\n");
 					goto error;
 				}
+
+				continue;
+
 			} else {
 				/* do update */
 				/* if the contact to be updated is not valid, it will be after
 				 * update, so need to compensate the total number of contact */
-				if ( !VALID_CONTACT(c,act_time) )
+				if ( !VALID_CONTACT(c,get_act_time()) )
 					num++;
 				while ( _sctx->max_contacts && num>_sctx->max_contacts ) {
 					if (_sctx->flags&REG_SAVE_FORCE_REG_FLAG) {
@@ -616,7 +489,7 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 						   (but not the one to be updated !) */
 						for( c_it=_r->contacts,c_last=NULL ; c_it ;
 						c_it=c_it->next )
-							if (VALID_CONTACT(c_it, act_time) && c_it!=c)
+							if (VALID_CONTACT(c_it, get_act_time()) && c_it!=c)
 								c_last=c_it;
 						if (c_last==NULL) {
 							LM_CRIT("BUG - overflow detected but no "
@@ -625,7 +498,7 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 						}
 						LM_DBG("overflow on update -> removing contact "
 							"<%.*s>\n", c_last->c.len, c_last->c.s);
-						if (ul.delete_ucontact( _r, c_last, 0)!=0) {
+						if (ul.delete_ucontact( _r, c_last, &_sctx->cmatch, 0)!=0) {
 							LM_ERR("failed to remove contact\n");
 							goto error;
 						}
@@ -639,18 +512,24 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 				}
 
 				/* pack the contact specific info */
-				if ( (ci=pack_ci( 0, _c, e, 0, _sctx->flags))==0 ) {
+				if ( (ci=pack_ci( 0, _c, e, 0, ul.nat_flag, _sctx->flags,
+								&_sctx->ownership_tag, &_sctx->cmatch))==0 ) {
 					LM_ERR("failed to pack contact specific info\n");
 					goto error;
 				}
 
-				if (ul.update_ucontact(_r, c, ci, 0) < 0) {
+				if (ul.update_ucontact(_r, c, ci, &_sctx->cmatch, 0) < 0) {
 					rerrno = R_UL_UPD_C;
 					LM_ERR("failed to update contact\n");
 					goto error;
 				}
 			}
 		}
+
+		if (pn_enable && pn_add_reply_purr(c) != 0)
+			LM_ERR("failed to add +sip.pnspurr for Contact: '%.*s'\n",
+			       _c->uri.len, _c->uri.s);
+
 		if (tcp_check) {
 			/* parse contact uri to see if transport is TCP */
 			if (parse_uri( _c->uri.s, _c->uri.len, &uri)<0) {
@@ -666,7 +545,7 @@ static inline int update_contacts(struct sip_msg* _m, urecord_t* _r,
 	}
 
 	if ( tcp_check && e_max>-1 ) {
-		if (e_max) e_max -= act_time;
+		if (e_max) e_max -= get_act_time();
 		trans_set_dst_attr( &_m->rcv, DST_FCNTL_SET_LIFETIME,
 			(void*)(long)(e_max + 10) );
 	}
@@ -698,12 +577,26 @@ static inline int add_contacts(struct sip_msg* _m, contact_t* _c,
 
 	if (res == 0) { /* Contacts found */
 		if (update_contacts(_m, r, _c, _sctx) < 0) {
+			if (_sctx->flags & REG_SAVE_REQ_CT_ONLY_FLAG)
+				filter_contacts(r, _m);
+
 			build_contact(r->contacts,_m);
+
+			if (_sctx->flags & REG_SAVE_REQ_CT_ONLY_FLAG)
+				restore_contacts(r);
 			ul.release_urecord(r, 0);
 			ul.unlock_udomain(_d, &_sctx->aor);
 			return -3;
 		}
+
+		if (_sctx->flags & REG_SAVE_REQ_CT_ONLY_FLAG)
+			filter_contacts(r, _m);
+
 		build_contact(r->contacts,_m);
+
+		if (_sctx->flags & REG_SAVE_REQ_CT_ONLY_FLAG)
+			restore_contacts(r);
+
 		ul.release_urecord(r, 0);
 	} else {
 		if (insert_contacts(_m, _c, _d, &_sctx->aor, _sctx) < 0) {
@@ -720,88 +613,40 @@ static inline int add_contacts(struct sip_msg* _m, contact_t* _c,
  * Process REGISTER request and save it's contacts
  */
 #define is_cflag_set(_name) ((sctx.flags)&(_name))
-int save_aux(struct sip_msg* _m, str* forced_binding, char* _d, char* _f, char* _s)
+int save_aux(struct sip_msg* _m, str* forced_binding, void* _d,
+	struct save_flags *flags, str* uri, str* _owtag)
 {
 	struct save_ctx  sctx;
 	contact_t* c;
 	contact_t* forced_c = NULL;
 	int st;
-	str uri;
-	str flags_s;
-	pv_value_t val;
 
 	rerrno = R_FINE;
 	memset( &sctx, 0 , sizeof(sctx));
-	sctx.max_contacts = -1;
 
-	sctx.flags = 0;
+	sctx.cmatch.mode = CT_MATCH_NONE;
 	sctx.min_expires = min_expires;
 	sctx.max_expires = max_expires;
-	if ( _f ) {
-		if (fixup_get_svalue( _m, (gparam_p)_f, &flags_s)!=0) {
-			LM_ERR("invalid flags parameter");
-			return -1;
-		}
-		for( st=0 ; st< flags_s.len ; st++ ) {
-			switch (flags_s.s[st]) {
-				case 'm': sctx.flags |= REG_SAVE_MEMORY_FLAG; break;
-				case 'r': sctx.flags |= REG_SAVE_NOREPLY_FLAG; break;
-				case 's': sctx.flags |= REG_SAVE_SOCKET_FLAG; break;
-				case 'v': sctx.flags |= REG_SAVE_PATH_RECEIVED_FLAG; break;
-				case 'f': sctx.flags |= REG_SAVE_FORCE_REG_FLAG; break;
-				case 'c':
-					sctx.max_contacts = 0;
-					while (st<flags_s.len-1 && isdigit(flags_s.s[st+1])) {
-						sctx.max_contacts = sctx.max_contacts*10 +
-							flags_s.s[st+1] - '0';
-						st++;
-					}
-					break;
-				case 'e':
-					sctx.min_expires = 0;
-					while (st<flags_s.len-1 && isdigit(flags_s.s[st+1])) {
-						sctx.min_expires = sctx.min_expires*10 +
-							flags_s.s[st+1] - '0';
-						st++;
-					}
-					break;
-				case 'E':
-					sctx.max_expires = 0;
-					while (st<flags_s.len-1 && isdigit(flags_s.s[st+1])) {
-						sctx.max_expires = sctx.max_expires*10 +
-							flags_s.s[st+1] - '0';
-						st++;
-					}
-					break;
-				case 'p':
-					if (st<flags_s.len-1) {
-						st++;
-						if (flags_s.s[st]=='2') {
-							sctx.flags |= REG_SAVE_PATH_STRICT_FLAG; break; }
-						if (flags_s.s[st]=='1') {
-							sctx.flags |= REG_SAVE_PATH_LAZY_FLAG; break; }
-						if (flags_s.s[st]=='0') {
-							sctx.flags |= REG_SAVE_PATH_OFF_FLAG; break; }
-					}
-				default: LM_WARN("unsupported flag %c \n",flags_s.s[st]);
-			}
-		}
+	sctx.max_contacts = max_contacts;
+	if (flags) {
+		sctx.flags = flags->flags;
+		sctx.max_contacts = flags->max_contacts;
+		sctx.min_expires = flags->min_expires;
+		sctx.max_expires = flags->max_expires;
+		sctx.cmatch = flags->cmatch;
 	}
-	if(route_type == ONREPLY_ROUTE)
+
+	if (route_type == ONREPLY_ROUTE)
 		sctx.flags |= REG_SAVE_NOREPLY_FLAG;
 
-	/* if no max_contact per AOR is defined, use the global one */
-	if (sctx.max_contacts == -1)
-		sctx.max_contacts = max_contacts;
-
-	if (parse_message(_m) < 0) {
+	if (parse_reg_headers(_m) < 0)
 		goto error;
-	}
 
 	if (forced_binding) {
 		if (parse_contacts(forced_binding, &forced_c) < 0) {
 			LM_ERR("Unable to parse forced binding [%.*s]\n",
 				forced_binding->len, forced_binding->s);
+			rerrno = R_INTERNAL;
 			goto error;
 		}
 		/* prevent processing all the headers from the message */
@@ -809,41 +654,47 @@ int save_aux(struct sip_msg* _m, str* forced_binding, char* _d, char* _f, char*
 		st = 0;
 		c = forced_c;
 	} else {
-		if (check_contacts(_m, &st) > 0) {
+		if (check_contacts(_m, &st) > 0)
 			goto error;
-		}
+
 		c = get_first_contact(_m);
 	}
 
-	get_act_time();
+	update_act_time();
 
-	if (_s) {
-		if (pv_get_spec_value( _m, (pv_spec_p)_s, &val)!=0) {
-			LM_ERR("failed to get PV value\n");
-			goto return_minus_one;
-		}
-		if ( (val.flags&PV_VAL_STR)==0 ) {
-			LM_ERR("PV vals is not string\n");
-			goto return_minus_one;
-		}
-		uri = val.rs;
-	} else {
-		uri = get_to(_m)->uri;
+	if (!uri)
+		uri = &get_to(_m)->uri;
+
+	if (_owtag)
+		sctx.ownership_tag = *_owtag;
+
+	if (extract_aor(uri, &sctx.aor, 0, 0, reg_use_domain) < 0) {
+		LM_ERR("failed to extract Address Of Record\n");
+		goto error;
 	}
 
-	if (extract_aor( &uri, &sctx.aor,0,0) < 0) {
-		LM_ERR("failed to extract Address Of Record\n");
+	if (sctx.aor.len == 0) {
+		LM_ERR("the AoR URI is missing the 'username' part!\n");
+		rerrno = R_AOR_PARSE;
 		goto error;
 	}
 
 	if (c == 0) {
 		if (st) {
-			if (star((udomain_t*)_d, &sctx,_m) < 0) goto error;
+			if (star((udomain_t*)_d, &sctx,_m) < 0)
+				goto error;
 		} else {
-			if (no_contacts((udomain_t*)_d, &sctx.aor,_m) < 0) goto error;
+			if (no_contacts((udomain_t*)_d, &sctx, _m) < 0)
+				goto error;
 		}
 	} else {
-		if (add_contacts(_m, c, (udomain_t*)_d, &sctx) < 0) goto error;
+		if (pn_enable && pn_inspect_request(_m, &c->uri, &sctx) != 0) {
+			LM_DBG("SIP PN processing failed (%d)\n", rerrno);
+			goto error;
+		}
+
+		if (add_contacts(_m, c, (udomain_t*)_d, &sctx) < 0)
+			goto error;
 	}
 
 	update_stat(accepted_registrations, 1);
@@ -870,8 +721,8 @@ return_minus_one:
 	return -1;
 }
 
-#define MAX_FORCED_BINDING_LEN 256
-int save(struct sip_msg* _m, char* _d, char* _f, char* _s)
+#define MAX_FORCED_BINDING_LEN 512
+int save(struct sip_msg* _m, void* _d, void* _f, str* _s, str* _owtag)
 {
 	struct sip_msg* msg = _m;
 	struct cell* t = NULL;
@@ -888,9 +739,10 @@ int save(struct sip_msg* _m, char* _d, char* _f, char* _s)
 	char forced_binding_buf[MAX_FORCED_BINDING_LEN];
 	str forced_binding = {NULL, 0};
 	str *binding_uri;
+	str path_bk;
 
 	if(_m->first_line.type != SIP_REPLY)
-		return save_aux(_m, NULL, _d, _f, _s);
+		return save_aux(_m, NULL, _d, _f, _s, _owtag);
 
 	memset(&val, 0, sizeof(int_str));
 	if(!tmb.t_gett) {
@@ -908,11 +760,19 @@ int save(struct sip_msg* _m, char* _d, char* _f, char* _s)
 		return -1;
 	}
 
-	if (parse_message(_m) < 0) return -1;
+	if (parse_reg_headers(_m) < 0) return -1;
 	if (check_contacts(_m, &st) > 0) return -1;
-	if (parse_message(msg) < 0) return -1;
+	if (parse_reg_headers(msg) < 0) return -1;
 	if (check_contacts(msg, &st) > 0) return -1;
 
+	/* detach the path vec from the msg as it is allocated in shm, and all
+	 * the parse/set ops (done below by save_aux) assume it is in pkg */
+	path_bk = msg->path_vec;
+	msg->path_vec.s = NULL;
+	msg->path_vec.len = 0;
+
+	ret = -1;
+
 	/* msg - request
 	   _m  - reply
 	*/
@@ -924,13 +784,13 @@ int save(struct sip_msg* _m, char* _d, char* _f, char* _s)
 			if (msg->expires && ((exp_body_t*)(msg->expires->parsed))->valid) {
 				requested_exp = ((exp_body_t*)(msg->expires->parsed))->val;
 			} else {
-				LM_WARN("No expired defined\n");
+				requested_exp = default_expires;
 			}
 		} else {
 			if (str2int(&(request_c->expires->body), (unsigned int*)&requested_exp)<0) {
 				LM_ERR("unable to get expires from [%.*s]\n",
 					request_c->expires->body.len, request_c->expires->body.s);
-				return -1;
+				goto done;
 			}
 		}
 		LM_DBG("Binding received from client [%.*s] with requested expires [%d]\n",
@@ -959,7 +819,7 @@ int save(struct sip_msg* _m, char* _d, char* _f, char* _s)
 							LM_ERR("unable to get expires from [%.*s]\n",
 								_c->expires->body.len,
 								_c->expires->body.s);
-							return -1;
+							goto done;
 						}
 						LM_DBG("Binding received from upper registrar"
 							" [%.*s] with imposed expires [%d]\n",
@@ -983,10 +843,10 @@ int save(struct sip_msg* _m, char* _d, char* _f, char* _s)
 								forced_binding.s);
 							break;
 						} else {
-							LM_ERR("forced binding to BIG:"
+							LM_ERR("forced binding too BIG:"
 								" %d > MAX_FORCED_BINDING_LEN\n",
 								forced_binding.len);
-							return -1;
+							goto done;
 						}
 					}
 				} else {
@@ -996,7 +856,7 @@ int save(struct sip_msg* _m, char* _d, char* _f, char* _s)
 				_c = get_next_contact(_c);
 			}
 		}
-		ret = save_aux(msg, forced_binding.s?&forced_binding:NULL, _d, _f, _s);
+		ret = save_aux(msg, forced_binding.s?&forced_binding:NULL, _d, _f, _s, _owtag);
 	} else {
 		LM_DBG("No Contact in request => this is an interogation\n");
 		ret = 1;
@@ -1033,197 +893,335 @@ int save(struct sip_msg* _m, char* _d, char* _f, char* _s)
 	}
 
 done:
+	/* remove whatever was parsed and attached as pkg to the shm cloned req */
 	clean_msg_clone(t->uas.request, t->uas.request, t->uas.end_request);
+	/* free and restore the shm path vec */
+	if (msg->path_vec.s) pkg_free(msg->path_vec.s);
+	msg->path_vec = path_bk;
 
 	return ret;
 }
 
+
 /**
- * _remove - Delete an entire AOR entry or just one or more of its Contacts
- * Parameter format: _remove(domain, AOR[, Contact URI or plain hostname])
+ * _remove - Delete an entire AOR entry or one or more of its Contacts
  *
- * @udomain:     (udomain_t *)
- * @aor_gp:      address-of-record as a SIP URI (plain string or pvar)
- * @contact_gp:  contact to be deleted or domain in front of multiple contacts
+ * @domain:          logical domain name (usually name of location table)
+ * @aor_gp:          address-of-record as a SIP URI (plain string or pvar)
+ * @contact_gp:      contact URI to be deleted
+ * @next_hop_gp:     IP/domain in front of contacts to be deleted
+ * @sip_instance_gp: delete contacts with given "+sip_instance"
  *
  * @return:      1 on success, negative on failure
  */
-int _remove(struct sip_msg *msg, char *udomain, char *aor_gp, char *contact_gp)
+int _remove(struct sip_msg *msg, void *udomain, str *aor_uri, str *match_ct,
+            str *match_next_hop, str *match_sin)
 {
-	struct sip_uri puri;
-	struct hostent delete_he, *he;
+	struct hostent delete_nh_he, *he;
 	urecord_t *record;
 	ucontact_t *contact, *it;
-	str uri, aor_user, delete_user = { NULL, 0 };
-	int err, count = 0;
-	int delete_by_hostname = 0;
-	unsigned short delete_port;
+	str aor_user;
+	int ret = 1;
+	unsigned short delete_port = 0;
 
-	memset(&delete_he, 0, sizeof delete_he);
-
-	if (fixup_get_svalue(msg, (gparam_p)aor_gp, &uri) != 0) {
-		LM_ERR("failed to get gparam_t value\n");
-		return E_UNSPEC;
-	}
-
-	if (extract_aor( &uri, &aor_user,0,0) < 0) {
+	if (extract_aor(aor_uri, &aor_user, 0, 0, reg_use_domain) < 0) {
 		LM_ERR("failed to extract Address Of Record\n");
 		return E_BAD_URI;
 	}
 
+	memset( &delete_nh_he, 0, sizeof(struct hostent));
+
 	ul.lock_udomain((udomain_t *)udomain, &aor_user);
 
 	if (ul.get_urecord((udomain_t *)udomain, &aor_user, &record) != 0) {
 		LM_DBG("no record '%.*s' found!\n", aor_user.len, aor_user.s);
-		err = 1;
 		goto out_unlock;
 	}
 
-	/* if no contact uri param is given, delete the whole urecord entry */
-	if (!contact_gp) {
+	/* without any additional filtering, delete the whole urecord entry */
+	if (!match_ct && !match_next_hop && !match_sin) {
 		if (ul.delete_urecord((udomain_t *)udomain, &aor_user, record, 0) != 0) {
 			LM_ERR("failed to delete urecord for aor '%.*s'\n",
 			        aor_user.len, aor_user.s);
-			err = E_UNSPEC;
+			ret = E_UNSPEC;
 			goto out_unlock;
 		}
 
-		err = 1;
 		goto out_unlock;
 	}
 
-	if (fixup_get_svalue(msg, (gparam_p)contact_gp, &uri) != 0) {
-		LM_ERR("failed to retrieve value of contact pv\n");
-		err = E_UNSPEC;
-		goto out_unlock;
-	}
+	if (match_ct)
+		LM_DBG("Delete by contact: [%.*s]\n", match_ct->len, match_ct->s);
 
-	/* minimum two-letters for the domain name */
-	if (uri.len < 5) {
-		LM_ERR("Invalid domain given: '%.*s'\n", uri.len, uri.s);
-		err = E_INVALID_PARAMS;
-		goto out_unlock;
-	}
+	if (match_sin)
+			LM_DBG("Delete by sip_instance: [%.*s]\n",
+				match_sin->len, match_sin->s);
 
-	/* a domain/IP address was given instead of a SIP contact URI */
-	if (uri.s[0] != 's' || uri.s[1] != 'i' ||
-	    uri.s[2] != 'p' || (uri.s[3] != ':' &&
-	                        (uri.s[3] != 's' || uri.s[4] != ':'))) {
-
-		delete_by_hostname = 1;
-
-		he = sip_resolvehost(&uri, &delete_port, NULL, 0, NULL);
+	if (match_next_hop) {
+		he = sip_resolvehost(match_next_hop, &delete_port, NULL, 0, NULL);
 		if (!he) {
-			LM_ERR("cannot resolve given host: '%.*s'\n", uri.len, uri.s);
-			err = E_UNSPEC;
+			LM_ERR("cannot resolve given host: '%.*s'\n",
+			       match_next_hop->len, match_next_hop->s);
+			ret = E_UNSPEC;
 			goto out_unlock;
 		}
 
+		struct sockaddr_in daddr;
+		memcpy(&daddr.sin_addr, he->h_addr_list[0], sizeof(daddr.sin_addr));
 		LM_DBG("Delete by host: '%s'\n",
-		        inet_ntoa(*(struct in_addr *)(he->h_addr_list[0])));
-	} else {
-		LM_DBG("parsing uri: %.*s\n", uri.len, uri.s);
+		        inet_ntoa(daddr.sin_addr));
 
-		if (parse_uri(uri.s, uri.len, &puri) != 0) {
-			LM_ERR("failed to parse contact uri: '%.*s'\n",
-			        uri.len, uri.s);
-			err = E_BAD_URI;
+		if (hostent_cpy(&delete_nh_he, he) != 0) {
+			LM_ERR("no more pkg mem\n");
+			ret = E_OUT_OF_MEM;
 			goto out_unlock;
 		}
-
-		delete_user = puri.user;
-
-		he = sip_resolvehost(&puri.host, &delete_port, &puri.proto, 0, NULL);
-		if (!he) {
-			LM_ERR("cannot resolve given uri: '%.*s'\n", uri.len, uri.s);
-			err = E_UNSPEC;
-			goto out_unlock;
-		}
-
-		if (puri.port_no > 0)
-			delete_port  = puri.port_no;
-
-		LM_DBG("Delete by contact: [ User %.*s | Host %s | Port %d ]\n",
-		        delete_user.len, delete_user.s,
-		        inet_ntoa(*(struct in_addr *)(he->h_addr_list[0])),
-				delete_port);
 	}
 
-	if (hostent_cpy(&delete_he, he) != 0) {
-		LM_ERR("no more pkg mem\n");
-		err = E_OUT_OF_MEM;
-		goto out_unlock;
-	}
 
 	for (it = record->contacts; it; ) {
 		contact = it;
 		it = it->next;
-		count++;
 
-		LM_DBG("parsing contact uri '%.*s'\n", contact->c.len, contact->c.s);
+		LM_DBG("checking contact uri '%.*s'\n", contact->c.len, contact->c.s);
 
-		if (parse_uri(contact->c.s, contact->c.len, &puri) != 0) {
-			LM_ERR("failed to parse contact uri: '%.*s'\n",
-			        contact->c.len, contact->c.s);
-			err = E_BAD_URI;
-			goto out_unlock;
-		}
-
-		/* if necessary, solve the next_hop towards the contact */
 		he = sip_resolvehost(&contact->next_hop.name,
 		                     &contact->next_hop.port,
 		                     &contact->next_hop.proto, 0, NULL);
 		if (!he) {
-			LM_ERR("failed to resolve next hop of contact '%.*s'\n",
-			        contact->c.len, contact->c.s);
+			LM_ERR("failed to resolve next hop %.*s of contact '%.*s'\n",
+				contact->next_hop.name.len, contact->next_hop.name.s,
+				contact->c.len, contact->c.s);
 			continue;
 		}
 
-		LM_DBG("Contact: [ User %.*s | Host %s | Port %d ]\n",
-		        puri.user.len, puri.user.s,
-		        inet_ntoa(*(struct in_addr *)(he->h_addr_list[0])),
-				puri.port_no);
+		struct sockaddr_in daddr;
+		memcpy(&daddr.sin_addr, he->h_addr_list[0], sizeof(daddr.sin_addr));
+		LM_DBG("next hop is [%.*s] resolving to [%s]\n",
+			contact->next_hop.name.len, contact->next_hop.name.s,
+			inet_ntoa(daddr.sin_addr));
 
-		if (delete_by_hostname) {
-			if (!memcmp(delete_he.h_addr_list[0],
-			            he->h_addr_list[0], he->h_length))
-			{
-				ul.delete_ucontact(record, contact, 0);
-				count--;
-			}
-		} else {
-			if (delete_user.len == puri.user.len &&
-			    delete_port == puri.port_no &&
-			    !memcmp(delete_he.h_addr_list[0],
-			            he->h_addr_list[0], he->h_length)
-				&& !memcmp(delete_user.s, puri.user.s, puri.user.len))
-			{
-				ul.delete_ucontact(record, contact, 0);
-				count--;
-			}
+		if (match_next_hop) {
+			if (memcmp(delete_nh_he.h_addr_list[0],
+			he->h_addr_list[0], he->h_length))
+				continue;
 		}
+
+		if (match_ct) {
+			if (match_ct->len != contact->c.len ||
+			memcmp(match_ct->s, contact->c.s, match_ct->len))
+				continue;
+		}
+
+		if (match_sin) {
+			if (str_strcmp(match_sin, &contact->instance))
+				continue;
+		}
+
+		ul.delete_ucontact(record, contact, NULL, 0);
 	}
 
-	err = 1;
-
-	/* remove the AOR if no more contacts are attached */
-	if (count == 0) {
-		if (ul.delete_urecord((udomain_t *)udomain, &aor_user, record, 0) != 0) {
-			LM_ERR("failed to delete urecord for aor '%.*s'\n",
-			        aor_user.len, aor_user.s);
-			err = 1;
-		}
-	}
+	ul.release_urecord(record, 0);
 
 out_unlock:
 	ul.unlock_udomain((udomain_t *)udomain, &aor_user);
-	free_hostent(&delete_he);
+	if (match_next_hop)
+		free_hostent(&delete_nh_he);
 
-	return err;
+	return ret;
 }
 
-int is_other_contact_f(struct sip_msg* msg, char* _d, char *_s)
+/* Assumes everything is locked */
+int _remove_ip_port_urecord(urecord_t* record, str *ip, int* port)
 {
-	LM_WARN("Deprecated! Use is_ip_registered() instead!\n");
-	return is_ip_registered(msg, _d, NULL, _s);
+	ucontact_t *contact, *it;
+	struct hostent *he;
+	str contact_ip;
+
+	for (it = record->contacts; it; ) {
+		contact = it;
+		it = it->next;
+
+		he = sip_resolvehost(&contact->next_hop.name,
+		                     &contact->next_hop.port,
+		                     &contact->next_hop.proto, 0, NULL);
+		if (!he) {
+			LM_ERR("failed to resolve next hop %.*s of contact '%.*s'\n",
+				contact->next_hop.name.len, contact->next_hop.name.s,
+				contact->c.len, contact->c.s);
+			continue;
+		}
+
+		contact_ip.s = inet_ntoa(*(struct in_addr *)(he->h_addr_list[0]));
+		contact_ip.len = strlen(contact_ip.s);
+
+		LM_DBG("next hop is [%.*s] resolving to [%s]\n",
+			contact->next_hop.name.len, contact->next_hop.name.s,
+			contact_ip.s);
+
+
+		if (ip->len == contact_ip.len &&
+		memcmp(ip->s,contact_ip.s,ip->len) == 0 &&
+		contact->next_hop.port == *port) {
+			LM_DBG("Removing contact \n");
+			ul.delete_ucontact(record, contact, NULL, 0);
+		}
+	}
+
+	return 0;
+}
+
+/**
+ * _remove - Delete ALL contacts by IP and Port
+ *
+ * @ip:		IP to filter for deletion
+ * @port	Port to filter for deletion
+ * @aor:	Optionally, AOR to filter for deletion
+ *
+ * @return:      1 on success, negative on failure
+ */
+int _remove_ip_port(struct sip_msg *msg, str *ip, int *port, void *udomain, str* aor)
+{
+	urecord_t *record;
+	str aor_user;
+	int ret = 1;
+	int i;
+	map_iterator_t it;
+	void ** dest;
+	udomain_t *dom = (udomain_t *)udomain;
+
+	if (!udomain || !ip || !port) {
+		LM_ERR("Mandatory parameters not provided \n");
+		return -1;
+	}
+
+	if (aor && aor->s ) {
+		LM_DBG("Removing %.*s : %d for AOR %.*s\n",ip->len,ip->s,
+		*port,aor->len,aor->s);
+		/* received AOR indication, we're lucky */
+		if (extract_aor(aor, &aor_user, 0, 0, reg_use_domain) < 0) {
+			LM_ERR("failed to extract Address Of Record\n");
+			return E_BAD_URI;
+		}
+
+
+		ul.lock_udomain(dom, &aor_user);
+
+		if (ul.get_urecord(dom, &aor_user, &record) != 0) {
+			LM_DBG("no record '%.*s' found!\n", aor_user.len, aor_user.s);
+			goto out_unlock;
+		}
+
+		if (_remove_ip_port_urecord(record,ip,port) != 0) {
+			LM_ERR("Failed to remove contacts \n");
+			ret = -1;
+		}
+
+		ul.release_urecord(record, 0);
+		goto out_unlock;
+	}
+
+	LM_DBG("Removing %.*s : %d for ALL AORs\n",ip->len,ip->s,*port);
+	
+	/* no AOR help, go through ALL registered AORs :( */
+	for(i=0; i<dom->size; i++) {
+		ul.lock_ulslot( dom, i);
+
+		map_first(dom->table[i].records, &it);
+		while (iterator_is_valid(&it)) {
+			dest = iterator_val(&it);
+			if (!dest) {
+				LM_ERR("Failed to get urecord\n");
+				goto error_unlock;
+			}
+
+			record = (urecord_t *)*dest;
+			iterator_next(&it);
+
+			if (_remove_ip_port_urecord(record,ip,port) != 0) {
+				LM_ERR("Failed to remove contacts \n");
+				/* continue here, might be more */
+			}
+
+			ul.release_urecord(record, 0);
+		}
+
+error_unlock:
+		ul.unlock_ulslot( dom, i);
+	}
+
+	return 1;
+
+out_unlock:
+	ul.unlock_udomain(dom, &aor_user);
+	return ret;
+}
+
+static ucontact_t **contacts_bak;
+static int contacts_bak_no;
+static int contacts_bak_sz;
+
+/* temporarily filter the contacts of a record */
+int filter_contacts(urecord_t *r, struct sip_msg *by_msg)
+{
+	contact_t *c;
+	ucontact_t *uc, *_uc;
+	int i;
+
+	/* back up the original list using a static array */
+	contacts_bak_no = 0;
+	for (i = 0, uc = r->contacts; uc; uc = uc->next, i++) {
+		if (i >= contacts_bak_sz) {
+			contacts_bak = pkg_realloc(contacts_bak,
+			                        (i ? 2 * contacts_bak_sz : 10) * sizeof r);
+			if (!contacts_bak) {
+				LM_ERR("oom\n");
+				return -1;
+			}
+
+			contacts_bak_sz = (i ? 2 * contacts_bak_sz : 10);
+		}
+
+		contacts_bak[i] = uc;
+	}
+	contacts_bak_no = i;
+
+	uc = NULL;
+	for (c = get_first_contact(by_msg); c; c = get_next_contact(c)) {
+		for (_uc = r->contacts; _uc; _uc = _uc->next) {
+			if (str_strcmp(&c->uri, &_uc->c))
+				continue;
+
+			if (!uc) {
+				uc = _uc;
+			} else {
+				uc->next = _uc;
+				uc = _uc;
+			}
+
+			break;
+		}
+	}
+
+	if (uc)
+		uc->next = NULL;
+
+	/* expose the filtered list */
+	r->contacts = uc;
+	return 0;
+}
+
+void restore_contacts(urecord_t *r)
+{
+	int i;
+
+	if (contacts_bak_no == 0)
+		return;
+
+	/* restore in-between links */
+	for (i = 0; i < contacts_bak_no - 1; i++)
+		contacts_bak[i]->next = contacts_bak[i + 1];
+
+	contacts_bak[contacts_bak_no - 1]->next = NULL;
+	r->contacts = contacts_bak[0];
 }
