@@ -106,9 +106,6 @@ int *sip_workers_pipes;
 /* pipe for the sangoma worker */
 int sangoma_pipe[2];
 
-static int *proc_counter;
-gen_lock_t *index_lock;
-
 /* generic module functions */
 static int mod_init(void);
 static int child_init(int rank);
@@ -121,36 +118,34 @@ static struct dlg_binds dlg_binds;
 /* module specific functions */
 static int sngtc_offer(struct sip_msg *msg);
 static int w_sngtc_callee_answer(struct sip_msg *msg,
-                                 char *gp_ip_a, char *gp_ip_b);
+                                 str *gp_ip_a, str *gp_ip_b);
 static int sngtc_callee_answer(struct sip_msg *msg);
 static int sngtc_caller_answer(struct sip_msg *msg);
 
 static int sng_logger(int level, char *fmt, ...);
 
-static proc_export_t procs[] = {
+static const proc_export_t procs[] = {
 	{ "sangoma_worker", NULL, NULL, sangoma_worker_loop, 1, 0 },
 	{ 0, 0, 0, 0, 0, 0 },
 };
 
-static param_export_t params[] = {
+static const param_export_t params[] = {
 	{ 0, 0, 0 }
 };
 
-static cmd_export_t cmds[] = {
-	{ "sngtc_offer",         (cmd_function)sngtc_offer, 0, 0, 0,
-		REQUEST_ROUTE|ONREPLY_ROUTE },
-	{ "sngtc_callee_answer", (cmd_function)w_sngtc_callee_answer, 0, 0, 0,
-		REQUEST_ROUTE|ONREPLY_ROUTE },
-	{ "sngtc_callee_answer", (cmd_function)w_sngtc_callee_answer, 1,
-	    fixup_sgp_null, 0, REQUEST_ROUTE|ONREPLY_ROUTE },
-	{ "sngtc_callee_answer", (cmd_function)w_sngtc_callee_answer, 2,
-	    fixup_sgp_sgp, 0, REQUEST_ROUTE|ONREPLY_ROUTE },
-	{ "sngtc_caller_answer", (cmd_function)sngtc_caller_answer, 0, 0, 0,
-		REQUEST_ROUTE|ONREPLY_ROUTE },
-	{ 0, 0, 0, 0, 0, 0 }
+static const cmd_export_t cmds[] = {
+	{"sngtc_offer", (cmd_function)sngtc_offer, {{0,0,0}},
+		REQUEST_ROUTE|ONREPLY_ROUTE},
+	{"sngtc_caller_answer", (cmd_function)sngtc_caller_answer, {{0,0,0}},
+		REQUEST_ROUTE|ONREPLY_ROUTE},
+	{"sngtc_callee_answer", (cmd_function)w_sngtc_callee_answer, {
+		{CMD_PARAM_STR,0,0},
+		{CMD_PARAM_STR,0,0}, {0,0,0}},
+		REQUEST_ROUTE|ONREPLY_ROUTE},
+	{0,0,{{0,0,0}},0}
 };
 
-static dep_export_t deps = {
+static const dep_export_t deps = {
 	{ /* OpenSIPS module dependencies */
 		{ MOD_TYPE_DEFAULT, "dialog", DEP_ABORT },
 		{ MOD_TYPE_NULL, NULL, 0 },
@@ -165,6 +160,7 @@ struct module_exports exports= {
 	MOD_TYPE_DEFAULT,/* class of this module */
 	MODULE_VERSION,
 	DEFAULT_DLFLAGS,
+	0,
 	&deps,           /* OpenSIPS module dependencies */
 	cmds,
 	0,
@@ -172,11 +168,14 @@ struct module_exports exports= {
 	0,
 	0,
 	0,
+	0,				/* exported transformations */
 	procs,
+	0,
 	mod_init,
 	(response_function) 0,
 	(destroy_function)mod_destroy,
-	child_init
+	child_init,
+	0               /* reload confirm function */
 };
 
 int sng_create_rtp(void * usr_priv, sngtc_codec_request_leg_t *codec_reg_leg,
@@ -263,17 +262,23 @@ free_mem:
 void sngtc_dlg_terminated(struct dlg_cell *dlg, int type,
                           struct dlg_cb_params *params)
 {
-	str info_ptr;
+	int_str info_ptr;
+	int val_type;
 	struct sngtc_info *info;
+	int rc;
+
+	rc = dlg_binds.fetch_dlg_value(dlg, &dlg_key_sngtc_info, &val_type,
+		&info_ptr, 0);
+
+	if (rc == -1) {
+		LM_ERR("failed to fetch caller sdp\n");
+		return;
+	} else  if (rc == -2)
+		return;
 
 	LM_DBG("freeing the sdp buffer\n");
 
-	if (dlg_binds.fetch_dlg_value(dlg, &dlg_key_sngtc_info, &info_ptr, 0) != 0) {
-		LM_ERR("failed to fetch caller sdp\n");
-		return;
-	}
-
-	info = *(struct sngtc_info **)info_ptr.s;
+	info = *(struct sngtc_info **)info_ptr.s.s;
 	LM_DBG("Info ptr: %p\n", info);
 
 	free_transcoding_sessions(info->sessions);
@@ -284,11 +289,16 @@ void sngtc_dlg_terminated(struct dlg_cell *dlg, int type,
 		shm_free(info->modified_caller_sdp.s);
 
 	shm_free(info);
+
+	if (dlg_binds.store_dlg_value(dlg, &dlg_key_sngtc_info, NULL,
+		DLG_VAL_TYPE_NONE) < 0)
+		LM_ERR("failed to clear dlg val with caller sdp\n");
 }
 
 static int mod_init(void)
 {
-	int i, sip_workers_no;
+	int i;
+	int max_processes = count_child_processes();
 
 	LM_INFO("initializing module\n");
 
@@ -304,36 +314,15 @@ static int mod_init(void)
 		return -1;
 	}
 
-	sip_workers_no = udp_count_processes() + tcp_count_processes();
+	LM_DBG("Children: %d\n", max_processes);
 
-	LM_DBG("Children: %d\n", sip_workers_no);
-
-    sip_workers_pipes = pkg_malloc(2 * sip_workers_no *
+    sip_workers_pipes = pkg_malloc(2 * max_processes *
 	                                sizeof(*sip_workers_pipes));
     if (!sip_workers_pipes) {
         LM_ERR("Not enough pkg mem\n");
         return -1;
     }
 
-	index_lock = shm_malloc(sizeof(*index_lock));
-	if (!index_lock) {
-		LM_ERR("No more shm mem\n");
-		return -1;
-	}
-
-	if (!lock_init(index_lock)) {
-		LM_ERR("Failed to init lock\n");
-		return -1;
-	}
-
-	proc_counter = shm_malloc(sizeof(*proc_counter));
-	if (!proc_counter) {
-		LM_ERR("Not enough shm mem\n");
-		return -1;
-	}
-
-	*proc_counter = 0;
-
 	if (pipe(sangoma_pipe) != 0) {
 		LM_ERR("Failed to create sangoma worker pipe\n");
 		return -1;
@@ -341,7 +330,7 @@ static int mod_init(void)
 
 	LM_DBG("Sangoma pipe: [%d %d]\n", sangoma_pipe[0], sangoma_pipe[1]);
 
-	for (i = 0; i < sip_workers_no; i++) {
+	for (i = 0; i < max_processes; i++) {
 		if (pipe(sip_workers_pipes + 2 * i) != 0) {
 			LM_ERR("Failed to create pipe for UDP receiver %d\n", i);
 			return -1;
@@ -379,17 +368,13 @@ static int child_init(int rank)
 {
 	LM_DBG("init child\n");
 
-	if (rank <= PROC_MAIN)
+	if (rank < 1 )
 		return 0;
 
-	lock_get(index_lock);
-
-	pipe_index = 2 * (*proc_counter)++;
+	pipe_index = 2 * (process_no);
 
 	close(sip_workers_pipes[pipe_index + WRITE_END]);
 
-	lock_release(index_lock);
-
 	LM_DBG("proc index: %d\n", pipe_index / 2);
 
 	return 0;
@@ -403,40 +388,32 @@ static void mod_destroy(void)
 static int sng_logger(int level, char *fmt, ...)
 {
 	va_list args;
-	static char buffer[256];
+	char buffer[256];
 
 	va_start(args, fmt);
 
 	vsnprintf(buffer, 256, fmt, args);
 
 	switch (level) {
-	case SNGTC_LOGLEVEL_DEBUG:
-		LM_GEN1(L_DBG, fmt, args);
 		LM_DBG("%s\n", buffer);
 		break;
-	case SNGTC_LOGLEVEL_WARN:
-		LM_GEN1(L_WARN, fmt, args);
-		LM_WARN("%s\n", buffer);
-		break;
+
 	case SNGTC_LOGLEVEL_INFO:
-		LM_GEN1(L_INFO, fmt, args);
-		LM_INFO("%s\n", buffer);
-		break;
 	case SNGTC_LOGLEVEL_STATS:
-		LM_GEN1(L_INFO, fmt, args);
 		LM_INFO("%s\n", buffer);
 		break;
-	case SNGTC_LOGLEVEL_ERROR:
-		LM_GEN1(L_ERR, fmt, args);
-		LM_ERR("%s\n", buffer);
-		break;
-	case SNGTC_LOGLEVEL_CRIT:
-		LM_GEN1(L_CRIT, fmt, args);
-		LM_CRIT("%s\n", buffer);
-		break;
 
+	case SNGTC_LOGLEVEL_WARN:
+		LM_WARN("%s\n", buffer);
+		break;
+
+	case SNGTC_LOGLEVEL_ERROR:
+		LM_ERR("%s\n", buffer);
+		break;
+
+	case SNGTC_LOGLEVEL_CRIT:
 	default:
-		LM_GEN1(L_WARN, fmt, args);
+		LM_CRIT("%s\n", buffer);
 	}
 
 	va_end(args);
@@ -451,7 +428,7 @@ static int sng_logger(int level, char *fmt, ...)
 int store_sngtc_info(struct dlg_cell *dlg, str *body)
 {
 	struct sngtc_info *info;
-	str st;
+	int_str st;
 
 	/* duplicate the body in shm and store the pointer in the dialog */
 	info = shm_malloc(sizeof(*info));
@@ -463,19 +440,20 @@ int store_sngtc_info(struct dlg_cell *dlg, str *body)
 	memset(info, 0, sizeof(*info));
 
 	lock_init(&info->lock);
-	info->caller_sdp.s = shm_malloc(body->len + 2);
+	info->caller_sdp.s = shm_malloc(body->len);
 	if (!info->caller_sdp.s) {
 		LM_ERR("no more shm\n");
 		goto exit;
 	}
 
-	info->caller_sdp.len = body->len + 2; /* SDP parser needs starting CRLF */
-	memcpy(info->caller_sdp.s, body->s - 2, info->caller_sdp.len);
+	info->caller_sdp.len = body->len; /* SDP parser needs starting CRLF */
+	memcpy(info->caller_sdp.s, body->s, info->caller_sdp.len);
 
-	st.s   = (void *)&info;
-	st.len = sizeof(void *);
-	LM_DBG("storing info ptr: %p\n", (void *) st.s);
-	if (dlg_binds.store_dlg_value(dlg, &dlg_key_sngtc_info, &st) != 0) {
+	st.s.s   = (void *)&info;
+	st.s.len = sizeof(void *);
+	LM_DBG("storing info ptr: %p\n", (void *) st.s.s);
+	if (dlg_binds.store_dlg_value(dlg, &dlg_key_sngtc_info, &st,
+		DLG_VAL_TYPE_STR) != 0) {
 		LM_ERR("failed to store msg body in dialog\n");
 		goto exit;
 	}
@@ -507,7 +485,9 @@ static int sngtc_offer(struct sip_msg *msg)
 	struct lump *lump;
 	struct dlg_cell *dlg;
 	struct sngtc_info *info = NULL;
-	str body, totag, st;
+	str body, totag;
+	int_str st;
+	int val_type;
 
 	if (dlg_binds.create_dlg(msg, 0) < 0) {
 		LM_ERR("failed to create dialog\n");
@@ -528,7 +508,8 @@ static int sngtc_offer(struct sip_msg *msg)
 	totag = get_to(msg)->tag_value;
 
 	/* INVITE retransmissions will skip this part */
-	if (dlg_binds.fetch_dlg_value(dlg, &dlg_key_sngtc_info, &st, 0) != 0) {
+	if (dlg_binds.fetch_dlg_value(dlg, &dlg_key_sngtc_info, &val_type,
+		&st, 0) != 0) {
 
 		if (store_sngtc_info(dlg, &body) != 0) {
 			LM_ERR("failed to create sngtc info struct\n");
@@ -537,7 +518,7 @@ static int sngtc_offer(struct sip_msg *msg)
 
 		/* register a callback to free the above */
 		if (dlg_binds.register_dlgcb(dlg,
-			                         DLGCB_EXPIRED|DLGCB_FAILED|DLGCB_TERMINATED,
+			DLGCB_EXPIRED|DLGCB_FAILED|DLGCB_TERMINATED|DLGCB_DESTROY,
 		    sngtc_dlg_terminated, NULL, NULL) != 0) {
 
 			LM_ERR("failed to register dialog callback\n");
@@ -546,7 +527,7 @@ static int sngtc_offer(struct sip_msg *msg)
 
 	/* for re-INVITES, just recreate the struct sngtc_info */
 	} else if (totag.s && totag.len != 0) {
-		info = *(struct sngtc_info **)(st.s);
+		info = *(struct sngtc_info **)(st.s.s);
 
 		free_transcoding_sessions(info->sessions);
 		if (info->caller_sdp.s)
@@ -1230,27 +1211,21 @@ out:
 }
 
 static int w_sngtc_callee_answer(struct sip_msg *msg,
-                                 char *gp_ip_a, char *gp_ip_b)
+                                 str *gp_ip_a, str *gp_ip_b)
 {
 	if (!gp_ip_a) {
 		card_ip_a.s = card_ip_b.s = NULL;
 		goto out;
 	}
 
-	if (fixup_get_svalue(msg, (gparam_p)gp_ip_a, &card_ip_a) != 0) {
-		LM_ERR("failed to get fixup value for caller: bad pvar\n");
-		return SNGTC_ERR;
-	}
+	card_ip_a = *gp_ip_a;
 
 	if (!gp_ip_b) {
 		card_ip_b.s = NULL;
 		goto out;
 	}
 
-	if (fixup_get_svalue(msg, (gparam_p)gp_ip_b, &card_ip_b) != 0) {
-		LM_ERR("failed to get fixup value for callee: bad pvar\n");
-		return SNGTC_ERR;
-	}
+	card_ip_b = *gp_ip_b;
 
 out:
 	return sngtc_callee_answer(msg);
@@ -1270,6 +1245,8 @@ static int sngtc_callee_answer(struct sip_msg *msg)
 	sdp_info_t sdp;
 	str *sdp_ptr;
 	int rc;
+	int_str isval;
+	int val_type;
 
 	LM_DBG("sngtc_callee_answer\n");
 
@@ -1280,12 +1257,13 @@ static int sngtc_callee_answer(struct sip_msg *msg)
 	}
 
 	/* get the pointer to the SDP body of the caller */
-	if (dlg_binds.fetch_dlg_value(dlg, &dlg_key_sngtc_info, &dst, 0) != 0) {
+	if (dlg_binds.fetch_dlg_value(dlg, &dlg_key_sngtc_info, &val_type,
+		&isval, 0) != 0) {
 		LM_ERR("failed to fetch caller sdp\n");
 		return SNGTC_ERR;
 	}
 
-	info = *(struct sngtc_info **)(dst.s);
+	info = *(struct sngtc_info **)(isval.s);
 	sdp_ptr = &info->caller_sdp;
 
 	LM_DBG("ptrs: %p %p\n", sdp_ptr, info->caller_sdp.s);
@@ -1305,7 +1283,7 @@ static int sngtc_callee_answer(struct sip_msg *msg)
 		goto out_free;
 	}
 
-	if (parse_sdp(msg) != 0) {
+	if (!parse_sdp(msg)) {
 		LM_ERR("failed to parse callee sdp body\n");
 		rc = SNGTC_SDP_ERR;
 		goto out_free;
@@ -1316,7 +1294,7 @@ static int sngtc_callee_answer(struct sip_msg *msg)
 
 	/* perform all 200 OK SDP changes and pre-compute the ACK SDP body */
 	rc = process_session(msg, info, &caller_sdp, &dst, sdp.sessions,
-	                     msg->sdp->sessions);
+	                     get_sdp(msg)->sessions);
 	if (rc != 0) {
 		LM_ERR("failed to rewrite SDP bodies of the endpoints\n");
 		goto out_free;
@@ -1341,7 +1319,7 @@ static int sngtc_callee_answer(struct sip_msg *msg)
 	lock_release(&info->lock);
 
 	if (sdp.sessions)
-		__free_sdp(&sdp);
+		free_sdp_content(&sdp);
 
 	return 1;
 
@@ -1350,7 +1328,7 @@ out_free:
 	lock_release(&info->lock);
 
 	if (sdp.sessions)
-		__free_sdp(&sdp);
+		free_sdp_content(&sdp);
 
 	return rc;
 }
@@ -1366,6 +1344,8 @@ static int sngtc_caller_answer(struct sip_msg *msg)
 	struct lump *lump;
 	struct sngtc_info *info;
 	int len;
+	int_str isval;
+	int val_type;
 
 	LM_DBG("processing ACK\n");
 
@@ -1381,12 +1361,13 @@ static int sngtc_caller_answer(struct sip_msg *msg)
 	}
 
 	/* get the SDP body from the INVITE which was mangled at 200 OK */
-	if (dlg_binds.fetch_dlg_value(dlg, &dlg_key_sngtc_info, &body, 0) != 0) {
+	if (dlg_binds.fetch_dlg_value(dlg, &dlg_key_sngtc_info, &val_type,
+		&isval, 0) != 0) {
 		LM_ERR("failed to fetch caller sdp\n");
 		return SNGTC_ERR;
 	}
 
-	info = *(struct sngtc_info **)(body.s);
+	info = *(struct sngtc_info **)(isval.s.s);
 
 	/* duplicate the SDP in pkg mem for the lumps mechanism */
 	if (pkg_str_dup(&body, &info->modified_caller_sdp) != 0) {
