@@ -138,11 +138,11 @@ int w_calculate_hooks(dlg_t* _d)
 /*
  * Create a new dialog - internal function
  */
-static int _internal_new_dlg_uac(str* _cid, str* _ltag, unsigned int _lseq, str* _luri, str* _ruri, struct socket_info* sock, dlg_t** _d)
+static int _internal_new_dlg_uac(str* _cid, str* _ltag, unsigned int _lseq, str* _luri, str* _turi, str* _ruri, struct socket_info* sock, dlg_t** _d)
 {
 	dlg_t* res;
 
-	if (!_cid || !_ltag || !_luri || !_ruri || !_d) {
+	if (!_cid || !_ltag || !_luri || !_turi || !_d) {
 		LM_ERR("Invalid parameter value\n");
 		return -1;
 	}
@@ -163,7 +163,9 @@ static int _internal_new_dlg_uac(str* _cid, str* _ltag, unsigned int _lseq, str*
 	     /* Make a copy of local URI (usually From) */
 	if (shm_str_dup(&res->loc_uri, _luri) < 0) return -5;
 	     /* Make a copy of remote URI (usually To) */
-	if (shm_str_dup(&res->rem_uri, _ruri) < 0) return -6;
+	if (shm_str_dup(&res->rem_uri, _turi) < 0) return -6;
+	     /* Make a copy of remote target (usually R-URI) */
+	if (_ruri && shm_str_dup(&res->rem_target, _ruri) < 0) return -7;
 	     /* Make a copy of local sequence (usually CSeq) */
 	res->loc_seq.value = _lseq;
 	     /* And mark it as set */
@@ -187,24 +189,28 @@ static int _internal_new_dlg_uac(str* _cid, str* _ltag, unsigned int _lseq, str*
 /*
  * Create a new dialog
  */
-int new_dlg_uac(str* _cid, str* _ltag, unsigned int _lseq, str* _luri, str* _ruri, dlg_t** _d)
+int new_dlg_uac(str* _cid, str* _ltag, unsigned int _lseq, str* _luri, str* _turi, str* _ruri, dlg_t** _d)
 {
-	return _internal_new_dlg_uac(_cid,_ltag,_lseq,_luri,_ruri,NULL,_d);
+	return _internal_new_dlg_uac(_cid,_ltag,_lseq,_luri,_turi,_ruri,NULL,_d);
 }
 
 
 /*
  * Create a new dialog (auto mode)
  */
-int new_auto_dlg_uac( str* _luri, str* _ruri, struct socket_info* _sock, dlg_t** _d)
+int new_auto_dlg_uac( str* _luri, str* _turi, str* _ruri, str *callid, struct socket_info* _sock, dlg_t** _d)
 {
-	str callid, fromtag;
+	str fromtag, clid;
 
-	generate_callid(&callid);
-	generate_fromtag(&fromtag, &callid);
+	if (!callid) {
+		generate_callid(&clid);
+		callid = &clid;
+	}
 
-	return _internal_new_dlg_uac(&callid, &fromtag, 13/*cseq*/,_luri,
-		_ruri,_sock,_d);
+	generate_fromtag(&fromtag, callid);
+
+	return _internal_new_dlg_uac(callid, &fromtag, 13/*cseq*/,_luri,
+		_turi, _ruri,_sock,_d);
 }
 
 
@@ -281,45 +287,6 @@ static inline int get_to_tag(struct sip_msg* _m, str* _tag)
 }
 
 
-/*
- * Extract tag from From header field of a request
- */
-static inline int get_from_tag(struct sip_msg* _m, str* _tag)
-{
-	if (parse_from_header(_m)<0) {
-		LM_ERR("failed to parse From header\n");
-		return -1;
-	}
-
-	if (get_from(_m)->tag_value.len) {
-		_tag->s = get_from(_m)->tag_value.s;
-		_tag->len = get_from(_m)->tag_value.len;
-	} else {
-		_tag->len = 0;
-	}
-
-	return 0;
-}
-
-
-/*
- * Extract Call-ID value
- * Doesn't parse headers !!
- */
-static inline int get_callid(struct sip_msg* _m, str* _cid)
-{
-	if (_m->callid == 0) {
-		LM_ERR("Call-ID not found\n");
-		return -1;
-	}
-
-	_cid->s = _m->callid->body.s;
-	_cid->len = _m->callid->body.len;
-	trim(_cid);
-	return 0;
-}
-
-
 /*
  * Create a copy of route set either in normal or reverse order
  */
@@ -594,250 +561,6 @@ int dlg_response_uac(dlg_t* _d, struct sip_msg* _m)
 }
 
 
-/*
- * Get CSeq number
- * Does not parse headers !!
- */
-static inline int get_cseq_value(struct sip_msg* _m, unsigned int* _cs)
-{
-	str num;
-
-	if (_m->cseq == 0) {
-		LM_ERR("CSeq header not found\n");
-		return -1;
-	}
-
-	num.s = get_cseq(_m)->number.s;
-	num.len = get_cseq(_m)->number.len;
-
-	trim_leading(&num);
-	if (str2int(&num, _cs) < 0) {
-		LM_ERR("converting cseq number failed\n");
-		return -2;
-	}
-	return 0;
-}
-
-
-/*
- * Copy To or From URI without tag parameter
- */
-static inline int get_dlg_uri(struct hdr_field* _h, str* _s)
-{
-	struct to_param* ptr, *prev;
-	struct to_body* body;
-	char* tag = 0; /* Makes gcc happy */
-	int tag_len = 0, len;
-
-	if (!_h) {
-		LM_ERR("header field not found\n");
-		return -1;
-	}
-
-	     /* From was already parsed when extracting tag
-	      * and To is parsed by default
-	      */
-
-	body = (struct to_body*)_h->parsed;
-
-	ptr = body->param_lst;
-	prev = 0;
-	while(ptr) {
-		if (ptr->type == TAG_PARAM) break;
-		prev = ptr;
-		ptr = ptr->next;
-	}
-
-	if (ptr) {
-		     /* Tag param found */
-		if (prev) {
-			tag = prev->value.s + prev->value.len;
-		} else {
-			tag = body->body.s + body->body.len;
-		}
-
-		if (ptr->next) {
-			tag_len = ptr->value.s + ptr->value.len - tag;
-		} else {
-			tag_len = _h->body.s + _h->body.len - tag;
-		}
-	}
-
-	_s->s = shm_malloc(_h->body.len - tag_len);
-	if (!_s->s) {
-		LM_ERR("No share memory left\n");
-		return -1;
-	}
-
-	if (tag_len) {
-		len = tag - _h->body.s;
-		memcpy(_s->s, _h->body.s, len);
-		memcpy(_s->s + len, tag + tag_len, _h->body.len - len - tag_len);
-		_s->len = _h->body.len - tag_len;
-	} else {
-		memcpy(_s->s, _h->body.s, _h->body.len);
-		_s->len = _h->body.len;
-	}
-
-	return 0;
-}
-
-
-/*
- * Extract all information from a request
- * and update a dialog structure
- */
-static inline int request2dlg(struct sip_msg* _m, dlg_t* _d)
-{
-	str contact, rtag, callid;
-
-	if (parse_headers(_m, HDR_EOH_F, 0) == -1) {
-		LM_ERR("failed to parse headers\n");
-		return -1;
-	}
-
-	if (get_contact_uri(_m, &contact) < 0) return -2;
-	if (contact.len && shm_str_dup(&_d->rem_target, &contact) < 0) return -3;
-
-	if (get_from_tag(_m, &rtag) < 0) goto err1;
-	if (rtag.len && shm_str_dup(&_d->id.rem_tag, &rtag) < 0) goto err1;
-
-	if (get_callid(_m, &callid) < 0) goto err2;
-	if (callid.len && shm_str_dup(&_d->id.call_id, &callid) < 0) goto err2;
-
-	if (get_cseq_value(_m, &_d->rem_seq.value) < 0) goto err3;
-	_d->rem_seq.is_set = 1;
-
-	if (get_dlg_uri(_m->from, &_d->rem_uri) < 0) goto err3;
-	if (get_dlg_uri(_m->to, &_d->loc_uri) < 0) goto err4;
-
-	if (get_route_set(_m, &_d->route_set, NORMAL_ORDER) < 0) goto err5;
-
-	return 0;
- err5:
-	if (_d->loc_uri.s) shm_free(_d->loc_uri.s);
-	_d->loc_uri.s = 0;
-	_d->loc_uri.len = 0;
- err4:
-	if (_d->rem_uri.s) shm_free(_d->rem_uri.s);
-	_d->rem_uri.s = 0;
-	_d->rem_uri.len = 0;
- err3:
-	if (_d->id.call_id.s) shm_free(_d->id.call_id.s);
-	_d->id.call_id.s = 0;
-	_d->id.call_id.len = 0;
- err2:
-	if (_d->id.rem_tag.s) shm_free(_d->id.rem_tag.s);
-	_d->id.rem_tag.s = 0;
-	_d->id.rem_tag.len = 0;
- err1:
-	if (_d->rem_target.s) shm_free(_d->rem_target.s);
-	_d->rem_target.s = 0;
-	_d->rem_target.len = 0;
-	return -4;
-}
-
-
-/*
- * Establishing a new dialog, UAS side
- */
-int new_dlg_uas(struct sip_msg* _req, int _code, /*str* _tag,*/ dlg_t** _d)
-{
-	dlg_t* res;
-	str tag;
-
-	if (!_req || /*!_tag ||*/ !_d) {
-		LM_ERR("Invalid parameter value\n");
-		return -1;
-	}
-
-	if ((_code < 200) || (_code > 299)) {
-		LM_DBG("not a 2xx, no dialog created\n");
-		return -2;
-	}
-
-	res = (dlg_t*)shm_malloc(sizeof(dlg_t));
-	if (res == 0) {
-		LM_ERR("no more share memory\n");
-		return -3;
-	}
-	     /* Clear everything */
-	memset(res, 0, sizeof(dlg_t));
-
-	if (request2dlg(_req, res) < 0) {
-		LM_ERR("converting request to dialog failed\n");
-		return -4;
-	}
-
-	tag.s = tm_tags;
-	tag.len = TOTAG_VALUE_LEN;
-	calc_crc_suffix(_req, tm_tag_suffix);
-	if (shm_str_dup(&res->id.loc_tag, &tag) < 0) {
-		free_dlg(res);
-		return -5;
-	}
-
-	*_d = res;
-
-	(*_d)->state = DLG_CONFIRMED;
-	if (calculate_hooks(*_d) < 0) {
-		LM_ERR("calculating hooks failed\n");
-		shm_free(*_d);
-		return -6;
-	}
-
-	return 0;
-}
-
-
-/*
- * UAS side - update a dialog from a request
- */
-int dlg_request_uas(dlg_t* _d, struct sip_msg* _m)
-{
-	str contact;
-	unsigned int cseq;
-
-	if (!_d || !_m) {
-		LM_ERR("Invalid parameter value\n");
-		return -1;
-	}
-
-	     /* We must check if the request is not out of order or retransmission
-	      * first, if so then we will not update anything
-	      */
-	if (parse_headers(_m, HDR_CSEQ_F, 0) == -1) {
-		LM_ERR("parsing headers failed\n");
-		return -2;
-	}
-	if (get_cseq_value(_m, &cseq) < 0) return -3;
-	if (_d->rem_seq.is_set && (cseq <= _d->rem_seq.value)) return 0;
-
-	     /* Neither out of order nor retransmission -> update */
-	_d->rem_seq.value = cseq;
-	_d->rem_seq.is_set = 1;
-
-	     /* We will als update remote target URI if the message
-	      * is target refresher
-	      */
-	if (_m->first_line.u.request.method_value == METHOD_INVITE) {
-		     /* target refresher */
-		if (parse_headers(_m, HDR_CONTACT_F, 0) == -1) {
-			LM_ERR("parsing headers failed\n");
-			return -4;
-		}
-
-		if (get_contact_uri(_m, &contact) < 0) return -5;
-		if (contact.len) {
-			if (_d->rem_target.s) shm_free(_d->rem_target.s);
-			if (shm_str_dup(&_d->rem_target, &contact) < 0) return -6;
-		}
-	}
-
-	return 0;
-}
-
-
 /*
  * Calculate length of the route set
  */
@@ -966,10 +689,10 @@ void print_dlg(FILE* out, dlg_t* _d)
 			_d->rem_target.len,_d->rem_target.s);
 	fprintf(out, "state         : ");
 	switch(_d->state) {
-	case DLG_NEW:       fprintf(out, "DLG_NEW\n");       break;
-	case DLG_EARLY:     fprintf(out, "DLG_EARLY\n");     break;
-	case DLG_CONFIRMED: fprintf(out, "DLG_CONFIRMED\n"); break;
-	case DLG_DESTROYED: fprintf(out, "DLG_DESTROYED\n"); break;
+	CASE_FPRINTENUM(out, DLG_NEW);
+	CASE_FPRINTENUM(out, DLG_EARLY);
+	CASE_FPRINTENUM(out, DLG_CONFIRMED);
+	CASE_FPRINTENUM(out, DLG_DESTROYED);
 	}
 	print_rr(out, _d->route_set);
 	if (_d->hooks.request_uri)
