@@ -66,7 +66,7 @@
 
 #define lump_clone( _new,_old,_ptr) \
 	{\
-		(_new) = (struct lump*)(_ptr);\
+		(_new) = (struct lump*)(void *)(_ptr);\
 		memcpy( (_new), (_old), sizeof(struct lump) );\
 		(_new)->flags|=LUMPFLAG_SHMEM; \
 		(_ptr)+=ROUND4(sizeof(struct lump));\
@@ -92,7 +92,7 @@ inline static struct via_body* via_body_cloner( char* new_buf,
 	do
 	{
 		/* clones the via_body structure */
-		new_via = (struct via_body*)(*p);
+		new_via = (struct via_body*)(void *)(*p);
 		memcpy( new_via , org_via , sizeof( struct via_body) );
 		(*p) += ROUND4(sizeof( struct via_body ));
 
@@ -125,7 +125,7 @@ inline static struct via_body* via_body_cloner( char* new_buf,
 			struct via_param *vp, *new_vp, *last_new_vp;
 			for( vp=org_via->param_lst, last_new_vp=0 ; vp ; vp=vp->next )
 			{
-				new_vp = (struct via_param*)(*p);
+				new_vp = (struct via_param*)(void *)(*p);
 				memcpy( new_vp , vp , sizeof(struct via_param));
 				(*p) += ROUND4(sizeof(struct via_param));
 				new_vp->name.s=translate_pointer(new_buf,org_buf,vp->name.s);
@@ -217,7 +217,7 @@ static inline struct auth_body* auth_body_cloner(char* new_buf, char *org_buf, s
 {
 	struct auth_body* new_auth;
 
-	new_auth = (struct auth_body*)(*p);
+	new_auth = (struct auth_body*)(void *)(*p);
 	memcpy(new_auth , auth , sizeof(struct auth_body));
 	(*p) += ROUND4(sizeof(struct auth_body));
 
@@ -378,7 +378,7 @@ do { \
 		struct lump_rpl   *_rpl_lump, **_rpl_lump_anchor; \
 		_rpl_lump_anchor = (_anchor); \
 		for(_rpl_lump=(_list);_rpl_lump;_rpl_lump=_rpl_lump->next) { \
-			*(_rpl_lump_anchor)=(struct lump_rpl*)(_p); \
+			*(_rpl_lump_anchor)=(struct lump_rpl*)(void *)(_p); \
 			(_p) += ROUND4(sizeof( struct lump_rpl )); \
 			(*_rpl_lump_anchor)->flags = LUMP_RPL_SHMEM | \
 				(_rpl_lump->flags&(~(LUMP_RPL_NODUP|LUMP_RPL_NOFREE))); \
@@ -487,18 +487,44 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 			case HDR_PPI_T:
 			case HDR_PAI_T:
 			case HDR_PRIVACY_T:
-			case HDR_RETRY_AFTER_T:
 			case HDR_CALL_INFO_T:
 			case HDR_WWW_AUTHENTICATE_T:
 			case HDR_PROXY_AUTHENTICATE_T:
+			case HDR_MIN_EXPIRES_T:
+			case HDR_FEATURE_CAPS_T:
+			case HDR_REPLACES_T:
 				/* we ignore them for now even if they have something parsed*/
 				break;
 
+			/* not used in sip_msg, used in MSRP */
+			case HDR_TO_PATH_T:
+			case HDR_FROM_PATH_T:
+			case HDR_MESSAGE_ID_T:
+			case HDR_BYTE_RANGE_T:
+			case HDR_FAILURE_REPORT_T:
+			case HDR_SUCCESS_REPORT_T:
+			case HDR_STATUS_T:
+			case HDR_USE_PATH_T:
+
+			/* not having shortcut */
+			case HDR_RETRY_AFTER_T:
+			case HDR_VIA2_T:
+
+			/* not actual hdrs */
+			case HDR_OTHER_T:
+			case HDR_ERROR_T:
+			case HDR_EOH_T:
+				break;
+
+			/* we do not have a "default" on purpose, so we get
+			 * a compile err/war where a new HDR is added and we do 
+			 * not handle it here.
 			default:
 				if (hdr->parsed) {
 					LM_WARN("header body ignored: %d\n", hdr->type );
 				}
 				break;
+			*/
 		}/*switch*/
 	}/*for all headers*/
 
@@ -540,14 +566,13 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 		*sip_msg_len = len;
 
 	/* filling up the new structure */
-	new_msg = (struct sip_msg*)p;
+	new_msg = (struct sip_msg*)(void *)p;
 	/* sip msg structure */
 	memcpy( new_msg , org_msg , sizeof(struct sip_msg));
 
 	/* avoid copying pointer to un-clonned structures */
-	new_msg->sdp = 0;
-	new_msg->multi = 0;
-	new_msg->msg_cb = 0;
+	new_msg->body = NULL;
+	new_msg->msg_cb = NULL;
 
 	new_msg->msg_flags |= FL_SHM_CLONE;
 	p += ROUND4(sizeof(struct sip_msg));
@@ -576,16 +601,11 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 			org_msg->first_line.u.request.version.s );
 		if(new_msg->parsed_orig_ruri_ok)
 			uri_trans(new_msg->buf, org_msg->buf, &new_msg->parsed_orig_ruri);
-		if(new_msg->parsed_uri_ok) {
-			if (new_msg->new_uri.s) {
-				/* uri string in a separate buffer */
-				uri_trans(new_msg->new_uri.s, org_msg->new_uri.s,
-					&new_msg->parsed_uri);
-			} else {
-				/* uri string still in buf */
-				uri_trans(new_msg->buf, org_msg->buf, &new_msg->parsed_uri);
-			}
-		}
+		if(new_msg->parsed_uri_ok && new_msg->new_uri.s==NULL) {
+			/* uri string still in buf */
+			uri_trans(new_msg->buf, org_msg->buf, &new_msg->parsed_uri);
+		} /* if parsed_uri_ok and RURI is in new_uri, we will translate it
+		   * later when setting the new_uri to the final value */
 	}
 	else if ( org_msg->first_line.type==SIP_REPLY )
 	{
@@ -605,7 +625,7 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 	new_msg->via2=0;
 	for( hdr=org_msg->headers,last_hdr=0 ; hdr ; hdr=hdr->next )
 	{
-		new_hdr = (struct hdr_field*)p;
+		new_hdr = (struct hdr_field*)(void *)p;
 		memcpy(new_hdr, hdr, sizeof(struct hdr_field) );
 		p += ROUND4(sizeof( struct hdr_field));
 		new_hdr->name.s = translate_pointer(new_msg->buf, org_msg->buf,
@@ -702,7 +722,7 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 				for(;to_prm;to_prm=to_prm->next)
 				{
 					/*alloc*/
-					new_to_prm = (struct to_param*)p;
+					new_to_prm = (struct to_param*)(void *)p;
 					p +=ROUND4(sizeof(struct to_param ));
 					/*coping*/
 					memcpy( new_to_prm, to_prm, sizeof(struct to_param ));
@@ -940,6 +960,21 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 					new_msg->proxy_authenticate = new_hdr;
 				}
 				break;
+			case HDR_MIN_EXPIRES_T:
+				if (HOOK_NOT_SET(min_expires)) {
+					new_msg->min_expires = new_hdr;
+				}
+				break;
+			case HDR_FEATURE_CAPS_T:
+				if (HOOK_NOT_SET(feature_caps)) {
+					new_msg->feature_caps = new_hdr;
+				}
+				break;
+			case HDR_REPLACES_T:
+				if (HOOK_NOT_SET(replaces)) {
+					new_msg->replaces = new_hdr;
+				}
+				break;
 			default:
 				/* ignore the rest*/
 				;
@@ -973,6 +1008,10 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 			new_msg->new_uri.s = p;
 			memcpy( p , org_msg->new_uri.s , org_msg->new_uri.len);
 			p += ROUND4(org_msg->new_uri.len);
+			/* if RURI was parsed, translate to new_uri buffer*/
+			if (new_msg->parsed_uri_ok)
+				uri_trans(new_msg->new_uri.s, org_msg->new_uri.s,
+					&new_msg->parsed_uri);
 		}
 		/* dst_uri to zero */
 		new_msg->dst_uri.s = 0;
@@ -1001,28 +1040,30 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 		new_msg->reply_lump = 0;
 		CLONE_RPL_LUMP_LIST( p, &(new_msg->reply_lump), org_msg->reply_lump);
 
+		/* fall through */
+
 	case 1: /* updatable and cloning now */
-		new_msg->msg_flags |= FL_SHM_UPDATABLE;
+		new_msg->msg_flags |= FL_SHM_UPDATABLE|FL_SHM_UPDATED;
 		/* msg is updatable -> the fields that can be updated are allocated in 
 		 * separate memory chunks */
-		tm_shm_lock();
+		shm_lock();
 		if (org_msg->new_uri.len)
-			new_msg->new_uri.s = (char*)tm_shm_malloc_unsafe( org_msg->new_uri.len );
+			new_msg->new_uri.s = (char*)shm_malloc_bulk( org_msg->new_uri.len );
 		if (org_msg->dst_uri.len)
-			new_msg->dst_uri.s = (char*)tm_shm_malloc_unsafe( org_msg->dst_uri.len );
+			new_msg->dst_uri.s = (char*)shm_malloc_bulk( org_msg->dst_uri.len );
 		if (org_msg->path_vec.len)
-			new_msg->path_vec.s = (char*)tm_shm_malloc_unsafe( org_msg->path_vec.len );
+			new_msg->path_vec.s = (char*)shm_malloc_bulk( org_msg->path_vec.len );
 		if (org_msg->set_global_address.len)
-			new_msg->set_global_address.s = (char*)tm_shm_malloc_unsafe( org_msg->set_global_address.len );
+			new_msg->set_global_address.s = (char*)shm_malloc_bulk( org_msg->set_global_address.len );
 		if (org_msg->set_global_port.len)
-			new_msg->set_global_port.s = (char*)tm_shm_malloc_unsafe( org_msg->set_global_port.len );
+			new_msg->set_global_port.s = (char*)shm_malloc_bulk( org_msg->set_global_port.len );
 		if (l1_len)
-			new_msg->add_rm = (struct lump*)tm_shm_malloc_unsafe(l1_len);
+			new_msg->add_rm = (struct lump*)shm_malloc_bulk(l1_len);
 		if (l2_len)
-			new_msg->body_lumps = (struct lump*)tm_shm_malloc_unsafe(l2_len);
+			new_msg->body_lumps = (struct lump*)shm_malloc_bulk(l2_len);
 		if (l3_len)
-			new_msg->reply_lump = (struct lump_rpl*)tm_shm_malloc_unsafe(l3_len);
-		tm_shm_unlock();
+			new_msg->reply_lump = (struct lump_rpl*)shm_malloc_bulk(l3_len);
+		shm_unlock();
 		/*check the malloc result*/
 		if ( (org_msg->new_uri.len && new_msg->new_uri.s==NULL)
 		  || (org_msg->dst_uri.len && new_msg->dst_uri.s==NULL)
@@ -1037,8 +1078,14 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 			return 0;
 		}
 		/* copy data */
-		if (org_msg->new_uri.len)
-			memcpy( new_msg->new_uri.s, org_msg->new_uri.s, org_msg->new_uri.len);
+		if (org_msg->new_uri.len) {
+			memcpy( new_msg->new_uri.s, org_msg->new_uri.s,
+				org_msg->new_uri.len);
+			/* if RURI was parsed, translate to new_uri buffer*/
+			if (new_msg->parsed_uri_ok)
+				uri_trans(new_msg->new_uri.s, org_msg->new_uri.s,
+					&new_msg->parsed_uri);
+		}
 		if (org_msg->dst_uri.len)
 			memcpy( new_msg->dst_uri.s, org_msg->dst_uri.s, org_msg->dst_uri.len);
 		if (org_msg->path_vec.len)
@@ -1054,11 +1101,20 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 		CLONE_LUMP_LIST( p, &(new_msg->body_lumps), org_msg->body_lumps);
 		p = (char*)new_msg->reply_lump;
 		CLONE_RPL_LUMP_LIST( p, &(new_msg->reply_lump), org_msg->reply_lump);
+		/* clone the body parts also */
+		if ( clone_sip_msg_body( org_msg, new_msg, &new_msg->body, 1)!=0 ) {
+			LM_ERR("failed to clone the body parts\n");
+			free_cloned_msg(new_msg);
+			return 0;
+		}
+
 		break;
 
 	case 2: /* updatable, but no cloning now */
 		new_msg->msg_flags |= FL_SHM_UPDATABLE;
 		/* new_uri to zero */
+		if (new_msg->parsed_uri_ok && new_msg->new_uri.s)
+			new_msg->parsed_uri_ok = 0;
 		new_msg->new_uri.s = 0;
 		new_msg->new_uri.len = 0;
 		/* dst_uri to zero */
@@ -1077,6 +1133,8 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 		new_msg->add_rm = 0;
 		new_msg->body_lumps = 0;
 		new_msg->reply_lump = 0;
+		/* set msg body parts */
+		new_msg->body = NULL;
 		break;
 	}
 
@@ -1088,13 +1146,13 @@ struct sip_msg*  sip_msg_cloner( struct sip_msg *org_msg, int *sip_msg_len,
 	do { \
 		if ( _new->_field.len==0) { \
 			if (_old->_field.len!=0) \
-				tm_shm_free_unsafe( _old->_field.s ); \
+				shm_free_bulk( _old->_field.s ); \
 		} else { \
 			if ( _old->_field.len==0 ) { \
-				_old->_field.s = (char*)tm_shm_malloc_unsafe(_new->_field.len);\
+				_old->_field.s = (char*)shm_malloc_bulk(_new->_field.len);\
 			} else if (_old->_field.len<_new->_field.len) { \
-				tm_shm_free_unsafe( _old->_field.s );\
-				_old->_field.s = (char*)tm_shm_malloc_unsafe(_new->_field.len);\
+				shm_free_bulk( _old->_field.s );\
+				_old->_field.s = (char*)shm_malloc_bulk(_new->_field.len);\
 			} \
 			copy_mask |= (1<<_bit);\
 			LM_DBG(#_field" must be copied old=%d, new=%d\n",_old->_field.len,_new->_field.len);\
@@ -1135,6 +1193,7 @@ int update_cloned_msg_from_msg(struct sip_msg *c_msg, struct sip_msg *msg)
 	char *p;
 	struct lump *add_rm_aux=NULL,*body_lumps_aux=NULL;
 	struct lump_rpl *reply_lump_aux=NULL;
+	struct sip_msg_body *body_bk=NULL;
 
 	if ( (c_msg->msg_flags & (FL_SHM_UPDATABLE|FL_SHM_CLONE))==0 ) {
 		LM_CRIT("BUG trying to update a msg not in SHM or not "
@@ -1148,7 +1207,7 @@ int update_cloned_msg_from_msg(struct sip_msg *c_msg, struct sip_msg *msg)
 	LUMP_LIST_LEN(l2_len, msg->body_lumps);
 	RPL_LUMP_LIST_LEN(l3_len, msg->reply_lump);
 
-	tm_shm_lock();
+	shm_lock();
 	/* SIP related strings */
 	REALLOC_CLONED_FIELD_unsafe( new_uri, c_msg, msg, 0);
 	REALLOC_CLONED_FIELD_unsafe( dst_uri, c_msg, msg, 1);
@@ -1168,23 +1227,23 @@ int update_cloned_msg_from_msg(struct sip_msg *c_msg, struct sip_msg *msg)
 	 *
 	 * That is why mem is not leaked after the following allocations:
 	 */
-	
+
 	if (l1_len) { 
 		add_rm_aux = c_msg->add_rm;
-		c_msg->add_rm = tm_shm_malloc_unsafe(l1_len); 
+		c_msg->add_rm = shm_malloc_bulk(l1_len); 
 	}
 	if (l2_len) {
 		body_lumps_aux = c_msg->body_lumps;
-		c_msg->body_lumps = tm_shm_malloc_unsafe(l2_len);
+		c_msg->body_lumps = shm_malloc_bulk(l2_len);
 	}
 
 	if (l3_len) {
 		reply_lump_aux = c_msg->reply_lump;
-		c_msg->reply_lump = tm_shm_malloc_unsafe(l3_len);
+		c_msg->reply_lump = shm_malloc_bulk(l3_len);
 	}
 
 	/* done with mem ops */
-	tm_shm_unlock();
+	shm_unlock();
 
 	/* copy data now */
 	COPY_CLONED_FIELD( new_uri, c_msg, msg, 0);
@@ -1230,19 +1289,32 @@ int update_cloned_msg_from_msg(struct sip_msg *c_msg, struct sip_msg *msg)
 	c_msg->msg_flags = msg->msg_flags|(FL_SHM_UPDATABLE|FL_SHM_CLONE);
 	c_msg->ruri_q = msg->ruri_q;
 	c_msg->ruri_bflags = msg->ruri_bflags;
+	/* reset this just it case, maybe the new_uri was updated */
+	c_msg->parsed_uri_ok = 0;
 
-	if (!(msg->msg_flags & FL_TM_FAKE_REQ)) {
-		/* if not a fake request, we should free old values right now, otherwise we leak
-		if it's a fake request, then we can't free old info now - we might still need
-		it ( eg. to build a reply from the faked req ) - let the freeing happen
-		when destryong the fake req */
-		tm_shm_lock();
-		if (add_rm_aux) tm_shm_free_unsafe(add_rm_aux);
-		if (body_lumps_aux) tm_shm_free_unsafe(body_lumps_aux);
-		if (reply_lump_aux) tm_shm_free_unsafe(reply_lump_aux);
-		tm_shm_unlock();
+	/* body - re-clone it */
+	body_bk = c_msg->body;
+	if ( clone_sip_msg_body( msg, c_msg, &c_msg->body, 1)!=0 ) {
+		LM_ERR("failed to re-clone the body parts, keeping old one\n");
+		/* if err, c_msg->body remains un-touched */
+	} else {
+		free_sip_body( body_bk );
 	}
 
+	if (!(msg->msg_flags & FL_TM_FAKE_REQ)) {
+		/* if not a fake request, we should free old values right now,
+		 * otherwise we leak if it's a fake request, then we can't free 
+		 * old info now - we might still need it ( eg. to build a reply 
+		 * from the faked req ) - let the freeing happen when destryong
+		 * the fake req */
+		shm_lock();
+		if (add_rm_aux) shm_free_bulk(add_rm_aux);
+		if (body_lumps_aux) shm_free_bulk(body_lumps_aux);
+		if (reply_lump_aux) shm_free_bulk(reply_lump_aux);
+		shm_unlock();
+	}
+
+	c_msg->msg_flags |= FL_SHM_UPDATED;
 	return 0;
 }
 
