@@ -50,6 +50,7 @@
 #include "../../usr_avp.h"
 #include "../../mem/mem.h"
 #include "../../parser/parser_f.h"
+#include "../../parser/parse_body.h"
 #include "t_funcs.h"
 #include "t_hooks.h"
 #include "t_msgbuilder.h"
@@ -59,34 +60,45 @@
 #include "t_fwd.h"
 #include "fix_lumps.h"
 #include "config.h"
+#include "cluster.h"
+#include "../usrloc/ul_evi.h"
 #include "../../msg_callbacks.h"
+#include "../../mod_fix.h"
+
+#define NO_BODY_CLONE_MARKER ((struct sip_msg_body*)-1)
 
 /* route to execute for the branches */
-static int goto_on_branch;
+static struct script_route_ref *goto_on_branch;
 int _tm_branch_index = 0;
 
-void t_on_branch( unsigned int go_to )
+void t_on_branch( struct script_route_ref *ref )
 {
 	struct cell *t = get_t();
+	struct script_route_ref **holder;
 
 	/* in MODE_REPLY and MODE_ONFAILURE T will be set to current transaction;
 	 * in MODE_REQUEST T will be set only if the transaction was already
 	 * created; if not -> use the static variable */
-	if (route_type==BRANCH_ROUTE || !t || t==T_UNDEFINED )
-		goto_on_branch=go_to;
-	else
-		t->on_branch = go_to;
+	holder = (!t || t==T_UNDEFINED ) ? &goto_on_branch : &t->on_branch ;
+
+	/* if something already set, free it first */
+	if (*holder)
+		shm_free( *holder );
+
+	*holder = ref ? dup_ref_script_route_in_shm( ref, 0) : NULL;
 }
 
 
-unsigned int get_on_branch(void)
+struct script_route_ref *get_on_branch(void)
 {
-	return goto_on_branch;
+	struct script_route_ref *ref = goto_on_branch;
+	goto_on_branch = NULL;
+	return ref;
 }
 
 
 static inline int pre_print_uac_request( struct cell *t, int branch,
-		struct sip_msg *request)
+					struct sip_msg *request, struct sip_msg_body **body_clone)
 {
 	int backup_route_type;
 	struct usr_avp **backup_list;
@@ -107,9 +119,9 @@ static inline int pre_print_uac_request( struct cell *t, int branch,
 	/* copy path vector into branch */
 	if (request->path_vec.len) {
 		t->uac[branch].path_vec.s =
-			shm_resize(t->uac[branch].path_vec.s, request->path_vec.len+1);
+			shm_realloc(t->uac[branch].path_vec.s, request->path_vec.len+1);
 		if (t->uac[branch].path_vec.s==NULL) {
-			LM_ERR("shm_resize failed\n");
+			LM_ERR("shm_realloc failed\n");
 			goto error;
 		}
 		t->uac[branch].path_vec.len = request->path_vec.len;
@@ -119,10 +131,10 @@ static inline int pre_print_uac_request( struct cell *t, int branch,
 
 	/* do the same for the advertised port & address */
 	if (request->set_global_address.len) {
-		t->uac[branch].adv_address.s = shm_resize(t->uac[branch].adv_address.s,
+		t->uac[branch].adv_address.s = shm_realloc(t->uac[branch].adv_address.s,
 			request->set_global_address.len+1);
 		if (t->uac[branch].adv_address.s==NULL) {
-			LM_ERR("shm_resize failed for storing the advertised address "
+			LM_ERR("shm_realloc failed for storing the advertised address "
 				"(len=%d)\n",request->set_global_address.len);
 			goto error;
 		}
@@ -131,10 +143,10 @@ static inline int pre_print_uac_request( struct cell *t, int branch,
 			request->set_global_address.len+1);
 	}
 	if (request->set_global_port.len) {
-		t->uac[branch].adv_port.s = shm_resize(t->uac[branch].adv_port.s,
+		t->uac[branch].adv_port.s = shm_realloc(t->uac[branch].adv_port.s,
 			request->set_global_port.len+1);
 		if (t->uac[branch].adv_port.s==NULL) {
-			LM_ERR("shm_resize failed for storing the advertised port "
+			LM_ERR("shm_realloc failed for storing the advertised port "
 				"(len=%d)\n",request->set_global_port.len);
 			goto error;
 		}
@@ -148,7 +160,7 @@ static inline int pre_print_uac_request( struct cell *t, int branch,
 
 	/* run branch route, if any; run it before RURI's DNS lookup
 	 * to allow to be changed --bogdan */
-	if (t->on_branch) {
+	if ( ref_script_route_check_and_update(t->on_branch)) {
 		/* need to pkg_malloc the dst_uri */
 		if ( request->dst_uri.s && request->dst_uri.len>0 ) {
 			if ( (p=pkg_malloc(request->dst_uri.len))==0 ) {
@@ -168,13 +180,19 @@ static inline int pre_print_uac_request( struct cell *t, int branch,
 		memcpy( p, request->new_uri.s, request->new_uri.len);
 		request->new_uri.s = p;
 		request->parsed_uri_ok = 0;
+		/* make a clone of the original body, to restore it later */
+		if (clone_sip_msg_body( request, NULL, body_clone, 0)!=0) {
+			LM_ERR("faile to clone the body, branch route changes will be"
+				" preserved\n");
+		}
 		/* make available the avp list from transaction */
 		backup_list = set_avp_list( &t->user_avps );
 		/* run branch route */
 		swap_route_type( backup_route_type, BRANCH_ROUTE);
 
 		_tm_branch_index = branch;
-		if (run_top_route(branch_rlist[t->on_branch].a, request)&ACT_FL_DROP) {
+		if (run_top_route( sroutes->branch[t->on_branch->idx],request)
+		&ACT_FL_DROP) {
 			LM_DBG("dropping branch <%.*s>\n", request->new_uri.len,
 					request->new_uri.s);
 			_tm_branch_index = 0;
@@ -193,15 +211,17 @@ static inline int pre_print_uac_request( struct cell *t, int branch,
 	}
 
 	/* run the specific callbacks for this transaction */
+	_tm_branch_index = branch;
 	run_trans_callbacks( TMCB_REQUEST_FWDED, t, request, 0,
 			-request->REQ_METHOD);
+	_tm_branch_index = 0;
 
 	/* copy dst_uri into branch (after branch route possible updated it) */
 	if (request->dst_uri.len) {
 		t->uac[branch].duri.s =
-			shm_resize(t->uac[branch].duri.s, request->dst_uri.len);
+			shm_realloc(t->uac[branch].duri.s, request->dst_uri.len);
 		if (t->uac[branch].duri.s==NULL) {
-			LM_ERR("shm_resize failed\n");
+			LM_ERR("shm_realloc failed\n");
 			goto error;
 		}
 		t->uac[branch].duri.len = request->dst_uri.len;
@@ -218,10 +238,14 @@ static inline char *print_uac_request(struct sip_msg *i_req, unsigned int *len,
 		struct socket_info *send_sock, enum sip_protos proto )
 {
 	char *buf;
+	str *cid = NULL;
+
+	if (1/* TODO: check if the cid should be used */)
+		cid = tm_via_cid();
 
 	/* build the shm buffer now */
 	buf=build_req_buf_from_sip_req( i_req, len, send_sock, proto,
-			MSG_TRANS_SHM_FLAG);
+			cid, MSG_TRANS_SHM_FLAG);
 	if (!buf) {
 		LM_ERR("no more shm_mem\n");
 		ser_error=E_OUT_OF_MEM;
@@ -233,7 +257,7 @@ static inline char *print_uac_request(struct sip_msg *i_req, unsigned int *len,
 
 
 static inline void post_print_uac_request(struct sip_msg *request,
-		str *org_uri, str *org_dst)
+			const str *org_uri, str *org_dst, struct sip_msg_body *body_clone)
 {
 	reset_init_lump_flags();
 	/* delete inserted branch lumps */
@@ -254,6 +278,11 @@ static inline void post_print_uac_request(struct sip_msg *request,
 		request->dst_uri.s = 0;
 		request->dst_uri.len = 0;
 	}
+	/* if cloned, restore the original body to the msg */
+	if (body_clone!=NO_BODY_CLONE_MARKER) {
+		free_sip_body(request->body);
+		request->body = body_clone;
+	}
 }
 
 
@@ -362,11 +391,15 @@ static inline unsigned int count_local_rr(struct sip_msg *req)
 /* introduce a new uac to transaction; returns its branch id (>=0)
    or error (<0); it doesn't send a message yet -- a reply to it
    might interfere with the processes of adding multiple branches
+
+   NOTICE: do NOT provide (str *) buffers pointing to @request itself, as this
+   may break the function logic!
 */
-static int add_uac( struct cell *t, struct sip_msg *request, str *uri,
+static int add_uac( struct cell *t, struct sip_msg *request, const str *uri,
 		str* next_hop, unsigned int bflags, str* path, struct proxy_l *proxy)
 {
 	unsigned short branch;
+	struct sip_msg_body *body_clone=NO_BODY_CLONE_MARKER;
 	int do_free_proxy;
 	int ret;
 
@@ -391,7 +424,7 @@ static int add_uac( struct cell *t, struct sip_msg *request, str *uri,
 	request->path_vec=*path;
 	request->ruri_bflags=bflags;
 
-	if ( pre_print_uac_request( t, branch, request)!= 0 ) {
+	if ( pre_print_uac_request( t, branch, request, &body_clone)!= 0 ) {
 		ret = -1;
 		goto error01;
 	}
@@ -426,6 +459,7 @@ static int add_uac( struct cell *t, struct sip_msg *request, str *uri,
 		&proxy->host, proxy->addr_idx, proxy->port ? proxy->port:SIP_PORT);
 	t->uac[branch].request.dst.proto = proxy->proto;
 
+	/* do print of the uac request */
 	if ( update_uac_dst( request, &t->uac[branch] )!=0) {
 		ret = ser_error;
 		goto error02;
@@ -448,146 +482,165 @@ error02:
 		pkg_free( proxy );
 	}
 error01:
-	post_print_uac_request( request, uri, next_hop);
+	post_print_uac_request( request, uri, next_hop, body_clone);
 	if (ret < 0) {
 		/* destroy all the bavps added, the path vector and the destination,
 		 * since this branch will never be properly added to
 		 * the UAC list, otherwise we'll have memory leaks - razvanc */
-		if (t->uac[branch].user_avps)
-			destroy_avp_list(&t->uac[branch].user_avps);
-		if (t->uac[branch].path_vec.s)
-			shm_free(t->uac[branch].path_vec.s);
-		if (t->uac[branch].adv_address.s)
-			shm_free(t->uac[branch].adv_address.s);
-		if (t->uac[branch].adv_port.s)
-			shm_free(t->uac[branch].adv_port.s);
-		if (t->uac[branch].duri.s)
-			shm_free(t->uac[branch].duri.s);
-		memset(&t->uac[branch],0,sizeof(t->uac[branch]));
+		clean_branch(t->uac[branch]);
+		memset(&t->uac[branch], 0, sizeof(t->uac[branch]));
+		init_branch(&t->uac[branch], branch, t->wait_tl.set, t);
 	}
 error:
 	return ret;
 }
 
 
-int e2e_cancel_branch( struct sip_msg *cancel_msg, struct cell *t_cancel,
-	struct cell *t_invite, int branch )
+int add_phony_uac( struct cell *t, int br_flags)
 {
-	int ret;
-	char *shbuf;
-	unsigned int len;
-	str bk_dst_uri;
-	str bk_path_vec;
-	str bk_adv_address;
-	str bk_adv_port;
+	str dummy_buffer = str_init("DUMMY");
+	unsigned short branch;
+	utime_t timer;
 
-	if (t_cancel->uac[branch].request.buffer.s) {
+	branch=t->nr_of_outgoings;
+	if (branch==MAX_BRANCHES) {
+		LM_ERR("maximum number of branches exceeded\n");
+		return E_CFG;
+	}
+
+	/* check existing buffer -- rewriting should never occur */
+	if (t->uac[branch].request.buffer.s) {
 		LM_CRIT("buffer rewrite attempt\n");
-		ret=ser_error=E_BUG;
-		goto error;
+		ser_error=E_BUG;
+		return E_BUG;
 	}
 
-	cancel_msg->new_uri = t_invite->uac[branch].uri;
-	cancel_msg->parsed_uri_ok=0;
-	bk_dst_uri = cancel_msg->dst_uri;
-	bk_path_vec = cancel_msg->path_vec;
-	bk_adv_address = cancel_msg->set_global_address;
-	bk_adv_port = cancel_msg->set_global_port;
+	/* we attach a dummy buffer just to pass all the "tests" for a 
+	 * valid branch */
+	t->uac[branch].request.buffer.s = (char*)shm_malloc(dummy_buffer.len);
+	if (t->uac[branch].request.buffer.s==NULL) {
+		LM_ERR("failed to alloc dummy buffer for phony branch\n");
+		/* there is nothing to reset on the branch */
+		return E_OUT_OF_MEM;
+	}
+	memcpy( t->uac[branch].request.buffer.s, dummy_buffer.s, dummy_buffer.len);
+	t->uac[branch].request.buffer.len = dummy_buffer.len;
 
-	/* force same path & advertising as for request */
-	cancel_msg->path_vec = t_invite->uac[branch].path_vec;
-	cancel_msg->set_global_address = t_invite->uac[branch].adv_address;
-	cancel_msg->set_global_port = t_invite->uac[branch].adv_port;
+	t->uac[branch].request.my_T = t;
+	t->uac[branch].request.branch = branch;
+	t->uac[branch].flags = T_UAC_IS_PHONY;
+	t->uac[branch].br_flags = br_flags;
 
-	if ( pre_print_uac_request( t_cancel, branch, cancel_msg)!= 0 ) {
-		ret = -1;
-		goto error01;
+	/* in invalid proto will prevent adding this retransmission buffer
+	 * to the retransmission timer (there is nothing to retransmit here :P */
+	t->uac[branch].request.dst.proto = PROTO_NONE;
+
+	t->nr_of_outgoings++;
+
+	/* we set here only FR (final response) timer, to be sure this branch
+	 * comes to an end - as timeout value we use exactly the same value the
+	 * transaction has set as FR_INV_TIMEOUT */
+	if (is_timeout_set(t->fr_inv_timeout)) {
+		timer = t->fr_inv_timeout;
+		set_1timer(&t->uac[branch].request.fr_timer, FR_INV_TIMER_LIST,&timer);
+	} else {
+		set_1timer(&t->uac[branch].request.fr_timer, FR_INV_TIMER_LIST, NULL);
 	}
 
-	/* force same uri as in INVITE */
-	if (cancel_msg->new_uri.s!=t_invite->uac[branch].uri.s) {
-		pkg_free(cancel_msg->new_uri.s);
-		cancel_msg->new_uri = t_invite->uac[branch].uri;
-		/* and just to be sure */
-		cancel_msg->parsed_uri_ok = 0;
-	}
+	set_kr(REQ_FWDED);
 
-	/* print */
-	shbuf=print_uac_request( cancel_msg, &len,
-		t_invite->uac[branch].request.dst.send_sock,
-		t_invite->uac[branch].request.dst.proto);
-	if (!shbuf) {
-		LM_ERR("printing e2e cancel failed\n");
-		ret=ser_error=E_OUT_OF_MEM;
-		goto error01;
-	}
-
-	/* install buffer */
-	t_cancel->uac[branch].request.dst=t_invite->uac[branch].request.dst;
-	t_cancel->uac[branch].request.buffer.s=shbuf;
-	t_cancel->uac[branch].request.buffer.len=len;
-	t_cancel->uac[branch].uri.s=t_cancel->uac[branch].request.buffer.s+
-		cancel_msg->first_line.u.request.method.len+1;
-	t_cancel->uac[branch].uri.len=t_invite->uac[branch].uri.len;
-	t_cancel->uac[branch].br_flags = cancel_msg->flags;
-
-	/* success */
-	ret=1;
-
-error01:
-	post_print_uac_request( cancel_msg, &t_invite->uac[branch].uri,
-		&bk_dst_uri);
-	cancel_msg->dst_uri = bk_dst_uri;
-	cancel_msg->path_vec = bk_path_vec;
-	cancel_msg->set_global_address = bk_adv_address;
-	cancel_msg->set_global_port = bk_adv_port;
-error:
-	return ret;
+	return 0;
 }
 
 
+static int _reason_avp_id = 0;
 
-void cancel_invite(struct sip_msg *cancel_msg,
-								struct cell *t_cancel, struct cell *t_invite )
+int t_set_reason(struct sip_msg *msg, str *val)
+{
+	str avp_name = str_init("_reason_avp_internal");
+	int_str reason;
+
+	if (_reason_avp_id==0) {
+		if (parse_avp_spec( &avp_name, &_reason_avp_id) ) {
+			LM_ERR("failed to init the internal AVP\n");
+			return -1;
+		}
+	}
+
+	reason.s = *val;
+	if (add_avp( AVP_VAL_STR, _reason_avp_id, reason)!=0) {
+		LM_ERR("failed to add the internal reason AVP\n");
+		return -1;
+	}
+	return 1;
+}
+
+
+int t_add_reason(struct sip_msg *msg, str *reason)
+{
+	return t_set_reason(msg, reason);
+}
+
+void get_cancel_reason(struct sip_msg *msg, int flags, str *reason)
 {
 #define CANCEL_REASON_SIP_487  \
 	"Reason: SIP;cause=487;text=\"ORIGINATOR_CANCEL\"" CRLF
+	int_str avp_reason;
+	struct hdr_field *hdr;
+
+	reason->s = NULL;
+	reason->len = 0;
+
+	if (search_first_avp( AVP_VAL_STR, _reason_avp_id, &avp_reason, NULL)) {
+		*reason = avp_reason.s;
+	} else {
+		/* propagate the REASON flag ? */
+		if ( flags&T_CANCEL_REASON_FLAG ) {
+			/* look for the Reason header */
+			if (parse_headers(msg, HDR_EOH_F, 0)<0) {
+				LM_ERR("failed to parse all hdrs - ignoring Reason hdr\n");
+				} else {
+				hdr = get_header_by_static_name(msg, "Reason");
+				if (hdr!=NULL) {
+					reason->s = hdr->name.s;
+					reason->len = hdr->len;
+				}
+			}
+		}
+	}
+
+	/* if no reason, use NORMAL CLEARING */
+	if (reason->s == NULL) {
+		reason->s = CANCEL_REASON_SIP_487;
+		reason->len = sizeof(CANCEL_REASON_SIP_487) - 1;
+	}
+}
+
+void cancel_invite(struct sip_msg *cancel_msg,
+					struct cell *t_cancel, struct cell *t_invite, int locked)
+{
 
 	branch_bm_t cancel_bitmap;
 	str reason;
-	struct hdr_field *hdr;
 
 	cancel_bitmap=0;
 
 	/* send back 200 OK as per RFC3261 */
 	reason.s = CANCELING;
 	reason.len = sizeof(CANCELING)-1;
-	t_reply( t_cancel, cancel_msg, 200, &reason );
+	if (locked)
+		t_reply_unsafe( t_cancel, cancel_msg, 200, &reason );
+	else
+		t_reply( t_cancel, cancel_msg, 200, &reason );
 
-	reason.s = NULL;
-	reason.len = 0;
-	/* propagate the REASON flag ? */
-	if ( t_cancel->flags&T_CANCEL_REASON_FLAG ) {
-		/* look for the Reason header */
-		if (parse_headers(cancel_msg, HDR_EOH_F, 0)<0) {
-			LM_ERR("failed to parse all hdrs - ignoring Reason hdr\n");
-		} else {
-			hdr = get_header_by_static_name(cancel_msg, "Reason");
-			if (hdr!=NULL) {
-				reason.s = hdr->name.s;
-				reason.len = hdr->len;
-			}
-		}
-	}
-
-	/* if no reason, use NORMAL CLEARING */
-	if (reason.s == NULL) {
-		reason.s = CANCEL_REASON_SIP_487;
-		reason.len = sizeof(CANCEL_REASON_SIP_487) - 1;
-	}
+	get_cancel_reason(cancel_msg, t_cancel->flags, &reason);
 
+	LOCK_REPLIES(t_invite);
+	/* we need to check which branches should be canceled under lock to avoid
+	 * concurrency with replies that are coming in the same time */
 	/* generate local cancels for all branches */
 	which_cancel(t_invite, &cancel_bitmap );
+	UNLOCK_REPLIES(t_invite);
 
 	set_cancel_extra_hdrs( reason.s, reason.len);
 	cancel_uacs(t_invite, cancel_bitmap );
@@ -599,6 +652,18 @@ void cancel_invite(struct sip_msg *cancel_msg,
 	 * timer; this helps with better coping with missed/lated provisional
 	 * replies in the context of cancelling the transaction
 	 */
+
+	 /* Handle a special case here, when there is only one PHONY
+	  * branch (not a real signaling branch, just a fake one
+	  * to force keeping the transaction alive) - if the case,
+	  * it means there are no other real signaling branches, 
+	  * so we can force a 487 reply on the PHONY branch, in order
+	  * to terminate the whole transaction on the spot */
+	if ( (t_invite->nr_of_outgoings-t_invite->first_branch)==1 &&
+	(t_invite->uac[t_invite->first_branch].flags & T_UAC_IS_PHONY) ) {
+		relay_reply( t_invite, FAKED_REPLY, t_invite->first_branch,
+			487, &cancel_bitmap);
+	}
 #if 0
 	/* internally cancel branches with no received reply */
 	for (i=t_invite->first_branch; i<t_invite->nr_of_outgoings; i++) {
@@ -620,7 +685,7 @@ void cancel_invite(struct sip_msg *cancel_msg,
  *      -1 - error during forward
  */
 int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
-	struct proxy_l * proxy)
+						struct proxy_l * proxy, int reset_bcounter, int locked)
 {
 	str reply_reason_487 = str_init("Request Terminated");
 	str backup_uri;
@@ -638,17 +703,14 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 	str path;
 	str bk_path;
 
-	/* make -Wall happy */
-	current_uri.s=0;
-
-	/* before doing enything, update the t falgs from msg */
+	/* before doing enything, update the t flags from msg */
 	t->uas.request->flags = p_msg->flags;
 
 	if (p_msg->REQ_METHOD==METHOD_CANCEL) {
 		t_invite=t_lookupOriginalT(  p_msg );
 		if (t_invite!=T_NULL_CELL) {
 			t_invite->flags |= T_WAS_CANCELLED_FLAG;
-			cancel_invite( p_msg, t, t_invite );
+			cancel_invite( p_msg, t, t_invite, locked );
 			return 1;
 		}
 	}
@@ -665,7 +727,10 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 			/* if no other signalling was performed on the transaction
 			 * and the transaction was already canceled, better
 			 * internally generate the 487 reply here */
-			t_reply( t, p_msg , 487 , &reply_reason_487);
+			if (locked)
+				t_reply_unsafe( t, p_msg , 487 , &reply_reason_487);
+			else
+				t_reply( t, p_msg , 487 , &reply_reason_487);
 		}
 		LM_INFO("discarding fwd for a cancelled transaction\n");
 		ser_error = E_NO_DESTINATION;
@@ -689,10 +754,19 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 	/* branches added */
 	added_branches=0;
 	/* branch to begin with */
-	t->first_branch=t->nr_of_outgoings;
+	if (reset_bcounter) {
+		t->first_branch=t->nr_of_outgoings;
+		/* check if the previous branch is a PHONY one and if yes
+		 * keep it in the set of active branches; that means the 
+		 * transaction had a t_wait_for_new_branches() call prior to relay() */
+		if ( t->first_branch>0 &&
+		(t->uac[t->first_branch-1].flags & T_UAC_IS_PHONY) )
+			t->first_branch--;
+	}
 
-	/* as first branch, use current uri */
-	current_uri = *GET_RURI(p_msg);
+	current_uri = *GET_RURI(p_msg); /* separate storage required! */
+
+	/* as first branch, use current R-URI, bflags, etc. */
 	branch_ret = add_uac( t, p_msg, &current_uri, &backup_dst,
 		getb0flags(p_msg), &p_msg->path_vec, proxy);
 	if (branch_ret>=0)
@@ -725,8 +799,11 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 	p_msg->path_vec = bk_path;
 	p_msg->ruri_bflags = bk_bflags;
 
-	/* update on_branch, if modified */
-	t->on_branch = get_on_branch();
+	/* update on_branch, _only_ if modified, otherwise it overwrites
+	 * whatever it is already in the transaction */
+	if (get_on_branch())
+		t->on_branch = get_on_branch();
+
 	/* update flags, if changed in branch route */
 	t->uas.request->flags = p_msg->flags;
 
@@ -745,6 +822,16 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 			if (t->uac[i].br_flags & tcp_no_new_conn_bflag)
 				tcp_no_new_conn = 1;
 
+			/* successfully sent out -> run callbacks */
+			if ( has_tran_tmcbs( t, TMCB_REQUEST_BUILT) ) {
+				_tm_branch_index = i;
+				set_extra_tmcb_params( &t->uac[i].request.buffer,
+					&t->uac[i].request.dst);
+				run_trans_callbacks( TMCB_REQUEST_BUILT, t,
+					p_msg, 0, 0);
+				_tm_branch_index = 0;
+			}
+
 			do {
 				if (check_blacklists( t->uac[i].request.dst.proto,
 				&t->uac[i].request.dst.to,
@@ -753,12 +840,17 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 					LM_DBG("blocked by blacklists\n");
 					ser_error=E_IP_BLOCKED;
 				} else {
+					set_bavp_list(&t->uac[i].user_avps);
+					set_extra_tmcb_params( &t->uac[i].request.buffer,
+							&t->uac[i].request.dst);
 					run_trans_callbacks(TMCB_PRE_SEND_BUFFER, t, p_msg, 0, i);
 
 					if (SEND_BUFFER( &t->uac[i].request)==0) {
+						reset_bavp_list();
 						ser_error = 0;
 						break;
 					}
+					reset_bavp_list();
 
 					LM_ERR("sending request failed\n");
 					ser_error=E_SEND;
@@ -789,11 +881,11 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 			set_kr(REQ_FWDED);
 
 			/* successfully sent out -> run callbacks */
-			if ( has_tran_tmcbs( t, TMCB_REQUEST_BUILT) ) {
+			if ( has_tran_tmcbs( t, TMCB_MSG_SENT_OUT) ) {
 				set_extra_tmcb_params( &t->uac[i].request.buffer,
 					&t->uac[i].request.dst);
-				run_trans_callbacks( TMCB_REQUEST_BUILT, t, p_msg,0,
-					-p_msg->REQ_METHOD);
+				run_trans_callbacks( TMCB_MSG_SENT_OUT, t,
+					p_msg, 0, 0);
 			}
 
 		}
@@ -803,6 +895,244 @@ int t_forward_nonack( struct cell *t, struct sip_msg* p_msg ,
 }
 
 
+static int ul_contact_event_to_msg(struct sip_msg *req)
+{
+	static enum ul_attrs { UL_URI, UL_RECEIVED, UL_PATH, UL_QVAL,
+		UL_SOCKET, UL_BFLAGS, UL_MAX } ul_attr;
+	static str ul_names[UL_MAX]= {str_init(UL_EV_PARAM_CT_URI),
+	                              str_init(UL_EV_PARAM_CT_RCV),
+	                              str_init(UL_EV_PARAM_CT_PATH),
+	                              str_init(UL_EV_PARAM_CT_QVAL),
+	                              str_init(UL_EV_PARAM_CT_SOCK),
+	                              str_init(UL_EV_PARAM_CT_BFL) };
+	static int avp_ids[UL_MAX] = { -1, -1, -1, -1, -1, -1 };
+	int_str vals[UL_MAX];
+	int proto, port;
+	str host;
+	str path_dst;
+
+	if (avp_ids[0]==-1) {
+		/* init the avp IDs mapping us on the UL event */
+		for( ul_attr=0 ; ul_attr<UL_MAX ; ul_attr++ ){
+			if (parse_avp_spec( &ul_names[ul_attr], &avp_ids[ul_attr])<0) {
+				LM_ERR("failed to init UL AVP %d/%s\n",
+					ul_attr,ul_names[ul_attr].s);
+				avp_ids[0] = -1;
+				return -1;
+			}
+		}
+	}
+
+	/* fetch the AVP values one by one */
+	for( ul_attr=0 ; ul_attr<UL_MAX ; ul_attr++ ) {
+		if (search_first_avp(0, avp_ids[ul_attr], &vals[ul_attr], NULL)==NULL){
+			LM_ERR("cannot find AVP(%d) for event attr %d/%s\n",
+				avp_ids[ul_attr], ul_attr, ul_names[ul_attr].s);
+			return -1;
+		}
+	}
+
+	/* OK, we have the values, lets inject them into the SIP msg */
+	LM_DBG("injecting new branch: uri=<%.*s>, received=<%.*s>,"
+		"path=<%.*s>, qval=%d, socket=<%.*s>, bflags=%X\n",
+		vals[UL_URI].s.len, vals[UL_URI].s.s,
+		vals[UL_RECEIVED].s.len, vals[UL_RECEIVED].s.s,
+		vals[UL_PATH].s.len, vals[UL_PATH].s.s,
+		vals[UL_QVAL].n,
+		vals[UL_SOCKET].s.len, vals[UL_SOCKET].s.s,
+		vals[UL_BFLAGS].n);
+
+	/* contact URI goes as RURI */
+	if (set_ruri( req, &vals[UL_URI].s)<0) {
+		LM_ERR("failed to set new RURI\n");
+		return -1;
+	}
+
+	/* contact PATH goes as path */
+	if (vals[UL_PATH].s.len) {
+		if (get_path_dst_uri(&vals[UL_PATH].s, &path_dst) < 0) {
+			LM_ERR("failed to get dst_uri for Path\n");
+			return -1;
+		}
+		if (set_dst_uri( req, &path_dst) < 0) {
+			LM_ERR("failed to set dst_uri of Path\n");
+			return -1;
+		}
+
+		if (set_path_vector( req, &vals[UL_PATH].s)<0) {
+			LM_ERR("failed to set PATH\n");
+			return -1;
+		}
+	} else
+	/* contact RECEIVED goes as DURI */
+	if (vals[UL_RECEIVED].s.len) {
+		if (set_dst_uri( req, &vals[UL_RECEIVED].s)<0) {
+			LM_ERR("failed to set DST URI\n");
+			return -1;
+		}
+	}
+
+	/* contact Qval goes as RURI Qval */
+	set_ruri_q( req, vals[UL_QVAL].n);
+
+	/* contact BFLAGS goes as RURI bflags */
+	setb0flags( req, vals[UL_BFLAGS].n);
+
+	/* socket info */
+	if (vals[UL_SOCKET].s.len) {
+		if ( parse_phostport( vals[UL_SOCKET].s.s, vals[UL_SOCKET].s.len,
+		&host.s, &host.len, &port, &proto) < 0) {
+			LM_ERR("failed to parse socket from Event attr <%.*s>\n",
+				vals[UL_SOCKET].s.len, vals[UL_SOCKET].s.s);
+		} else {
+			req->force_send_socket = grep_sock_info( &host,
+				(unsigned short)port, (unsigned short)proto);
+		}
+	}
+
+	return 0;
+}
+
+
+static int dst_to_msg(struct sip_msg *s_msg, struct sip_msg *d_msg)
+{
+	/* move RURI */
+	if (set_ruri( d_msg, GET_RURI(s_msg))<0) {
+		LM_ERR("failed to set new RURI\n");
+		return -1;
+	}
+
+	/* move DURI (empty is accepted as reset) */
+	if (set_dst_uri( d_msg, &s_msg->dst_uri)<0) {
+		LM_ERR("failed to set DST URI\n");
+		return -1;
+	}
+
+	/* move PATH  (empty is accepted as reset) */
+	if (set_path_vector( d_msg, & s_msg->path_vec)<0) {
+		LM_ERR("failed to set PATH\n");
+		return -1;
+	}
+
+	/* Qval */
+	set_ruri_q( d_msg, get_ruri_q(s_msg) );
+
+	/* BFLAGS */
+	setb0flags( d_msg, getb0flags(s_msg) );
+
+	/* socket info */
+	d_msg->force_send_socket = s_msg->force_send_socket;
+
+	return 0;
+}
+
+
+int t_wait_no_more_branches( struct cell *t, int extra)
+{
+	int b;
+
+	/* look back (in the set of active branches for a PHONY branch
+	 * that might control the EBR waiting. If found, update it
+	 * (the br_flags field), so that this is the last allowed injected
+	 * branch (the max number of allowed branches is set to the current
+	 * number of branches) */
+	for ( b=t->nr_of_outgoings-1; b>=t->first_branch ; b-- ) {
+		if (t->uac[b].flags & T_UAC_IS_PHONY) {
+			t->uac[b].br_flags=t->nr_of_outgoings+extra;
+			return 0;
+		}
+	}
+
+	return -1;
+}
+
+
+int t_inject_branch( struct cell *t, struct sip_msg *msg, int flags)
+{
+	static struct sip_msg faked_req;
+	branch_bm_t cancel_bm = 0;
+	str reason = str_init(CANCEL_REASON_200);
+	struct cell *bk_t = NULL;
+	int rc;
+
+	/* does the transaction state still accept new branches ? */
+	if (t->uas.status >= 200) {
+		LM_DBG("cannot add branches to a transaction with %d UAS\n",
+			t->uas.status);
+		return -2;
+	}
+
+	if (!fake_req( &faked_req, t->uas.request, &t->uas, NULL)) {
+		LM_ERR("fake_req failed\n");
+		return -1;
+	}
+
+	/* do we have the branches to be injected ? */
+	if (flags&TM_INJECT_SRC_EVENT) {
+		/* get the branch from Event AVPs, as populated by EVI/EBR */
+		if (ul_contact_event_to_msg( &faked_req )<0) {
+			LM_ERR("failed to grab new branch from Event\n");
+			goto error;
+		}
+	} else {
+		/* use the RURI+dset array as destinations to be injected */
+		if (msg->first_line.type==SIP_REQUEST) {
+			/* take the RURI branch from the script msg and move it
+			 * into the faked msg (that will be used by t_fwd function) */
+			if (dst_to_msg( msg, &faked_req )<0) {
+				LM_ERR("failed to grab new branch from Event\n");
+				goto error;
+			}
+		} else {
+			/* current message is a reply, so take the first branch from dset
+			 * and move it into the faked msg (that will be used by t_fwd 
+			 * function)*/
+			if (move_branch_to_ruri( 0, &faked_req)<0) {
+				LM_ERR("no branch found to be moved as new destination\n");
+				goto error;
+			}
+			/* remove it from set */
+			remove_branch(0);
+		}
+	}
+
+	/* t_forward_nonack() relies on T for various internal stuff, so be
+	 * we advertise the correct one */
+	bk_t = get_t();
+	set_t( t );
+
+	/* do we have to cancel the existing branches before injecting new ones? */
+	if (flags&TM_INJECT_FLAG_CANCEL) {
+		which_cancel( t, &cancel_bm );
+	}
+
+	if (flags&TM_INJECT_FLAG_LAST)
+		t_wait_no_more_branches(t, 1/*the branch we will create*/);
+
+	/* generated the new branches, without branch counter reset */
+	rc = t_forward_nonack( t, &faked_req , NULL, 0, 1/*locked*/ );
+
+	/* do we have to cancel the existing branches before injecting new ones? */
+	if (flags&TM_INJECT_FLAG_CANCEL) {
+		set_cancel_extra_hdrs( reason.s, reason.len);
+		cancel_uacs( t, cancel_bm );
+		set_cancel_extra_hdrs( NULL, 0);
+	}
+
+	set_t( bk_t );
+
+	/* cleanup the faked request */
+	free_faked_req( &faked_req, t);
+
+	return (rc==1)?1:-3;
+
+error:
+	/* cleanup the faked request */
+	free_faked_req( &faked_req, t);
+	return -1;
+}
+
+
 int t_replicate(struct sip_msg *p_msg, str *dst, int flags)
 {
 	/* this is a quite horrible hack -- we just take the message
@@ -836,7 +1166,7 @@ int t_replicate(struct sip_msg *p_msg, str *dst, int flags)
 			LM_CRIT("BUG - undefined transaction in failure route\n");
 			return -1;
 		}
-		return t_relay_to( p_msg, NULL, flags|TM_T_REPLY_repl_FLAG);
+		return t_relay_to( p_msg, NULL, flags|TM_T_RELAY_repl_FLAG);
 	} else {
 		/* transaction already created */
 		if (p_msg->REQ_METHOD==METHOD_ACK)
@@ -845,11 +1175,19 @@ int t_replicate(struct sip_msg *p_msg, str *dst, int flags)
 
 		t->flags|=T_IS_LOCAL_FLAG;
 
-		return t_forward_nonack( t, p_msg, NULL );
+		return t_forward_nonack( t, p_msg, NULL, 1/*reset*/, 0/*unlocked*/);
 	}
 }
 
+
 int get_branch_index(void)
 {
 	return _tm_branch_index;
 }
+
+
+int t_inject_ul_event_branch(void)
+{
+	return w_t_inject_branches(NULL,
+	            (void *)(unsigned long)TM_INJECT_SRC_EVENT, (void *)0);
+}
