@@ -87,6 +87,7 @@
 #include "dlg.h" /* for t_lookup_callid */
 #include "t_msgbuilder.h" /* for t_lookup_callid */
 #include "t_fwd.h" /* for get_on_branch */
+#include "cluster.h" /*for messages replication */
 
 #define EQ_VIA_LEN(_via)\
 	( (p_msg->via1->bsize-(p_msg->_via->name.s-(p_msg->_via->hdr.s+p_msg->_via->hdr.len)))==\
@@ -148,7 +149,16 @@ static struct cell *cancelled_T;
  */
 static struct cell *e2eack_T;
 
-static str relay_reason_100 = str_init("Giving a try");
+
+/* a transaction ID in string format "hash.label", in reversed hexa;
+ * This is to be used to point TM to a remote transaction, other
+ * than the one currently processes (or outside the current processing)
+ * Functions pushing a transaction via ext_T must attache a valid str
+ * and keep it until the TM handling is done */
+struct tm_id* remote_T = NULL;
+
+
+static str relay_reason_100 = str_init("Giving it a try");
 
 
 struct cell *get_t(void) { return T; }
@@ -310,7 +320,8 @@ static int matching_3261( struct sip_msg *p_msg, struct cell **trans,
 		p_cell; p_cell = p_cell->next_cell )
 	{
 		t_msg=p_cell->uas.request;
-		if (!t_msg) continue;  /* don't try matching UAC transactions */
+		/* don't try matching UAC transactions */
+		if (is_local(p_cell) || !t_msg) continue;
 		if (skip_method & t_msg->REQ_METHOD) continue;
 
 		/* here we do an exercise which will be removed from future code
@@ -542,6 +553,8 @@ found:
 	set_kr(REQ_EXIST);
 	UNLOCK_HASH( p_msg->hash_index );
 	LM_DBG("transaction found (T=%p)\n",T);
+	if (has_tran_tmcbs( T, TMCB_MSG_MATCHED_IN) )
+		run_trans_callbacks( TMCB_MSG_MATCHED_IN, T, p_msg, 0,0);
 	return 1;
 }
 
@@ -565,6 +578,9 @@ struct cell* t_lookupOriginalT(  struct sip_msg* p_msg )
 		return cancelled_T;
 
 	/* start searching in the table */
+	if (!p_msg->hash_index)
+		p_msg->hash_index = tm_hash( p_msg->callid->body,
+			get_cseq(p_msg)->number);
 	hash_index = p_msg->hash_index;
 	LM_DBG("searching on hash entry %d\n",hash_index );
 
@@ -683,9 +699,9 @@ found:
 int t_reply_matching( struct sip_msg *p_msg , int *p_branch )
 {
 	struct cell*  p_cell;
-	int hash_index   = 0;
-	int entry_label  = 0;
-	int branch_id    = 0;
+	unsigned int hash_index   = 0;
+	unsigned int entry_label  = 0;
+	unsigned int branch_id    = 0;
 	char  *hashi, *branchi, *p, *n;
 	int hashl, branchl;
 	int scan_space;
@@ -754,19 +770,19 @@ int t_reply_matching( struct sip_msg *p_msg , int *p_branch )
 	branchi=p;
 
 	/* sanity check */
-	if ((hash_index=reverse_hex2int(hashi, hashl))<0
+	if (reverse_hex2int(hashi, hashl, &hash_index)<0
 		||hash_index>=TM_TABLE_ENTRIES
-		|| (branch_id=reverse_hex2int(branchi, branchl))<0
+		|| reverse_hex2int(branchi, branchl, &branch_id)<0
 		||branch_id>=MAX_BRANCHES
-		|| (syn_branch ? (entry_label=reverse_hex2int(syni, synl))<0
+		|| (syn_branch ? reverse_hex2int(syni, synl, &entry_label)<0
 			: loopl!=MD5_LEN )
 	) {
-		LM_DBG("poor reply labels %d label %d branch %d\n",
+		LM_DBG("poor reply labels %u label %u branch %u\n",
 				hash_index, entry_label, branch_id );
 		goto nomatch2;
 	}
 
-	LM_DBG("hash %d label %d branch %d\n",hash_index, entry_label, branch_id);
+	LM_DBG("hash %u label %d branch %u\n",hash_index, entry_label, branch_id);
 
 	cseq = get_cseq(p_msg);
 
@@ -827,11 +843,7 @@ int t_reply_matching( struct sip_msg *p_msg , int *p_branch )
 				LM_ERR("to parsing failed\n");
 			}
 		}
-		if (!is_local(p_cell) &&
-		!(cseq->method_id==METHOD_CANCEL && is_invite(p_cell)) ) {
-			run_trans_callbacks( TMCB_RESPONSE_IN, T, T->uas.request, p_msg,
-				p_msg->REPLY_STATUS);
-		}
+
 		return 1;
 	} /* for cycle */
 
@@ -851,9 +863,9 @@ nomatch2:
 
 /* Determine current transaction
  *
- *                   Found      Not Found     Error (e.g. parsing)
- *  Return Value     1          0             -1
- *  T                ptr        0             T_UNDEFINED
+ *                   Found      Not Found     Error (e.g. parsing)     Not Here (e.g. anycast)
+ *  Return Value     1          0             -1                       -2
+ *  T                ptr        0             T_UNDEFINED              T_UNDEFINED
  */
 int t_check( struct sip_msg* p_msg , int *param_branch )
 {
@@ -899,9 +911,13 @@ int t_check( struct sip_msg* p_msg , int *param_branch )
 						return -1;
 					}
 			}
-
-			t_reply_matching( p_msg ,
-				param_branch!=0?param_branch:&local_branch );
+			/* first check if the transaction is addressed to us */
+			if (!tm_reply_replicate(p_msg))
+				t_reply_matching(p_msg ,
+						param_branch!=0?param_branch:&local_branch);
+			else
+				return -2; /* reply replicated
+					should have never got here */
 
 		}
 #ifdef EXTRA_DEBUG
@@ -1050,6 +1066,8 @@ int t_newtran( struct sip_msg* p_msg, int full_uas )
 		} else {
 			t_retransmit_reply(T);
 		}
+		/* hide the transaction from the upper layer */
+		t_unref( p_msg );
 		/* things are done -- return from script */
 		return 0;
 	}
@@ -1110,9 +1128,9 @@ int t_newtran( struct sip_msg* p_msg, int full_uas )
 
 	if (auto_100trying && p_msg->REQ_METHOD==METHOD_INVITE) {
 		ctx_backup = current_processing_ctx;
-		current_processing_ctx = NULL;
+		set_global_context(NULL);
 		t_reply( T, p_msg , 100 , &relay_reason_100);
-		current_processing_ctx = ctx_backup;
+		set_global_context(ctx_backup);
 	}
 
 	return 1;
@@ -1172,7 +1190,11 @@ int t_get_trans_ident(struct sip_msg* p_msg, unsigned int* hash_index,
 }
 
 
-
+/* Looks for the transaction with the given coordinates (index/label).
+ * If found, the transaction is ref'ed and retuned via parameter; note that
+ * the global T holder is not set.
+ * Returns 1 if transaction found, -1 otherwise.
+ */
 int t_lookup_ident(struct cell ** trans, unsigned int hash_index,
 															unsigned int label)
 {
@@ -1192,7 +1214,6 @@ int t_lookup_ident(struct cell ** trans, unsigned int hash_index,
 		if(p_cell->label == label){
 			REF_UNSAFE(p_cell);
 			UNLOCK_HASH(hash_index);
-			set_t(p_cell);
 			*trans=p_cell;
 			LM_DBG("transaction found\n");
 			return 1;
@@ -1200,7 +1221,6 @@ int t_lookup_ident(struct cell ** trans, unsigned int hash_index,
 	}
 
 	UNLOCK_HASH(hash_index);
-	set_t(0);
 	*trans=p_cell;
 
 	LM_DBG("transaction not found\n");
