@@ -43,6 +43,7 @@
 #include "../../parser/contact/parse_contact.h"
 #include "t_funcs.h"
 #include "t_msgbuilder.h"
+#include "cluster.h"
 #include "uac.h"
 
 
@@ -52,8 +53,11 @@
 #define ROUTE_SEPARATOR ", "
 #define ROUTE_SEPARATOR_LEN (sizeof(ROUTE_SEPARATOR) - 1)
 
-#define LOCAL_MAXFWD_HEADER "Max-Forwards: " LOCAL_MAXFWD_VALUE CRLF
-#define LOCAL_MAXFWD_HEADER_LEN (sizeof(LOCAL_MAXFWD_HEADER) - 1)
+#define LOCAL_MAXFWD_PREFIX "Max-Forwards: "
+#define LOCAL_MAXFWD_PREFIX_LEN  (sizeof(LOCAL_MAXFWD_PREFIX) - 1)
+
+#define LOCAL_MAXFWD_HEADER  LOCAL_MAXFWD_PREFIX LOCAL_MAXFWD_VALUE CRLF
+#define LOCAL_MAXFWD_HEADER_LEN  (sizeof(LOCAL_MAXFWD_HEADER) - 1)
 
 /* convenience macros */
 #define  append_string(_d,_s,_len) \
@@ -111,6 +115,7 @@ char *build_local(struct cell *Trans,unsigned int branch,
 	str from;
 	str to;
 	str cseq_n;
+	str *cid = NULL;
 
 	req = Trans->uas.request;
 	cseq_n = Trans->cseq_n;
@@ -152,6 +157,9 @@ char *build_local(struct cell *Trans,unsigned int branch,
 	*len=SIP_VERSION_LEN + method->len + 2 /* spaces */ + CRLF_LEN;
 	*len+=Trans->uac[branch].uri.len;
 
+	if (is_anycast(Trans->uac[branch].request.dst.send_sock))
+		cid = tm_via_cid();
+
 	/*via*/
 	branch_str.s=branch_buf;
 	if (!t_calc_branch(Trans,  branch, branch_str.s, &branch_str.len ))
@@ -162,7 +170,7 @@ char *build_local(struct cell *Trans,unsigned int branch,
 	if (Trans->uac[branch].adv_port.len)
 		hp.port = &Trans->uac[branch].adv_port;
 	via=via_builder(&via_len, Trans->uac[branch].request.dst.send_sock,
-		&branch_str, 0, Trans->uac[branch].request.dst.proto, &hp );
+		&branch_str, cid, Trans->uac[branch].request.dst.proto, &hp );
 	if (!via){
 		LM_ERR("no via header got from builder\n");
 		goto error;
@@ -172,9 +180,9 @@ char *build_local(struct cell *Trans,unsigned int branch,
 	*len+=from.len+Trans->callid.len+to.len+cseq_n.len+1+method->len+CRLF_LEN;
 
 	/* copy'n'paste Route headers that were sent out */
-	if (!is_local(Trans) &&
-	( (req && req->route) || /* at least one route was received*/
-	(Trans->uac[branch].path_vec.len!=0)) ) /* path was forced */
+	if (is_local(Trans) || (!is_local(Trans) &&
+	/* if there was a ROUTE received or added */
+	( req && (req->route || req->msg_flags&FL_HAS_ROUTE_LUMP))) )
 	{
 		buf_hdrs = extract_parsed_hdrs(Trans->uac[branch].request.buffer.s,
 			Trans->uac[branch].request.buffer.len );
@@ -189,7 +197,7 @@ char *build_local(struct cell *Trans,unsigned int branch,
 
 	/* User Agent */
 	if (server_signature) {
-		*len += user_agent_header.len + CRLF_LEN;
+		*len += user_agent_header->len + CRLF_LEN;
 	}
 	/* Content Length, MaxFwd, EoM */
 	*len+=LOCAL_MAXFWD_HEADER_LEN + CONTENT_LENGTH_LEN+1 + (extra?extra->len:0)
@@ -236,7 +244,7 @@ char *build_local(struct cell *Trans,unsigned int branch,
 
 	/* User Agent header, Content Length, EoM */
 	if (server_signature) {
-		append_string(p, user_agent_header.s, user_agent_header.len);
+		append_string(p, user_agent_header->s, user_agent_header->len);
 		append_string(p, CRLF CONTENT_LENGTH "0" CRLF CRLF ,
 			CRLF_LEN+CONTENT_LENGTH_LEN+1 + CRLF_LEN + CRLF_LEN);
 	} else {
@@ -509,7 +517,7 @@ char *build_dlg_ack(struct sip_msg* rpl, struct cell *Trans,
 
 	/* User Agent */
 	if (server_signature)
-		*len += user_agent_header.len + CRLF_LEN;
+		*len += user_agent_header->len + CRLF_LEN;
 
 	/* Content Length, EoM */
 	*len += CONTENT_LENGTH_LEN + 1 + CRLF_LEN + CRLF_LEN;
@@ -543,7 +551,7 @@ char *build_dlg_ack(struct sip_msg* rpl, struct cell *Trans,
 
 	/* User Agent header, Content Length, EoM */
 	if (server_signature) {
-		append_string(p, user_agent_header.s, user_agent_header.len);
+		append_string(p, user_agent_header->s, user_agent_header->len);
 		append_string(p, CRLF CONTENT_LENGTH "0" CRLF CRLF,
 			CRLF_LEN+CONTENT_LENGTH_LEN + 1 + CRLF_LEN + CRLF_LEN);
 	} else {
@@ -600,7 +608,7 @@ static inline int print_cseq_num(str* _s, dlg_t* _d)
 /*
  * Create Via header
  */
-static inline int assemble_via(str* dest, struct cell* t, struct socket_info* sock, int branch)
+static inline int assemble_via(str* dest, struct cell* t, struct socket_info* sock, int branch, str *extra)
 {
 	static char branch_buf[MAX_BRANCH_PARAM_LEN];
 	char* via;
@@ -622,7 +630,7 @@ static inline int assemble_via(str* dest, struct cell* t, struct socket_info* so
 #endif
 
 	set_hostport(&hp, 0);
-	via = via_builder(&via_len, sock, &branch_str, 0, sock->proto, &hp);
+	via = via_builder(&via_len, sock, &branch_str, extra, sock->proto, &hp);
 	if (!via) {
 		LM_ERR("via building failed\n");
 		return -2;
@@ -658,7 +666,7 @@ static inline char* print_request_uri(char* w, str* method, dlg_t* dialog, struc
 static inline char* print_to(char* w, dlg_t* dialog, struct cell* t)
 {
 	t->to.s = w;
-	t->to.len = TO_LEN + dialog->rem_uri.len + CRLF_LEN;
+	t->to.len = TO_LEN + dialog->rem_uri.len + 2 + CRLF_LEN;
 
 	append_string(w, TO, TO_LEN);
 
@@ -668,18 +676,9 @@ static inline char* print_to(char* w, dlg_t* dialog, struct cell* t)
 		*(w++) = ' ';
 	}
 
-	if(dialog->rem_dname.len || dialog->id.rem_tag.len) {
-		t->to.len +=1;
-		*(w++) = '<';
-	}
-
+	*(w++) = '<';
 	append_string(w, dialog->rem_uri.s, dialog->rem_uri.len);
-
-	if(dialog->rem_dname.len || dialog->id.rem_tag.len) {
-		t->to.len +=1;
-		*(w++) = '>';
-	}
-
+	*(w++) = '>';
 
 	if (dialog->id.rem_tag.len) {
 		t->to.len += TOTAG_LEN + dialog->id.rem_tag.len ;
@@ -698,7 +697,7 @@ static inline char* print_to(char* w, dlg_t* dialog, struct cell* t)
 static inline char* print_from(char* w, dlg_t* dialog, struct cell* t)
 {
 	t->from.s = w;
-	t->from.len = FROM_LEN + dialog->loc_uri.len + CRLF_LEN;
+	t->from.len = FROM_LEN + dialog->loc_uri.len + 2 + CRLF_LEN;
 
 	append_string(w, FROM, FROM_LEN);
 
@@ -708,17 +707,9 @@ static inline char* print_from(char* w, dlg_t* dialog, struct cell* t)
 		*(w++) = ' ';
 	}
 
-	if(dialog->loc_dname.len || dialog->id.loc_tag.len) {
-		t->from.len +=1;
-		*(w++) = '<';
-	}
-
+	*(w++) = '<';
 	append_string(w, dialog->loc_uri.s, dialog->loc_uri.len);
-
-	if(dialog->loc_dname.len || dialog->id.loc_tag.len) {
-		t->from.len += 1;
-		*(w++) = '>';
-	}
+	*(w++) = '>';
 
 	if (dialog->id.loc_tag.len) {
 		t->from.len += FROMTAG_LEN + dialog->id.loc_tag.len;
@@ -782,7 +773,8 @@ char* build_uac_req(str* method, str* headers, str* body, dlg_t* dialog,
 										int branch, struct cell *t, int* len)
 {
 	char* buf, *w;
-	str content_length, cseq, via;
+	str content_length, cseq, via, mf;
+	str *cid = tm_via_cid();
 
 	if (!method || !dialog) {
 		LM_ERR("inalid parameter value\n");
@@ -799,7 +791,7 @@ char* build_uac_req(str* method, str* headers, str* body, dlg_t* dialog,
 	*len = method->len + 1 + dialog->hooks.request_uri->len + 1 +
 		SIP_VERSION_LEN + CRLF_LEN;
 
-	if (assemble_via(&via, t, dialog->send_sock, branch) < 0) {
+	if (assemble_via(&via, t, dialog->send_sock, branch, cid) < 0) {
 		LM_ERR("failed to assemble Via\n");
 		return 0;
 	}
@@ -810,15 +802,13 @@ char* build_uac_req(str* method, str* headers, str* body, dlg_t* dialog,
 		+ (dialog->rem_dname.len ? dialog->rem_dname.len+1 : 0)
 		+ dialog->rem_uri.len
 		+ (dialog->id.rem_tag.len ? TOTAG_LEN + dialog->id.rem_tag.len : 0)
-		+ (dialog->rem_dname.len || dialog->id.rem_tag.len ? 2 : 0)
-		+ CRLF_LEN;
+		+ 2 /* <> for URI */ + CRLF_LEN;
 	/* From */
 	*len += FROM_LEN
 		+ (dialog->loc_dname.len ? dialog->loc_dname.len+1 : 0)
 		+ dialog->loc_uri.len
 		+ (dialog->id.loc_tag.len ? FROMTAG_LEN + dialog->id.loc_tag.len : 0)
-		+ (dialog->loc_dname.len || dialog->id.loc_tag.len ? 2 : 0)
-		+ CRLF_LEN;
+		+ 2 /* <> for URI */ + CRLF_LEN;
 	/* Call-ID */
 	*len += CALLID_LEN + dialog->id.call_id.len + CRLF_LEN;
 	/* CSeq */
@@ -826,11 +816,17 @@ char* build_uac_req(str* method, str* headers, str* body, dlg_t* dialog,
 	/* Route set */
 	*len += calculate_routeset_length(dialog);
 	/* Max-Forwards */
-	*len += LOCAL_MAXFWD_HEADER_LEN;
+	if (dialog->mf_enforced) {
+		mf.s = int2str( dialog->mf_value, &mf.len);
+		*len += LOCAL_MAXFWD_PREFIX_LEN + mf.len + CRLF_LEN;
+	} else {
+		mf.len = 0; mf.s = NULL;
+		*len += LOCAL_MAXFWD_HEADER_LEN;
+	}
 	/* Content-Length */
 	*len += CONTENT_LENGTH_LEN + content_length.len + CRLF_LEN;
 	/* Signature */
-	*len += (server_signature ? (user_agent_header.len + CRLF_LEN) : 0);
+	*len += (server_signature ? (user_agent_header->len + CRLF_LEN) : 0);
 	/* Additional headers */
 	*len += (headers ? headers->len : 0);
 	/* Message body */
@@ -855,7 +851,13 @@ char* build_uac_req(str* method, str* headers, str* body, dlg_t* dialog,
 	w = print_routeset(w, dialog);                        /* Route set */
 
 	/* Max-Forwards */
-	append_string(w, LOCAL_MAXFWD_HEADER, LOCAL_MAXFWD_HEADER_LEN);
+	if (dialog->mf_enforced) {
+		append_string(w, LOCAL_MAXFWD_PREFIX, LOCAL_MAXFWD_PREFIX_LEN);
+		append_string(w, mf.s, mf.len);
+		append_string(w, CRLF, CRLF_LEN);
+	} else {
+		append_string(w, LOCAL_MAXFWD_HEADER, LOCAL_MAXFWD_HEADER_LEN);
+	}
 
 	/* Content-Length */
 	append_string(w, CONTENT_LENGTH, CONTENT_LENGTH_LEN);
@@ -864,7 +866,7 @@ char* build_uac_req(str* method, str* headers, str* body, dlg_t* dialog,
 
 	/* Server signature */
 	if (server_signature) {
-		append_string(w, user_agent_header.s, user_agent_header.len);
+		append_string(w, user_agent_header->s, user_agent_header->len);
 		append_string(w, CRLF, CRLF_LEN);
 	}
 
