@@ -100,7 +100,7 @@ int onreply_avp_mode = 0;
 /* disable the 6xx fork-blocking - default no (as per RFC3261) */
 int disable_6xx_block = 0;
 
-/* flag for marching minor branches */
+/* flag for marking minor branches */
 int minor_branch_flag = -1;
 char *minor_branch_flag_str = 0;
 
@@ -112,9 +112,9 @@ char *tm_tag_suffix;
 static int picked_branch=-1;
 
 /* where to go if there is no positive reply */
-static int goto_on_negative=0;
+static struct script_route_ref *goto_on_negative = NULL;
 /* where to go on receipt of reply */
-static int goto_on_reply=0;
+static struct script_route_ref *goto_on_reply = NULL;
 
 /* currently processed branch */
 extern int _tm_branch_index;
@@ -138,46 +138,56 @@ int t_get_picked_branch(void)
 */
 
 
-void t_on_negative( unsigned int go_to )
+void t_on_negative( struct script_route_ref *ref )
 {
 	struct cell *t = get_t();
+	struct script_route_ref **holder;
 
 	/* in MODE_REPLY and MODE_ONFAILURE T will be set to current transaction;
 	 * in MODE_REQUEST T will be set only if the transaction was already
 	 * created; if not -> use the static variable */
-	if (!t || t==T_UNDEFINED )
-		goto_on_negative=go_to;
-	else
-		t->on_negative = go_to;
+	holder = (!t || t==T_UNDEFINED ) ? &goto_on_negative : &t->on_negative ;
+
+	/* if something already set, free it first */
+	if (*holder)
+		shm_free( *holder );
+
+	*holder = ref ? dup_ref_script_route_in_shm( ref, 0) : NULL;
 }
 
 
-void t_on_reply( unsigned int go_to )
+void t_on_reply( struct script_route_ref *ref  )
 {
 	struct cell *t = get_t();
 
+	struct script_route_ref **holder;
+
 	/* in MODE_REPLY and MODE_ONFAILURE T will be set to current transaction;
 	 * in MODE_REQUEST T will be set only if the transaction was already
 	 * created; if not -> use the static variable */
-	if (!t || t==T_UNDEFINED ) {
-		goto_on_reply=go_to;
-	} else {
-		if (route_type==BRANCH_ROUTE) {
-			t->uac[_tm_branch_index].on_reply = go_to;
-		} else {
-			t->on_reply = go_to;
-		}
-	}
+	holder = (!t || t==T_UNDEFINED ) ? &goto_on_reply :
+		( route_type==BRANCH_ROUTE ?
+			&t->uac[_tm_branch_index].on_reply : &t->on_reply );
+
+	/* if something already set, free it first */
+	if (*holder)
+		shm_free( *holder );
+
+	*holder = ref ? dup_ref_script_route_in_shm( ref, 0) : NULL;
 }
 
 
-unsigned int get_on_negative(void)
+struct script_route_ref *get_on_negative(void)
 {
-	return goto_on_negative;
+	struct script_route_ref *ref = goto_on_negative;
+	goto_on_negative = NULL;
+	return ref;
 }
-unsigned int get_on_reply(void)
+struct script_route_ref *get_on_reply(void)
 {
-	return goto_on_reply;
+	struct script_route_ref *ref = goto_on_reply;
+	goto_on_reply = NULL;
+	return ref;
 }
 
 void tm_init_tags(void)
@@ -259,16 +269,13 @@ inline static int update_totag_set(struct cell *t, struct sip_msg *ok)
 			return 1;
 		}
 	}
+
 	/* that's a new to-tag -- record it */
-#ifndef HP_MALLOC
 	shm_lock();
-	n=(struct totag_elem*) shm_malloc_unsafe(sizeof(struct totag_elem));
-	s=(char *)shm_malloc_unsafe(tag->len);
+	n = shm_malloc_bulk(sizeof *n);
+	s = shm_malloc_bulk(tag->len);
 	shm_unlock();
-#else
-	n=(struct totag_elem*) shm_malloc(sizeof(struct totag_elem));
-	s=(char *)shm_malloc(tag->len);
-#endif
+
 	if (!s || !n) {
 		LM_ERR("no more share memory \n");
 		if (n) shm_free(n);
@@ -287,13 +294,16 @@ inline static int update_totag_set(struct cell *t, struct sip_msg *ok)
 
 /*
  * Build and send an ACK to a negative reply
+ * On successful send:
+ *   - return 0
+ *   - populate @ack_buf, for callback purposes, which *must* be SHM freed!
  */
-static int send_ack(struct sip_msg* rpl, struct cell *trans, int branch)
+static int send_ack(struct sip_msg* rpl, struct cell *trans, int branch, str *ack_buf)
 {
 	str method = str_init(ACK);
 	str to;
-	char *ack_p;
-	unsigned int  ack_len;
+	struct usr_avp **backup_list;
+	int rc;
 
 	if(parse_headers(rpl,is_local(trans)?HDR_EOH_F:(HDR_TO_F|HDR_FROM_F),0)==-1
 	|| !rpl->to || !rpl->from ) {
@@ -303,19 +313,30 @@ static int send_ack(struct sip_msg* rpl, struct cell *trans, int branch)
 	to.s=rpl->to->name.s;
 	to.len=rpl->to->len;
 
-	ack_p = is_local(trans)?
-		build_dlg_ack(rpl, trans, branch, &to, &ack_len):
-		build_local( trans, branch, &method, NULL, rpl, &ack_len );
-	if (ack_p==0) {
+	ack_buf->s = is_local(trans)?
+		build_dlg_ack(rpl, trans, branch, &to, (unsigned int*)&ack_buf->len):
+		build_local( trans, branch, &method, NULL, rpl, (unsigned int*)&ack_buf->len );
+	if (ack_buf->s==0) {
 		LM_ERR("failed to build ACK\n");
 		goto error;
 	}
 
-	SEND_PR_BUFFER(&trans->uac[branch].request, ack_p, ack_len);
-	shm_free(ack_p);
+	if (trans->uac[branch].br_flags & tcp_no_new_conn_bflag)
+		tcp_no_new_conn = 1;
 
-	return 0;
+	set_bavp_list(&trans->uac[branch].user_avps);
+	backup_list = set_avp_list( &trans->user_avps );
+
+	rc = SEND_PR_BUFFER(&trans->uac[branch].request, ack_buf->s, ack_buf->len);
+
+	set_avp_list(backup_list);
+	reset_bavp_list();
+
+	tcp_no_new_conn = 0;
+
+	return rc;
 error:
+	memset(ack_buf, 0, sizeof *ack_buf);
 	return -1;
 }
 
@@ -326,7 +347,7 @@ static int _reply_light( struct cell *trans, char* buf, unsigned int len,
 {
 	struct retr_buf *rb;
 	unsigned int buf_len;
-	branch_bm_t cancel_bitmap;
+	branch_bm_t cancel_bitmap = 0;
 	str cb_s;
 
 	if (!buf)
@@ -342,14 +363,14 @@ static int _reply_light( struct cell *trans, char* buf, unsigned int len,
 		goto error;
 	}
 
-	cancel_bitmap=0;
 	if (lock) LOCK_REPLIES( trans );
-	if ( is_invite(trans) ) which_cancel(trans, &cancel_bitmap );
 	if (trans->uas.status>=200) {
 		LM_ERR("failed to generate %d reply when a final %d was sent out\n",
 				code, trans->uas.status);
 		goto error2;
 	}
+	if ( is_invite(trans) && code>=200 )
+		which_cancel(trans, &cancel_bitmap );
 
 
 	rb = & trans->uas.response;
@@ -357,7 +378,7 @@ static int _reply_light( struct cell *trans, char* buf, unsigned int len,
 
 	trans->uas.status = code;
 	buf_len = rb->buffer.s ? len : len + REPLY_OVERBUFFER_LEN;
-	rb->buffer.s = (char*)shm_resize( rb->buffer.s, buf_len );
+	rb->buffer.s = shm_realloc( rb->buffer.s, buf_len );
 	/* puts the reply's buffer to uas.response */
 	if (! rb->buffer.s ) {
 			LM_ERR("failed to allocate shmem buffer\n");
@@ -415,12 +436,28 @@ static int _reply_light( struct cell *trans, char* buf, unsigned int len,
 	if (!trans->uas.response.dst.send_sock) {
 		LM_CRIT("send_sock is NULL\n");
 	}
-	SEND_PR_BUFFER( rb, buf, len );
-	LM_DBG("reply sent out. buf=%p: %.9s..., "
-		"shmem=%p: %.9s\n", buf, buf, rb->buffer.s, rb->buffer.s );
+
+	if(trans->uas.request && trans->uas.request->flags&tcp_no_new_conn_rplflag)
+		tcp_no_new_conn = 1;
+
+	if ( SEND_PR_BUFFER( rb, buf, len )==0 ) {
+		LM_DBG("reply sent out. buf=%p: %.9s..., "
+			"shmem=%p: %.9s\n", buf, buf, rb->buffer.s, rb->buffer.s );
+
+		if (has_tran_tmcbs(trans, TMCB_MSG_SENT_OUT) ) {
+			cb_s.s = buf;
+			cb_s.len = len;
+			set_extra_tmcb_params( &cb_s, &rb->dst);
+			run_trans_callbacks( TMCB_MSG_SENT_OUT, trans,
+				NULL, FAKED_REPLY, code);
+		}
+		stats_trans_rpl( code, 1 /*local*/ );
+	}
+
+	tcp_no_new_conn = 0;
 
 	/* run the POST send callbacks */
-	if (code>=200&&!is_local(trans)&&has_tran_tmcbs(trans,TMCB_RESPONSE_OUT)) {
+	if (code>=200&&!is_local(trans)&&has_tran_tmcbs(trans,TMCB_RESPONSE_OUT)){
 		cb_s.s = buf;
 		cb_s.len = len;
 		set_extra_tmcb_params( &cb_s, &rb->dst);
@@ -432,9 +469,7 @@ static int _reply_light( struct cell *trans, char* buf, unsigned int len,
 				trans->uas.request, FAKED_REPLY, code);
 	}
 
-
 	pkg_free( buf ) ;
-	stats_trans_rpl( code, 1 /*local*/ );
 	LM_DBG("finished\n");
 	return 1;
 
@@ -456,7 +491,7 @@ error:
  * returns 1 if everything was OK or -1 for error
  */
 static int _reply( struct cell *trans, struct sip_msg* p_msg,
-									unsigned int code, str *text, int lock )
+    unsigned int code, const str *text, int lock )
 {
 	unsigned int len;
 	char * buf, *dset;
@@ -475,6 +510,11 @@ static int _reply( struct cell *trans, struct sip_msg* p_msg,
 		}
 	}
 
+	if (has_tran_tmcbs(trans, TMCB_LOCAL_RESPONSE)) {
+		run_trans_callbacks(TMCB_LOCAL_RESPONSE,
+			trans, p_msg, 0, code);
+	}
+
 	/* check if the UAS retranmission port needs to be updated */
 	if ( (p_msg->msg_flags ^ trans->uas.request->msg_flags) & FL_FORCE_RPORT )
 		su_setport( &trans->uas.response.dst.to, p_msg->rcv.src_port );
@@ -482,7 +522,7 @@ static int _reply( struct cell *trans, struct sip_msg* p_msg,
 	if (code>=180 && p_msg->to
 				&& (get_to(p_msg)->tag_value.s==0
 			    || get_to(p_msg)->tag_value.len==0)) {
-		calc_crc_suffix( p_msg, tm_tag_suffix );
+		calc_tag_suffix( p_msg, tm_tag_suffix );
 		buf = build_res_buf_from_sip_req(code,text, &tm_tag, p_msg, &len, &bm);
 		return _reply_light( trans, buf, len, code,
 			tm_tag.s, TOTAG_VALUE_LEN, lock, &bm);
@@ -538,26 +578,30 @@ static inline int run_failure_handlers(struct cell *t)
 	static struct sip_msg faked_req;
 	struct sip_msg *shmem_msg;
 	struct ua_client *uac;
-	int on_failure;
+	int on_failure_idx;
+	int old_route_type;
 
 	shmem_msg = t->uas.request;
 	uac = &t->uac[picked_branch];
 
 	/* failure_route for a local UAC? */
 	if (!shmem_msg || REQ_LINE(shmem_msg).method_value==METHOD_CANCEL ) {
-		LM_WARN("no UAC or CANCEL support (%d, %d) \n",
-				t->on_negative, t->tmcb_hl.reg_types);
+		LM_WARN("no UAC or CANCEL support (%s, %d) \n",
+				ref_script_route_name(t->on_negative), t->tmcb_hl.reg_types);
 		return 0;
 	}
 
 	/* don't start faking anything if we don't have to */
-	if ( !has_tran_tmcbs( t, TMCB_ON_FAILURE) && !t->on_negative ) {
-		LM_WARN("no negative handler (%d, %d)\n",t->on_negative,
+	if ( !ref_script_route_check_and_update(t->on_negative) &&
+	!has_tran_tmcbs( t, TMCB_ON_FAILURE) ) {
+		LM_WARN("no failure route (%s/%d), callbacks (%d)\n",
+			ref_script_route_name(t->on_negative),
+			ref_script_route_idx(t->on_negative),
 			t->tmcb_hl.reg_types);
 		return 1;
 	}
 
-	if (!fake_req(&faked_req, shmem_msg, &t->uas, uac, 0/*no dst_uri*/)) {
+	if (!fake_req(&faked_req, shmem_msg, &t->uas, NULL)) {
 		LM_ERR("fake_req failed\n");
 		return 0;
 	}
@@ -569,16 +613,18 @@ static inline int run_failure_handlers(struct cell *t)
 		run_trans_callbacks( TMCB_ON_FAILURE, t, &faked_req,
 				uac->reply, uac->last_received);
 	}
-	if (t->on_negative) {
+	if (ref_script_route_is_valid(t->on_negative)) {
 		/* update flags in transaction if changed by callbacks */
 		shmem_msg->flags = faked_req.flags;
 		/* avoid recursion -- if failure_route forwards, and does not
 		 * set next failure route, failure_route will not be reentered
 		 * on failure */
-		on_failure = t->on_negative;
-		t->on_negative=0;
+		on_failure_idx = t->on_negative->idx;
+		t_on_negative(NULL);
 		/* run a reply_route action if some was marked */
-		run_top_route(failure_rlist[on_failure].a, &faked_req);
+		swap_route_type(old_route_type, FAILURE_ROUTE);
+		run_top_route(sroutes->failure[on_failure_idx], &faked_req);
+		set_route_type(old_route_type);
 	}
 
 	/* restore original environment and free the fake msg */
@@ -657,17 +703,17 @@ static inline int do_dns_failover(struct cell *t)
 		if (t->uas.request==NULL) {
 			LM_ERR("cloning failed\n");
 			free_sip_msg(req);
+			pkg_free(req);
 			return -1;
 		}
 		t->uas.end_request = ((char*)t->uas.request) + sip_msg_len;
 		/* free the actual SIP message, keep the clone only */
 		free_sip_msg(req);
-		/* the sip_msg structure is static in buf_to_sip_msg,
-		   so no need to free it */
+		pkg_free(req);
 	}
 	shmem_msg = t->uas.request;
 
-	if (!fake_req(&faked_req, shmem_msg, &t->uas, uac, 1/*with dst_uri*/)) {
+	if (!fake_req(&faked_req, shmem_msg, &t->uas, uac)) {
 		LM_ERR("fake_req failed\n");
 		return -1;
 	}
@@ -683,7 +729,7 @@ static inline int do_dns_failover(struct cell *t)
 	faked_req.force_send_socket = shmem_msg->force_send_socket;
 
 	/* send it out */
-	if (t_forward_nonack( t, &faked_req, uac->proxy)==1)
+	if (t_forward_nonack( t, &faked_req, uac->proxy,1/*reset*/,1/*locked*/)==1)
 		ret = 0;
 
 done:
@@ -745,6 +791,8 @@ static inline int branch_prio( short ret_code, unsigned int is_cancelled)
  */
 static inline int t_pick_branch( struct cell *t, int *res_code, int *do_cancel)
 {
+	#define PHONY_NO_WAIT(_t,_b) \
+		(_t->uac[_b].br_flags && _t->uac[_b].br_flags<=_t->nr_of_outgoings)
 	int lowest_b, lowest_s, b, prio;
 	unsigned int cancelled;
 
@@ -752,6 +800,20 @@ static inline int t_pick_branch( struct cell *t, int *res_code, int *do_cancel)
 	cancelled = was_cancelled(t);
 	*do_cancel = 0;
 	for ( b=t->first_branch; b<t->nr_of_outgoings ; b++ ) {
+		if ( (t->uac[b].flags & T_UAC_IS_PHONY) && (
+		/* skip PHONY branches if the transaction was canceled by UAC;
+		 * a phony branch is used just to force the transaction to wait for
+		 * more branches, but if canceled, it makes no sense to wait anymore;
+		 * Exception - do not ignore the branch if there is reply pushed
+		 * on that branch, like an internal timeout or so */
+		(t->flags & T_WAS_CANCELLED_FLAG && t->uac[b].last_received<299)
+		||
+		/* also skip a PHONY branch if it has a true setting about how
+		 * many branches are to be waited. The max branch idx to be waited
+		 * is stored in the 'br_flags' flags of the uac */
+		PHONY_NO_WAIT(t,b)
+		))
+			continue;
 		/* skip 'empty branches' */
 		if (!t->uac[b].request.buffer.s) continue;
 		/* there is still an unfinished UAC transaction; wait now! */
@@ -777,6 +839,20 @@ static inline int t_pick_branch( struct cell *t, int *res_code, int *do_cancel)
 }
 
 
+/* Quick and harmless test to see if the transaction still has 
+   pending branches */
+static inline int tran_is_completed( struct cell *t )
+{
+	int i;
+
+	for( i=t->first_branch ; i<t->nr_of_outgoings ; i++ )
+		if ( t->uac[i].last_received<200 )
+			return 0;
+
+	return 1;
+}
+
+
 /* This is the neurological point of reply processing -- called
  * from within a REPLY_LOCK, t_should_relay_response decides
  * how a reply shall be processed and how transaction state is
@@ -804,6 +880,31 @@ static enum rps t_should_relay_response( struct cell *Trans , int new_code,
 	   out
 	*/
 	LM_DBG("T_code=%d, new_code=%d\n", Trans->uas.status,new_code);
+
+	/* a final reply after 200 OK on a transaction with multi branch 200 OK */
+	if ( is_invite(Trans) && new_code>=200
+	&& Trans->flags&T_MULTI_200OK_FLAG
+	&& Trans->uas.status>=200 && Trans->uas.status<300) {
+		*should_store=0;
+		picked_branch=-1;
+		Trans->uac[branch].last_received=new_code;
+		if (new_code>=300) {
+			/* negative reply, we simply discard (no relay, no save) */
+			*should_relay = -1;
+			return RPS_DISCARDED;
+		}
+		/* 2xx reply - is this the last branch to complete?? */
+		if (tran_is_completed(Trans)) {
+			/* last branch gets also a 200OK, no more pending branches */
+			*should_relay = branch;
+			return RPS_COMPLETED;
+		} else {
+			/* 200 OK, but still having ongoing branches */
+			*should_relay = branch;
+			return RPS_RELAY;
+		}
+	}
+
 	inv_through=new_code>=200 && new_code<300 && is_invite(Trans);
 	/* if final response sent out, allow only INVITE 2xx  */
 	if ( Trans->uas.status >= 200 ) {
@@ -831,7 +932,7 @@ static enum rps t_should_relay_response( struct cell *Trans , int new_code,
 #ifdef EXTRA_DEBUG
 		/* don't report on retransmissions */
 		if (Trans->uac[branch].last_received==new_code) {
-			LM_DBG("final reply retransmission\n");
+			LM_DBG("final rely retransmission\n");
 		} else
 		/* if you FR-timed-out, faked a local 408 and 487 came, don't
 		 * report on it either */
@@ -956,6 +1057,14 @@ static enum rps t_should_relay_response( struct cell *Trans , int new_code,
 		*should_store=0;
 		*should_relay= new_code==100? -1 : branch;
 		if (new_code>=200 ) {
+			/* if a a multi-200OK transaction, prevent the transaction
+			   completion if we still have pending branches */
+			if (Trans->flags&T_MULTI_200OK_FLAG) {
+				if (tran_is_completed(Trans))
+					return RPS_COMPLETED;
+				else
+					return RPS_RELAY;
+			}
 			which_cancel( Trans, cancel_bitmap );
 			return RPS_COMPLETED;
 		} else return RPS_PROVISIONAL;
@@ -976,6 +1085,7 @@ int t_retransmit_reply( struct cell *t )
 {
 	static char b[BUF_SIZE];
 	int len;
+	str cb_s;
 
 	/* we need to lock the transaction as messages from
 	   upstream may change it continuously */
@@ -1000,9 +1110,26 @@ int t_retransmit_reply( struct cell *t )
 	}
 	memcpy( b, t->uas.response.buffer.s, len );
 	UNLOCK_REPLIES( t );
-	SEND_PR_BUFFER( & t->uas.response, b, len );
-	LM_DBG("buf=%p: %.9s..., shmem=%p: %.9s\n",b, b, t->uas.response.buffer.s,
-			t->uas.response.buffer.s );
+
+	/* send the buffer out */
+	if (t->uas.request && t->uas.request->flags & tcp_no_new_conn_rplflag)
+		tcp_no_new_conn = 1;
+
+	if (SEND_PR_BUFFER( & t->uas.response, b, len )==0) {
+		/* success */
+		LM_DBG("buf=%p: %.9s..., shmem=%p: %.9s\n",b, b,
+			t->uas.response.buffer.s, t->uas.response.buffer.s );
+		if (has_tran_tmcbs( t, TMCB_MSG_SENT_OUT) ) {
+			cb_s.s = b;
+			cb_s.len = len;
+			set_extra_tmcb_params( &cb_s, &t->uas.response.dst);
+			run_trans_callbacks( TMCB_MSG_SENT_OUT, t,
+					NULL, FAKED_REPLY, t->uas.status);
+		}
+	}
+
+	tcp_no_new_conn = 0;
+
 	return 1;
 
 error:
@@ -1011,14 +1138,13 @@ error:
 }
 
 
-
-
 int t_reply( struct cell *t, struct sip_msg* p_msg, unsigned int code,
-	str * text )
+	const str * text )
 {
 	return _reply( t, p_msg, code, text, 1 /* lock replies */ );
 }
 
+
 int t_reply_unsafe( struct cell *t, struct sip_msg* p_msg, unsigned int code,
 	str * text )
 {
@@ -1026,7 +1152,13 @@ int t_reply_unsafe( struct cell *t, struct sip_msg* p_msg, unsigned int code,
 }
 
 
+int t_gen_totag(struct sip_msg *msg, str *totag)
+{
+	calc_tag_suffix( msg, tm_tag_suffix );
+	*totag = tm_tag;
 
+	return 1;
+}
 
 
 void set_final_timer( /* struct s_table *h_table, */ struct cell *t )
@@ -1124,8 +1256,8 @@ enum rps relay_reply( struct cell *t, struct sip_msg *p_msg, int branch,
 	/* *** store and relay message as needed *** */
 	reply_status = t_should_relay_response(t, msg_status, branch,
 		&save_clone, &relay, cancel_bitmap, p_msg );
-	LM_DBG("branch=%d, save=%d, relay=%d\n",
-		branch, save_clone, relay );
+	LM_DBG("T_state=%d, branch=%d, save=%d, relay=%d, cancel_BM=%X\n",
+		reply_status, branch, save_clone, relay, *cancel_bitmap );
 
 	/* store the message if needed */
 	if (save_clone) /* save for later use, typically branch picking */
@@ -1154,7 +1286,7 @@ enum rps relay_reply( struct cell *t, struct sip_msg *p_msg, int branch,
 			if (relayed_code>=180 && t->uas.request->to
 					&& (get_to(t->uas.request)->tag_value.s==0
 					|| get_to(t->uas.request)->tag_value.len==0)) {
-				calc_crc_suffix( t->uas.request, tm_tag_suffix );
+				calc_tag_suffix( t->uas.request, tm_tag_suffix );
 				buf = build_res_buf_from_sip_req(
 						relayed_code,
 						&text,
@@ -1195,7 +1327,7 @@ enum rps relay_reply( struct cell *t, struct sip_msg *p_msg, int branch,
 		      larger messages are likely to follow and we will be
 		      able to reuse the memory frag
 		*/
-		uas_rb->buffer.s = (char*)shm_resize( uas_rb->buffer.s, res_len +
+		uas_rb->buffer.s = shm_realloc( uas_rb->buffer.s, res_len +
 			(msg_status<200 ?  REPLY_OVERBUFFER_LEN : 0));
 		if (!uas_rb->buffer.s) {
 			LM_ERR("no more share memory\n");
@@ -1244,9 +1376,27 @@ enum rps relay_reply( struct cell *t, struct sip_msg *p_msg, int branch,
 			run_trans_callbacks_locked(TMCB_RESPONSE_PRE_OUT,t,t->uas.request,
 				relayed_msg, relayed_code);
 		}
-		SEND_PR_BUFFER( uas_rb, buf, res_len );
-		LM_DBG("sent buf=%p: %.9s..., shmem=%p: %.9s\n",
-			buf, buf, uas_rb->buffer.s, uas_rb->buffer.s );
+
+		if (t->uas.request && t->uas.request->flags & tcp_no_new_conn_rplflag)
+			tcp_no_new_conn = 1;
+
+		/* send it out*/
+		if (SEND_PR_BUFFER( uas_rb, buf, res_len)==0) {
+			/* success */
+			LM_DBG("sent buf=%p: %.9s..., shmem=%p: %.9s\n",
+				buf, buf, uas_rb->buffer.s, uas_rb->buffer.s );
+
+			if (has_tran_tmcbs( t, TMCB_MSG_SENT_OUT) ) {
+				cb_s.s = buf;
+				cb_s.len = res_len;
+				set_extra_tmcb_params( &cb_s, &uas_rb->dst);
+				run_trans_callbacks( TMCB_MSG_SENT_OUT, t,
+					NULL, relayed_msg, relayed_code);
+			}
+		}
+
+		tcp_no_new_conn = 0;
+
 		/* run the POST sending out callback */
 		if (!totag_retr && has_tran_tmcbs(t, TMCB_RESPONSE_OUT) ) {
 			cb_s.s = buf;
@@ -1343,7 +1493,7 @@ enum rps local_reply( struct cell *t, struct sip_msg *p_msg, int branch,
 		} else {
 			LM_DBG("local transaction completed\n");
 			if (!totag_retr && has_tran_tmcbs(t,TMCB_LOCAL_COMPLETED) ) {
-				run_trans_callbacks( TMCB_LOCAL_COMPLETED, t, 0,
+				run_trans_callbacks( TMCB_LOCAL_COMPLETED, t, t->uas.request,
 					winning_msg, winning_code );
 			}
 		}
@@ -1379,11 +1529,16 @@ int reply_received( struct sip_msg  *p_msg )
 	struct cell *t;
 	struct usr_avp **backup_list;
 	unsigned int has_reply_route;
+	int old_route_type, ack_sent = 0;
+	str ack_buf;
 
 	set_t(T_UNDEFINED);
 
 	/* make sure we know the associated transaction ... */
-	if (t_check(p_msg, &branch ) == -1) goto not_found;
+	switch (t_check(p_msg, &branch )) {
+		case -1: goto not_found;
+		case -2: return 0; /* reply forwarded elsewhere */
+	}
 
 	/*... if there is none, tell the core router to fwd statelessly */
 	t = get_t();
@@ -1407,6 +1562,11 @@ int reply_received( struct sip_msg  *p_msg )
 				reset_timer( &uac->local_cancel.fr_timer);
 		}
 		LM_DBG("reply to local CANCEL processed\n");
+
+		if (has_tran_tmcbs( t, TMCB_MSG_MATCHED_IN) )
+			run_trans_callbacks( TMCB_MSG_MATCHED_IN, t, 0,
+				p_msg, p_msg->REPLY_STATUS);
+
 		goto done;
 	}
 
@@ -1419,17 +1579,34 @@ int reply_received( struct sip_msg  *p_msg )
 		reset_timer( &uac->request.fr_timer);
 	}
 
-	/* acknowledge negative INVITE replies (do it before detailed
+	/* acknowledge negative INVITE replies ASAP! (do it before detailed
 	 * on_reply processing, which may take very long, like if it
 	 * is attempted to establish a TCP connection to a fail-over dst */
 	if (is_invite(t) && ((msg_status >= 300) ||
 	(is_local(t) && !no_autoack(t) && msg_status >= 200) )) {
-		if (send_ack(p_msg, t, branch)!=0)
+		if (!(ack_sent = (send_ack(p_msg, t, branch, &ack_buf)>=0)))
 			LM_ERR("failed to send ACK (local=%s)\n", is_local(t)?"yes":"no");
 	}
 
 	_tm_branch_index = branch;
 
+	if (has_tran_tmcbs( t, TMCB_MSG_MATCHED_IN) )
+		run_trans_callbacks( TMCB_MSG_MATCHED_IN, t, 0,
+			p_msg, p_msg->REPLY_STATUS);
+
+	if (!is_local(t))
+		run_trans_callbacks( TMCB_RESPONSE_IN, t, t->uas.request, p_msg,
+			p_msg->REPLY_STATUS);
+
+	if (ack_sent) {
+		if ( has_tran_tmcbs( t, TMCB_MSG_SENT_OUT) ) {
+			set_extra_tmcb_params( &ack_buf, &t->uac[branch].request.dst);
+			run_trans_callbacks( TMCB_MSG_SENT_OUT,
+				t, t->uas.request, 0, 0);
+		}
+		shm_free(ack_buf.s);
+	}
+
 	/* processing of on_reply block */
 	has_reply_route = (t->on_reply) || (t->uac[branch].on_reply);
 	if (has_reply_route) {
@@ -1444,10 +1621,13 @@ int reply_received( struct sip_msg  *p_msg )
 		/* transfer transaction flag to branch context */
 		p_msg->flags = t->uas.request ? t->uas.request->flags : 0;
 		setb0flags( p_msg, t->uac[branch].br_flags);
+
+		swap_route_type(old_route_type, ONREPLY_ROUTE);
 		/* run block - first per branch and then global one */
-		if ( t->uac[branch].on_reply &&
-		(run_top_route(onreply_rlist[t->uac[branch].on_reply].a,p_msg)
+		if ( ref_script_route_check_and_update(t->uac[branch].on_reply) &&
+		(run_top_route(sroutes->onreply[t->uac[branch].on_reply->idx],p_msg)
 		&ACT_FL_DROP) && (msg_status<200) ) {
+			set_route_type(old_route_type);
 			if (onreply_avp_mode) {
 				UNLOCK_REPLIES( t );
 				set_avp_list( backup_list );
@@ -1455,8 +1635,10 @@ int reply_received( struct sip_msg  *p_msg )
 			LM_DBG("dropping provisional reply %d\n", msg_status);
 			goto done;
 		}
-		if ( t->on_reply && (run_top_route(onreply_rlist[t->on_reply].a,p_msg)
+		if ( ref_script_route_check_and_update(t->on_reply) &&
+		(run_top_route(sroutes->onreply[t->on_reply->idx],p_msg)
 		&ACT_FL_DROP) && (msg_status<200) ) {
+			set_route_type(old_route_type);
 			if (onreply_avp_mode) {
 				UNLOCK_REPLIES( t );
 				set_avp_list( backup_list );
@@ -1464,6 +1646,7 @@ int reply_received( struct sip_msg  *p_msg )
 			LM_DBG("dropping provisional reply %d\n", msg_status);
 			goto done;
 		}
+		set_route_type(old_route_type);
 		/* transfer current message context back to t */
 		t->uac[branch].br_flags = getb0flags(p_msg);
 		if (t->uas.request)
@@ -1556,67 +1739,8 @@ not_found:
 	return 1;
 }
 
-int w_t_reply_body(struct sip_msg* msg, str* code, str *text,
-								str *body)
-{
-	struct cell *t;
-	int r;
-	str code_s;
-	unsigned int code_i;
-	str body_s;
-
-	if(body==0)
-	{
-		LM_ERR("Wrong argument, body must not be NULL\n");
-		return -1;
-	}
-
-	if(((pv_elem_p)code)->spec.getf!=NULL) {
-		if(pv_printf_s(msg, (pv_elem_p)code, &code_s)!=0)
-			return -1;
-		if(str2int(&code_s, &code_i)!=0 || code_i<100 || code_i>699)
-			return -1;
-	} else {
-		code_i = ((pv_elem_p)code)->spec.pvp.pvn.u.isname.name.n;
-	}
-
-	if(((pv_elem_p)text)->spec.getf!=NULL) {
-		if(pv_printf_s(msg, (pv_elem_p)text, &code_s)!=0 || code_s.len <=0)
-			return -1;
-	} else {
-		code_s = ((pv_elem_p)text)->text;
-	}
-
-	if(((pv_elem_p)body)->spec.getf!=NULL) {
-		if(pv_printf_s(msg, (pv_elem_p)body, &body_s)!=0 || body_s.len <=0)
-			return -1;
-	} else {
-		body_s = ((pv_elem_p)body)->text;
-	}
-
-	t=get_t();
-	if ( t==0 || t==T_UNDEFINED ) {
-		/* t_reply_with_body() is a bit of a weird function as it 
-		 * receiving as parameter the actual msg, but the transaction
-		 * (and uses the saved msg from transaction).
-		 * So we need to take care and save everything into transaction,
-		 * otherwise we will loose the rpl lumps. --bogdan */
-		r = t_newtran( msg, 1/*full uas cloning*/ );
-		if (r==0) {
-			/* retransmission -> break the script */
-			return 0;
-		} else if (r<0) {
-			LM_ERR("could not create a new transaction\n");
-			return -1;
-		}
-		t=get_t();
-	}
-	return t_reply_with_body(t, code_i, &code_s, &body_s, 0, 0);
-}
-
-
-int t_reply_with_body( struct cell *trans, unsigned int code, str *text,
-									str *body, str *new_header, str *to_tag )
+static int _reply_with_body( struct cell *trans, unsigned int code, str *text,
+						str *body, str *new_header, str *to_tag, int lock_replies)
 {
 	struct lump_rpl *hdr_lump;
 	struct lump_rpl *body_lump;
@@ -1649,6 +1773,11 @@ int t_reply_with_body( struct cell *trans, unsigned int code, str *text,
 	} else {
 		body_lump = 0;
 	}
+	if (has_tran_tmcbs(trans, TMCB_LOCAL_RESPONSE)) {
+		set_extra_tmcb_params(body, NULL);
+		run_trans_callbacks(TMCB_LOCAL_RESPONSE,
+			trans, p_msg, 0, code);
+	}
 
 	if(to_tag && to_tag->len) {
 		rpl.s = build_res_buf_from_sip_req(code, text, to_tag, p_msg,
@@ -1658,7 +1787,7 @@ int t_reply_with_body( struct cell *trans, unsigned int code, str *text,
 	else
 	if (code>=180 && p_msg->to && (get_to(p_msg)->tag_value.s==0
 			|| get_to(p_msg)->tag_value.len==0)) {
-		calc_crc_suffix( p_msg, tm_tag_suffix );
+		calc_tag_suffix( p_msg, tm_tag_suffix );
 		rpl.s = build_res_buf_from_sip_req(code,text, &tm_tag, p_msg,
 				(unsigned int*)&rpl.len, &bm);
 		to_tag_rpl.s = tm_tag.s;
@@ -1685,7 +1814,7 @@ int t_reply_with_body( struct cell *trans, unsigned int code, str *text,
 		goto error;
 	}
 	ret=_reply_light( trans, rpl.s, rpl.len, code, to_tag_rpl.s, to_tag_rpl.len,
-			1 /* lock replies */, &bm );
+			lock_replies, &bm );
 
 	/* mark the transaction as replied */
 	if (code>=200) set_kr(REQ_RPLD);
@@ -1700,3 +1829,59 @@ error:
 	return -1;
 }
 
+
+int w_t_reply_body(struct sip_msg* msg, unsigned int* code, str *text,
+				str *body)
+{
+	struct cell *t;
+	int r, lock_replies = 1;
+
+	if (msg->REQ_METHOD==METHOD_ACK) {
+		LM_DBG("ACKs are not replied\n");
+		return 0;
+	}
+
+	switch (route_type) {
+		case FAILURE_ROUTE:
+			t=get_t();
+			if ( t==0 || t==T_UNDEFINED ) {
+				LM_BUG("no transaction found in Failure Route\n");
+				return -1;
+			}
+			lock_replies = 0;
+			break;
+		case REQUEST_ROUTE:
+			t=get_t();
+			if ( t==0 || t==T_UNDEFINED ) {
+				/* t_reply_with_body() is a bit of a weird function as it is
+				 * receiving as parameter the actual msg, but the transaction
+				 * (and uses the saved msg from transaction).
+				 * So we need to take care and save everything into transaction,
+				 * otherwise we will loose the rpl lumps. --bogdan */
+				r = t_newtran( msg, 1/*full uas cloning*/ );
+				if (r==0) {
+					/* retransmission -> break the script */
+					return 0;
+				} else if (r<0) {
+					LM_ERR("could not create a new transaction\n");
+					return -1;
+				}
+				t=get_t();
+			} else {
+				update_cloned_msg_from_msg( t->uas.request, msg);
+			}
+			break;
+		default:
+			LM_CRIT("unsupported route_type (%d)\n", route_type);
+			return -1;
+	}
+	return _reply_with_body(t, *code, text, body, 0, 0, lock_replies);
+}
+
+
+int t_reply_with_body( struct cell *trans, unsigned int code, str *text,
+									str *body, str *new_header, str *to_tag)
+{
+	return _reply_with_body(trans, code, text, body, new_header,
+			to_tag, 1 /* lock replies */);
+}
