@@ -35,9 +35,12 @@
 #include "../../parser/parse_authenticate.h"
 #include "../../parser/contact/parse_contact.h"
 #include "../../parser/parse_min_expires.h"
+#include "../../parser/parse_expires.h"
 #include "../uac_auth/uac_auth.h"
+#include "../../lib/digest_auth/digest_auth.h"
 #include "reg_records.h"
 #include "reg_db_handler.h"
+#include "clustering.h"
 
 
 #define UAC_REGISTRAR_URI_PARAM              1
@@ -60,6 +63,8 @@
 #define UAC_REG_INTERNAL_ERROR_STATE    "INTERNAL_ERROR_STATE"
 #define UAC_REG_WRONG_CREDENTIALS_STATE "WRONG_CREDENTIALS_STATE"
 #define UAC_REG_REGISTRAR_ERROR_STATE   "REGISTRAR_ERROR_STATE"
+#define UAC_REG_UNREGISTERING_STATE		"UNREGISTERING_STATE"
+#define UAC_REG_AUTHENTICATING_UNREGISTER_STATE	"AUTHENTICATING_UNREGISTER_STATE"
 
 const str uac_reg_state[]={
 	str_init(UAC_REG_NOT_REGISTERED_STATE),
@@ -70,6 +75,8 @@ const str uac_reg_state[]={
 	str_init(UAC_REG_INTERNAL_ERROR_STATE),
 	str_init(UAC_REG_WRONG_CREDENTIALS_STATE),
 	str_init(UAC_REG_REGISTRAR_ERROR_STATE),
+	str_init(UAC_REG_UNREGISTERING_STATE),
+	str_init(UAC_REG_AUTHENTICATING_UNREGISTER_STATE),
 };
 
 /** Functions declarations */
@@ -78,10 +85,24 @@ static void mod_destroy(void);
 static int child_init(int rank);
 
 void timer_check(unsigned int ticks, void* hash_counter);
+void handle_shtag_change(str *tag_name, int state, int c_id, void *param);
+
+static mi_response_t *mi_reg_list(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_reg_list_record(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_reg_reload(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_reg_reload_record(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_reg_enable(const mi_params_t *params,
+								struct mi_handler *async_hdl);
+static mi_response_t *mi_reg_disable(const mi_params_t *params,
+								struct mi_handler *async_hdl);
 
-static struct mi_root* mi_reg_list(struct mi_root* cmd, void* param);
-static struct mi_root* mi_reg_reload(struct mi_root* cmd, void* param);
 int send_register(unsigned int hash_index, reg_record_t *rec, str *auth_hdr);
+int send_unregister(unsigned int hash_index, reg_record_t *rec, str *auth_hdr,
+	unsigned int all_contacts);
 
 
 /** Global variables */
@@ -114,20 +135,14 @@ typedef struct reg_tm_cb {
 	reg_record_t *uac;
 }reg_tm_cb_t;
 
-/** Exported functions */
-static cmd_export_t cmds[]=
-{
-	{0,0,0,0,0,0}
-};
-
-
 /** Exported parameters */
-static param_export_t params[]= {
+static const param_export_t params[]= {
 	{"hash_size",		INT_PARAM,			&reg_hsize},
 	{"default_expires",	INT_PARAM,			&default_expires},
 	{"timer_interval",	INT_PARAM,			&timer_interval},
+	{"enable_clustering",	INT_PARAM,			&enable_clustering},
 	{"db_url",		STR_PARAM,			&db_url.s},
-	{"table_name",		STR_PARAM,			&reg_table_name},
+	{"table_name",		STR_PARAM,			&reg_table_name.s},
 	{"registrar_column",	STR_PARAM,			&registrar_column.s},
 	{"proxy_column",	STR_PARAM,			&proxy_column.s},
 	{"aor_column",		STR_PARAM,			&aor_column.s},
@@ -138,24 +153,43 @@ static param_export_t params[]= {
 	{"binding_params_column",	STR_PARAM,	&binding_params_column.s},
 	{"expiry_column",	STR_PARAM,		&expiry_column.s},
 	{"forced_socket_column",	STR_PARAM,	&forced_socket_column.s},
+	{"cluster_shtag_column",	STR_PARAM,	&cluster_shtag_column.s},
+	{"state_column",	STR_PARAM,		&state_column.s},
 	{0,0,0}
 };
 
 
 /** MI commands */
-static mi_export_t mi_cmds[] = {
-	{"reg_list",   0, mi_reg_list,   0, 0, 0},
-	{"reg_reload", 0, mi_reg_reload, 0,	0, 0},
-	{0,            0, 0,             0, 0, 0}
+static const mi_export_t mi_cmds[] = {
+	{ "reg_list", 0, 0, 0, {
+		{mi_reg_list, {0}},
+		{mi_reg_list_record, {"aor", "contact", "registrar", 0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{ "reg_reload", 0, 0, 0, {
+		{mi_reg_reload, {0}},
+		{mi_reg_reload_record, {"aor", "contact", "registrar", 0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{ "reg_enable", 0, 0, 0, {
+		{mi_reg_enable, {"aor", "contact", "registrar", 0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{ "reg_disable", 0, 0, 0, {
+		{mi_reg_disable, {"aor", "contact", "registrar", 0}},
+		{EMPTY_MI_RECIPE}}
+	},
+	{EMPTY_MI_EXPORT}
 };
 
-static dep_export_t deps = {
+static const dep_export_t deps = {
 	{ /* OpenSIPS module dependencies */
 		{ MOD_TYPE_DEFAULT, "tm",       DEP_ABORT },
 		{ MOD_TYPE_DEFAULT, "uac_auth", DEP_ABORT },
 		{ MOD_TYPE_NULL, NULL, 0 },
 	},
 	{ /* modparam dependencies */
+		{ "enable_clustering", get_deps_clusterer},
 		{ NULL, NULL },
 	},
 };
@@ -166,18 +200,22 @@ struct module_exports exports= {
 	MOD_TYPE_DEFAULT,       /* class of this module */
 	MODULE_VERSION,			/* module version */
 	DEFAULT_DLFLAGS,		/* dlopen flags */
-	&deps,                  /* OpenSIPS module dependencies */
-	cmds,				/* exported functions */
+	0,						/* load function */
+	&deps,					/* OpenSIPS module dependencies */
+	0,					/* exported functions */
 	0,					/* exported async functions */
 	params,				/* exported parameters */
 	NULL,				/* exported statistics */
 	mi_cmds,			/* exported MI functions */
 	NULL,				/* exported pseudo-variables */
-	0,				/* extra processes */
+	0,					/* exported transformations */
+	0,					/* extra processes */
+	0,					/* module pre-initialization function */
 	mod_init,			/* module initialization function */
 	(response_function) NULL,	/* response handling function */
 	(destroy_function) mod_destroy,	/* destroy function */
-	child_init			/* per-child init function */
+	child_init,			/* per-child init function */
+	0					/* reload confirm function */
 };
 
 
@@ -198,6 +236,11 @@ static int mod_init(void)
 		return -1;
 	}
 
+	if (enable_clustering && ureg_init_cluster( handle_shtag_change )<0) {
+		LM_ERR("failed to initialized clustering support\n");
+		return -1;
+	}
+
 	if(default_expires<15){
 		LM_ERR("default_expires to short: [%d]<15\n", default_expires);
 		return -1;
@@ -228,6 +271,8 @@ static int mod_init(void)
 	binding_params_column.len = strlen(binding_params_column.s);
 	expiry_column.len = strlen(expiry_column.s);
 	forced_socket_column.len = strlen(forced_socket_column.s);
+	cluster_shtag_column.len = strlen(cluster_shtag_column.s);
+	state_column.len = strlen(state_column.s);
 	init_db_url(db_url , 0 /*cannot be null*/);
 	if (init_reg_db(&db_url) != 0) {
 		LM_ERR("failed to initialize the DB support\n");
@@ -248,7 +293,7 @@ static int mod_init(void)
 		register_timer("uac_reg_check", timer_check, (void*)(long)param, _timer,
 			TIMER_FLAG_DELAY_ON_DELAY);
 	} else {
-		LM_ERR("timer_interval=[%d] MUST be bigger then reg_hsize=[%d]\n",
+		LM_ERR("timer_interval=[%d] MUST be at least as big as reg_hsize=[%d]\n",
 			timer_interval, reg_hsize);
 		return -1;
 	}
@@ -275,8 +320,6 @@ static int child_init(int rank)
 }
 
 
-void shm_free_param(void* param) {shm_free(param);}
-
 struct reg_tm_cback_data {
 	struct cell *t;
 	struct tmcb_params *ps;
@@ -284,23 +327,49 @@ struct reg_tm_cback_data {
 	reg_tm_cb_t *cb_param;
 };
 
+int get_cur_time_s(str *str_now, time_t now)
+{
+	char *p;
+	int len;
+
+	p = int2str((unsigned long)(now), &len);
+	if (p && len>0) {
+		str_now->s = (char *)pkg_malloc(len);
+		if (str_now->s) {
+			memcpy(str_now->s, p, len);
+			str_now->len = len;
+		} else {
+			LM_ERR("oom\n");
+			return -1;
+		}
+	} else {
+		return -1;
+	}
+
+	return 0;
+}
+
 int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 {
 	struct sip_msg *msg;
+	str msg_body;
 	int statuscode = 0;
 	unsigned int exp = 0;
+	exp_body_t *header_exp = NULL;
+	unsigned int bindings_counter = 0;
 	reg_record_t *rec = (reg_record_t*)e_data;
 	struct hdr_field *c_ptr, *head_contact;
 	struct uac_credential crd;
 	contact_t *contact;
 	struct authenticate_body *auth = NULL;
 	static struct authenticate_nc_cnonce auth_nc_cnonce;
-	HASHHEX response;
+	struct digest_auth_response response;
 	str *new_hdr;
 	struct reg_tm_cback_data *tm_cback_data = (struct reg_tm_cback_data*)data;
 	struct cell *t;
 	struct tmcb_params *ps;
 	time_t now;
+	str str_now = {NULL, 0};
 	reg_tm_cb_t *cb_param;
 
 	cb_param = tm_cback_data->cb_param;
@@ -344,9 +413,72 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 				c_ptr = c_ptr->next;
 			}
 		} else {
-			LM_ERR("No contact header in received 200ok\n");
-			goto done;
+			switch(rec->state) {
+			case UNREGISTERING_STATE:
+			case AUTHENTICATING_UNREGISTER_STATE:
+				if (!(rec->flags&REG_ENABLED)) {
+					/* succesfully unREGISTERED */
+					rec->state = NOT_REGISTERED_STATE;
+					rec->registration_timeout = 0;
+				} else {
+					/* registrant got enabled while waiting for a reply to
+					 * a previous unregister */
+					if (get_cur_time_s(&str_now, now) < 0) {
+						LM_ERR("Failed to get current time string\n");
+						goto done;
+					}
+					new_call_id_ftag_4_record(rec, &str_now);
+					if(send_register(cb_param->hash_index, rec, NULL)==1) {
+						rec->last_register_sent = now;
+						rec->state = REGISTERING_STATE;
+					} else {
+						rec->registration_timeout = now + rec->expires - timer_interval;
+						rec->state = INTERNAL_ERROR_STATE;
+					}
+				}
+
+				break;
+			default:
+				LM_ERR("No contact header in received 200ok in state [%d]\n",
+					rec->state);
+				goto done;
+			}
+			break; /* done with 200ok handling */
 		}
+
+		if (rec->flags&FORCE_SINGLE_REGISTRATION &&
+			(rec->state == REGISTERING_STATE ||
+			 rec->state == AUTHENTICATING_STATE)) {
+			head_contact = msg->contact;
+			contact = ((contact_body_t*)msg->contact->parsed)->contacts;
+			while (contact) {
+				bindings_counter++;
+				if (contact->next == NULL) {
+					contact = NULL;
+					c_ptr = head_contact->next;
+					while(c_ptr) {
+						if (c_ptr->type == HDR_CONTACT_T) {
+							head_contact = c_ptr;
+							contact = ((contact_body_t*)c_ptr->parsed)->contacts;
+							break;
+						}
+						c_ptr = c_ptr->next;
+					}
+				} else {
+					contact = contact->next;
+				}
+			}
+			if (bindings_counter>1) {
+				LM_DBG("got [%d] bindings\n", bindings_counter);
+				if(send_unregister(cb_param->hash_index, rec, NULL, 1)==1) {
+					rec->state = UNREGISTERING_STATE;
+				} else {
+					rec->state = INTERNAL_ERROR_STATE;
+				}
+				break;
+			}
+		}
+
 		head_contact = msg->contact;
 		contact = ((contact_body_t*)msg->contact->parsed)->contacts;
 		while (contact) {
@@ -381,15 +513,62 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 				contact = contact->next;
 			}
 		}
-		rec->state = REGISTERED_STATE;
-		if (exp) rec->expires = exp;
-		if (rec->expires <= timer_interval) {
-			LM_ERR("Please decrease timer_interval=[%u]"
-				" - imposed server expires [%u] to small for AOR=[%.*s]\n",
-				timer_interval, rec->expires,
-				rec->td.rem_uri.len, rec->td.rem_uri.s);
+		// if there is no expires in contact, try parse expires from header
+		if (exp == 0 && msg->expires) {
+			if (parse_expires(msg->expires) < 0) {
+				LM_ERR("failed to parse Expires header\n");
+			} else {
+				header_exp = (exp_body_t*)msg->expires->parsed;
+				exp = header_exp->val;
+			}
 		}
-		rec->registration_timeout = now + rec->expires - timer_interval;
+
+		if (!(rec->flags&REG_ENABLED)) {
+			if (rec->state==REGISTERING_STATE ||
+				rec->state==AUTHENTICATING_STATE) {
+				/* registrant got disabled while waiting for a reply to
+				 * a previous register */
+				if(send_unregister(cb_param->hash_index, rec, NULL, 0)==1) {
+					rec->state = UNREGISTERING_STATE;
+				} else {
+					rec->state = INTERNAL_ERROR_STATE;
+				}
+			} else {
+				/* succesfully unREGISTERED */
+				rec->state = NOT_REGISTERED_STATE;
+				rec->registration_timeout = 0;
+			}
+		} else {
+			if (rec->state == UNREGISTERING_STATE ||
+				rec->state == AUTHENTICATING_UNREGISTER_STATE) {
+				/* registrant got enabled while waiting for a reply to
+				 * a previous unregister */
+				if (get_cur_time_s(&str_now, now) < 0) {
+					LM_ERR("Failed to get current time string\n");
+					goto done;
+				}
+				new_call_id_ftag_4_record(rec, &str_now);
+				if(send_register(cb_param->hash_index, rec, NULL)==1) {
+					rec->last_register_sent = now;
+					rec->state = REGISTERING_STATE;
+				} else {
+					rec->registration_timeout = now + rec->expires - timer_interval;
+					rec->state = INTERNAL_ERROR_STATE;
+				}
+			} else {
+				/* succesfully REGISTERED */
+				rec->state = REGISTERED_STATE;
+				if (exp) rec->expires = exp;
+				if (rec->expires <= timer_interval) {
+					LM_ERR("Please decrease timer_interval=[%u]"
+						" - imposed server expires [%u] to small for AOR=[%.*s]\n",
+						timer_interval, rec->expires,
+						rec->td.rem_uri.len, rec->td.rem_uri.s);
+				}
+				rec->registration_timeout = now + rec->expires - timer_interval;
+			}
+		}
+
 		break;
 
 	case WWW_AUTH_CODE:
@@ -410,11 +589,11 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 		}
 
 		if (statuscode==WWW_AUTH_CODE) {
-			if (0 == parse_www_authenticate_header(msg))
-				auth = get_www_authenticate(msg);
+			parse_www_authenticate_header(msg,
+			    DAUTH_AHFM_ANYSUP, &auth);
 		} else if (statuscode==PROXY_AUTH_CODE) {
-			if (0 == parse_proxy_authenticate_header(msg))
-				auth = get_proxy_authenticate(msg);
+			parse_proxy_authenticate_header(msg,
+			    DAUTH_AHFM_ANYSUP, &auth);
 		}
 		if (auth == NULL) {
 			LM_ERR("Unable to extract authentication info\n");
@@ -431,8 +610,10 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 
 		switch(rec->state) {
 		case REGISTERING_STATE:
+		case UNREGISTERING_STATE:
 			break;
 		case AUTHENTICATING_STATE:
+		case AUTHENTICATING_UNREGISTER_STATE:
 			/* We already sent an authenticated REGISTER and we are still challanged! */
 			LM_WARN("Wrong credentials for [%.*s]\n",
 				rec->td.rem_uri.len, rec->td.rem_uri.s);
@@ -448,27 +629,50 @@ int run_reg_tm_cback(void *e_data, void *data, void *r_data)
 
 		/* perform authentication */
 		if (auth->realm.s && auth->realm.len) {
-			crd.realm.s = auth->realm.s; crd.realm.len = auth->realm.len;
+			crd.auth_data.realm = auth->realm;
 		} else {
 			LM_ERR("No realm found\n");
 			goto done;
 		}
-		crd.user.s = rec->auth_user.s; crd.user.len = rec->auth_user.len;
-		crd.passwd.s = rec->auth_password.s; crd.passwd.len = rec->auth_password.len;
+		crd.auth_data.user = rec->auth_user;
+		crd.auth_data.passwd = rec->auth_password;
+
+		if ((auth->flags & QOP_AUTH_INT) && get_body(msg, &msg_body) < 0) {
+			LM_ERR("Failed to get message body\n");
+			goto done;
+		}
 
 		memset(&auth_nc_cnonce, 0, sizeof(struct authenticate_nc_cnonce));
-		uac_auth_api._do_uac_auth(&register_method, &rec->td.rem_target, &crd,
-					auth, &auth_nc_cnonce, response);
+		if (uac_auth_api._do_uac_auth(&msg_body, &register_method,
+		    &rec->td.rem_target, &crd, auth, &auth_nc_cnonce, &response) != 0) {
+			LM_ERR("Failed in do_uac_auth()\n");
+			goto done;
+		}
 		new_hdr = uac_auth_api._build_authorization_hdr(statuscode, &rec->td.rem_target,
-					&crd, auth, &auth_nc_cnonce, response);
+					&crd, auth, &auth_nc_cnonce, &response);
 		if (!new_hdr) {
 			LM_ERR("failed to build authorization hdr\n");
 			goto done;
 		}
-		if(send_register(cb_param->hash_index, rec, new_hdr)==1) {
-			rec->state = AUTHENTICATING_STATE;
-		} else {
-			rec->state = INTERNAL_ERROR_STATE;
+		switch(rec->state) {
+		case REGISTERING_STATE:
+			if(send_register(cb_param->hash_index, rec, new_hdr)==1) {
+				rec->state = AUTHENTICATING_STATE;
+			} else {
+				rec->state = INTERNAL_ERROR_STATE;
+			}
+			break;
+		case UNREGISTERING_STATE:
+			if(send_unregister(cb_param->hash_index, rec, new_hdr, 0)==1) {
+				rec->state = AUTHENTICATING_UNREGISTER_STATE;
+			} else {
+				rec->state = INTERNAL_ERROR_STATE;
+			}
+			break;
+		default:
+			LM_ERR("Unexpected [%d] notification cb in state [%d]\n",
+				statuscode, rec->state);
+			goto done;
 		}
 		pkg_free(new_hdr->s);
 		new_hdr->s = NULL; new_hdr->len = 0;
@@ -615,6 +819,87 @@ int send_register(unsigned int hash_index, reg_record_t *rec, str *auth_hdr)
 	}
 	extra_hdrs.len = (int)(p - extra_hdrs.s);
 
+	LM_DBG("extra_hdrs=[%p][%d]->[%.*s]\n",
+		extra_hdrs.s, extra_hdrs.len, extra_hdrs.len, extra_hdrs.s);
+
+	if ( !push_new_global_context() ) {
+
+		LM_ERR("failed to alloc new ctx in pkg\n");
+		result = 0;
+
+	} else {
+
+		/* reset the new to-be-used CTX */
+		memset( current_processing_ctx, 0, context_size(CONTEXT_GLOBAL) );
+
+		/* send the request within the new context */
+		result=tmb.t_request_within(
+			&register_method,	/* method */
+			&extra_hdrs,		/* extra headers*/
+			NULL,			/* body */
+			&rec->td,		/* dialog structure*/
+			reg_tm_cback,		/* callback function */
+			(void *)cb_param,	/* callback param */
+			osips_shm_free);	/* function to release the parameter */
+
+		pop_pushed_global_context();
+	}
+
+	if (result < 1)
+		shm_free(cb_param);
+
+	LM_DBG("result=[%d]\n", result);
+	return result;
+}
+
+int send_unregister(unsigned int hash_index, reg_record_t *rec, str *auth_hdr,
+	unsigned int all_contacts)
+{
+	int result;
+	reg_tm_cb_t *cb_param;
+	char *p;
+
+	/* Allocate space for tm callback params */
+	cb_param = shm_malloc(sizeof(reg_tm_cb_t));
+	if (!cb_param) {
+		LM_ERR("oom\n");
+		return -1;
+	}
+	cb_param->hash_index = hash_index;
+	cb_param->uac = rec;
+
+	p = extra_hdrs.s;
+	memcpy(p, contact_hdr.s, contact_hdr.len);
+	p += contact_hdr.len;
+	if (all_contacts) {
+		*p = '*'; p++;
+		memcpy(p, CRLF, CRLF_LEN); p += CRLF_LEN;
+
+		/* adding exires header */
+		memcpy(p, expires_hdr.s, expires_hdr.len);
+		p += expires_hdr.len;
+		*p = '0'; p++;
+		memcpy(p, CRLF, CRLF_LEN); p += CRLF_LEN;
+	} else {
+		*p = '<'; p++;
+		memcpy(p, rec->contact_uri.s, rec->contact_uri.len);
+		p += rec->contact_uri.len;
+		*p = '>'; p++;
+		memcpy(p, rec->contact_params.s, rec->contact_params.len);
+		p += rec->contact_params.len;
+		/* adding exiration time as a parameter */
+		memcpy(p, expires_param.s, expires_param.len);
+		p += expires_param.len;
+		*p = '0'; p++;
+		memcpy(p, CRLF, CRLF_LEN); p += CRLF_LEN;
+	}
+
+	if (auth_hdr) {
+		memcpy(p, auth_hdr->s, auth_hdr->len);
+		p += auth_hdr->len;
+	}
+	extra_hdrs.len = (int)(p - extra_hdrs.s);
+
 	LM_DBG("extra_hdrs=[%p][%d]->[%.*s]\n",
 		extra_hdrs.s, extra_hdrs.len, extra_hdrs.len, extra_hdrs.s);
 
@@ -625,12 +910,15 @@ int send_register(unsigned int hash_index, reg_record_t *rec, str *auth_hdr)
 		&rec->td,		/* dialog structure*/
 		reg_tm_cback,		/* callback function */
 		(void *)cb_param,	/* callback param */
-		shm_free_param);	/* function to release the parameter */
+		osips_shm_free);	/* function to release the parameter */
+
+	if (result < 1)
+		shm_free(cb_param);
+
 	LM_DBG("result=[%d]\n", result);
 	return result;
 }
 
-
 struct timer_check_data {
 	time_t now;
 	str *s_now;
@@ -645,22 +933,35 @@ int run_timer_check(void *e_data, void *data, void *r_data)
 	str *s_now = t_check_data->s_now;
 	unsigned int i = t_check_data->hash_counter;
 
+	if (!ureg_cluster_shtag_is_active( &rec->cluster_shtag, rec->cluster_id))
+		return 0;
+
 	switch(rec->state){
 	case REGISTERING_STATE:
+	case UNREGISTERING_STATE:
 	case AUTHENTICATING_STATE:
+	case AUTHENTICATING_UNREGISTER_STATE:
 		break;
 	case WRONG_CREDENTIALS_STATE:
 	case REGISTER_TIMEOUT_STATE:
 	case INTERNAL_ERROR_STATE:
 	case REGISTRAR_ERROR_STATE:
 		reg_print_record(rec);
-		new_call_id_ftag_4_record(rec, s_now);
-		if(send_register(i, rec, NULL)==1) {
-			rec->last_register_sent = now;
-			rec->state = REGISTERING_STATE;
+		if (rec->flags&REG_ENABLED) {
+			new_call_id_ftag_4_record(rec, s_now);
+			if(send_register(i, rec, NULL)==1) {
+				rec->last_register_sent = now;
+				rec->state = REGISTERING_STATE;
+			} else {
+				rec->registration_timeout = now + rec->expires - timer_interval;
+				rec->state = INTERNAL_ERROR_STATE;
+			}
 		} else {
-			rec->registration_timeout = now + rec->expires - timer_interval;
-			rec->state = INTERNAL_ERROR_STATE;
+			if(send_unregister(i, rec, NULL, 0)==1) {
+				rec->state = UNREGISTERING_STATE;
+			} else {
+				rec->state = INTERNAL_ERROR_STATE;
+			}
 		}
 		break;
 	case REGISTERED_STATE:
@@ -669,12 +970,14 @@ int run_timer_check(void *e_data, void *data, void *r_data)
 			break;
 		}
 	case NOT_REGISTERED_STATE:
-		if(send_register(i, rec, NULL)==1) {
-			rec->last_register_sent = now;
-			rec->state = REGISTERING_STATE;
-		} else {
-			rec->registration_timeout = now + rec->expires - timer_interval;
-			rec->state = INTERNAL_ERROR_STATE;
+		if (rec->flags&REG_ENABLED) {
+			if(send_register(i, rec, NULL)==1) {
+				rec->last_register_sent = now;
+				rec->state = REGISTERING_STATE;
+			} else {
+				rec->registration_timeout = now + rec->expires - timer_interval;
+				rec->state = INTERNAL_ERROR_STATE;
+			}
 		}
 		break;
 	default:
@@ -684,29 +987,20 @@ int run_timer_check(void *e_data, void *data, void *r_data)
 	return 0; /* continue list traversal */
 }
 
-
 void timer_check(unsigned int ticks, void* hash_counter)
 {
 	unsigned int i=*(unsigned*)(unsigned long*)hash_counter;
-	char *p;
-	int len, ret;
+	int ret;
 	time_t now;
 	str str_now = {NULL, 0};
 	struct timer_check_data t_check_data;
 
-	now = time(0);
 	*(unsigned*)(unsigned long*)hash_counter = (i+1)%reg_hsize;
 
-	p = int2str((unsigned long)(time(0)), &len);
-	if (p && len>0) {
-		str_now.s = (char *)pkg_malloc(len);
-		if (str_now.s) {
-			memcpy(str_now.s, p, len);
-			str_now.len = len;
-		} else {
-			LM_ERR("oom\n");
-			return;
-		}
+	now = time(0);
+	if (get_cur_time_s(&str_now, now) < 0) {
+		LM_ERR("Failed to get current time string\n");
+		return;
 	}
 
 	/* Initialize slinkedl run traversal data */
@@ -727,111 +1021,268 @@ void timer_check(unsigned int ticks, void* hash_counter)
 }
 
 
-/*** MI **/
+struct shtag_check_data {
+	str *tag;
+	int c_id;
+};
 
+static int cluster_shtag_check(void *e_data, void *data, void *r_data)
+{
+	reg_record_t *rec = (reg_record_t*)e_data;
+	struct shtag_check_data *shtag_data = (struct shtag_check_data*)data;
+
+	LM_DBG("checking record with cluster_id [%d] and shtag [%.*s]\n",
+		rec->cluster_id, rec->cluster_shtag.len, rec->cluster_shtag.s);
+	if (rec->cluster_id==shtag_data->c_id &&
+		rec->cluster_shtag.s && rec->cluster_shtag.len &&
+		rec->cluster_shtag.len==shtag_data->tag->len &&
+		0==memcmp(rec->cluster_shtag.s, shtag_data->tag->s, shtag_data->tag->len)) {
+		/* this record matches the shtag + cluster_id, so we need to de-active it */
+		LM_DBG("Moving record to NOT_REGISTERED_STATE\n");
+		rec->state = NOT_REGISTERED_STATE;
+	}
+
+
+	return 0;
+}
+
+
+void handle_shtag_change(str *tag_name, int state, int c_id, void *param)
+{
+	struct shtag_check_data shtag_data;
+	int ret, i;
+
+	if (state!=SHTAG_STATE_BACKUP)
+		return;
+
+	shtag_data.c_id = c_id;
+	shtag_data.tag = tag_name;
+
+	/* a shatg in cluster became backup on local node-> check if
+	 * one of our uac reg depends on it */
+	LM_DBG("checking for shtag [%.*s] in cluster [%d]\n",
+		shtag_data.tag->len, shtag_data.tag->s, shtag_data.c_id);
+
+	for( i=0 ; i<reg_hsize ; i++) {
+
+		lock_get(&reg_htable[i].lock);
+		ret = slinkedl_traverse(reg_htable[i].p_list, &cluster_shtag_check,
+							(void*)&shtag_data, NULL);
+		if (ret<0) LM_CRIT("Unexpected return code %d\n", ret);
+		lock_release(&reg_htable[i].lock);
+
+	}
+}
+
+
+/*** MI **/
 int run_mi_reg_list(void *e_data, void *data, void *r_data)
 {
-	struct mi_root* rpl_tree = (struct mi_root*)data;
-	struct mi_node *node, *node1;
-	struct mi_attr* attr;
 	reg_record_t *rec = (reg_record_t*)e_data;
 	int len;
 	char* p;
+	char cbuf[26];
 	struct ip_addr addr;
+	mi_item_t *arr_obj_item = (mi_item_t *)data;
+	mi_item_t *record_item;
 
-	node = add_mi_node_child(&rpl_tree->node, MI_DUP_VALUE, "AOR", 3,
-							rec->td.rem_uri.s, rec->td.rem_uri.len);
-	if(node == NULL) goto error;
-	p = int2str(rec->expires, &len);
-	attr = add_mi_attr(node, MI_DUP_VALUE, "expires", 7, p, len);
-	if(attr == NULL) goto error;
+	if (MI_ITEM_IS_ARRAY(arr_obj_item)) {
+		record_item = add_mi_object(arr_obj_item, NULL, 0);
+		if (!record_item)
+			goto error;
+	} else
+		record_item = arr_obj_item;
 
-	node1 = add_mi_node_child(node, MI_DUP_VALUE, "state", 5,
-							uac_reg_state[rec->state].s, uac_reg_state[rec->state].len);
-	if(node1 == NULL) goto error;
+	if (add_mi_string(record_item, MI_SSTR("AOR"),
+		rec->td.rem_uri.s, rec->td.rem_uri.len) < 0)
+		goto error;
 
-	p = ctime(&rec->last_register_sent);
-	len = strlen(p)-1;
-	node1 = add_mi_node_child(node, MI_DUP_VALUE, "last_register_sent", 18, p, len);
-	if(node1 == NULL) goto error;
+	if (add_mi_number(record_item, MI_SSTR("expires"), rec->expires) < 0)
+		goto error;
 
-	p = ctime(&rec->registration_timeout);
-	len = strlen(p)-1;
-	node1 = add_mi_node_child(node, MI_DUP_VALUE, "registration_t_out", 18, p, len);
-	if(node1 == NULL) goto error;
+	if (add_mi_string(record_item, MI_SSTR("state"),
+		uac_reg_state[rec->state].s, uac_reg_state[rec->state].len) < 0)
+		goto error;
 
-	node1 = add_mi_node_child(node, MI_DUP_VALUE, "registrar", 9,
-							rec->td.rem_target.s, rec->td.rem_target.len);
-	if(node1 == NULL) goto error;
+	if (add_mi_string_fmt(record_item, MI_SSTR("enabled"),
+		"%s", rec->flags&REG_ENABLED ? "yes" : "no") < 0)
+		goto error;
 
-	node1 = add_mi_node_child(node, MI_DUP_VALUE, "binding", 7,
-							rec->contact_uri.s, rec->contact_uri.len);
-	if(node1 == NULL) goto error;
+	ctime_r(&rec->last_register_sent, cbuf);
+	if (add_mi_string(record_item, MI_SSTR("last_register_sent"),
+		cbuf, strlen(cbuf) - 1) < 0)
+		goto error;
 
-	if(rec->td.loc_uri.s != rec->td.rem_uri.s) {
-		node1 = add_mi_node_child(node, MI_DUP_VALUE,
-								"third_party_registrant", 12,
-								rec->td.loc_uri.s, rec->td.loc_uri.len);
-		if(node1 == NULL) goto error;
-	}
+	ctime_r(&rec->registration_timeout, cbuf);
+	if (add_mi_string(record_item, MI_SSTR("registration_t_out"),
+		cbuf, strlen(cbuf) - 1) < 0)
+		goto error;
 
-	if (rec->td.obp.s && rec->td.obp.len) {
-		node1 = add_mi_node_child(node, MI_DUP_VALUE,
-								"proxy", 5, rec->td.obp.s, rec->td.obp.len);
-		if(node1 == NULL) goto error;
-	}
+	if (add_mi_string(record_item, MI_SSTR("registrar"),
+		rec->td.rem_target.s, rec->td.rem_target.len) < 0)
+		goto error;
+
+	if (add_mi_string(record_item, MI_SSTR("binding"),
+		rec->contact_uri.s, rec->contact_uri.len) < 0)
+		goto error;
+
+	if(rec->contact_params.s && rec->contact_params.len)
+		if (add_mi_string(record_item, MI_SSTR("binding_params"),
+			rec->contact_params.s, rec->contact_params.len) < 0)
+			goto error;
+
+	if(rec->td.loc_uri.s != rec->td.rem_uri.s)
+		if (add_mi_string(record_item, MI_SSTR("third_party_registrant"),
+			rec->td.loc_uri.s, rec->td.loc_uri.len) < 0)
+			goto error;
+
+	if (rec->td.obp.s && rec->td.obp.len)
+		if (add_mi_string(record_item, MI_SSTR("proxy"),
+			rec->td.obp.s, rec->td.obp.len) < 0)
+			goto error;
 
 	switch(rec->td.forced_to_su.s.sa_family) {
 	case AF_UNSPEC:
 		break;
 	case AF_INET:
 	case AF_INET6:
-		node1 = add_mi_node_child(node, MI_DUP_VALUE, "dst_IP", 6,
-			(rec->td.forced_to_su.s.sa_family==AF_INET)?"IPv4":"IPv6", 4);
+		if (add_mi_string(record_item, MI_SSTR("dst_IP"),
+			(rec->td.forced_to_su.s.sa_family==AF_INET)?"IPv4":"IPv6", 4) < 0)
+			goto error;
 		sockaddr2ip_addr(&addr, &rec->td.forced_to_su.s);
 		p = ip_addr2a(&addr);
 		if (p == NULL) goto error;
 		len = strlen(p);
-		attr = add_mi_attr(node1, MI_DUP_VALUE, "ip", 2, p, len);
-		if(attr == NULL) goto error;
+		if (add_mi_string(record_item, MI_SSTR("ip"), p, len) < 0)
+			goto error;
 		break;
 	default:
 		LM_ERR("unexpected sa_family [%d]\n", rec->td.forced_to_su.s.sa_family);
-		node1 = add_mi_node_child(node, MI_DUP_VALUE, "dst_IP", 6, "Error", 5);
-		p = int2str(rec->td.forced_to_su.s.sa_family, &len);
-		attr = add_mi_attr(node, MI_DUP_VALUE, "sa_family", 9, p, len);
-		if(attr == NULL) goto error;
+		if (add_mi_string(record_item, MI_SSTR("dst_IP"), "Error", 5) < 0)
+			goto error;
+
+		if (add_mi_number(record_item, MI_SSTR("sa_family"),
+			rec->td.forced_to_su.s.sa_family) < 0)
+			goto error;
 	}
 
+	if (rec->cluster_shtag.s && rec->cluster_shtag.len) {
+		if (add_mi_string(record_item, MI_SSTR("shtag"),
+			rec->cluster_shtag.s, rec->cluster_shtag.len) < 0)
+			goto error;
+		if (add_mi_number(record_item, MI_SSTR("cluster_id"), rec->cluster_id) < 0)
+			goto error;
+		}
+
 	/* action successfully completed on current list element */
 	return 0; /* continue list traversal */
+
 error:
 	LM_ERR("Unable to create reply\n");
 	return -1; /* exit list traversal */
 }
 
 
-static struct mi_root* mi_reg_list(struct mi_root* cmd, void* param)
+static mi_response_t *mi_reg_list(const mi_params_t *params,
+								struct mi_handler *async_hdl)
 {
-	struct mi_root *rpl_tree;
 	int i, ret;
+	mi_response_t *resp;
+	mi_item_t *resp_obj;
+	mi_item_t *records_arr;
 
-	rpl_tree = init_mi_tree( 200, MI_OK_S, MI_OK_LEN);
-	if (rpl_tree==NULL) return NULL;
-	rpl_tree->node.flags |= MI_IS_ARRAY;
+	resp = init_mi_result_object(&resp_obj);
+	if (!resp)
+		return NULL;
+
+	records_arr = add_mi_array(resp_obj, MI_SSTR("Records"));
+	if (!records_arr)
+		goto error;
 
 	for(i=0; i<reg_hsize; i++) {
 		lock_get(&reg_htable[i].lock);
 		ret = slinkedl_traverse(reg_htable[i].p_list,
-						&run_mi_reg_list, (void*)rpl_tree, NULL);
+						&run_mi_reg_list, (void*)records_arr, NULL);
 		lock_release(&reg_htable[i].lock);
 		if (ret<0) {
 			LM_ERR("Unable to create reply\n");
-			free_mi_tree(rpl_tree);
-			return NULL;
+			goto error;
 		}
 	}
-	return rpl_tree;
+
+	return resp;
+
+error:
+	free_mi_response(resp);
+	return NULL;
+}
+
+int run_mi_reg_list_record(void *e_data, void *data, void *r_data)
+{
+	reg_record_t *rec = (reg_record_t*)e_data;
+	record_coords_t *coords = (record_coords_t *)data;
+
+	if (!str_strcmp(&coords->contact, &rec->contact_uri) &&
+		!str_strcmp(&coords->registrar, &rec->td.rem_target)) {
+		return run_mi_reg_list(rec, coords->extra, NULL) ? -1 : 1;
+	} else
+		return 0;  /* continue search */
+}
+
+static mi_response_t *mi_get_coords(const mi_params_t *params, record_coords_t *coords)
+{
+	if (get_mi_string_param(params, "aor", &coords->aor.s, &coords->aor.len) < 0)
+		return init_mi_param_error();
+	if (get_mi_string_param(params, "contact",
+		&coords->contact.s, &coords->contact.len) < 0)
+		return init_mi_param_error();
+	if (get_mi_string_param(params, "registrar",
+		&coords->registrar.s, &coords->registrar.len) < 0)
+		return init_mi_param_error();
+
+	return NULL;
+}
+
+static mi_response_t *mi_reg_list_record(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	mi_response_t *resp;
+	mi_item_t *resp_obj, *reg_obj;
+	record_coords_t coords;
+	int rc;
+	unsigned int hash_code;
+
+	if ((resp = mi_get_coords(params, &coords)))
+		return resp;
+
+	resp = init_mi_result_object(&resp_obj);
+	if (!resp)
+		return NULL;
+
+	reg_obj = add_mi_object(resp_obj, MI_SSTR("Registrant"));
+	if (!reg_obj)
+		goto error;
+
+	coords.extra = reg_obj;
+
+	hash_code = core_hash(&coords.aor, NULL, reg_hsize);
+
+	lock_get(&reg_htable[hash_code].lock);
+	rc = slinkedl_traverse(reg_htable[hash_code].p_list,
+					&run_mi_reg_list_record, &coords, NULL);
+	lock_release(&reg_htable[hash_code].lock);
+
+	if (rc < 0)
+		goto error;
+	else if (rc == 0) {
+		free_mi_response(resp);
+		return init_mi_error(404, MI_SSTR("No such registrant"));
+	}
+
+	return resp;
+error:
+	free_mi_response(resp);
+	return NULL;
 }
 
 int run_compare_rec(void *e_data, void *data, void *r_data)
@@ -862,15 +1313,12 @@ int run_find_same_rec(void *e_data, void *data, void *r_data)
 	return 0;
 }
 
-static struct mi_root* mi_reg_reload(struct mi_root* cmd, void* param)
+static mi_response_t *mi_reg_reload(const mi_params_t *params,
+								struct mi_handler *async_hdl)
 {
-	struct mi_root *rpl_tree;
 	int i;
 	int err = 0;
 
-	rpl_tree = init_mi_tree( 200, MI_OK_S, MI_OK_LEN);
-	if (rpl_tree==NULL) return NULL;
-
 	for(i=0; i<reg_hsize; i++) {
 		lock_get(&reg_htable[i].lock);
 		if (reg_htable[i].s_list!=NULL) {
@@ -887,9 +1335,8 @@ static struct mi_root* mi_reg_reload(struct mi_root* cmd, void* param)
 		if (err) goto error;
 	}
 	/* Load registrants into the secondary list */
-	if(load_reg_info_from_db(1) !=0){
+	if(load_reg_info_from_db(REG_DB_RELOAD, NULL) !=0){
 		LM_ERR("unable to reload the registrant data\n");
-		free_mi_tree(rpl_tree);
 		goto error;
 	}
 	/* Swap the lists: secondary will become primary */
@@ -904,7 +1351,7 @@ static struct mi_root* mi_reg_reload(struct mi_root* cmd, void* param)
 		lock_release(&reg_htable[i].lock);
 	}
 
-	return rpl_tree;
+	return init_mi_result_ok();
 
 error:
 	for(i=0; i<reg_hsize; i++) {
@@ -916,3 +1363,139 @@ error:
 	return NULL;
 }
 
+static mi_response_t *mi_reg_reload_record(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	record_coords_t coords;
+	mi_response_t *resp;
+	int rc;
+
+	if ((resp = mi_get_coords(params, &coords)))
+		return resp;
+
+	rc = load_reg_info_from_db(REG_DB_LOAD_RECORD, &coords);
+	if (rc == -2) {
+		return init_mi_error(404, MI_SSTR("No such registrant in database"));
+	} else if (rc < 0) {
+		LM_ERR("unable to reload the registrant data\n");
+		return NULL;
+	}
+
+	return init_mi_result_ok();
+}
+
+int run_mi_reg_enable(void *e_data, void *data, void *r_data)
+{
+	reg_record_t *rec = (reg_record_t*)e_data;
+	record_coords_t *coords = (record_coords_t *)data;
+	str str_now = {NULL, 0};
+	time_t now;
+
+	if (!str_strcmp(&coords->contact, &rec->contact_uri) &&
+		!str_strcmp(&coords->registrar, &rec->td.rem_target)) {
+		if (!(rec->flags&REG_ENABLED)) {
+			if (rec->state == NOT_REGISTERED_STATE) {
+				now = time(0);
+				if (get_cur_time_s(&str_now, now) < 0) {
+					LM_ERR("Failed to get current time string\n");
+					return -1;
+				}
+				new_call_id_ftag_4_record(rec, &str_now);
+
+				if(send_register((unsigned long)coords->extra, rec, NULL)==1) {
+					rec->last_register_sent = now;
+					rec->state = REGISTERING_STATE;
+				} else {
+					rec->registration_timeout = now + rec->expires - timer_interval;
+					rec->state = INTERNAL_ERROR_STATE;
+				}
+			}
+
+			rec->flags |= REG_ENABLED;
+
+			reg_update_db_state(rec);
+		}
+
+		return 1;
+	} else
+		return 0;  /* continue search */
+}
+
+int run_mi_reg_disable(void *e_data, void *data, void *r_data)
+{
+	reg_record_t *rec = (reg_record_t*)e_data;
+	record_coords_t *coords = (record_coords_t *)data;
+
+	if (!str_strcmp(&coords->contact, &rec->contact_uri) &&
+		!str_strcmp(&coords->registrar, &rec->td.rem_target)) {
+		if (rec->flags&REG_ENABLED) {
+			if (rec->state == REGISTERED_STATE) {
+				if(send_unregister((unsigned long)coords->extra, rec, NULL, 0)==1)
+					rec->state = UNREGISTERING_STATE;
+				else
+					rec->state = INTERNAL_ERROR_STATE;
+			}
+
+			rec->flags &= ~REG_ENABLED;
+
+			reg_update_db_state(rec);
+		}
+
+		return 1;
+	} else
+		return 0;  /* continue search */
+}
+
+static mi_response_t *mi_reg_enable(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	mi_response_t *resp;
+	record_coords_t coords;
+	int rc;
+	unsigned int hash_code;
+
+	if ((resp = mi_get_coords(params, &coords)))
+		return resp;
+
+	hash_code = core_hash(&coords.aor, NULL, reg_hsize);
+	coords.extra = (void*)(unsigned long)hash_code;
+
+	lock_get(&reg_htable[hash_code].lock);
+	rc = slinkedl_traverse(reg_htable[hash_code].p_list,
+					&run_mi_reg_enable, &coords, NULL);
+	lock_release(&reg_htable[hash_code].lock);
+
+	if (rc < 0)
+		return NULL;
+	else if (rc == 0)
+		return init_mi_error(404, MI_SSTR("No such registrant"));
+
+	return init_mi_result_ok();
+}
+
+static mi_response_t *mi_reg_disable(const mi_params_t *params,
+								struct mi_handler *async_hdl)
+{
+	mi_response_t *resp;
+	record_coords_t coords;
+	int rc;
+	unsigned int hash_code;
+
+	if ((resp = mi_get_coords(params, &coords)))
+		return resp;
+
+	hash_code = core_hash(&coords.aor, NULL, reg_hsize);
+	coords.extra = (void*)(unsigned long)hash_code;
+
+	lock_get(&reg_htable[hash_code].lock);
+	rc = slinkedl_traverse(reg_htable[hash_code].p_list,
+					&run_mi_reg_disable, &coords, NULL);
+	lock_release(&reg_htable[hash_code].lock);
+
+	if (rc < 0)
+		return NULL;
+	else if (rc == 0)
+		return init_mi_error(404, MI_SSTR("No such registrant"));
+
+	return init_mi_result_ok();
+}
