@@ -44,6 +44,7 @@
 #include "../ut.h"
 #include "../error.h"
 #include "../dprint.h"
+#include "../data_lump.h"
 #include "../data_lump_rpl.h"
 #include "../mem/mem.h"
 #include "../error.h"
@@ -227,6 +228,16 @@ char* get_hdr_field(char* buf, char* end, struct hdr_field* hdr)
 		case HDR_CALL_INFO_T:
 		case HDR_WWW_AUTHENTICATE_T:
 		case HDR_PROXY_AUTHENTICATE_T:
+		case HDR_FEATURE_CAPS_T:
+		case HDR_REPLACES_T:
+		case HDR_TO_PATH_T:
+		case HDR_FROM_PATH_T:
+		case HDR_MESSAGE_ID_T:
+		case HDR_BYTE_RANGE_T:
+		case HDR_FAILURE_REPORT_T:
+		case HDR_SUCCESS_REPORT_T:
+		case HDR_STATUS_T:
+		case HDR_USE_PATH_T:
 		case HDR_OTHER_T:
 			/* just skip over it */
 			hdr->body.s=tmp;
@@ -237,7 +248,8 @@ char* get_hdr_field(char* buf, char* end, struct hdr_field* hdr)
 				if (match){
 					match++;
 				}else {
-					LM_ERR("bad body for <%s>(%d)\n", hdr->name.s, hdr->type);
+					LM_ERR("bad body for <%.*s>(%d)\n",
+					         hdr->name.len, hdr->name.s, hdr->type);
 					tmp=end;
 					goto error_bad_hdr;
 				}
@@ -329,6 +341,14 @@ int parse_headers(struct sip_msg* msg, hdr_flags_t flags, int next)
 				pkg_free(hf);
 				goto skip;
 			case HDR_OTHER_T: /*do nothing*/
+			case HDR_TO_PATH_T:
+			case HDR_FROM_PATH_T:
+			case HDR_MESSAGE_ID_T:
+			case HDR_BYTE_RANGE_T:
+			case HDR_FAILURE_REPORT_T:
+			case HDR_SUCCESS_REPORT_T:
+			case HDR_STATUS_T:
+			case HDR_USE_PATH_T:
 				break;
 			case HDR_CALLID_T:
 				if (msg->callid==0) msg->callid=hf;
@@ -475,11 +495,11 @@ int parse_headers(struct sip_msg* msg, hdr_flags_t flags, int next)
 				msg->parsed_flag |= HDR_MIN_EXPIRES_F;
 				break;
 			case HDR_PPI_T:
-				if (msg->ppi==0) msg->ppi = hf;
+				link_sibling_hdr(ppi,hf);
 				msg->parsed_flag|=HDR_PPI_F;
 				break;
 			case HDR_PAI_T:
-				if (msg->pai==0) msg->pai = hf;
+				link_sibling_hdr(pai,hf);
 				msg->parsed_flag|=HDR_PAI_F;
 				break;
 			case HDR_PRIVACY_T:
@@ -507,6 +527,14 @@ int parse_headers(struct sip_msg* msg, hdr_flags_t flags, int next)
 					LM_DBG("parse_headers: this is the second via\n");
 				}
 				break;
+			case HDR_FEATURE_CAPS_T:
+				link_sibling_hdr(feature_caps, hf);
+				msg->parsed_flag |= HDR_FEATURE_CAPS_F;
+				break;
+			case HDR_REPLACES_T:
+				link_sibling_hdr(replaces, hf);
+				msg->parsed_flag |= HDR_REPLACES_F;
+				break;
 			default:
 				LM_CRIT("unknown header type %d\n",	hf->type);
 				goto error;
@@ -538,9 +566,175 @@ error:
 	return -1;
 }
 
+/* clones the headers list from the `from` sip_msg
+ * into the `to` sip_msg structure */
+int clone_headers(struct sip_msg *from_msg, struct sip_msg *to_msg)
+{
+	int hdrs_no, i;
+	struct hdr_field *hdrs;
+	struct hdr_field *hdr;
+	struct hdr_field *itr;
+
+#define link_sibling_hdr_case(_hook, _hdr_type) \
+	case _hdr_type: \
+		if (to_msg->_hook==0) to_msg->_hook=&hdrs[i];\
+		else {\
+			for(itr=to_msg->_hook;itr->sibling;itr=itr->sibling);\
+			itr->sibling = &hdrs[i];\
+		}\
+		break
+#define link_hdr_case(_hook, _hdr_type) \
+	case _hdr_type: \
+		to_msg->_hook=&hdrs[i];\
+		break
+
+	 /*
+	 * we need to duplicate the headers because the hdr->parsed field resides
+	 * in shm memory, so there might be a different process that access the
+	 * parsed field and find the structure parsed by the current process in
+	 * pkg
+	 */
+	for (hdrs_no = 0, hdr = from_msg->headers; hdr; hdr = hdr->next)
+		hdrs_no++;
+
+	hdrs = pkg_malloc(hdrs_no * sizeof(struct hdr_field));
+	if (!hdrs) {
+		LM_ERR("could not allocate %d contact headers!\n", hdrs_no);
+		return -1;
+	}
+
+	/* reset all header fields before populating new ones */
+	to_msg->h_via1 = NULL;
+	to_msg->callid = NULL;
+	to_msg->to = NULL;
+	to_msg->cseq = NULL;
+	to_msg->from = NULL;
+	to_msg->contact = NULL;
+	to_msg->maxforwards = NULL;
+	to_msg->route = NULL;
+	to_msg->record_route = NULL;
+	to_msg->path = NULL;
+	to_msg->content_type = NULL;
+	to_msg->content_length = NULL;
+	to_msg->authorization = NULL;
+	to_msg->expires = NULL;
+	to_msg->proxy_auth = NULL;
+	to_msg->supported = NULL;
+	to_msg->proxy_require = NULL;
+	to_msg->unsupported = NULL;
+	to_msg->allow = NULL;
+	to_msg->event = NULL;
+	to_msg->accept = NULL;
+	to_msg->accept_language = NULL;
+	to_msg->organization = NULL;
+	to_msg->priority = NULL;
+	to_msg->subject = NULL;
+	to_msg->user_agent = NULL;
+	to_msg->content_disposition = NULL;
+	to_msg->accept_disposition = NULL;
+	to_msg->diversion = NULL;
+	to_msg->rpid = NULL;
+	to_msg->refer_to = NULL;
+	to_msg->session_expires = NULL;
+	to_msg->min_se = NULL;
+	to_msg->ppi = NULL;
+	to_msg->pai = NULL;
+	to_msg->privacy = NULL;
+	to_msg->call_info = NULL;
+	to_msg->www_authenticate = NULL;
+	to_msg->proxy_authenticate = NULL;
+	to_msg->min_expires = NULL;
+	to_msg->feature_caps = NULL;
+	to_msg->replaces = NULL;
+
+	for (i = 0, hdr = from_msg->headers; hdr; i++, hdr = hdr->next) {
+		memcpy(&hdrs[i], hdr, sizeof(struct hdr_field));
+		/* fix next and sibling */
+		hdrs[i].next = &hdrs[i + 1];
+		hdrs[i].sibling = NULL;
+		switch(hdr->type) {
+			link_sibling_hdr_case(h_via1, HDR_VIA_T);
+			link_hdr_case        (callid, HDR_CALLID_T);
+			link_hdr_case        (to, HDR_TO_T);
+			link_hdr_case        (cseq, HDR_CSEQ_T);
+			link_hdr_case        (from, HDR_FROM_T);
+			link_sibling_hdr_case(contact, HDR_CONTACT_T);
+			link_hdr_case        (maxforwards, HDR_MAXFORWARDS_T);
+			link_sibling_hdr_case(route, HDR_ROUTE_T);
+			link_sibling_hdr_case(record_route, HDR_RECORDROUTE_T);
+			link_sibling_hdr_case(path, HDR_PATH_T);
+			link_hdr_case        (content_type, HDR_CONTENTTYPE_T);
+			link_hdr_case        (content_length, HDR_CONTENTLENGTH_T);
+			link_sibling_hdr_case(authorization, HDR_AUTHORIZATION_T);
+			link_hdr_case        (expires, HDR_EXPIRES_T);
+			link_sibling_hdr_case(proxy_auth, HDR_PROXYAUTH_T);
+			link_sibling_hdr_case(supported, HDR_SUPPORTED_T);
+			link_sibling_hdr_case(proxy_require, HDR_PROXYREQUIRE_T);
+			link_sibling_hdr_case(unsupported, HDR_UNSUPPORTED_T);
+			link_sibling_hdr_case(allow, HDR_ALLOW_T);
+			link_sibling_hdr_case(event, HDR_EVENT_T);
+			link_sibling_hdr_case(accept, HDR_ACCEPT_T);
+			link_sibling_hdr_case(accept_language, HDR_ACCEPTLANGUAGE_T);
+			link_hdr_case        (organization, HDR_ORGANIZATION_T);
+			link_hdr_case        (priority, HDR_PRIORITY_T);
+			link_hdr_case        (subject, HDR_SUBJECT_T);
+			link_hdr_case        (user_agent, HDR_USERAGENT_T);
+			link_hdr_case        (content_disposition, HDR_CONTENTDISPOSITION_T);
+			link_sibling_hdr_case(accept_disposition, HDR_ACCEPTDISPOSITION_T);
+			link_sibling_hdr_case(diversion, HDR_DIVERSION_T);
+			link_hdr_case        (rpid, HDR_RPID_T);
+			link_hdr_case        (refer_to, HDR_REFER_TO_T);
+			link_hdr_case        (session_expires, HDR_SESSION_EXPIRES_T);
+			link_hdr_case        (min_se, HDR_MIN_SE_T);
+			link_sibling_hdr_case(ppi, HDR_PPI_T);
+			link_sibling_hdr_case(pai, HDR_PAI_T);
+			link_hdr_case        (privacy, HDR_PRIVACY_T);
+			link_sibling_hdr_case(call_info, HDR_CALL_INFO_T);
+			link_sibling_hdr_case(www_authenticate, HDR_WWW_AUTHENTICATE_T);
+			link_sibling_hdr_case(proxy_authenticate, HDR_PROXY_AUTHENTICATE_T);
+			link_hdr_case        (min_expires, HDR_MIN_EXPIRES_T);
+			link_sibling_hdr_case(feature_caps, HDR_FEATURE_CAPS_T);
+			link_hdr_case        (replaces, HDR_REPLACES_T);
+
+			/* not used in sip_msg, used in MSRP */
+			case HDR_TO_PATH_T:
+			case HDR_FROM_PATH_T:
+			case HDR_MESSAGE_ID_T:
+			case HDR_BYTE_RANGE_T:
+			case HDR_FAILURE_REPORT_T:
+			case HDR_SUCCESS_REPORT_T:
+			case HDR_STATUS_T:
+			case HDR_USE_PATH_T:
+
+			/* not having shortcut */
+			case HDR_RETRY_AFTER_T:
+			case HDR_VIA2_T:
+
+			/* not actual hdrs */
+			case HDR_OTHER_T:
+			case HDR_ERROR_T:
+			case HDR_EOH_T:
+				break;
+			/* we do not have a "default" on purpose, so we get
+			 * a compile err/war where a new HDR is added and we do 
+			 * not handle it here.
+			default:
+				LM_ERR("unknown header type %d\n", hdr->type);
+				break;
+			*/
+		}
+	}
+	hdrs[i - 1].next = 0;
+	to_msg->headers = hdrs;
+#undef link_hdr_case
+#undef link_sibling_hdr_case
+	return 0;
+}
+
 
 /* returns 0 if ok, -1 for errors */
-int parse_msg(char* buf, unsigned int len, struct sip_msg* msg)
+int parse_msg_opt(char* buf, unsigned int len, struct sip_msg* msg,
+															int free_on_err)
 {
 
 	char *tmp;
@@ -550,8 +744,8 @@ int parse_msg(char* buf, unsigned int len, struct sip_msg* msg)
 	hdr_flags_t flags;
 
 	/* eat crlf from the beginning */
-	for (tmp=buf; (*tmp=='\n' || *tmp=='\r')&&
-			(unsigned int)(tmp-buf) < len ; tmp++);
+	for (tmp=buf; (unsigned int)(tmp-buf) < len
+	        && (*tmp=='\n' || *tmp=='\r'); tmp++);
 	offset=tmp-buf;
 	fl=&(msg->first_line);
 	rest=parse_first_line(tmp, len-offset, fl);
@@ -576,7 +770,7 @@ int parse_msg(char* buf, unsigned int len, struct sip_msg* msg)
 				ZSW(fl->u.request.uri.s));
 			LM_DBG(" version: <%.*s>\n",fl->u.request.version.len,
 				ZSW(fl->u.request.version.s));
-			flags=HDR_VIA_F;
+			flags=HDR_EOH_F;
 			break;
 		case SIP_REPLY:
 			LM_DBG("SIP Reply  (status):\n");
@@ -586,7 +780,7 @@ int parse_msg(char* buf, unsigned int len, struct sip_msg* msg)
 					ZSW(fl->u.reply.status.s));
 			LM_DBG(" reason:  <%.*s>\n", fl->u.reply.reason.len,
 					ZSW(fl->u.reply.reason.s));
-			flags=HDR_VIA_F;
+			flags=HDR_EOH_F;
 			break;
 		default:
 			LM_DBG("unknown type %d\n",fl->type);
@@ -594,12 +788,12 @@ int parse_msg(char* buf, unsigned int len, struct sip_msg* msg)
 	}
 	msg->unparsed=tmp;
 	/*find first Via: */
-	if (parse_headers(msg, flags, 0)==-1) goto error;
+	if (parse_headers(msg, flags, 0)==-1) goto parse_error;
 
 #ifdef EXTRA_DEBUG
 	/* dump parsed data */
 	if (msg->via1){
-		LM_DBG(" first  via: <%.*s/%.*s/%.*s> <%.*s:%.*s(%d)>",
+		LM_DBG(" first  via: <%.*s/%.*s/%.*s> <%.*s:%.*s(%d)>\n",
 			msg->via1->name.len,
 			ZSW(msg->via1->name.s),
 			msg->via1->version.len,
@@ -611,15 +805,15 @@ int parse_msg(char* buf, unsigned int len, struct sip_msg* msg)
 			msg->via1->port_str.len,
 			ZSW(msg->via1->port_str.s),
 			msg->via1->port);
-		if (msg->via1->params.s)  LM_DBG(";<%.*s>",
+		if (msg->via1->params.s)  LM_DBG(";<%.*s>\n",
 				msg->via1->params.len, ZSW(msg->via1->params.s));
 		if (msg->via1->comment.s)
-				LM_DBG(" <%.*s>",
+				LM_DBG(" <%.*s>\n",
 					msg->via1->comment.len, ZSW(msg->via1->comment.s));
 		LM_DBG ("\n");
 	}
 	if (msg->via2){
-		LM_DBG(" first  via: <%.*s/%.*s/%.*s> <%.*s:%.*s(%d)>",
+		LM_DBG(" first  via: <%.*s/%.*s/%.*s> <%.*s:%.*s(%d)>\n",
 			msg->via2->name.len,
 			ZSW(msg->via2->name.s),
 			msg->via2->version.len,
@@ -631,9 +825,9 @@ int parse_msg(char* buf, unsigned int len, struct sip_msg* msg)
 			msg->via2->port_str.len,
 			ZSW(msg->via2->port_str.s),
 			msg->via2->port);
-		if (msg->via2->params.s)  LM_DBG(";<%.*s>",
+		if (msg->via2->params.s)  LM_DBG(";<%.*s>\n",
 				msg->via2->params.len, ZSW(msg->via2->params.s));
-		if (msg->via2->comment.s) LM_DBG(" <%.*s>",
+		if (msg->via2->comment.s) LM_DBG(" <%.*s>\n",
 				msg->via2->comment.len, ZSW(msg->via2->comment.s));
 		LM_DBG ("\n");
 	}
@@ -646,6 +840,14 @@ int parse_msg(char* buf, unsigned int len, struct sip_msg* msg)
 
 	return 0;
 
+parse_error:
+	if (free_on_err) {
+		/* Free the SIP message if it has already been created to
+		 * prevent memory leaks */
+		free_sip_msg(msg);
+		memset(msg, 0, sizeof(*msg));
+	}
+
 error:
 	/* more debugging, msg->orig is/should be null terminated*/
 	LM_ERR("message=<%.*s>\n", (int)len, ZSW(buf));
@@ -666,31 +868,33 @@ void free_reply_lump( struct lump_rpl *lump)
 }
 
 
-/*only the content*/
+/* Free only the content, not the msg structure itself
+ * NOTE: the function doesn't do any cleanup/reset of the subfields */
 void free_sip_msg(struct sip_msg* msg)
 {
-	if (msg->msg_cb) { msg_callback_process(msg, MSG_DESTROY, NULL); }
-	if (msg->new_uri.s) { pkg_free(msg->new_uri.s); msg->new_uri.len=0; }
-	if (msg->set_global_address.s) {
+	if (msg->msg_cb)
+		msg_callback_process(msg, MSG_DESTROY, NULL);
+	if (msg->new_uri.s)
+		pkg_free(msg->new_uri.s);
+	if (msg->set_global_address.s)
 		pkg_free(msg->set_global_address.s);
-		msg->set_global_address.s = NULL;
-	}
-	if (msg->set_global_port.s) {
+	if (msg->set_global_port.s)
 		pkg_free(msg->set_global_port.s);
-		msg->set_global_port.s = NULL;
-	}
-	if (msg->dst_uri.s) { pkg_free(msg->dst_uri.s); msg->dst_uri.len=0; }
-	if (msg->path_vec.s) { pkg_free(msg->path_vec.s); msg->path_vec.len=0; }
-	if (msg->headers)     free_hdr_field_lst(msg->headers);
-	if (msg->sdp)         free_sdp(&(msg->sdp));
-	if (msg->add_rm)      free_lump_list(msg->add_rm);
-	if (msg->body_lumps)  free_lump_list(msg->body_lumps);
-	if (msg->reply_lump)   free_reply_lump(msg->reply_lump);
-	if (msg->multi )  { free_multi_body(msg->multi);msg->multi = 0;}
+	if (msg->dst_uri.s)
+		pkg_free(msg->dst_uri.s);
+	if (msg->path_vec.s)
+		pkg_free(msg->path_vec.s);
+	if (msg->headers)
+		free_hdr_field_lst(msg->headers);
+	if (msg->add_rm)
+		free_lump_list(msg->add_rm);
+	if (msg->body_lumps)
+		free_lump_list(msg->body_lumps);
+	if (msg->reply_lump)
+		free_reply_lump(msg->reply_lump);
+	if (msg->body )
+		free_sip_body(msg->body);
 	/* don't free anymore -- now a pointer to a static buffer */
-#	ifdef DYN_BUF
-	pkg_free(msg->buf);
-#	endif
 }
 
 
@@ -782,6 +986,284 @@ int set_dst_uri(struct sip_msg *msg, str *uri)
 	return 0;
 }
 
+void reset_dst_uri(struct sip_msg *msg)
+{
+	if(msg->dst_uri.s!=0)
+		pkg_free(msg->dst_uri.s);
+	msg->dst_uri.s = 0;
+	msg->dst_uri.len = 0;
+}
+
+int set_dst_host_port(struct sip_msg *msg, str *host, str *port)
+{
+	char *tmp, *end, *crt, *new_uri;
+	int len;
+	struct sip_uri uri;
+	int user = 0;
+
+	tmp = msg->dst_uri.s;
+	len = msg->dst_uri.len;
+
+	if (tmp == NULL || len == 0) {
+		LM_ERR("failure - null uri\n");
+		return -1;
+	}
+	if (host && (host->s == NULL || host->len == 0)) {
+		LM_ERR("cannot set a null uri domain\n");
+		return -1;
+	}
+	if (parse_uri(tmp, len, &uri)<0) {
+		LM_ERR("bad uri <%.*s>, dropping packet\n", len, tmp);
+		return -1;
+	}
+	new_uri=pkg_malloc(MAX_URI_SIZE);
+	if (new_uri == NULL) {
+		LM_ERR("memory allocation failure\n");
+		return -1;
+	}
+	end=new_uri+MAX_URI_SIZE;
+	crt=new_uri;
+	len = (uri.user.len?uri.user.s:uri.host.s) - tmp;
+	if (crt+len>end) goto error_uri;
+	memcpy(crt,tmp,len);
+	crt += len;
+	/* user */
+	tmp = uri.user.s;
+	len = uri.user.len;
+	if (tmp) {
+		if (crt+len>end) goto error_uri;
+		memcpy(crt,tmp,len);
+		crt += len;
+		user = 1;
+	}
+	/* passwd */
+	tmp = uri.passwd.s;
+	len = uri.passwd.len;
+	if (tmp) {
+		if (crt+len+1>end) goto error_uri;
+		*crt++=':';
+		memcpy(crt, tmp, len);
+		crt += len;
+	}
+	/* host */
+	if (host) {
+		tmp = host->s;
+		len = host->len;
+	} else {
+		tmp = uri.host.s;
+		len = uri.host.len;
+	}
+	if (tmp) {
+		if (user) {
+			if (crt+1>end) goto error_uri;
+			*crt++='@';
+		}
+		if (crt+len+1>end) goto error_uri;
+		memcpy(crt, tmp, len);
+		crt += len;
+	}
+	/* port */
+	if (port) {
+		tmp = port->s;
+		len = port->len;
+	} else {
+		tmp = uri.port.s;
+		len = uri.port.len;
+	}
+	if (tmp && len > 0) {
+		if (crt+len+1>end) goto error_uri;
+		*crt++=':';
+		memcpy(crt, tmp, len);
+		crt += len;
+	}
+	/* params */
+	tmp=uri.params.s;
+	if (tmp){
+		len=uri.params.len; if(crt+len+1>end) goto error_uri;
+		*crt++=';';
+		memcpy(crt,tmp,len);
+		crt += len;
+	}
+	/* headers */
+	tmp=uri.headers.s;
+	if (tmp){
+		len=uri.headers.len; if(crt+len+1>end) goto error_uri;
+		*crt++='?';
+		memcpy(crt,tmp,len);
+		crt += len;
+	}
+	*crt=0; /* null terminate the thing */
+	/* copy it to the msg */
+	pkg_free(msg->dst_uri.s);
+	msg->dst_uri.s=new_uri;
+	msg->dst_uri.len=crt-new_uri;
+	
+	return 0;
+
+error_uri:
+	pkg_free(new_uri);
+	return -1;
+}
+
+int rewrite_ruri(struct sip_msg *msg, str *sval, int ival,
+				enum rw_ruri_part part)
+{
+	int user = 0;
+	char *tmp, *new_uri, *end, *crt;
+	int len;
+	struct sip_uri uri;
+
+	if (msg->new_uri.s) {
+		tmp=msg->new_uri.s;
+		len=msg->new_uri.len;
+	}else{
+		tmp=msg->first_line.u.request.uri.s;
+		len=msg->first_line.u.request.uri.len;
+	}
+	if (parse_uri(tmp, len, &uri)<0){
+		LM_ERR("bad uri <%.*s>, dropping packet\n", len, tmp);
+		return -1;
+	}
+
+	new_uri=pkg_malloc(MAX_URI_SIZE);
+	if (new_uri==0){
+		LM_ERR("memory allocation failure\n");
+		return -1;
+	}
+	end=new_uri+MAX_URI_SIZE;
+	crt=new_uri;
+	/* begin copying */
+	len = (uri.user.len?uri.user.s:uri.host.s) - tmp;
+	if (crt+len>end) goto error;
+	memcpy(crt,tmp,len);crt+=len;
+
+	if (part==RW_RURI_PREFIX) {
+		if (crt+sval->len>end) goto error;
+		memcpy( crt, sval->s, sval->len);
+		crt+=sval->len;
+		/* whatever we had before, with prefix we have username
+		   now */
+		user=1;
+	}
+
+	if ((part==RW_RURI_USER)||(part==RW_RURI_USERPASS)) {
+		tmp=sval->s;
+		len=sval->len;
+	} else if (part==RW_RURI_STRIP) {
+		if (ival>uri.user.len) {
+			LM_WARN("too long strip asked; "
+					" deleting username: %d of <%.*s>\n",
+					ival, uri.user.len, uri.user.s);
+			len=0;
+		} else if (ival==uri.user.len) {
+			len=0;
+		} else {
+			tmp=uri.user.s + ival;
+			len=uri.user.len - ival;
+		}
+	} else if (part==RW_RURI_STRIP_TAIL) {
+		if (ival>uri.user.len) {
+			LM_WARN("too long strip_tail asked;"
+					" deleting username: %d of <%.*s>\n",
+					ival, uri.user.len, uri.user.s);
+			len=0;
+		} else if (ival==uri.user.len) {
+			len=0;
+		} else {
+			tmp=uri.user.s;
+			len=uri.user.len - ival;
+		}
+	} else {
+		tmp=uri.user.s;
+		len=uri.user.len;
+	}
+
+	if (len){
+		if(crt+len>end) goto error;
+		memcpy(crt,tmp,len);crt+=len;
+		user=1; /* we have an user field so mark it */
+	}
+
+	if (part==RW_RURI_USERPASS) tmp=0;
+	else tmp=uri.passwd.s;
+	/* passwd */
+	if (tmp){
+		len=uri.passwd.len; if(crt+len+1>end) goto error;
+		*crt=':'; crt++;
+		memcpy(crt,tmp,len);crt+=len;
+	}
+	/* host */
+	if (user || tmp){ /* add @ */
+		if(crt+1>end) goto error;
+		*crt='@'; crt++;
+	}
+	if ((part==RW_RURI_HOST) ||(part==RW_RURI_HOSTPORT)) {
+		tmp=sval->s;
+		len=sval->len;
+	} else {
+		tmp=uri.host.s;
+		len = uri.host.len;
+	}
+	if (tmp){
+		if(crt+len>end) goto error;
+		memcpy(crt,tmp,len);crt+=len;
+	}
+	/* port */
+	if (part==RW_RURI_HOSTPORT) tmp=0;
+	else if (part==RW_RURI_PORT) {
+		tmp=sval->s;
+		len=sval->len;
+	} else {
+		tmp=uri.port.s;
+		len = uri.port.len;
+	}
+	if (tmp && len>0){
+		if(crt+len+1>end) goto error;
+		*crt=':'; crt++;
+		memcpy(crt,tmp,len);crt+=len;
+	}
+	/* params */
+	tmp=uri.params.s;
+	if (tmp){
+		/* include in param string the starting ';' */
+		len=uri.params.len+1;
+		tmp--;
+		if(crt+len+1>end) goto error;
+		/* if a maddr param is present, strip it out */
+		if (uri.maddr.len &&
+		(part==RW_RURI_HOSTPORT || part==RW_RURI_HOST)) {
+			memcpy(crt,tmp,uri.maddr.s-tmp-1);
+			crt+=uri.maddr.s-tmp-1;
+			memcpy(crt,uri.maddr_val.s+uri.maddr_val.len,
+				tmp+len-uri.maddr_val.s-uri.maddr_val.len);
+			crt+=tmp+len-uri.maddr_val.s-uri.maddr_val.len;
+		} else {
+			memcpy(crt,tmp,len);crt+=len;
+		}
+	}
+	/* headers */
+	tmp=uri.headers.s;
+	if (tmp){
+		len=uri.headers.len; if(crt+len+1>end) goto error;
+		*crt='?'; crt++;
+		memcpy(crt,tmp,len);crt+=len;
+	}
+	*crt=0; /* null terminate the thing */
+	/* copy it to the msg */
+	if (msg->new_uri.s) pkg_free(msg->new_uri.s);
+	msg->new_uri.s=new_uri;
+	msg->new_uri.len=crt-new_uri;
+	msg->parsed_uri_ok=0;
+
+	return 0;
+
+error:
+	LM_ERR("uri too long\n");
+	if (new_uri)
+		pkg_free(new_uri);
+	return -1;
+}
+
 /*
  * Make a private copy of the string and assign it to path_vec
  */
@@ -816,6 +1298,13 @@ int set_path_vector(struct sip_msg *msg, str *path)
 	return 0;
 }
 
+void clear_path_vector(struct sip_msg *msg)
+{
+	if (msg->path_vec.s) {
+		pkg_free(msg->path_vec.s);
+		memset(&msg->path_vec, 0, sizeof msg->path_vec);
+	}
+}
 
 /* convenience macros */
 #define LC(_cp) ((*(_cp))|0x20)
@@ -862,7 +1351,8 @@ int extract_ftc_hdrs( char *buf, int len, str *from, str *to, str *cseq,str *cal
 			case '\r':
 				switch (state) {
 					case 4: state=5;break;
-					case 5: case 6: state=6;break;
+					case 5: state=6;break;
+					case 6: if(!(*p=='\n' && *(p-1)=='\r')) SET_FOUND(1);break;
 					default : state=2;break;
 				}
 				break;
@@ -958,3 +1448,15 @@ done:
 }
 
 
+int delete_headers(struct sip_msg *msg, struct hdr_field *hdr)
+{
+	for (; hdr; hdr = hdr->sibling) {
+		if (!del_lump(msg, hdr->name.s - msg->buf, hdr->len, hdr->type)) {
+			LM_ERR("failed to delete contact '%.*s'\n", hdr->name.len,
+			       hdr->name.s);
+			return -1;
+		}
+	}
+
+	return 0;
+}
