@@ -41,6 +41,7 @@
 #define MSG_PARSER_H
 
 #include <strings.h>
+#include <sys/time.h>
 
 #include "../str.h"
 #include "../lump_struct.h"
@@ -54,14 +55,15 @@
 #include "parse_content.h"
 #include "parse_via.h"
 #include "parse_fline.h"
-#include "parse_multipart.h"
+#include "parse_body.h"
 #include "hf.h"
-#include "sdp/sdp.h"
+#include "../trim.h"
 
 
 /* convenience short-cut macros */
 #define REQ_LINE(_msg) ((_msg)->first_line.u.request)
-#define REQ_METHOD first_line.u.request.method_value
+#define REQ_METHOD   first_line.u.request.method_value
+#define REQ_METHOD_S first_line.u.request.method
 #define REPLY_STATUS first_line.u.reply.statuscode
 #define REPLY_CLASS(_reply) ((_reply)->REPLY_STATUS/100)
 
@@ -103,23 +105,39 @@ enum request_method {
                                       * callback had already been registered */
 #define FL_NAT_TRACK_DIALOG  (1<<13) /* trigger dialog tracking from the
                                       * nat_traversal module */
-#define FL_USE_SIPTRACE      (1<<14) /* used by siptrace to check if the tm
+#define FL_USE_SIPTRACE      (1<<14) /* used by tracer to check if the tm
                                       * callbacks were registered */
 #define FL_SHM_UPDATABLE     (1<<15) /* a SHM cloned message can be updated
                                       * (TM used, requires FL_SHM_CLONE) */
-#define FL_TM_CB_REGISTERED  (1<<16) /* tm callbacks for this message have been
-									 * registered (by setting this flag, you
-									 * will know if any tm callbacks for this
-									 * message have been registered) */
-#define FL_TM_FAKE_REQ       (1<<17) /* the SIP request is a fake one, generated based on the transaction, 
-					either in failure route or resume route */	
+#define FL_SHM_UPDATED       (1<<16) /* an updatable SHM cloned message that 
+                                      * had at least one update; if the flag is
+                                      * missing, it means the cloned msg was
+                                      * never updated.
+                                      * (TM used, requires FL_SHM_UPDATABLE) */
+#define FL_TM_CB_REGISTERED  (1<<17) /* tm callbacks for this message have been
+                                      * registered (by setting this flag, you
+                                      * will know if any tm callbacks for this
+                                      * message have been registered) */
+#define FL_TM_FAKE_REQ       (1<<18) /* the SIP request is a fake one,
+                                      * generated based on the transaction,
+                                      * either in failure route or resume 
+                                      * route */
+#define FL_TM_REPLICATED	 (1<<19) /* message received due to a tm replication */
+#define FL_BODY_NO_SDP       (1<<20) /* message does not have an SDP body */
+#define FL_IS_LOCAL          (1<<21) /* the message is a locally generated
+                                      * one, not received */
+#define FL_HAS_ROUTE_LUMP    (1<<22) /* the message had Route headers added
+                                      * as lumps */
+#define FL_USE_SIPTRACE_B2B  (1<<23) /* used by tracer to check if the b2b
+                                      * tracing was enabled */
+#define FL_ACK_WITH_BODY     (1<<24) /* ACK message has SDP body */
 
 /* define the # of unknown URI parameters to parse */
 #define URI_MAX_U_PARAMS 10
 
 #define IFISMETHOD(methodname,firstchar)                                  \
 if (  (*tmp==(firstchar) || *tmp==((firstchar) | 32)) &&                  \
-        strncasecmp( tmp+1, #methodname +1, methodname##_LEN-1)==0 &&     \
+        strncasecmp( tmp+1, (char *)#methodname+1, methodname##_LEN-1)==0 &&     \
         *(tmp+methodname##_LEN)==' ') {                                   \
                 fl->type=SIP_REQUEST;                                     \
                 fl->u.request.method.len=methodname##_LEN;                \
@@ -150,7 +168,7 @@ if (  (*tmp==(firstchar) || *tmp==((firstchar) | 32)) &&                  \
 (((m)->new_uri.s && (m)->new_uri.len) ? (&(m)->new_uri) : (&(m)->first_line.u.request.uri))
 
 
-enum _uri_type{ERROR_URI_T=0, SIP_URI_T, SIPS_URI_T, TEL_URI_T, TELS_URI_T, URN_SERVICE_URI_T};
+enum _uri_type{ERROR_URI_T=0, SIP_URI_T, SIPS_URI_T, TEL_URI_T, TELS_URI_T, URN_SERVICE_URI_T, URN_NENA_SERVICE_URI_T};
 typedef enum _uri_type uri_type;
 
 struct sip_uri {
@@ -163,7 +181,8 @@ struct sip_uri {
 	unsigned short port_no;
 	unsigned short proto; /* from transport */
 	uri_type type; /* uri scheme */
-	/* parameters */
+
+	/* parameters [+ "=value" parts, if any] */
 	str transport;
 	str ttl;
 	str user_param;
@@ -172,7 +191,12 @@ struct sip_uri {
 	str lr;
 	str r2; /* ser specific rr parameter */
 	str gr; /* GRUU */
-	/* values */
+	str pn_provider; /* RFC 8599 (SIP PN) */
+	str pn_prid;
+	str pn_param;
+	str pn_purr;
+
+	/* just values */
 	str transport_val;
 	str ttl_val;
 	str user_param_val;
@@ -181,6 +205,14 @@ struct sip_uri {
 	str lr_val; /* lr value placeholder for lr=on a.s.o*/
 	str r2_val;
 	str gr_val;
+	str pn_provider_val;
+	str pn_prid_val;
+	str pn_param_val;
+	str pn_purr_val;
+	/* XXX - in the future when adding params as special links
+	 * in the list above, make sure to also update compare_uris() function
+	 * to explicitly compare these here */
+
 	/* unknown params */
 	str u_name[URI_MAX_U_PARAMS]; /* Unknown param names */
 	str u_val[URI_MAX_U_PARAMS];  /* Unknown param valss */
@@ -248,10 +280,10 @@ struct sip_msg {
 	struct hdr_field* www_authenticate;
 	struct hdr_field* proxy_authenticate;
 	struct hdr_field* min_expires;
+	struct hdr_field* feature_caps;
+	struct hdr_field* replaces;
 
-	struct sdp_info* sdp;
-
-	struct multi_body * multi;
+	struct sip_msg_body *body;
 
 	char* eoh;        /* pointer to the end of header (if found) or null */
 	char* unparsed;   /* here we stopped parsing*/
@@ -311,6 +343,12 @@ struct sip_msg {
 	str set_global_address;
 	str set_global_port;
 
+	/* used to store a particular time of the message - note that the time is
+	 * not stored when the message was received, but only the first time
+	 * someone gets interested in it - this way we have a consistent timestamp
+	 * of the message, without being affected by lazy callbacks */
+	struct timeval time;
+
 	struct msg_callback *msg_cb;
 };
 
@@ -323,14 +361,23 @@ struct sip_msg {
 
 extern int via_cnt;
 
-int parse_msg(char* buf, unsigned int len, struct sip_msg* msg);
+#define parse_msg( _buf, _len, _msg) \
+	parse_msg_opt( _buf, _len, _msg, 1)
+
+int parse_msg_opt(char* buf, unsigned int len, struct sip_msg* msg,
+		int free_on_err);
 
 int parse_headers(struct sip_msg* msg, hdr_flags_t flags, int next);
 
 char* get_hdr_field(char* buf, char* end, struct hdr_field* hdr);
 
+/* add DEL lumps for all headers matching the given @hdr */
+int delete_headers(struct sip_msg *msg, struct hdr_field *hdr);
+
 void free_sip_msg(struct sip_msg* msg);
 
+int clone_headers(struct sip_msg *from_msg, struct sip_msg *to_msg);
+
 /* make sure all HFs needed for transaction identification have been
    parsed; return 0 if those HFs can't be found
  */
@@ -402,12 +449,14 @@ inline static int get_body(struct sip_msg *msg, str *body)
 	body->len = msg->buf + msg->len - body->s;
 
 	/* double check the len against content-length hdr
-	   (if present, it must be already parsed) */
+	   (if present, it must be already parsed);
+	   NOTE that the CT hdr may be missing if using UDP, so 
+	   we do not consider its missing an err case */
 	if (msg->content_length) {
 		ct_len = get_content_length( msg );
 		if (ct_len<body->len)
 			body->len = ct_len;
-	} else {
+	} else if (msg->rcv.proto!=PROTO_UDP) {
 		/* no ct -> no body */
 		body->s = NULL;
 		body->len = 0;
@@ -416,6 +465,28 @@ inline static int get_body(struct sip_msg *msg, str *body)
 	return 0;
 }
 
+/*
+ * Get the callid of a message. If returned value is 0, the callid is stored
+ * in the _cid field, otherwise -1 is returned on error
+ */
+inline static int get_callid(struct sip_msg* _m, str* _cid)
+{
+	if ((parse_headers(_m, HDR_CALLID_F, 0) == -1)) {
+		LM_ERR("failed to parse call-id header\n");
+		return -1;
+	}
+
+	if (_m->callid == NULL) {
+		LM_ERR("call-id not found\n");
+		return -1;
+	}
+
+	_cid->s = _m->callid->body.s;
+	_cid->len = _m->callid->body.len;
+	trim(_cid);
+	return 0;
+}
+
 
 /*
  * Search through already parsed headers (no parsing done) a non-standard
@@ -448,6 +519,27 @@ int set_ruri(struct sip_msg* msg, str* uri);
 int set_dst_uri(struct sip_msg* msg, str* uri);
 
 
+void reset_dst_uri(struct sip_msg *msg);
+
+
+int set_dst_host_port(struct sip_msg *msg, str *host, str *port);
+
+
+enum rw_ruri_part {
+	RW_RURI_HOST = 1,
+	RW_RURI_HOSTPORT,
+	RW_RURI_USER,
+	RW_RURI_USERPASS,
+	RW_RURI_PORT,
+	RW_RURI_PREFIX,
+	RW_RURI_STRIP,
+	RW_RURI_STRIP_TAIL
+};
+
+int rewrite_ruri(struct sip_msg *msg, str *sval, int ival,
+				enum rw_ruri_part part);
+
+
 /*
  * Set the q value of the Request-URI
  */
@@ -480,6 +572,7 @@ int set_dst_uri(struct sip_msg* msg, str* uri);
  * Make a private copy of the string and assign it to path_vec
  */
 int set_path_vector(struct sip_msg* msg, str* path);
+void clear_path_vector(struct sip_msg* msg);
 
 
 /*
@@ -488,4 +581,16 @@ int set_path_vector(struct sip_msg* msg, str* path);
  */
 int extract_ftc_hdrs( char *buf, int len, str *from, str *to, str *cseq,str *callid);
 
+inline static struct timeval *get_msg_time(struct sip_msg *msg)
+{
+	static struct timeval static_time;
+	if (!msg || msg == (struct sip_msg *)-1) {
+		gettimeofday(&static_time, NULL);
+		return &static_time;
+	}
+	if (msg->time.tv_sec == 0 && msg->time.tv_usec == 0)
+		gettimeofday(&msg->time, NULL);
+	return &msg->time;
+}
+
 #endif
