@@ -81,14 +81,18 @@ void free_to_params(struct to_body* tb)
 
 void free_to(struct to_body* tb)
 {
-	free_to_params(tb);
-	pkg_free(tb);
+	if (tb) {
+		free_to( tb->next );
+		free_to_params(tb);
+		pkg_free(tb);
+	}
 }
 
 
 static inline char* parse_to_param(char *buffer, char *end,
 					struct to_body *to_b,
-					int *returned_status)
+					int *returned_status,
+					int multi)
 {
 	struct to_param *param;
 	int status;
@@ -197,9 +201,12 @@ static inline char* parse_to_param(char *buffer, char *end,
 						goto parse_error;
 				}
 				break;
-			case 0:
+			case  0:
+			case ',':
 				switch (status)
 				{
+					case PARA_VALUE_QUOTED:
+						break;
 					case PARA_NAME:
 						param->name.len = tmp-param->name.s;
 					case S_EQUAL:
@@ -213,6 +220,8 @@ static inline char* parse_to_param(char *buffer, char *end,
 						add_param( param , to_b );
 					case E_PARA_VALUE:
 						saved_status = status;
+						if ( !multi && *tmp==',')
+							goto parse_error;
 						goto endofheader;
 						break;
 					default:
@@ -223,6 +232,8 @@ static inline char* parse_to_param(char *buffer, char *end,
 				switch (status)
 				{
 					case PARA_VALUE_QUOTED:
+						if (tmp+1==end)
+							goto parse_error;
 						switch (*(tmp+1))
 						{
 							case '\r':
@@ -444,8 +455,13 @@ static inline char* parse_to_param(char *buffer, char *end,
 		}/*switch*/
 	}/*for*/
 
+	if (status==PARA_VALUE_QUOTED) {
+			LM_ERR("unexpected end of header in state %d\n", status);
+			goto parse_error;
+	}
 
 endofheader:
+	LM_DBG("end of header reached, state=%d\n", status);
 	if (param) {
 		if (saved_status==S_EQUAL||saved_status==S_PARA_VALUE) {
 			saved_status = E_PARA_VALUE;
@@ -463,7 +479,7 @@ endofheader:
 
 parse_error:
 	LM_ERR("unexpected char [%c] in status %d: <<%.*s>> .\n",
-		*tmp,status, (int)(tmp-buffer), ZSW(buffer));
+	    tmp < end? *tmp : *(end-1),status, (int)(tmp-buffer), ZSW(buffer));
 error:
 	if (param) pkg_free(param);
 	free_to_params(to_b);
@@ -475,12 +491,14 @@ error:
 
 
 
-char* parse_to(char* buffer, char *end, struct to_body *to_b)
+static inline char* _parse_to(char* buffer, char *end, struct to_body *to_b,
+																	int multi)
 {
 	int status;
 	int saved_status;
 	char  *tmp;
 	char  *end_mark;
+	struct to_body *first_b = to_b;
 
 	status=START_TO;
 	saved_status=START_TO;
@@ -547,6 +565,7 @@ char* parse_to(char* buffer, char *end, struct to_body *to_b)
 					case URI_OR_TOKEN:
 						end_mark = tmp;
 						status = MAYBE_URI_END;
+						/* fall through */
 					case MAYBE_URI_END:
 					case DISPLAY_TOKEN:
 					case DISPLAY_TOKEN2:
@@ -570,6 +589,7 @@ char* parse_to(char* buffer, char *end, struct to_body *to_b)
 					case URI_OR_TOKEN:
 					case MAYBE_URI_END:
 						to_b->uri.len = tmp - to_b->uri.s;
+						/* fall through */
 					case END:
 						saved_status = status = END;
 						goto endofheader;
@@ -577,6 +597,45 @@ char* parse_to(char* buffer, char *end, struct to_body *to_b)
 						goto parse_error;
 				}
 				break;
+			case ',':
+				switch (status)
+				{
+					case DISPLAY_QUOTED:
+					case URI_ENCLOSED:
+						break;
+					case URI_OR_TOKEN:
+						/* the next transition cannot be determined here. The
+						 * ',' maybe part of the username inside URI, or 
+						 * it can be separator between 2 hdr parts. As this
+						 * parsed is not URI aware (we do not actually parse
+						 * the URI, but we simply skip it), we have no idea
+						 * in which care we are..... For the moment, if the
+						 * header is marked as single part, at least let's
+						 * consider the ',' as part of the URI */
+						if (multi==0)
+							break;
+					case MAYBE_URI_END:
+						to_b->uri.len = tmp - to_b->uri.s;
+						/* fall through */
+					case END:
+						if (multi==0)
+							goto parse_error;
+						to_b->next = (struct to_body*)
+							pkg_malloc(sizeof(struct to_body));
+						if (to_b->next==NULL) {
+							LM_ERR("failed to allocate new TO body\n");
+							goto error;
+						}
+						to_b = to_b->next;
+						memset(to_b, 0, sizeof(struct to_body));
+						to_b->error = PARSE_OK;
+						saved_status = status = START_TO;
+						end_mark=0;
+						break;
+					default:
+						goto parse_error;
+				}
+				break;
 			case '\\':
 				switch (status)
 				{
@@ -602,6 +661,7 @@ char* parse_to(char* buffer, char *end, struct to_body *to_b)
 					case URI_OR_TOKEN:
 					case DISPLAY_TOKEN:
 						end_mark = tmp;
+						/* fall through */
 					case DISPLAY_TOKEN2:
 					case MAYBE_URI_END:
 						to_b->display.len=end_mark-to_b->display.s;
@@ -623,6 +683,7 @@ char* parse_to(char* buffer, char *end, struct to_body *to_b)
 						break;
 					case URI_ENCLOSED:
 						to_b->uri.len = tmp - to_b->uri.s;
+						/* fall through */
 					case E_URI_ENCLOSED:
 						status = END;
 						break;
@@ -665,12 +726,30 @@ char* parse_to(char* buffer, char *end, struct to_body *to_b)
 						break;
 					case URI_OR_TOKEN:
 						end_mark = tmp;
+						/* fall through */
 					case MAYBE_URI_END:
 						to_b->uri.len = end_mark - to_b->uri.s;
+						/* fall through */
 					case END:
 						to_b->body.len = tmp-to_b->body.s;
-						tmp = parse_to_param(tmp,end,to_b,&saved_status);
-						goto endofheader;
+						tmp = parse_to_param(tmp,end,to_b,&saved_status,multi);
+						if (to_b->error!=PARSE_ERROR && multi && *tmp==',') {
+							/* continue with a new body instance */
+							to_b->next = (struct to_body*)
+								pkg_malloc(sizeof(struct to_body));
+							if (to_b->next==NULL) {
+								LM_ERR("failed to allocate new TO body\n");
+								goto error;
+							}
+							to_b = to_b->next;
+							memset(to_b, 0, sizeof(struct to_body));
+							to_b->error=PARSE_OK;
+							saved_status = status = START_TO;
+							end_mark=0;
+							break;
+						} else {
+							goto endofheader;
+						}
 					case F_CRLF:
 					case F_LF:
 					case F_CR:
@@ -737,15 +816,28 @@ endofheader:
 
 parse_error:
 	LM_ERR("unexpected char [%c] in status %d: <<%.*s>> .\n",
-		*tmp,status, (int)(tmp-buffer), buffer);
+	    tmp < end? *tmp : *(end-1), status, (int)(tmp-buffer), buffer);
 error:
-	to_b->error=PARSE_ERROR;
-	free_to_params(to_b);
+	first_b->error=PARSE_ERROR;
+	free_to_params(first_b);
+	free_to(first_b->next);
 	return tmp;
 
 }
 
 
+char* parse_to(char* buffer, char *end, struct to_body *to_b)
+{
+	return _parse_to( buffer, end, to_b, 0/*multi*/);
+}
+
+
+char* parse_multi_to(char* buffer, char *end, struct to_body *to_b)
+{
+	return _parse_to( buffer, end, to_b, 1/*multi*/);
+}
+
+
 /**
  *
  */
@@ -811,3 +903,20 @@ int parse_to_header( struct sip_msg *msg)
 error:
 	return -1;
 }
+
+
+/*
+ * Parses the URI from a generic to_body structure
+ * Helper function (not specific to TO hdr)
+ */
+int parse_to_body_uri(struct to_body *to_b)
+{
+	if (to_b==NULL)
+		return -1;
+
+	if (parse_uri(to_b->uri.s, to_b->uri.len, &to_b->parsed_uri) < 0) {
+		memset( &to_b->parsed_uri, 0, sizeof(struct sip_uri));
+		return -1;
+	}
+	return 0;
+}
