@@ -43,6 +43,55 @@
 #include "../core_stats.h"
 #include "../strcommon.h"
 
+static const str uri_type_names[7] = {
+	{NULL, 0}, /*This is the error type*/
+	str_init("sip"),
+	str_init("sips"),
+	str_init("tel"),
+	str_init("tels"),
+	str_init("urn:service"),
+	str_init("urn:nena:service")
+};
+
+char* uri_type2str(const uri_type type, char *result)
+{
+	if (type == ERROR_URI_T)
+		return NULL;
+
+	memcpy(result, uri_type_names[type].s, uri_type_names[type].len);
+	return result + uri_type_names[type].len;
+}
+
+int uri_typestrlen(const uri_type type)
+{
+	return uri_type_names[type].len;
+}
+
+uri_type str2uri_type(char * buf)
+{
+	int scheme = 0;
+	uri_type type = ERROR_URI_T;
+	scheme=buf[0]+(buf[1]<<8)+(buf[2]<<16)+(buf[3]<<24);
+	scheme|=0x20202020;
+	if (scheme==SIP_SCH){
+		type=SIP_URI_T;
+	}else if(scheme==SIPS_SCH){
+		if(buf[4]==':')
+			type=SIPS_URI_T;
+		else type = ERROR_URI_T;
+	}else if (scheme==TEL_SCH){
+		type=TEL_URI_T;
+	}else if (scheme==URN_SERVICE_SCH){
+		if (memcmp(buf+3,URN_SERVICE_STR,URN_SERVICE_STR_LEN) == 0) {
+			type=URN_SERVICE_URI_T;
+		}
+		else if (memcmp(buf+3,URN_NENA_SERVICE_STR,URN_NENA_SERVICE_STR_LEN) == 0) {
+			type=URN_NENA_SERVICE_URI_T;
+		}
+	}
+	return type;
+}
+
 int parse_uri_headers(str headers, str h_name[], str h_val[], int h_size)
 {
 	enum states {URI_H_HEADER, URI_H_VALUE};
@@ -148,6 +197,108 @@ int parse_uri_headers(str headers, str h_name[], str h_val[], int h_size)
 	return 0;
 }
 
+int print_uri(struct sip_uri *uri, str *out_buf)
+{
+#define append_str_chunk(field) \
+	do { \
+		if (bytes + uri->field.len > out_buf->len) { \
+			LM_ERR("no more space left! printed so far: '%.*s'\n", \
+		           bytes, out_buf->s); \
+			return -1; \
+		} \
+		memcpy(out_buf->s + bytes, uri->field.s, uri->field.len); \
+		bytes += uri->field.len; \
+	} while (0)
+
+#define append_char(ch) \
+	do { \
+		if (bytes + 1 > out_buf->len) { \
+			LM_ERR("no more space left! printed so far: '%.*s'\n", \
+		           bytes, out_buf->s); \
+			return -1; \
+		} \
+		out_buf->s[bytes++] = ch; \
+	} while (0)
+
+#define VAL(p) p##_val
+
+#define append_param(p) \
+	do { \
+		if (uri->p.s) { \
+			append_char(';'); \
+			append_str_chunk(p); \
+		} \
+	} while (0)
+
+#define append_uk_param(idx) \
+	do { \
+		if (uri->u_name[idx].s) { \
+			append_char(';'); \
+			if (bytes + uri->u_name[idx].len > out_buf->len) { \
+				LM_ERR("no more space left! printed so far: '%.*s'\n", \
+			           bytes, out_buf->s); \
+				return -1; \
+			} \
+			memcpy(out_buf->s + bytes, uri->u_name[idx].s, uri->u_name[idx].len); \
+			bytes += uri->u_name[idx].len; \
+			if (uri->u_val[idx].s) { \
+				append_char('='); \
+				if (bytes + uri->u_val[idx].len > out_buf->len) { \
+					LM_ERR("no more space left! printed so far: '%.*s'\n", \
+				           bytes, out_buf->s); \
+					return -1; \
+				} \
+				memcpy(out_buf->s + bytes, uri->u_val[idx].s, uri->u_val[idx].len); \
+				bytes += uri->u_val[idx].len; \
+			} \
+		} \
+	} while (0)
+
+	int bytes = 0;
+	int i;
+
+	memcpy(out_buf->s, uri_type_names[uri->type].s, uri_type_names[uri->type].len);
+	bytes += uri_type_names[uri->type].len;
+	append_char(':');
+	append_str_chunk(user);
+	if (uri->passwd.s) {
+		append_char(':');
+		append_str_chunk(passwd);
+	}
+	if (uri->host.s) {
+		append_char('@');
+		append_str_chunk(host);
+	}
+	if (uri->port.s) {
+		append_char(':');
+		append_str_chunk(port);
+	}
+
+	append_param(transport);
+	append_param(ttl);
+	append_param(user_param);
+	append_param(maddr);
+	append_param(method);
+	append_param(lr);
+	append_param(r2);
+	append_param(gr);
+	append_param(pn_provider);
+	append_param(pn_prid);
+	append_param(pn_param);
+	append_param(pn_purr);
+
+	for (i = 0; i < uri->u_params_no; i++)
+		append_uk_param(i);
+
+	out_buf->len = bytes;
+
+	return 0;
+#undef append_str_chunk
+#undef append_char
+#undef VAL
+#undef append_param
+#undef append_uk_param
+}
 
 /* buf= pointer to beginning of uri (sip:x@foo.bar:5060;a=b?h=i)
  * len= len of uri
@@ -188,7 +339,15 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 					/* sctp */
 					VS_S, VS_C, VS_T, VS_P_FIN,
 					/* ws */
-					VW_W, VW_S, VW_S_FIN, VWS_S_FIN
+					VW_W, VW_S, VW_S_FIN, VWS_S_FIN,
+
+					/* pn-{provider, prid, param, purr} (RFC 8599 - SIP PN) */
+					PN_P, PN_N, PN_dash, PN_P2, PN_PR,
+					PN1_O, PN1_V, PN1_I, PN1_D, PN1_E, PN1_FIN, PN1_eq,
+					PN2_I, PN2_D, PN2_eq,
+					PN3_A, PN3_R, PN3_A2, PN3_M, PN3_eq,
+					PN4_U, PN4_R, PN4_R2, PN4_eq,
+
 	};
 	register enum states state;
 	char* s;
@@ -198,7 +357,7 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 	str* param_val; /* current param val */
 	str user;
 	str password;
-	int port_no;
+	unsigned int port_no;
 	register char* p;
 	char* end;
 	char* pass;
@@ -210,17 +369,13 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 	int i;
 #endif
 
-#define SIP_SCH			0x3a706973
-#define SIPS_SCH		0x73706973
-#define TEL_SCH			0x3a6c6574
-#define URN_SERVICE_SCH		0x3a6e7275
-#define URN_SERVICE_STR 	":service:"
-#define URN_SERVICE_STR_LEN	(sizeof(URN_SERVICE_STR) - 1)
-
-#define case_port( ch, var) \
+#define case_port( ch, var, ovf_check1, ovf_check2) \
 	case ch: \
-			 (var)=(var)*10+ch-'0'; \
-			 break
+			if (ovf_check1) \
+				(var)=(var)*10+(ch-'0'); \
+			if (ovf_check2 && (var) > USHRT_MAX) \
+				goto error_bad_port; \
+			break
 
 #define still_at_user  \
 						if (found_user==0){ \
@@ -391,6 +546,16 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 						state=URI_PARAM_P; \
 				} \
 				break
+#define param_xswitch1(old_state, c1, new_state) \
+			case old_state: \
+				switch(*p){ \
+					case c1: \
+						state=(new_state); \
+						break; \
+					default: \
+						goto error_bad_char; \
+				} \
+				break
 #define param_switch_big(old_state, c1, c2, d1, d2, new_state_c, new_state_d) \
 			case old_state : \
 				switch(*p){ \
@@ -407,6 +572,26 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 						state=URI_PARAM_P; \
 				} \
 				break
+#define param_switch_bigger(old_state, c1, c2, d1, d2, e1, e2, new_state_c, new_state_d, new_state_e) \
+			case old_state : \
+				switch(*p){ \
+					case c1: \
+					case c2: \
+						state=(new_state_c); \
+						break; \
+					case d1: \
+					case d2: \
+						state=(new_state_d); \
+						break; \
+					case e1: \
+					case e2: \
+						state=(new_state_e); \
+						break; \
+					u_param_common_cases; \
+					default: \
+						state=URI_PARAM_P; \
+				} \
+				break
 #define value_switch(old_state, c1, c2, new_state) \
 			case old_state: \
 				switch(*p){ \
@@ -474,7 +659,10 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 	memset(uri, 0, sizeof(struct sip_uri)); /* zero it all, just to be sure*/
 	/*look for sip:, sips: or tel:*/
 	if (len<5) goto error_too_short;
-	scheme=buf[0]+(buf[1]<<8)+(buf[2]<<16)+(buf[3]<<24);
+	scheme=(unsigned)(unsigned char)buf[0]
+			+ (((unsigned)(unsigned char)buf[1])<<8)
+			+ (((unsigned)(unsigned char)buf[2])<<16)
+			+ (((unsigned)(unsigned char)buf[3])<<24);
 	scheme|=0x20202020;
 	if (scheme==SIP_SCH){
 		uri->type=SIP_URI_T;
@@ -484,10 +672,16 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 	}else if (scheme==TEL_SCH){
 		uri->type=TEL_URI_T;
 	}else if (scheme==URN_SERVICE_SCH){
-		if (memcmp(buf+3,URN_SERVICE_STR,URN_SERVICE_STR_LEN) == 0) {
+		if ((end-(buf+3)) >= URN_SERVICE_STR_LEN
+		        && memcmp(buf+3,URN_SERVICE_STR,URN_SERVICE_STR_LEN) == 0) {
 			p+= URN_SERVICE_STR_LEN-1;
 			uri->type=URN_SERVICE_URI_T;
-		} else goto error_bad_uri;
+		}
+		else if ((end-(buf+3)) >= URN_NENA_SERVICE_STR_LEN
+		        && memcmp(buf+3,URN_NENA_SERVICE_STR,URN_NENA_SERVICE_STR_LEN) == 0) {
+			p+= URN_NENA_SERVICE_STR_LEN-1;
+			uri->type=URN_NENA_SERVICE_URI_T;
+		}else goto error_bad_uri;
 	}else goto error_bad_uri;
 
 	s=p;
@@ -587,16 +781,16 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 						found_user=1; /*  there is no user part */
 						s=p+1;
 						break;
-					case_port('0', port_no);
-					case_port('1', port_no);
-					case_port('2', port_no);
-					case_port('3', port_no);
-					case_port('4', port_no);
-					case_port('5', port_no);
-					case_port('6', port_no);
-					case_port('7', port_no);
-					case_port('8', port_no);
-					case_port('9', port_no);
+					case_port('0', port_no, port_no < INT_MAX / 10, 0);
+					case_port('1', port_no, port_no < INT_MAX / 10, 0);
+					case_port('2', port_no, port_no < INT_MAX / 10, 0);
+					case_port('3', port_no, port_no < INT_MAX / 10, 0);
+					case_port('4', port_no, port_no < INT_MAX / 10, 0);
+					case_port('5', port_no, port_no < INT_MAX / 10, 0);
+					case_port('6', port_no, port_no < INT_MAX / 10, 0);
+					case_port('7', port_no, port_no < INT_MAX / 10, 0);
+					case_port('8', port_no, port_no < INT_MAX / 10, 0);
+					case_port('9', port_no, port_no < INT_MAX / 10, 0);
 					case '[':
 					case ']':
 					case ':':
@@ -682,16 +876,16 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 						state=URI_HEADERS;
 						s=p+1;
 						break;
-					case_port('0', port_no);
-					case_port('1', port_no);
-					case_port('2', port_no);
-					case_port('3', port_no);
-					case_port('4', port_no);
-					case_port('5', port_no);
-					case_port('6', port_no);
-					case_port('7', port_no);
-					case_port('8', port_no);
-					case_port('9', port_no);
+					case_port('0', port_no, 1, 1);
+					case_port('1', port_no, 1, 1);
+					case_port('2', port_no, 1, 1);
+					case_port('3', port_no, 1, 1);
+					case_port('4', port_no, 1, 1);
+					case_port('5', port_no, 1, 1);
+					case_port('6', port_no, 1, 1);
+					case_port('7', port_no, 1, 1);
+					case_port('8', port_no, 1, 1);
+					case_port('9', port_no, 1, 1);
 					case '&':
 					case '@':
 					case ':':
@@ -733,6 +927,11 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 						b=p;
 						state=PR2_R;
 						break;
+					case 'p':
+					case 'P':
+						b=p;
+						state=PN_P;
+						break;
 					default:
 						b=p;
 						state=URI_PARAM_P;
@@ -956,7 +1155,7 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 						state=URI_PARAM_P;
 				}
 				break;
-				/* handle lr=something case */
+				/* handle r2=something case */
 			case PR2_eq:
 				param=&uri->r2;
 				param_val=&uri->r2_val;
@@ -1006,6 +1205,102 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 				break;
 
 
+			/* pn-* */
+			param_switch(PN_P, 'n', 'N', PN_N);
+			param_switch1(PN_N, '-', PN_dash);
+			param_switch(PN_dash, 'p', 'P', PN_P2);
+
+			param_switch_bigger(PN_P2, 'r', 'R', 'a', 'A', 'u', 'U',
+			                    PN_PR, PN3_A, PN4_U);
+			param_switch_big(PN_PR, 'o', 'O', 'i', 'I', PN1_O, PN2_I);
+
+			/* pn-provider */
+			param_switch(PN1_O, 'v', 'V', PN1_V);
+			param_switch(PN1_V, 'i', 'I', PN1_I);
+			param_switch(PN1_I, 'd', 'D', PN1_D);
+			param_switch(PN1_D, 'e', 'E', PN1_E);
+			param_switch(PN1_E, 'r', 'R', PN1_FIN);
+			case PN1_FIN:
+				param=&uri->pn_provider;
+				switch(*p){
+					case '@':
+						still_at_user;
+						break;
+					case '=':
+						state=PN1_eq;
+						break;
+					semicolon_case;
+						uri->pn_provider.s=b;
+						uri->pn_provider.len=(p-b);
+						break;
+					question_case;
+						uri->pn_provider.s=b;
+						uri->pn_provider.len=(p-b);
+						break;
+					colon_case;
+						break;
+					default:
+						state=URI_PARAM_P;
+				}
+				break;
+				/* handle pn-provider=something case */
+			case PN1_eq:
+				param=&uri->pn_provider;
+				param_val=&uri->pn_provider_val;
+				switch(*p){
+					param_common_cases;
+					default:
+						v=p;
+						state=URI_VAL_P;
+				}
+				break;
+
+			/* pn-prid */
+			param_switch(PN2_I, 'd', 'D', PN2_D);
+			param_xswitch1(PN2_D, '=', PN2_eq);
+			case PN2_eq:
+				param=&uri->pn_prid;
+				param_val=&uri->pn_prid_val;
+				switch(*p){
+					param_common_cases;
+					default:
+						v=p;
+						state=URI_VAL_P;
+				}
+				break;
+
+			/* pn-param */
+			param_switch(PN3_A, 'r', 'R', PN3_R);
+			param_switch(PN3_R, 'a', 'A', PN3_A2);
+			param_switch(PN3_A2, 'm', 'M', PN3_M);
+			param_xswitch1(PN3_M, '=', PN3_eq);
+			case PN3_eq:
+				param=&uri->pn_param;
+				param_val=&uri->pn_param_val;
+				switch(*p){
+					param_common_cases;
+					default:
+						v=p;
+						state=URI_VAL_P;
+				}
+				break;
+
+			/* pn-purr */
+			param_switch(PN4_U, 'r', 'R', PN4_R);
+			param_switch(PN4_R, 'r', 'R', PN4_R2);
+			param_xswitch1(PN4_R2, '=', PN4_eq);
+			case PN4_eq:
+				param=&uri->pn_purr;
+				param_val=&uri->pn_purr_val;
+				switch(*p){
+					param_common_cases;
+					default:
+						v=p;
+						state=URI_VAL_P;
+				}
+				break;
+
+
 			case URI_HEADERS:
 				/* for now nobody needs them so we completely ignore the
 				 * headers (they are not allowed in request uri) --andrei */
@@ -1061,6 +1356,7 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 		case URI_PASSWORD:
 			/* this is the port, it can't be the passwd */
 			if (found_user) goto error_bad_port;
+			if (port_no > USHRT_MAX) goto error_bad_port;
 			uri->port.s=s;
 			uri->port.len=p-s;
 			uri->port_no=port_no;
@@ -1142,6 +1438,13 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 			uri->gr.s=b;
 			uri->gr.len=p-b;
 			break;
+		case PN1_FIN:
+		case PN1_eq:
+			uri->params.s=s;
+			uri->params.len=p-s;
+			uri->pn_provider.s=b;
+			uri->pn_provider.len=p-b;
+			break;
 		case URI_VAL_P:
 		/* intermediate value states */
 		case VU_U:
@@ -1201,44 +1504,54 @@ int parse_uri(char* buf, int len, struct sip_uri* uri)
 			uri->headers.len=p-s;
 			if (error_headers) goto error_headers;
 			break;
+		/* intermediate PN param states */
+		case PN_P:
+		case PN_N:
+		case PN_dash:
+		case PN_P2:
+		case PN_PR:
+		case PN1_O:
+		case PN1_V:
+		case PN1_I:
+		case PN1_D:
+		case PN1_E:
+		case PN2_I:
+		case PN3_A:
+		case PN3_R:
+		case PN3_A2:
+		case PN4_U:
+		case PN4_R:
+			uri->params.s=s;
+			uri->params.len=p-s;
+			break;
+		case PN2_D:
+		case PN2_eq:
+		case PN3_M:
+		case PN3_eq:
+		case PN4_R2:
+		case PN4_eq:
+			goto error_bad_uri;
 		default:
 			goto error_bug;
 	}
 	switch(uri->type){
-		case SIP_URI_T:
-			if ((uri->user_param_val.len == 5) &&
-				(strncasecmp(uri->user_param_val.s, "phone", 5) == 0)) {
-				uri->type = TEL_URI_T;
-				/* move params from user into uri->params */
-				p=q_memchr(uri->user.s, ';', uri->user.len);
-				if (p){
-					uri->params.s=p+1;
-					uri->params.len=uri->user.s+uri->user.len-uri->params.s;
-					uri->user.len=p-uri->user.s;
-				}
-			}
-			break;
-		case SIPS_URI_T:
-			if ((uri->user_param_val.len == 5) &&
-				(strncasecmp(uri->user_param_val.s, "phone", 5) == 0)) {
-				uri->type = TELS_URI_T;
-				p=q_memchr(uri->user.s, ';', uri->user.len);
-				if (p){
-					uri->params.s=p+1;
-					uri->params.len=uri->user.s+uri->user.len-uri->params.s;
-					uri->user.len=p-uri->user.s;
-				}
-			}
-			break;
 		case TEL_URI_T:
 		case TELS_URI_T:
 			/* fix tel uris, move the number in uri and empty the host */
 			uri->user=uri->host;
-			uri->host.s="";
+			/* TEL does not have a host part, still most of the code expects
+			 * one, so lets keep the pointer, but set a 0 length */
 			uri->host.len=0;
 			break;
+		case SIP_URI_T:
+		case SIPS_URI_T:
 		case URN_SERVICE_URI_T:
-			/* leave the actual service name in the URI domain part */
+			/* nothing to do for these URIs */
+			break;
+		case URN_NENA_SERVICE_URI_T:
+			uri->user.s=0;
+			uri->user.len=0;
+			/* keep the service name as host part */
 			break;
 		case ERROR_URI_T:
 			LM_ERR("unexpected error (BUG?)\n");
@@ -1293,20 +1606,20 @@ error_too_short:
 error_bad_char:
 	LM_ERR("bad char '%c' in state %d"
 			" parsed: <%.*s> (%d) / <%.*s> (%d)\n",
-			*p, state, (int)(p-buf), ZSW(buf), (int)(p-buf),
-			len, ZSW(buf), len);
+			p < end ? *p : *(buf+len-1), state, (int)(p-buf), ZSW(buf),
+			(int)(p-buf), len, ZSW(buf), len);
 	goto error_exit;
 error_bad_host:
 	LM_ERR("bad host in uri (error at char %c in"
 			" state %d) parsed: <%.*s>(%d) /<%.*s> (%d)\n",
-			*p, state, (int)(p-buf), ZSW(buf), (int)(p-buf),
-			len, ZSW(buf), len);
+			p < end ? *p : *(buf+len-1), state, (int)(p-buf), ZSW(buf),
+			(int)(p-buf), len, ZSW(buf), len);
 	goto error_exit;
 error_bad_port:
-	LM_ERR("bad port in uri (error at char %c in"
+	LM_ERR("bad port in uri (error at char '%c' in"
 			" state %d) parsed: <%.*s>(%d) /<%.*s> (%d)\n",
-			*p, state, (int)(p-buf), ZSW(buf), (int)(p-buf),
-			len, ZSW(buf), len);
+			p < end ? *p : *(buf+len-1), state, (int)(p-buf), ZSW(buf),
+			(int)(p-buf), len, ZSW(buf), len);
 	goto error_exit;
 error_bad_uri:
 	LM_ERR("bad uri, state %d parsed: <%.*s> (%d) / <%.*s> (%d)\n",
@@ -1319,7 +1632,7 @@ error_headers:
 			len, ZSW(buf), len);
 	goto error_exit;
 error_bug:
-	LM_CRIT("bad state %d parsed: <%.*s> (%d) / <%.*s> (%d)\n",
+	LM_ERR("bad state %d parsed: <%.*s> (%d) / <%.*s> (%d)\n",
 			 state, (int)(p-buf), ZSW(buf), (int)(p-buf), len, ZSW(buf), len);
 error_exit:
 	ser_error=E_BAD_URI;
@@ -1486,6 +1799,11 @@ int compare_uris(str *raw_uri_a,struct sip_uri* parsed_uri_a,
 	compare_uri_val(method_val,strncasecmp);
 	compare_uri_val(lr_val,strncasecmp);
 	compare_uri_val(r2_val,strncasecmp);
+	compare_uri_val(gr_val,strncasecmp);
+	compare_uri_val(pn_provider_val,strncasecmp);
+	compare_uri_val(pn_prid_val,strncasecmp);
+	compare_uri_val(pn_param_val,strncasecmp);
+	compare_uri_val(pn_purr_val,strncasecmp);
 
 	if (first.u_params_no == 0 || second.u_params_no == 0)
 		/* one URI doesn't have other params,
@@ -1542,48 +1860,3 @@ headers_check:
 	compare_uri_val(headers,strncasecmp);
 	return 0;
 }
-
-static const str uri_type_names[6] = {
-	{NULL, 0}, /*This is the error type*/
-	str_init("sip"),
-	str_init("sips"),
-	str_init("tel"),
-	str_init("tels"),
-	str_init("urn:service")
-};
-
-char* uri_type2str(const uri_type type, char *result)
-{
-	if (type == ERROR_URI_T)
-		return NULL;
-
-	memcpy(result, uri_type_names[type].s, uri_type_names[type].len);
-	return result + uri_type_names[type].len;
-}
-
-int uri_typestrlen(const uri_type type)
-{
-	return uri_type_names[type].len;
-}
-
-uri_type str2uri_type(char * buf)
-{
-	int scheme = 0;
-	uri_type type = ERROR_URI_T;
-	scheme=buf[0]+(buf[1]<<8)+(buf[2]<<16)+(buf[3]<<24);
-	scheme|=0x20202020;
-	if (scheme==SIP_SCH){
-		type=SIP_URI_T;
-	}else if(scheme==SIPS_SCH){
-		if(buf[4]==':')
-			type=SIPS_URI_T;
-		else type = ERROR_URI_T;
-	}else if (scheme==TEL_SCH){
-		type=TEL_URI_T;
-	}else if (scheme==URN_SERVICE_SCH){
-		if (memcmp(buf+3,URN_SERVICE_STR,URN_SERVICE_STR_LEN) == 0) {
-			type=URN_SERVICE_URI_T;
-		}
-	}
-	return type;
-}
