@@ -89,6 +89,9 @@ static char hostbuf[MAX_BUFF_SIZE];
 static char *h_addr_ptrs[MAXADDRS];
 static char *host_aliases[MAXALIASES];
 
+stat_var *dns_total_queries;
+stat_var *dns_slow_queries;
+
 typedef union {
 	int32_t al;
 	char ac;
@@ -113,14 +116,14 @@ typedef union {
 
 # define DNS_GET16(s, cp) \
 	do { \
-		const uint16_t *t_cp = (const uint16_t *) (cp); \
-		(s) = ntohs (*t_cp); \
+		uint16_t t_cp; memcpy(&t_cp, cp, sizeof(uint16_t)); \
+		(s) = ntohs (t_cp); \
 	} while (0)
 
 # define DNS_GET32(s, cp) \
 	do { \
-		const uint32_t *t_cp = (const uint32_t *) (cp); \
-		(s) = ntohl (*t_cp); \
+		uint32_t t_cp; memcpy(&t_cp, cp, sizeof(uint32_t)); \
+		(s) = ntohl (t_cp); \
 	} while (0)
 
 static inline unsigned int dns_get16(const u_char *src) {
@@ -447,58 +450,59 @@ query:
 
 inline struct hostent* resolvehost(char* name, int no_ip_test)
 {
-        static struct hostent* he=0;
+	static struct hostent *he = NULL;
 #ifdef HAVE_GETIPNODEBYNAME
-        int err;
-        static struct hostent* he2=0;
+	int err;
+	static struct hostent *he2 = NULL;
 #endif
-        struct ip_addr* ip;
-        str s;
+	struct timeval start;
+	struct ip_addr *ip;
+	str s;
 
-        if (!no_ip_test) {
-                s.s = (char*)name;
-                s.len = strlen(name);
+	if (!no_ip_test) {
+		s.s = (char *)name;
+		s.len = strlen(name);
 
-                /* check if it's an ip address */
-                if ( ((ip=str2ip(&s))!=0)
-                        || ((ip=str2ip6(&s))!=0)
-                ){
-                        /* we are lucky, this is an ip address */
-                        return ip_addr2he(&s, ip);
-                }
-        }
+		/* check if it's an ip address */
+		if ((ip = str2ip(&s)) || (ip = str2ip6(&s))) {
+			/* we are lucky, this is an ip address */
+			return ip_addr2he(&s, ip);
+		}
+	}
 
-        if(dns_try_ipv6){
-                /*try ipv6*/
-        #ifdef HAVE_GETHOSTBYNAME2
-                if (dnscache_fetch_func != NULL) {
-                        he = own_gethostbyname2(name,AF_INET6);
-                }
-                else {
-                        he=gethostbyname2(name, AF_INET6);
-                }
+	start_expire_timer(start, execdnsthreshold);
 
-        #elif defined HAVE_GETIPNODEBYNAME
-                /* on solaris 8 getipnodebyname has a memory leak,
-                 * after some time calls to it will fail with err=3
-                 * solution: patch your solaris 8 installation */
-                if (he2) freehostent(he2);
-                he=he2=getipnodebyname(name, AF_INET6, 0, &err);
-        #else
-                #error neither gethostbyname2 or getipnodebyname present
-        #endif
-                if (he != 0)
-                        /* return the inet6 result if exists */
-                        return he;
-        }
+	if (dns_try_ipv6) {
+		/* try ipv6 */
+	#ifdef HAVE_GETHOSTBYNAME2
+		if (dnscache_fetch_func)
+	        he = own_gethostbyname2(name,AF_INET6);
+		else
+			he = gethostbyname2(name, AF_INET6);
 
-        if (dnscache_fetch_func != NULL) {
-                he = own_gethostbyname2(name,AF_INET);
-        }
-        else {
-                he=gethostbyname(name);
-        }
-        return he;
+	#elif defined HAVE_GETIPNODEBYNAME
+		/* on solaris 8 getipnodebyname has a memory leak,
+		 * after some time calls to it will fail with err=3
+		 * solution: patch your solaris 8 installation */
+		if (he2) freehostent(he2);
+		he = he2 = getipnodebyname(name, AF_INET6, 0, &err);
+	#else
+		#error "neither gethostbyname2 or getipnodebyname present"
+	#endif
+		if (he != 0)
+			/* return the inet6 result if exists */
+			goto out;
+	}
+
+	if (dnscache_fetch_func)
+		he = own_gethostbyname2(name,AF_INET);
+	else
+		he = gethostbyname(name);
+
+out:
+	_stop_expire_timer(start, execdnsthreshold, "dns",
+	            name, strlen(name), 0, dns_slow_queries, dns_total_queries);
+	return he;
 }
 
 struct hostent * own_gethostbyaddr(void *addr, socklen_t len, int af)
@@ -611,35 +615,19 @@ struct hostent* rev_resolvehost(struct ip_addr *ip)
 int check_ip_address(struct ip_addr* ip, str *name,
 				unsigned short port, unsigned short proto, int resolver)
 {
+	struct ip_addr *ip2;
 	struct hostent* he;
-	int i, len;
-	char* s;
+	int i;
 
-	/* maybe we are lucky and name it's an ip */
-	s=ip_addr2a(ip);
-	if (s){
-		LM_DBG("params %s, %.*s, %d\n", s, name->len, name->s, resolver);
-		len=strlen(s);
-
-		/* check if name->s is an ipv6 address or an ipv6 address ref. */
-		if ((ip->af==AF_INET6) &&
-				(	((len==name->len)&&(strncasecmp(name->s, s, name->len)==0))
-					||
-					((len==(name->len-2))&&(name->s[0]=='[')&&
-						(name->s[name->len-1]==']')&&
-						(strncasecmp(name->s+1, s, len)==0))
-				)
-		   )
+	/* maybe we are lucky and host (name) is an IP */
+	if ( (ip2=str2ip(name))!=NULL || (ip2=str2ip6(name))!=NULL ) {
+		/* It's an IP :D */
+		if (ip_addr_cmp(ip, ip2))
 			return 0;
-		else
-
-			if (strncmp(name->s, s, name->len)==0)
-				return 0;
-	}else{
-		LM_CRIT("could not convert ip address\n");
 		return -1;
 	}
 
+	/* host is not an IP, do the DNS lookups on it :(*/
 	if (port==0) port=SIP_PORT;
 	if (resolver&DO_DNS){
 		LM_DBG("doing dns lookup\n");
@@ -698,6 +686,13 @@ int resolv_init(void)
 #warning "no resolv timeout support"
 	LM_WARN("no resolv options support - resolv options will be ignored\n");
 #endif
+
+	if (register_stat("dns", "dns_total_queries", &dns_total_queries, 0) ||
+	    register_stat("dns", "dns_slow_queries", &dns_slow_queries, 0)) {
+		LM_ERR("failed to register DNS stats\n");
+		return -1;
+	}
+
 	return 0;
 }
 
@@ -709,7 +704,7 @@ int resolv_blacklist_init(void)
 	str name = str_init(DNS_REVOLVER_BL_NAME);
 
 	if (!disable_dns_blacklist) {
-		failover_bl = create_bl_head( DNS_REVOLVER_BL_ID,
+		failover_bl = create_bl_head(_str("dns"),
 			BL_DO_EXPIRE|BL_BY_DEFAULT, 0, 0, &name);
 		if (failover_bl==NULL) {
 			LM_ERR("failed to create blacklist\n");
@@ -957,8 +952,12 @@ struct txt_rdata* dns_txt_parser( unsigned char* msg, unsigned char* end,
 	len = *rdata;
 	if (rdata + 1 + len >= end)
 		goto error;	/*  something fishy in the record */
+
+#if 0
+	/* Comparison is always false because len <= 255. */
 	if (len >= sizeof(txt->txt))
 		goto error; /* not enough space? */
+#endif
 	memcpy(txt->txt, rdata+1, len);
 	txt->txt[len] = 0;		/* 0-terminate string */
 	return txt;
@@ -1094,7 +1093,9 @@ struct rdata* get_record(char* name, int type)
 query:
 	start_expire_timer(start,execdnsthreshold);
 	size=res_search(name, C_IN, type, buff.buff, sizeof(buff));
-	stop_expire_timer(start,execdnsthreshold,"dns",name,strlen(name),0);
+	_stop_expire_timer(start, execdnsthreshold, "dns",
+	            name, strlen(name), 0, dns_slow_queries, dns_total_queries);
+
 	if (size<0) {
 		LM_DBG("lookup(%s, %d) failed\n", name, type);
 		if (dnscache_put_func != NULL) {
@@ -1450,7 +1451,7 @@ static inline void sort_srvs(struct rdata **head)
 				/* order the elements between rd and crt */
 				while (rd->next) {
 					rand_no = (unsigned int)
-						(weight_sum*((float)rand()/RAND_MAX));
+						(weight_sum*((float)rand()/(float)RAND_MAX));
 					for( crt=rd,crt2=NULL ; crt ; crt2=crt,crt=crt->next) {
 						if (rd2srv(crt)->running_sum>=rand_no) break;
 					}
@@ -1751,14 +1752,18 @@ struct hostent* sip_resolvehost( str* name, unsigned short* port,
 	struct rdata *head;
 	struct rdata *rd;
 	struct hostent* he;
+	unsigned short local_proto=PROTO_NONE;
 
 	if (dn)
 		*dn = 0;
 
+	if (proto==NULL)
+		proto = &local_proto;
+
 	/* check if it's an ip address */
 	if ( ((ip=str2ip(name))!=0) || ((ip=str2ip6(name))!=0) ){
 		/* we are lucky, this is an ip address */
-		if (proto && *proto==PROTO_NONE)
+		if (*proto==PROTO_NONE)
 			*proto = (is_sips)?PROTO_TLS:PROTO_UDP;
 		if (port && *port==0)
 			*port = protos[*proto].default_port;
@@ -1770,13 +1775,13 @@ struct hostent* sip_resolvehost( str* name, unsigned short* port,
 		/* have port -> no NAPTR, no SRV lookup, just A record lookup */
 		LM_DBG("has port -> do A record lookup!\n");
 		/* set default PROTO if not set */
-		if (proto && *proto==PROTO_NONE)
+		if (*proto==PROTO_NONE)
 			*proto = (is_sips)?PROTO_TLS:PROTO_UDP;
 		goto do_a;
 	}
 
 	/* no port... what about proto? */
-	if ( proto && (*proto)!=PROTO_NONE ) {
+	if ( (*proto)!=PROTO_NONE ) {
 		/* have proto, but no port -> do SRV lookup */
 		LM_DBG("no port, has proto -> do SRV lookup!\n");
 		if (is_sips && (*proto)!=PROTO_TLS) {
@@ -1787,8 +1792,7 @@ struct hostent* sip_resolvehost( str* name, unsigned short* port,
 	}
 
 	if ( dns_try_naptr==0 ) {
-		if (proto)
-			*proto = (is_sips)?PROTO_TLS:PROTO_UDP;
+		*proto = (is_sips)?PROTO_TLS:PROTO_UDP;
 		goto do_srv;
 	}
 	LM_DBG("no port, no proto -> do NAPTR lookup!\n");
@@ -2061,7 +2065,7 @@ static inline struct dns_node *dns_node_copy(struct dns_node *s)
 		return 0;
 	}
 	memcpy( d, s, s->size);
-	d->vals = (struct dns_val*)((char*)d + ((char*)s->vals-(char*)s));
+	d->vals = (struct dns_val*)(void *)((char*)d + ((char*)s->vals-(char*)s));
 	for( i=0 ; i<s->no ; i++ )
 		d->vals[i].sval = (char*)d + ((char*)s->vals[i].sval-(char*)s);
 	return d;
@@ -2085,3 +2089,40 @@ struct dns_node *dns_res_copy(struct dns_node *s)
 	return d;
 }
 
+int resolve_hostport(str *in, unsigned short default_port,
+                     union sockaddr_union *dst)
+{
+	struct proxy_l* proxy;
+	char *p;
+	unsigned int port;
+	str st;
+
+	p = memchr(in->s, ':', in->len);
+	if (p != NULL) {
+		st.s = p+1;
+		st.len = in->len - (p + 1 - in->s);
+
+		if (str2int(&st, &port) != 0) {
+			LM_ERR("failed to parse port '%.*s' %d in host '%.*s'\n",
+			       st.len, st.s, st.len, in->len, in->s);
+			return -1;
+		}
+		st.s = in->s;
+		st.len = p - in->s;
+	} else {
+		st = *in;
+		port = default_port;
+	}
+
+	proxy = mk_proxy(&st, port, PROTO_NONE, 0);
+	if (proxy == NULL) {
+		LM_ERR("could not resolve hostname '%.*s'\n", in->len, in->s);
+		return -1;
+	}
+
+	hostent2su(dst, &proxy->host, proxy->addr_idx, proxy->port);
+
+	free_proxy(proxy);
+	pkg_free(proxy);
+	return 0;
+}
