@@ -1,11 +1,12 @@
-INSERT INTO version (table_name, table_version) values ('dispatcher','7');
+INSERT INTO version (table_name, table_version) values ('dispatcher','9');
 CREATE TABLE dispatcher (
     id NUMBER(10) PRIMARY KEY,
     setid NUMBER(10) DEFAULT 0 NOT NULL,
     destination VARCHAR2(192) DEFAULT '',
     socket VARCHAR2(128) DEFAULT NULL,
     state NUMBER(10) DEFAULT 0 NOT NULL,
-    weight NUMBER(10) DEFAULT 1 NOT NULL,
+    probe_mode NUMBER(10) DEFAULT 0 NOT NULL,
+    weight VARCHAR2(64) DEFAULT 1 NOT NULL,
     priority NUMBER(10) DEFAULT 0 NOT NULL,
     attrs VARCHAR2(128) DEFAULT '',
     description VARCHAR2(64) DEFAULT ''
